--                    -*- Mode: Ada -*-
-- Filename        : opengl-clock.adb
-- Description     : Types and subprograms for analog clock.
-- Author          : J. Schrer
-- Created On      : 9. Apr. 1998
-- Last Modified By:
-- Last Modified On: 
-- Update Count    : 
-- Status          :

--with Ada.Exceptions;
with Ada.Unchecked_Deallocation;
with Ada.Finalization;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded;

with Opengl.Logging;
with Opengl.Glu;
with Opengl.Colors;
with Opengl.Fonts;
with Opengl.Command_Line;
with Opengl.Timer;
with Opengl.Earth;
with Opengl.Float_32.Arrays.Vector_Operations;
with Opengl.Float_32.Arrays.Types;
with Opengl.Coordinate_System;
with Opengl.Win;

with Portable.Standard;
with Portable.Operating_System;

with Utilities.Files;
with Utilities.Generics.Block_Stack;
with Utilities.Generics.Sorting_And_Searching;

with Windows.Utils;
with Windows.Messages;

with Win32.Winuser;
with Win32.Windef;

package body Opengl.Clock is

  use Opengl.Float_32;
  use Opengl.Float_32.Arrays.Vector_Operations;
  use type Opengl.Float_32.Arrays.Types.Vector;
  use type Models.Strings.String_2d_Access;
  use type Models.Root_Model_Access;
  use type Bitmaps.Byte_Array_Access, Bitmaps.Pixel_Array_Access;
  use type Ada.Calendar.Time;
  use type Types.Points_3d;
  use type Gl.Gluint;
  use type Gl.Glboolean;
  use type Win.Ada_Called_Windows_Callback;
  use type Portable.Standard.Float_64;
  
  subtype Float_64 is Portable.Standard.Float_64;
   
  package Os                 renames Portable.Operating_System;
  package Calendar_Utilities renames Utilities.Calendar_Utilities;

  Message_Filename           : constant String  := "messages.txt";   
  Message_Font_Size          : constant         := 0.05;
  Alpha                      : Gl.Glfloat       := 0.67;
  Render_Mode                : Render_Mode_Type := Render_Both;
  Resize_Happened            : Boolean          := False; 
  Change_Alpha               : Boolean          := False;    
  Render_Lists               : Boolean          := True;    
  Initialized                : Boolean          := False;    
  Render_C_sys               : Boolean          := False;    
  
  Clock_Position             : Types.Vector_3d  := (0.0, 0.0, 0.0);
  Timezones_Position         : Types.Vector_3d  := (0.0, 0.0, 0.0);
  Bmp_X, Bmp_X0              : Gl.Glfloat       := 0.0;

  Resize_Callback            : Win.Ada_Called_Windows_Callback;
  Keypressed_Callback        : Win.Ada_Called_Windows_Callback;
  Middle_Mouse_Down_Callback : Win.Ada_Called_Windows_Callback;
    
  --==========================================================================
  
  procedure Edit_Messages is
  begin
    Windows.Utils.Start_Application(Executable => "notepad.exe",
                                    Parameter  => Message_Filename);
  end Edit_Messages;
  
  --==========================================================================
  
  function Keypressed
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin        
    case Wparam is      
      when  27    => case Render_Mode is                        -- Esc  
                       when Render_Both  => Render_Mode := Render_Clock;
                       when Render_Clock => Render_Mode := Render_Earth;
                       when Render_Earth => Render_Mode := Render_Both;
                     end case;
                     Initialized   := False;
      when 187    => if Change_Alpha then                       -- +
                       Alpha       := Gl.Glfloat'Min(Alpha + 0.1, 1.0);
                       Initialized := False;
                     end if;  
      when 189    => if Change_Alpha then                       -- -
                       Alpha := Gl.Glfloat'Max(Alpha - 0.1, 0.0); 
                       Initialized := False;
                     end if;  
      when  46    => Render_Lists  := not Render_Lists;         -- del
      when  65    => Change_Alpha  := not Change_Alpha;         -- a
      when  67    => Render_C_sys  := not Render_C_sys;         -- c
      when others => null;
    end case;    
    Timer.Set("Clock.Alpha:" & Gl.Glfloat'Image(Alpha) & 
              ", Clock.Lists: " & Boolean'Image(Render_Lists), 3, Append => False);
    
    if Keypressed_Callback /= null then
      return Keypressed_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Keypressed;
                          
  --==========================================================================

  function Middle_Mouse_Down
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin         
    Edit_Messages;
    
    if Middle_Mouse_Down_Callback /= null then
      return Middle_Mouse_Down_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Middle_Mouse_Down;
  
  --==========================================================================             
  
  function Resize
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    New_Width       : constant Natural := 
                      Natural(Win32.Windef.Loword(Win32.Dword(Lparam)));
    New_Height      : constant Natural := 
                      Natural(Win32.Windef.Hiword(Win32.Dword(Lparam)));                   
    Height_To_Width : constant Gl.Glfloat := 
                      Gl.Glfloat(New_Height) / Gl.Glfloat(New_Width);
    Screen_Height,
    Screen_Width    : Gl.Glfloat;
  begin    
    Resize_Happened := True;
    
    if New_Width < New_Height then
      Screen_Height := Screen.Width  * Height_To_Width;
      Screen_Width  := Screen.Width;
    else
      Screen_Height := Screen.Height;
      Screen_Width  := Screen.Height / Height_To_Width;
    end if;    

    Bmp_X := Bmp_X0 * Screen_Width / Screen.Width;
    if Clock_Position(X .. Y) /= (0.0, 0.0) then
      Clock_Position(X) := Screen_Width  / 2.0; 
      Clock_Position(Y) := Screen_Height / 2.0; 
    end if;                                     
    if Timezones_Position(X .. Y) /= (0.0, 0.0) then
      Timezones_Position(X) := Screen_Width  / 2.0; 
      Timezones_Position(Y) := Screen_Height / 2.0; 
    end if;                                     
    
    if Resize_Callback /= null then
      return Resize_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Resize;
                          
  --==========================================================================             

  task Error_Manager is
  
    entry Handle(The_File : in Ada.Text_Io.File_Type;
                 The_Line : in String);
                 
  end Error_Manager;
  
  task body Error_Manager is
    --------------------------------------------------------------------------- 
    procedure Set(Source : in     String;
                  Target :    out String;
                  Length :    out Natural) is
    begin
      if Target'Length <= Source'Length then
        Length := Target'Length;
      else
        Length := Source'Length;
      end if;  
      Target(1 .. Length) := Source(Source'First .. Source'First + Length - 1);            
    end Set;              
    ---------------------------------------------------------------------------   
    Nl          : constant String := Os.New_Line;
    Filename    : String(1 .. 256);
    Line        : String(1 .. 64);  
    Name_Length,
    Length,
    Line_Number : Natural := 0;   
  begin
    loop
      accept Handle(The_File : in Ada.Text_Io.File_Type;
                    The_Line : in String) do
        Set(Source => The_Line, 
            Target => Line, 
            Length => Length); 
            
        Set(Source => Ada.Text_Io.Name(The_File), 
            Target => Filename, 
            Length => Name_Length);
 
        Line_Number := Natural(Ada.Text_Io.Line(The_File));    
      end Handle;              
      
      Windows.Messages.Display_Message
        (The_Title           => "Time conversion layout-error",
         The_Message         => "Line " & Natural'Image(Line_Number - 1) & 
           ", " & Filename(1 .. Name_Length)                             & Nl & 
           "Wrong time format: '" & Line(1 .. Length) & "'"              & Nl &
           "correct format   : '[dd.mm.yyyy,]hh:mm[,d][,n]'"             & Nl &
           "dd and mm may have 2 or 1 digits, yyyy 4, 2 or 1"            & Nl &
           "d is the delta in hours, after that the message is repeated" & Nl &
           "n is the number of times the message is repeated, default 1",
         Add_Windows_Message => False);
            
      Edit_Messages;
    end loop;                                  
  end Error_Manager;
  
  --==========================================================================
  
  procedure Read(Alarm : in out Alarm_Type) is
    --------------------------------------------------------------------
    procedure Read_File(Now : in Ada.Calendar.Time) is
      --------------------------------------------------------------------
      procedure Free(Item : in out Alarm_Type) is
        procedure Free is new Ada.Unchecked_Deallocation
          (Name => Message_Array_Access, Object => Message_Array);
      begin
        if Item.Messages /= null then
          for I in Item.Messages'range loop
            Models.Strings.Free(Item.Messages(I).Text.String);
          end loop;
          Free(Item.Messages);
        end if;
      end Free;  
      --------------------------------------------------------------------
      procedure Extract(Text       : in     String;
                        Time       :    out Ada.Calendar.Time;
                        Times      :    out Positive;
                        Offset_Sec :    out Gl.Glfloat) is
     
        First  : Positive := Text'First;
        Last   : Natural  := Ada.Strings.Fixed.Index(Text, ".");
        ------------------------------------------------
        procedure Next_Separator(Key : in String) is
        begin
          First := Last + 1;
          Last  := Ada.Strings.Fixed.Index(Text(First + 1 .. Text'Last), Key);          
        end Next_Separator;  
        ------------------------------------------------
        Year   : Natural;
        Month  : Ada.Calendar.Month_Number;
        Day    : Ada.Calendar.Day_Number;
        Hour   : Natural range 0 .. 23;
        Minute : Natural range 0 .. 59;
        Sec : Ada.Calendar.Day_Duration := 0.0;
      begin
        if Last > 0 then -- contains date and time
          Day   := Ada.Calendar.Year_Number'Value(Text(First .. Last - 1));
          Next_Separator(".");          
          Month := Ada.Calendar.Year_Number'Value(Text(First .. Last - 1));
          Next_Separator(",");          
          Year  := Natural'Value(Text(First .. Last - 1));
          if Year < 2000 then
            Year := Year + 2000;
          end if;
        else -- only contains time
          Ada.Calendar.Split(Date    => Now, 
                             Year    => Year, 
                             Month   => Month, 
                             Day     => Day, 
                             Seconds => Sec);
        end if;
        
        Next_Separator(":");          
        Hour   := Natural'Value(Text(First .. Last - 1));
        Next_Separator(",");
 
        Offset_Sec := Gl.Glfloat'Last;
        Times      := 1;  

        if Last > 0 then          
          Minute := Natural'Value(Text(First .. Last - 1));          
          Next_Separator(",");
          if Last > 0 then
            Offset_Sec := Gl.Glfloat(Positive'Value
                                    (Text(First .. Last - 1)) * 3600);
            Times      := Natural'Value(Text(Last + 1 .. Text'Last));
          else
            Offset_Sec := Gl.Glfloat(Positive'Value
                                    (Text(First .. Text'Last)) * 3600);            
          end if;
        else
          Minute := Natural'Value(Text(First .. Text'Last));
        end if;       
        
        Time := Ada.Calendar.Time_Of(Year    => Year, 
                                     Month   => Month, 
                                     Day     => Day, 
                                     Seconds => Duration(Hour)   * 3600.0 + 
                                                Duration(Minute) * 60.0); 
      end Extract;
      -------------------------------------------------------------------            
      function "<" (Left, Right : in Message) return Boolean is
      begin
        return Left.Time < Right.Time;
      end "<";
      -------------------------------------------------------------------            
      function "<="(Left, Right : in Message) return Boolean is
      begin
        return Left.Time <= Right.Time;
      end "<=";
      -------------------------------------------------------------------                               
      package Message_Sorting_And_Searching is new 
              Utilities.Generics.Sorting_And_Searching
             (Element       => Message,
              Index         => Natural,
              Vector        => Message_Array,
              "<"           => "<",
              "<="          => "<=");
              
      package Message_Stack is new Utilities.Generics.Block_Stack
             (Element       => Message,
              Index         => Natural,
              Vector        => Message_Array,
              Vector_Access => Message_Array_Access);
 
      The_Stack   : Message_Stack.Stack;
      The_Message : Message;
      File        : Ada.Text_Io.File_Type;
      Line        : String(1 .. 256);
      Length      : Natural := 0;
      Limit       : constant Ada.Calendar.Time := Now - 86400.0;
      --------------------------------------------------------------------
      procedure Free(The_Stack : in out Message_Stack.Stack) is
      begin
        for I in 1 .. Message_Stack.Depth_Of(The_Stack) loop
          Message_Stack.Pop(Item => The_Message, From => The_Stack);
          Models.Strings.Free(The_Message.Text.String);
        end loop;  
      end Free;  
      --------------------------------------------------------------------
      function Image_Of(Item : in Ada.Calendar.Time) return String is
      
        Time : constant Calendar_Utilities.Time := 
                        Calendar_Utilities.Time_Of(Item);                          
        -----------------------------------                          
        function Separator return String is        
        begin
          if Time.The_Minute < 10 then
            return ":0";
          else         
            return ":";
          end if;         
        end Separator;  
        -----------------------------------                          
        Hour   : constant String := Natural'Image(Time.The_Hour);
        Minute : constant String := Natural'Image(Time.The_Minute);
      begin
        return Hour(Hour'First + 1 .. Hour'Last) & Separator &
               Minute(Minute'First + 1 .. Minute'Last);
      end Image_Of;
      -------------------------------------------------------------------
      function String_Of(Text : in String) return Models.Strings.String_2d is
      
        The_String : Models.Strings.String_2d :=
                     (Ada.Finalization.Controlled with
                      Size_Scaling => Alarm.Scale,
                      Color        => Colors.Red,
                      Position     => (X => 0.0, Y => 0.0, Z => 0.0),
                      Attitude     => 0.0,
                      Length       => Text'Length,
                      Text         => Text,
                      Alignment    => (Horizontal => Models.Strings.Center,
                                       Vertical   => Models.Strings.Middle),
                      Font         => Fonts.Font,
                      Font_Size    => Message_Font_Size,
                      Hidden_Data  => Models.Strings.Default_Hidden_Data);                                                          
      begin
        Models.Strings.Initialize(The_String);
        return The_String;
      end String_Of;        
      --------------------------------------------------------------------
    begin
      Ada.Text_Io.Open(File, Ada.Text_Io.In_File, Message_Filename);
     
      while not Ada.Text_Io.End_Of_File(File) loop
        begin
          Ada.Text_Io.Get_Line(File, Line, Length);
          
          if Length > 0 and then 
             Ada.Strings.Fixed.Index_Non_Blank(Line(1 .. Length)) > 0 then
             
            begin
              Extract(Text       => Line(1 .. Length),
                      Time       => The_Message.Time,
                      Offset_Sec => The_Message.Offset_Sec,
                      Times      => The_Message.Times);                    
            exception 
              when others =>                                             
                Free(The_Stack);
                Error_Manager.Handle(File, Line(1 .. Length));
                Ada.Text_Io.Close(File);
                exit;
            end;    
       
            if not Ada.Text_Io.End_Of_File(File) then
              Ada.Text_Io.Get_Line(File, Line, Length);
            else
              Length := 0;
            end if;  
            
            if The_Message.Time > Limit then 
              The_Message.Text.String := new Models.Strings.String_2d'
                (String_Of(Image_Of(The_Message.Time) & " " & Line(1 .. Length)));
                
              The_Message.Text.Background := 
                Models.Strings.Background_Of
                  (The_String       => The_Message.Text.String.all,
                   Background_Color => Colors.White,
                   Frame            => True,
                   Frame_Color      => Colors.Black,
                   Blanks           => False);
                   
              The_Message.Count := 0;
              The_Message.On    := True;
              
              Message_Stack.Push(Item => The_Message, Onto => The_Stack);        
            end if; 
          end if;
        exception 
          when others => null;
        end;                                              
      end loop;

      if Ada.Text_Io.Is_Open(File) then
        Ada.Text_Io.Close(File);
        Free(Alarm);
        Message_Stack.Convert(The_Stack, To_Vector => Alarm.Messages);
        Message_Sorting_And_Searching.Quicksort(Alarm.Messages.all);
      end if;  
      Timer.Set(Calendar_Utilities.Image_Of(Now) & 
                ", read " & Message_Filename, 1, Append => False);
    end Read_File;
    --------------------------------------------------------------------
    Time_Stamp : Ada.Calendar.Time := Ada.Calendar.Clock;
  begin
    if Os.Is_File(Message_Filename) then        
      Time_Stamp := Os.File_Time_Stamp(Message_Filename);   
      if Time_Stamp > Alarm.Time_Stamp then 
        Read_File(Now => Time_Stamp);
        Alarm.Time_Stamp := Time_Stamp;
        Alarm.On         := True;
      end if;  
    end if;
  end Read;
  
  --==========================================================================
  
  function The_Text(Text             : in String;
                    Position         : in Types.Vector_3d;
                    Scale            : in Gl.Glfloat;
                    Font_Size        : in Gl.Glfloat       := 0.1;
                    Text_Color       : in Types.Color_Type := Colors.Black;
                    Background_Color : in Types.Color_Type := Colors.Yellow;
                    Frame            : in Boolean          := True;
                    Frame_Color      : in Types.Color_Type := Colors.Black)
           return Text_Type is
                      
    The_String : Models.Strings.String_2d :=
                 (Ada.Finalization.Controlled with
                  Size_Scaling => Scale,
                  Color        => Text_Color,
                  Position     => Scale * Position,
                  Attitude     => 0.0,
                  Length       => Text'Length,
                  Text         => Text,
                  Alignment    => (Horizontal => Models.Strings.Center,
                                   Vertical   => Models.Strings.Middle),
                  Font         => Fonts.Font,
                  Font_Size    => Font_Size,
                  Hidden_Data  => Models.Strings.Default_Hidden_Data);                                    
  begin
    Models.Strings.Initialize(The_String);
    
    return Text_Type'(Background => Models.Strings.Background_Of
                                   (The_String       => The_String,
                                    Background_Color => Background_Color,
                                    Frame            => Frame,
                                    Frame_Color      => Frame_Color),                        
                      String     => new Models.Strings.String_2d'(The_String));
  end The_Text; 
  
  --==========================================================================
  
  procedure Render(The : in out Text_Type) is
  begin
    if The.String /= null then
      Models.Forms.Render  (The.Background);
      Models.Strings.Render(The.String.all);
    end if;  
  end Render;
  
  --==========================================================================
  
  procedure Render(The : in out Clock_Dialface_Type) is
  begin    
    case The.Tick_Number is
      when 12     => for I in 1 .. The.Tick_Number loop
                       Models.Dialfaces.Render(The.Small_Ticks);
                       Models.Rotate_Z(30.0);
                     end loop;  
      when others => null;
    end case;  
    
    Models.Dialfaces.Render(The.Big_Ticks);    
    Models.Dialfaces.Render(The.H_Digits);
    Models.Forms.Render    (The.Outer_Disk);
    
    case The.Render_Mode is 
      when Render_Clock => Render(The.Name);
      when others       => null;
    end case;  
  end Render;
  
  --==========================================================================
    
  Text_Colors : array(Color_Index_Type) of Types.Color_Type := 
                 (0 => Colors.White, 
                  1 => Colors.Red,
                  2 => Colors.Yellow);  
                  
  Bg_Colors   : array(Color_Index_Type) of Types.Color_Type := 
                 (0 => Colors.Red, 
                  1 => Colors.Green,
                  2 => Colors.Blue);  
                  
  Frame_Color : Types.Color_Type := Colors.Black;
  
  --==========================================================================
    
  procedure Set_Alpha(The : in out Text_Type;
                      To  : in     Gl.Glfloat) is                    
  begin
    if The.String /= null then  
      The.Background.Color.Alpha       := To;
      The.Background.Frame_Color.Alpha := To;
      The.String.Color.Alpha           := To;
    end if;  
  end Set_Alpha;  
  
  --==========================================================================

  procedure Check_List(The  : in out Text_Type;
                       List : in out Gl.Gluint) is            
  begin
    if The.String /= null then
      Display_Lists.Generate(List);      
      Gl.Glnewlist(List => List, Mode => Gl.Gl_Compile);
      Render(The);
      Gl.Glendlist;    
    end if;  
  end Check_List;

  --==========================================================================
    
  procedure Set_Alpha(The_Model : in out Model;
                      To        : in     Gl.Glfloat) is
  begin  
    The_Model.Clock_Parts.Hour_Pointer.Color.Alpha   := To;     
    The_Model.Clock_Parts.Minute_Pointer.Color.Alpha := To;         
    The_Model.Clock_Parts.Second_Pointer.Color.Alpha := To;
    The_Model.Clock_Parts.Center_Disk.Color.Alpha    := To;            
    for I in The_Model.Clock_Parts.Dialface.H_Digits.Strings'range loop
      The_Model.Clock_Parts.Dialface.H_Digits.Strings(I).Color.Alpha := To;
    end loop;  
    
    The_Model.Clock_Parts.Dialface.H_Digits.Color.Alpha    := To;
    The_Model.Clock_Parts.Dialface.Big_Ticks.Color.Alpha   := To;
    The_Model.Clock_Parts.Dialface.Small_Ticks.Color.Alpha := To;    
    The_Model.Clock_Parts.Dialface.Outer_Disk.Color.Alpha  := To;      
          
    Set_Alpha(The_Model.Clock_Parts.Dialface.Name, To);
    Set_Alpha(The_Model.Location, To);
                     
    The_Model.Digital_Date.Color.Alpha := To;     
    The_Model.Digital_Time.Color.Alpha := To;              
    
    for I in Text_Colors'range loop
      Text_Colors(I).Alpha := To;
      Bg_Colors(I).Alpha   := To;
    end loop;             
    
    Frame_Color.Alpha := To;
  end Set_Alpha;  
            
  --==========================================================================
  
  function Alpha_Of(The_Model : in Model) return Gl.Glfloat is
  begin  
    case The_Model.Clock_Parts.Dialface.Render_Mode is
      when Render_Both  => if The_Model.Earth /= null then
                             return Alpha;
                           else
                             return 1.0;
                           end if;  
      when Render_Clock => return 1.0;
      when Render_Earth => return Alpha;
    end case;  
  end Alpha_Of;  

  --==========================================================================
  
  procedure Set_Alpha(The_Model : in out Model) is
  begin  
    Set_Alpha(The_Model, To => Alpha_Of(The_Model));
  end Set_Alpha;  

  --==========================================================================
  
  procedure Check_Lists(The_Model : in out Model) is        
    -----------------------------------------------------------
    procedure Generate(Object : in out Models.Forms.Disk_2d_Z;
                       List   : in out Gl.Gluint) is    
    begin                     
      Display_Lists.Generate(List);
      Gl.Glnewlist(List, Mode => Gl.Gl_Compile);
      Models.Forms.Render(Object);
      Gl.Glendlist;
    end Generate;
    -----------------------------------------------------------      
    Parts : Clock_Parts_Type      renames The_Model.Clock_Parts;
    Lists : Clock_Part_Lists_Type renames The_Model.Clock_Part_Lists;
  begin
    if not Initialized and then The_Model.Index = 1 then
      Set_Alpha(The_Model);
      
      Display_Lists.Generate(Lists.Dialface);
      Gl.Glnewlist(Lists.Dialface, Mode => Gl.Gl_Compile);
      Render(Parts.Dialface.all);
      Gl.Glendlist;

      Generate(Object => Parts.Background, List => Lists.Background);
      Generate(Object => Parts.Center_Disk, List => Lists.Center_Disk);
      
      Check_List(The_Model.Location, List => The_Model.Location_List);
  
      Parts.Hour_Pointer.Attitude   := 0.0;
      Parts.Minute_Pointer.Attitude := 0.0;
      Parts.Second_Pointer.Attitude := 0.0;
      
      Lists.Hour_Pointer   := Models.Display_List_Model_Of
                              (The_Model => Parts.Hour_Pointer,
                               The_List  => Lists.Hour_Pointer.Display_List);
           
      Lists.Minute_Pointer := Models.Display_List_Model_Of
                              (The_Model => Parts.Minute_Pointer,
                               The_List  => Lists.Minute_Pointer.Display_List);
           
      Lists.Second_Pointer := Models.Display_List_Model_Of
                              (The_Model => Parts.Second_Pointer,
                               The_List  => Lists.Second_Pointer.Display_List);
           
      Initialized := True;
    end if;           
  end Check_Lists;
          
  --==========================================================================
    
  procedure Render(The_Alarm : in out Alarm_Type;
                   Now       : in     Ada.Calendar.Time) is
  
    Y_Offset   : constant   := -0.96;
    Blink_Time : constant   := 10.0;
    Beep_Time  : constant   := 60.0;
    On_Time    : constant   := 24.0 * 3600.0;
    Count      : Positive   := 1;
    Chars      : Positive   := Positive'Last;
    Char_Width : Gl.Glfloat := 0.0;      
    --------------------------------------------------------------------------
    procedure Render(The_Message : in out Message) is     
      ------------------------------------------------------------------------  
      procedure Render(Text : in out Text_Type) is              
      begin
        if Count = 1 then
          Char_Width := Text.String.Size_Scaling * Message_Font_Size * 
                        Fonts.Mean_Character_Width_Of(Text.String.Font) / 
                        Fonts.Max_Character_Height_Of(Text.String.Font);
          Chars      := Integer(The_Alarm.Window_Width / Char_Width);            
        end if;
        
        Text.String.Position(Y)     := (Y_Offset + Gl.Glfloat(Count - 1) * 1.6 * 
                                       Message_Font_Size) * The_Alarm.Scale;
        Text.String.Color           := Text_Colors(The_Message.Color_Index);
        Text.Background.Position(Y) := Text.String.Position(Y);
        Text.Background.Color       := Bg_Colors  (The_Message.Color_Index);
        Count                       := Count + 1;
    
        Models.Forms.Render  (Text.Background);
        Models.Strings.Render(Text.String.all, Chars => Chars, Increment => 1);
      end Render;  
      ------------------------------------------------------------------------                                        
      D : Gl.Glfloat := Gl.Glfloat(Float_64'
                       (Calendar_Utilities."-"(Now, The_Message.Time)));
    begin      
      if D >= 0.0 then                    
        D := D mod The_Message.Offset_Sec;
        
        if D in 0.0 .. Beep_Time then
          Windows.Utils.Beep;
          if D in 0.0 .. Blink_Time then
            The_Message.On := not The_Message.On;
            Win.Make_Opengl_Foreground_Window;
          else
            The_Message.On := True;  
          end if;
        elsif D >= Gl.Glfloat'Min(The_Message.Offset_Sec, On_Time) then
          The_Message.Count := The_Message.Count + 1;
        end if;
        
        if The_Message.On and then The_Message.Text.String /= null then            
          The_Message.Color_Index := The_Message.Color_Index + 1;            
          Render(The_Message.Text);
        end if;
      end if;
    end Render;
    --------------------------------------------------------------------------
  begin  
    Read(The_Alarm);

    if The_Alarm.Messages /= null then      
      for I in The_Alarm.Messages'range loop
        if The_Alarm.Messages(I).Count < The_Alarm.Messages(I).Times then
          Render(The_Alarm.Messages(I));
        end if;  
      end loop;
    end if;
  end Render;
  
  --==========================================================================
  
  Bitmap_Color : Color_Index_Type := Color_Index_Type'First;
  
  --==========================================================================
  
  procedure Render(The_Model : in out Model) is
    --------------------------------------------------------------------------
    function The_Time return Ada.Calendar.Time is
    begin
      if The_Model.Calculate_Time then
        The_Model.Now := Ada.Calendar.Clock;
      end if;  
      return The_Model.Now + The_Model.Time_Offset;
    end The_Time;  
    --------------------------------------------------------------------------
    Time_Now   : constant Ada.Calendar.Time := The_Time;
    Now        : constant Calendar_Utilities.Time := 
                          Calendar_Utilities.Time_Of(Time_Now);
    Date_Image : constant String := Calendar_Utilities.Date_Image_Of(Now);
    Time_Image : constant String := Calendar_Utilities.Time_Image_Of(Now);
    Color1, 
    Color2     : Color_Index_Type;
    Bmp_Pos    : constant Types.Vector_3d := The_Model.Bitmaps(1).Position;
    --------------------------------------------------------------------------
    procedure Render_Bitmaps is
    begin  
      if The_Model.Index = 1 then
        for I in The_Model.Bitmaps'range loop
          if The_Model.Bitmaps(I).Bitmap.Bytes  /= null or else
             The_Model.Bitmaps(I).Bitmap.Pixels /= null then
            Bitmap_Color := Bitmap_Color + 1; 
            Color1       := Bitmap_Color; 
            Color2       := Bitmap_Color + 1;         
            Bitmaps.Exchange_Colors(The_Model.Bitmaps(I).Bitmap, 
                                    First_Color  => Natural(Color1) + 1, 
                                    Second_Color => Natural(Color2) + 1); 
            Bitmaps.Render(The_Model.Bitmaps(I));
          end if;                  
        end loop;  
      end if;  
    end Render_Bitmaps;
    --------------------------------------------------------------------------
    procedure Render_Coordinate_Systems is
                         
      C_System : Coordinate_System.Model :=
        (Models.Root_Model with
         Position           => (0.0, 0.0, 0.0),
         Attitude           => The_Model.Attitude,
         Size_Scaling       => Coordinate_System.Max_Length,
         Color              => Colors.Black,
         Axis_Width         => 1.0,
                            --  1234567890123456
         Name               => " Clock          ",
         Transformation     => 0);
    begin
      Coordinate_System.Render(C_System);             
      
      C_System.Size_Scaling := 0.5 * Coordinate_System.Max_Length;    
      C_System.Name         := " Sec            ";
      C_System.Attitude(Z)  := The_Model.Clock_Part_Lists.Second_Pointer.Attitude;
      Coordinate_System.Render(C_System);             
      
      C_System.Size_Scaling := 0.16 * Coordinate_System.Max_Length;    
      C_System.Name         := " Time           ";
      C_System.Attitude(Z)  := 0.0;
      C_System.Position     := The_Model.Digital_Time.Position;
      Coordinate_System.Render(C_System);             
    end Render_Coordinate_Systems;  
    --------------------------------------------------------------------------
    procedure Set_Attitudes(Second :   out Gl.Glfloat;
                            Minute :   out Gl.Glfloat;
                            Hour   :   out Gl.Glfloat;                            
                            Now   : in     Calendar_Utilities.Time;
                            Ticks : in     Positive) is
    begin
      Second := Gl.Glfloat(90 - Now.The_Second * 6);
      Minute := Gl.Glfloat(90 - Now.The_Minute * 6) -
                Gl.Glfloat(Now.The_Second) / 10.0;                                         
      case Ticks is
        when 12 => Hour := Gl.Glfloat(90 - Now.The_Hour * 30) -
                           Gl.Glfloat(Now.The_Minute) / 2.0   -
                           Gl.Glfloat(Now.The_Second) / 120.0;
        when 24 => Hour := Gl.Glfloat(90 - Now.The_Hour * 15) -
                           Gl.Glfloat(Now.The_Minute) / 4.0   -
                           Gl.Glfloat(Now.The_Second) / 240.0;
        when others => null;                                     
      end case;    
    end Set_Attitudes;  
    --------------------------------------------------------------------------                                  
    procedure Render_Clock_Parts is
    begin  
      Render_Bitmaps;    
      
      Gl.Glcalllist(The_Model.Clock_Part_Lists.Dialface);
      
      Models.Strings.Render(The_Model.Digital_Date);
      Models.Strings.Render(The_Model.Digital_Time);
      
      if Render_Lists then
        Set_Attitudes(Second => The_Model.Clock_Part_Lists.Second_Pointer.Attitude, 
                      Minute => The_Model.Clock_Part_Lists.Minute_Pointer.Attitude, 
                      Hour   => The_Model.Clock_Part_Lists.Hour_Pointer.Attitude, 
                      Now    => Now,
                      Ticks  => The_Model.Clock_Parts.Dialface.H_Digits.Number);                                          
                    
        Models.Render(The_Model.Clock_Part_Lists.Hour_Pointer);
        Models.Render(The_Model.Clock_Part_Lists.Minute_Pointer);
        Models.Render(The_Model.Clock_Part_Lists.Second_Pointer);        
      else
        Set_Attitudes(Second => The_Model.Clock_Parts.Second_Pointer.Attitude, 
                      Minute => The_Model.Clock_Parts.Minute_Pointer.Attitude, 
                      Hour   => The_Model.Clock_Parts.Hour_Pointer.Attitude, 
                      Now    => Now,
                      Ticks  => The_Model.Clock_Parts.Dialface.H_Digits.Number);                                          
                    
        Models.Forms.Render(The_Model.Clock_Parts.Hour_Pointer);
        Models.Forms.Render(The_Model.Clock_Parts.Minute_Pointer);
        Models.Forms.Render(The_Model.Clock_Parts.Second_Pointer);        
      end if;
        
      Gl.Glcalllist(The_Model.Clock_Part_Lists.Center_Disk);      
      
      if Render_C_sys then
        Render_Coordinate_Systems;
      end if;      
    end Render_Clock_Parts;  
    --------------------------------------------------------------------------
    procedure Render_Earth is
    begin
      if The_Model.Earth /= null then
        Earth.Set_Time_Offset(The_Model => Earth.Model(The_Model.Earth.all), 
                              To        => The_Model.Time_Offset);         
        Earth.Set_Location(The_Model => Earth.Model(The_Model.Earth.all),
                           Latitude  => The_Model.Latitude,
                           Longitude => The_Model.Longitude);
        Earth.Render(Earth.Model(The_Model.Earth.all));
      end if;  
    end Render_Earth;    
    --------------------------------------------------------------------------
  begin
    The_Model.Bitmaps(1).Position(X)           := Bmp_X; 
    The_Model.Position                         := Clock_Position;
    The_Model.Clock_Parts.Dialface.Render_Mode := Render_Mode;
    
    Check_Lists(The_Model);
        
    The_Model.Digital_Date.Text := Date_Image;
    The_Model.Digital_Time.Text := Time_Image(Time_Image'First ..
                                              Time_Image'Last - 3);
       
    Models.Initialize_Render(The_Model);
    
    case The_Model.Clock_Parts.Dialface.Render_Mode is
      when Render_Both  => 
        Gl.Glcalllist(The_Model.Clock_Part_Lists.Background);
        Render_Earth;
        Render_Clock_Parts;
      when Render_Clock => 
        The_Model.Bitmaps(1).Position := (-20.0, 40.0, 0.0);
        Gl.Glcalllist(The_Model.Clock_Part_Lists.Background);
        Render_Clock_Parts;
        The_Model.Bitmaps(1).Position := Bmp_Pos;
      when Render_Earth => 
        Render_Earth;
        Render_Bitmaps;
    end case;  
      
    Gl.Glcalllist(The_Model.Location_List);
    
    if The_Model.Alarm.On then
      The_Model.Alarm.Window_Width := 2.0 * The_Model.Position(X);         
      Render(The_Model.Alarm, Now => The_Model.Now);
    end if;  
    
    Models.Finalize_Render(The_Model);
  end Render;
  
  --==========================================================================
  
  function Model_With
    (Position       : in Types.Vector_3d := (X => Screen.Width  / 2.0,
     	                                       Y => Screen.Height / 2.0,
                                             Z => 0.0);
     Attitude       : in Types.Vector_3d := (X => 0.0,
     	                                       Y => 0.0,
                                             Z => 0.0);
     Scale          : in Gl.Glfloat      := Screen.Height / 2.0;
     Dialface       : in Dialface_Type   := Hours_12;
     Time_Offset    : in Duration        := 0.0;
     Earth_Scale    : in Gl.Glfloat      := 0.0;
     Earth_Location : in String          := "";
     Latitude       : in Gl.Glfloat      := 0.0; -- of geographic location
     Longitude      : in Gl.Glfloat      := 0.0; -- of geographic location
     Sunlight       : in Boolean         := False) return Model is

    ------------------------------------------------- 
    function The_Number return Positive is
    begin
      case Dialface is
        when Hours_12 => return 12;
        when Hours_24 => return 24;
      end case;
    end The_Number;  
    ------------------------------------------------- 
    function The_Start_Angle return Gl.Glfloat is
    begin
      case Dialface is
        when Hours_12 => return 60.0;
        when Hours_24 => return 75.0;
      end case;
    end The_Start_Angle;  
    ------------------------------------------------- 
    function The_Angle_Increment return Gl.Glfloat is
    begin
      case Dialface is
        when Hours_12 => return -30.0;
        when Hours_24 => return -15.0;
      end case;
    end The_Angle_Increment;  
    ------------------------------------------------- 
    function The_Digit_Value return Types.Integer_Array is
      The_Digits : Types.Integer_Array(1 .. The_Number); 
    begin
      for I in The_Digits'range loop
        The_Digits(I) := I;
      end loop;
      return The_Digits;
    end The_Digit_Value;  
    ------------------------------------------------- 
    function The_Width return Gl.Glfloat is
    begin
      case Dialface is
        when Hours_12 => return 0.025;
        when Hours_24 => return 0.02;
      end case;
    end The_Width;  
    ------------------------------------------------- 
    function The_Font_Size return Gl.Glfloat is
    begin
      case Dialface is
        when Hours_12 => return 0.08;
        when Hours_24 => return 0.06;
      end case;
    end The_Font_Size;  
    ------------------------------------------------- 
    Tick_Outer_Radius        : constant := 0.99;
    Tick_Inner_Radius        : constant := 0.85;
    Small_Tick_Inner_Radius  : constant := 0.90;
    Small_Tick_Width         : constant := 0.01;
    String_Center            : constant := 0.92;
    Name_Fontsize            : constant := 0.1;
    Date_Fontsize            : constant := 0.06;
    Time_Fontsize            : constant := 0.12;
    Name_Y                   : constant := -0.52;
    Date_Y                   : constant := -0.2;
    Time_Y                   : constant := -0.35;
    H_Rectangle_Width        : constant := 0.035;
    H_Triangle_Base_Radius   : constant := 0.4;
    H_Triangle_Peak_Radius   : constant := 0.67;
    H_Triangle_Base_Width    : constant := 0.035;
    M_Rectangle_Width        : constant := 0.03;
    M_Triangle_Base_Radius   : constant := 0.8;
    M_Triangle_Peak_Radius   : constant := 0.85;
    M_Triangle_Base_Width    : constant := 0.03;
    S_Inner_Radius           : constant := -0.1;
    S_Rectangle_Width        : constant := 0.01;
    S_Triangle_Base_Radius   : constant := 0.85;
    S_Triangle_Peak_Radius   : constant := 0.85;
    S_Triangle_Base_Width    : constant := 0.01;
    Center_Disk_Outer_Radius : constant := 0.025;
    Location_Y               : constant := 0.92;
    Background_Disk_Color    : constant Types.Color_Type := Colors.White;
    Center_Disk_Color        : constant Types.Color_Type := Colors.Black;
    Name_Color               : constant Types.Color_Type := Colors.Brown;
    Rectangle_Color          : constant Types.Color_Type := Colors.Grey9;
    Text_Color               : constant Types.Color_Type := Colors.Red;
    Tick_Color               : constant Types.Color_Type := Colors.Light_Blue;
    Digit_Color              : constant Types.Color_Type := Colors.Green;
    Pointer_Color            : constant Types.Color_Type := Colors.Orange;
    Clock_Text               : constant String := "Ada 95 & OpenGL";  
    ------------------------------------------------- 
    
    Clock_Dialface : Clock_Dialface_Access := new Clock_Dialface_Type'

      (Tick_Number => The_Number,
        
       H_Digits    => Models.Dialfaces.Ticks_With_Digits'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
    	   Color              => Digit_Color,
    	   Position           => (0.0, 0.0, 0.0),
    	   Attitude           => 0.0,
    	   Number             => The_Number,
    	   Start_Angle        => The_Start_Angle,
    	   Angle_Increment    => The_Angle_Increment,
    	   Min_Radius         => Tick_Inner_Radius,
    	   Max_Radius         => Tick_Outer_Radius,
    	   Width              => 0.0,
    	   Digit_Value        => The_Digit_Value,
         Strings            => (others => null),
         String_Orientation => Models.Dialfaces.Vertical,
         String_Center      => String_Center,
    	   Font               => Fonts.Font,
    	   Font_Size          => The_Font_Size),

       Big_Ticks => Models.Dialfaces.Ticks'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
    	   Color              => Tick_Color,
    	   Position           => (0.0, 0.0, 0.0),
    	   Attitude           => 0.0,
    	   Number             => The_Number,
    	   Start_Angle        => The_Start_Angle,
    	   Angle_Increment    => The_Angle_Increment,
    	   Min_Radius         => Tick_Inner_Radius,
    	   Max_Radius         => Tick_Outer_Radius,
    	   Width              => The_Width),

       Small_Ticks => Models.Dialfaces.Ticks'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
         Color              => Tick_Color,
         Position           => (0.0, 0.0, 0.0),
         Attitude           => 0.0,
         Number             => 4,
         Start_Angle        => 6.0,
         Angle_Increment    => 6.0,
         Min_Radius         => Small_Tick_Inner_Radius,
         Max_Radius         => Tick_Outer_Radius,
         Width              => Small_Tick_Width),
       
       Outer_Disk => Models.Forms.Disk_2d_Z'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
         Color              => Tick_Color,
         Position           => (0.0, 0.0, 0.0),
         Attitude           => 0.0,
         Inner_Radius       => Tick_Outer_Radius,
         Outer_Radius       => 1.0,
         Slices             => 360,
         Loops              => 1,
         Glu_Quadric_Object => Glu.Glunewquadric),

       Name => The_Text(Text             => Clock_Text,
                        Position         => (X => 0.0, Y => Name_Y, Z => 0.0),
                        Scale            => Scale,
                        Font_Size        => Name_Fontsize,
                        Text_Color       => Name_Color,
                        Background_Color => Rectangle_Color,
                        Frame            => False),
                        
       Render_Mode => Render_Clock); 
    ------------------------------------------------- 
       
    Clock_Parts : Clock_Parts_Type :=
      
      (Background   => Models.Forms.Disk_2d_Z'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
         Color              => Background_Disk_Color,
         Position           => (0.0, 0.0, -1.0),
         Attitude           => 0.0,
         Inner_Radius       => 0.0,
         Outer_Radius       => Tick_Outer_Radius,
         Slices             => 360,
         Loops              => 1,
         Glu_Quadric_Object => Glu.Glunewquadric),
         
       Dialface     => Clock_Dialface,
      
       Hour_Pointer => Models.Forms.Radial_Arrow_2d_Z'
        (Ada.Finalization.Controlled with
	       Size_Scaling         => Scale,
	       Color                => Pointer_Color,
	       Position             => (0.0, 0.0, 0.0),
	       Attitude             => 0.0,
	       Inner_Radius         => 0.0,
	       Rectangle_Width      => H_Rectangle_Width,
	       Triangle_Base_Radius => H_Triangle_Base_Radius,
	       Triangle_Peak_Radius => H_Triangle_Peak_Radius,
	       Triangle_Base_Width  => H_Triangle_Base_Width,
	       Parts                => Models.Forms.Default_Radial_Arrow_Parts),

       Minute_Pointer => Models.Forms.Radial_Arrow_2d_Z'
        (Ada.Finalization.Controlled with
	       Size_Scaling         => Scale,
	       Color                => Pointer_Color,
	       Position             => (0.0, 0.0, 0.0),
	       Attitude             => 0.0,
	       Inner_Radius         => 0.0,
	       Rectangle_Width      => M_Rectangle_Width,
	       Triangle_Base_Radius => M_Triangle_Base_Radius,
	       Triangle_Peak_Radius => M_Triangle_Peak_Radius,
	       Triangle_Base_Width  => M_Triangle_Base_Width,
	       Parts                => Models.Forms.Default_Radial_Arrow_Parts),

       Second_Pointer => Models.Forms.Radial_Arrow_2d_Z'
        (Ada.Finalization.Controlled with
	       Size_Scaling         => Scale,
	       Color                => Pointer_Color,
	       Position             => (0.0, 0.0, 0.0),
	       Attitude             => 0.0,
	       Inner_Radius         => S_Inner_Radius,
	       Rectangle_Width      => S_Rectangle_Width,
	       Triangle_Base_Radius => S_Triangle_Base_Radius,
	       Triangle_Peak_Radius => S_Triangle_Peak_Radius,
	       Triangle_Base_Width  => S_Triangle_Base_Width,
	       Parts                => Models.Forms.Default_Radial_Arrow_Parts),
         
       Center_Disk => Models.Forms.Disk_2d_Z'
        (Ada.Finalization.Controlled with
         Size_Scaling       => Scale,
         Color              => Center_Disk_Color,
         Position           => (0.0, 0.0, 1.0),
         Attitude           => 0.0,
         Inner_Radius       => 0.0,
         Outer_Radius       => Center_Disk_Outer_Radius,
         Slices             => 36,
         Loops              => 1,
         Glu_Quadric_Object => Glu.Glunewquadric));
         
    Lists : Clock_Part_Lists_Type;
    ------------------------------------------------- 
         
    Digital_Date : Models.Strings.String_2d :=      
	      (Ada.Finalization.Controlled with
	       Size_Scaling => Scale,
	       Color        => Text_Color,
	       Position     => Scale * (X => 0.0, Y => Date_Y, Z => 0.0),
	       Attitude     => 0.0,
	       Length       => Date_Image'Length,
	       Text         => Date_Image,
             Alignment    => (Horizontal => Models.Strings.Center,
                          Vertical   => Models.Strings.Middle),
	       Font         => Fonts.Font,
	       Font_Size    => Date_Fontsize,
         Hidden_Data  => Models.Strings.Default_Hidden_Data);

    Digital_Time : Models.Strings.String_2d :=
	      (Ada.Finalization.Controlled with
	       Size_Scaling => Scale,
	       Color        => Text_Color,
	       Position     => Scale * (X => 0.0, Y => Time_Y, Z => 0.0),
	       Attitude     => 0.0,
	       Length       => Time_Image'Length,
	       Text         => Time_Image,
         Alignment    => (Horizontal => Models.Strings.Center,
                          Vertical   => Models.Strings.Middle),
	       Font         => Fonts.Font("Courier New"),
	       Font_Size    => Time_Fontsize,
         Hidden_Data  => Models.Strings.Default_Hidden_Data);
    ------------------------------------------------- 
      
    The_Earth       : Models.Root_Model_Access;                                          
    Location        : Text_Type;
    Location_List   : Gl.Gluint := 0;
    
    Ada_Excel  : aliased String := "AdaExcellenceLarge.bmp";
    Elke       : aliased String := "\privat\Elke.bmp";
    Ada_Logo   : aliased String := "Adalogo-new-small.bmp";
    Ada_Powered: aliased String := "Adapowered2.bmp";
    Lady_Earth : aliased String := "Lady-earth.bmp";
    Lady_Moon  : aliased String := "Lady-moon.bmp";
    Pictures   : Bitmaps_Array(1 .. 4);       
    Filenames  : constant array(Pictures'range) of 
                 Ada.Strings.Unbounded.String_Access :=
                 (1 => Ada_Excel'Unchecked_Access,      
--                  2 => Elke'Unchecked_Access,      
--                  2 => Ada_Logo'Unchecked_Access,      
                  2 => Ada_Powered'Unchecked_Access,      
                  3 => Lady_Moon'Unchecked_Access,      
                  4 => Lady_Earth'Unchecked_Access);
    Positions  : constant Types.Points_3d(Pictures'range) := 0.75 * 
                 (1 => (X => - Screen.Width  / 2.0 - 18.0, 
                        Y => + Screen.Height / 2.0,         
                        Z => 0.0),                    
                  2 => (X => + Screen.Width  / 2.0 + 42.0, 
                        Y => + Screen.Height / 2.0 + 33.0,    
                        Z => 0.0),                    
                  3 => (X => - Screen.Width  / 2.0 - 8.0,  
                        Y => - Screen.Height / 2.0 - 6.0,     
                        Z => 0.0),                    
                  4 => (X => + Screen.Width  / 2.0 + 6.0, 
                        Y => - Screen.Height * 5.0 / 12.0 - 16.0, 
                        Z => 0.0));                    
    Scales     : constant Types.Float_Vector(Pictures'range) :=
                 (1 => 1.0,
                  2 => 1.0,
                  3 => 1.0,
                  4 => 1.0);                  
                  
    Window_Area: constant Screen.Window_Area_Type := Screen.Current_Window_Area;              
    Alarm      : Alarm_Type;    
    The        : Model;
 begin
    Keypressed_Callback := 
      Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Keydown);
    Middle_Mouse_Down_Callback := 
      Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Mbuttondown);
    Resize_Callback := Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Size);
      
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Keydown)     :=
      Keypressed'access;
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Mbuttondown) :=
      Middle_Mouse_Down'access;
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Size) := Resize'access;  
    
    for I in Pictures'range loop    
      if Utilities.Files.Exists(Filenames(I).all) then
        if Pictures(I).Bitmap.Bytes = null and then 
           Pictures(I).Bitmap.Pixels = null then
             
          Pictures(I) := Bitmaps.Model_With
                        (Filename => Filenames(I).all,
                         Position => Positions(I),
                         Attitude => 0.0,
                         Scale    => 0.5 * Scales(I) * 
                                     Gl.Glfloat(Window_Area.Height) / 
                                     Gl.Glfloat(Screen.Pixel_Height));

          Bitmaps.Generate_Pixels(Pictures(I).Bitmap);
        end if;  
      end if;
    end loop;
    
    if Earth_Location /= "" then
      Location := The_Text(Text     => Earth_Location,
                           Position => (X => 0.0, Y => Location_Y, Z => 0.0),
                           Scale    => Scale);
    end if;  
                                       
    Alarm.Scale := Scale;
    Read(Alarm);

    if Earth_Scale > 0.0 then
      The_Earth := Earth.Model_With
        (Bitmap_File     => Command_Line.Earth_File,
         Position        => (0.0, 0.0, 0.0),
         Attitude        => (0.0, 0.0, 0.0),
         Scale           => Scale * Gl.Glfloat(Earth_Scale),
         Longitude_Lines => 36,
         Latitude_Lines  => 18,
         Time_Offset     => Time_Offset,
         Time_Factor     => Gl.Glfloat(Command_Line.Time_Factor),
         Sunlight        => Sunlight);
         
      Clock_Parts.Dialface.Render_Mode := Render_Both;   
    end if;

    The := Model'(Ada.Finalization.Controlled with
                  Size_Scaling     => Scale,
                  Color            => Colors.White,
                  Position         => Position,
                  Attitude         => Attitude,
                   
                  Clock_Parts      => Clock_Parts,
                  Clock_Part_Lists => Lists,
                   
                  Digital_Date     => Digital_Date,
                  Digital_Time     => Digital_Time,
                    
                  Now              => Ada_Now,
                  Time_Offset      => Time_Offset,
                  Calculate_Time   => True,
                  Index            => 1,       
                                 
                  Location         => Location,
                  Location_List    => Location_List,
                  Latitude         => Latitude,
                  Longitude        => Longitude,
                  Earth            => The_Earth,
                  Bitmaps          => Pictures,
                   
                  Alarm            => Alarm);    
                  
    Models.Dialfaces.Initialize(The.Clock_Parts.Dialface.H_Digits);  
    Check_Lists(The);                  
    
    Clock_Position := Position;
    Bmp_X0         := The.Bitmaps(1).Position(X); 
      
    return The;
  end Model_With;
  
  --==========================================================================
  
  procedure Render(The_Model : in out Timezone_Models) is
    
    Alpha   : constant Gl.Glfloat := Alpha_Of(The_Model.Clock); 
    Is_Init : constant Boolean    := Initialized;                                       
    --------------------------------------------------------------------------
    procedure Set(Clock    : in out Model;
                  Timezone : in out Timezone_Info) is
    begin
      if not Is_Init then
        Set_Alpha(The => Timezone.Location, To => Alpha);               
        Check_List(The  => Timezone.Location, 
                   List => Timezone.Location_List);   
      end if;             
                       
      Clock.Time_Offset   := Timezone.Time_Offset;
      Clock.Position      := Timezone.Position;
      Clock.Location_List := Timezone.Location_List;
      Clock.Latitude      := Timezone.Latitude;
      Clock.Longitude     := Timezone.Longitude;
    end Set;
    --------------------------------------------------------------------------
  begin
    if The_Model.Timezones /= null then
      The_Model.Position       := Timezones_Position;
      The_Model.Clock.Now      := Ada.Calendar.Clock;
      The_Model.Clock.Index    := 1;   
      The_Model.Clock.Alarm.On := False;
      
      Models.Initialize_Render(The_Model);
                  
      for Row in The_Model.Timezones'range(1) loop        
        for Col in The_Model.Timezones'range(2) loop
          Set(The_Model.Clock, Timezone => The_Model.Timezones(Row, Col));
          Render(The_Model.Clock);                            
          The_Model.Clock.Index := The_Model.Clock.Index + 1;   
        end loop;
      end loop;
      
      Render(The_Model.Clock.Alarm, Now => The_Model.Clock.Now);
      The_Model.Clock.Alarm.On := True;
        
      Models.Finalize_Render(The_Model);
    end if;  
  end Render;  
  
  --==========================================================================
  
  function Model_With
    (Position       : in Types.Vector_3d := (X => Screen.Width  / 2.0,
     	                                       Y => Screen.Height / 2.0,
                                             Z => 0.0);
     Attitude       : in Types.Vector_3d := (X => 0.0,
     	                                       Y => 0.0,
                                             Z => 0.0);
     Scale          : in Gl.Glfloat      := Screen.Height / 2.0;
     Dialface       : in Dialface_Type   := Hours_12;
     Earth_Scale    : in Gl.Glfloat      := 0.0;
     Number         : in Positive        := 4)
     return Timezone_Models is
  
    Alaska          : aliased String := "Alaska (Anchorage)";
    Los_Angeles     : aliased String := "Los Angeles";
    Arizona         : aliased String := "Arizona (Phoenix)";
    Denver          : aliased String := "Denver";
    Chicago         : aliased String := "Chicago";
    New_York        : aliased String := "New York";
    Santiago        : aliased String := "Santiago";
    Buenos_Aires    : aliased String := "Buenos Aires";
    Middle_Atlantic : aliased String := "Middle Atlantic (St. Paul)";
    Acores          : aliased String := "Acores";
    London          : aliased String := "London";
    Berlin          : aliased String := "Berlin";
    Kairo           : aliased String := "Kairo";
    Moskau          : aliased String := "Moskau";
    Baku            : aliased String := "Baku";
    Islamabad       : aliased String := "Islamabad";
    Novosibirsk     : aliased String := "Novosibirsk";
    Jakarta         : aliased String := "Jakarta";
    Beijing         : aliased String := "Beijing";
    Tokyo           : aliased String := "Tokyo";
    Sydney          : aliased String := "Sydney";
    Salomon_Islands : aliased String := "Salomon Islands";
    Auckland        : aliased String := "Auckland";
    Hawaii          : aliased String := "Hawaii";
    
    type Geographic_Location is record
      Latitude  : Gl.Glfloat := 0.0;
      Longitude : Gl.Glfloat := 0.0;
      Name      : Ada.Strings.Unbounded.String_Access;
    end record;  
    
    Location   : constant array(1 .. 24) of Geographic_Location :=
                  (1 => (Latitude  => 20.0,
                         Longitude => -155.0,
                         Name      => Hawaii'Unchecked_Access),
                   2 => (Latitude  => 61.25,
                         Longitude => -149.0,
                         Name      => Alaska'Unchecked_Access),
                   3 => (Latitude  => 34.0,
                         Longitude => -118.0,
                         Name      => Los_Angeles'Unchecked_Access),
                   4 => (Latitude  => 33.5,
                         Longitude => -112.5,
                         Name      => Arizona'Unchecked_Access),
                   5 => (Latitude  => 39.0,
                         Longitude => -105.0,
                         Name      => Denver'Unchecked_Access),
                   6 => (Latitude  => 42.0,
                         Longitude => -87.5,
                         Name      => Chicago'Unchecked_Access),
                   7 => (Latitude  => 41.0,
                         Longitude => -74.0,
                         Name      => New_York'Unchecked_Access),
                   8 => (Latitude  => -33.0,
                         Longitude => -72.0,
                         Name      => Santiago'Unchecked_Access),
                   9 => (Latitude  => -35.0,
                         Longitude => -58.0,
                         Name      => Buenos_Aires'Unchecked_Access),
                  10 => (Latitude  => 1.0,
                         Longitude => -29.0,
                         Name      => Middle_Atlantic'Unchecked_Access),
                  11 => (Latitude  => 38.0,
                         Longitude => -28.0,
                         Name      => Acores'Unchecked_Access),
                  12 => (Latitude  => 51.5,
                         Longitude => 0.0,
                         Name      => London'Unchecked_Access),
                  13 => (Latitude  => 52.5,
                         Longitude => 13.3,
                         Name      => Berlin'Unchecked_Access),
                  14 => (Latitude  => 30.0,
                         Longitude => 31.5,
                         Name      => Kairo'Unchecked_Access),
                  15 => (Latitude  => 55.8,
                         Longitude => 37.5,
                         Name      => Moskau'Unchecked_Access),
                  16 => (Latitude  => 40.5,
                         Longitude => 50.0,
                         Name      => Baku'Unchecked_Access),
                  17 => (Latitude  => 34.0,
                         Longitude => 73.0,
                         Name      => Islamabad'Unchecked_Access),
                  18 => (Latitude  => 55.0,
                         Longitude => 83.0,
                         Name      => Novosibirsk'Unchecked_Access),
                  19 => (Latitude  => -6.0,
                         Longitude => 107.0,
                         Name      => Jakarta'Unchecked_Access),
                  20 => (Latitude  => 40.0,
                         Longitude => 117.0,
                         Name      => Beijing'Unchecked_Access),
                  21 => (Latitude  => 36.0,
                         Longitude => 140.0,
                         Name      => Tokyo'Unchecked_Access),
                  22 => (Latitude  => -34.0,
                         Longitude => 151.5,
                         Name      => Sydney'Unchecked_Access),
                  23 => (Latitude  => -12.0,
                         Longitude => 162.0,
                         Name      => Salomon_Islands'Unchecked_Access),
                  24 => (Latitude  => -37.0,
                         Longitude => 175.0,
                         Name      => Auckland'Unchecked_Access));
                  
    type Rows_Cols_Type is record
      Rows, Cols : Positive := 1;
    end record;
    
    type Location_Info is record
      Index : Positive := 1; -- for location array
      X, Y  : Integer  := 0; -- on screen * R
    end record;

    type Location_Info_Array is array(Positive range <>) of Location_Info;
    --------------------------------------------------------------------------    
    function The_Rows_Cols return Rows_Cols_Type is
    begin
      case Number is
        when  1       => return (Rows => 1, Cols => 1);
        when  2       => return (Rows => 1, Cols => 2);
        when  3 ..  4 => return (Rows => 2, Cols => 2);       
        when  5 ..  6 => return (Rows => 2, Cols => 3);
        when  7 .. 12 => return (Rows => 3, Cols => 4);
        when 13 .. 18 => return (Rows => 3, Cols => 6);
        when 19 .. 24 => return (Rows => 4, Cols => 6);
        when others   => return (Rows => 4, Cols => 6);        
      end case;        
    end The_Rows_Cols;
    --------------------------------------------------------------------------                  
    Rows_Cols : constant Rows_Cols_Type := The_Rows_Cols;
    Radius    : constant Gl.Glfloat := Scale / (Screen.Height / 2.0) * 
                Gl.Glfloat'Min(Screen.Width  / Gl.Glfloat(Rows_Cols.Cols),                                             
                               Screen.Height / Gl.Glfloat(Rows_Cols.Rows)) / 2.0;                  
    Timezones : Timezone_Infos_Access;
    Index     : Natural := 0;
    H_Offset  : Integer := 0;
    Clock_N   : constant Positive := Rows_Cols.Rows * Rows_Cols.Cols;
    Clock     : Model := Model_With(Position       => (0.0, 0.0, 0.0),
                                    Attitude       => Attitude,
                                    Scale          => Radius,
                                    Dialface       => Dialface,
                                    Time_Offset    => 0.0,
                                    Earth_Scale    => Earth_Scale,
                                    Earth_Location => "",
                                    Sunlight       => True);
    Alpha     : constant Gl.Glfloat := Alpha_Of(Clock);                                    
    --------------------------------------------------------------------------                  
    procedure Select_Special(Info : in Location_Info_Array) is
    begin
      for I in Info'range loop
        Timezones(1, I) := 
            (Latitude  => Location(Info(I).Index).Latitude,
             Longitude => Location(Info(I).Index).Longitude,
             Location  => The_Text(Text     => Location(Info(I).Index).Name.all,
                                   Position => (X => 0.0, Y => 0.95, Z => 0.0),
                                   Scale    => Radius),
             Location_List => 0,                        
             Time_Offset   => 3600.0 * Duration(Info(I).Index - 13),
             Position      => (X => Gl.Glfloat(Info(I).X),
                               Y => Gl.Glfloat(Info(I).Y), 
                               Z => 0.0) * Radius);                                          
                               
        Set_Alpha(The => Timezones(1, I).Location, To => Alpha);
          
        Check_List(The  => Timezones(1, I).Location, 
                   List => Timezones(1, I).Location_List);   
      end loop;    
    end Select_Special;         
    --------------------------------------------------------------------------                  
  begin        
    Timezones_Position   := Position;
    Clock.Alarm.Scale    := Scale;
    Clock.Alarm.On       := True;  
    Clock.Calculate_Time := False;
    
    if Number = Clock_N then
      Timezones := new Timezone_Infos(1 .. Rows_Cols.Rows, 1 .. Rows_Cols.Cols);
    else     
      Timezones := new Timezone_Infos(1 .. 1, 1 .. Number);
    end if;
    
    if Number = 1 and then Command_Line.Earth_Location /= "" then
      Timezones(1, 1) := 
        (Latitude      => Gl.Glfloat(Command_Line.Latitude),
         Longitude     => Gl.Glfloat(Command_Line.Longitude),
         Location      => The_Text(Text     => Command_Line.Earth_Location,
                                   Position => (X => 0.0, Y => 0.95, Z => 0.0),
                                   Scale    => Radius),
         Location_List => 0,                        
         Time_Offset   => 0.0,
         Position      => Position);
         
      Set_Alpha(The => Timezones(1, 1).Location, To => Alpha);
      
      Check_List(The  => Timezones(1, 1).Location, 
                 List => Timezones(1, 1).Location_List);   
    elsif Number = 5 then
      Select_Special(Info => (1 => (Index => 7,  -- New York
                                    X     => -2, 
                                    Y     => 1),
                              2 => (Index => 12, -- London
                                    X     => 0, 
                                    Y     => 1),
                              3 => (Index => 21, -- Tokyo
                                    X     => 2, 
                                    Y     => 1),
                              4 => (Index => 14, -- Kairo
                                    X     => -1, 
                                    Y     => -1),
                              5 => (Index => 22, -- Sidney
                                    X     => 1, 
                                    Y     => -1)));             
    else  
      for Row in Timezones'range(1) loop                             
        for Col in Timezones'range(2) loop
          H_Offset := (Index - Clock_N / 2) * (24 / Clock_N);   
          Index    := Index + 1;
  
          Timezones(Row, Col) := 
            (Latitude  => Location(H_Offset + 13).Latitude,
             Longitude => Location(H_Offset + 13).Longitude,
             Location  => The_Text(Text     => Location(H_Offset + 13).Name.all,
                                   Position => (X => 0.0, Y => 0.95, Z => 0.0),
                                   Scale    => Radius),
             Location_List => 0,                        
             Time_Offset   => 3600.0 * Duration(H_Offset),
             Position      => (X =>   Gl.Glfloat(2 * Col - Rows_Cols.Cols - 1),
                               Y => - Gl.Glfloat(2 * Row - Rows_Cols.Rows - 1), 
                               Z => 0.0) * Radius);                                          
                               
          Set_Alpha(The => Timezones(Row, Col).Location, To => Alpha);
          
          Check_List(The  => Timezones(Row, Col).Location, 
                     List => Timezones(Row, Col).Location_List);   
        end loop;
      end loop;
    end if;    
    
    case Clock_N is
      when  2 => Clock.Bitmaps(1).Position := ( 65.0,  101.0, 0.0);
      when  4 => Clock.Bitmaps(1).Position := (210.0,  -80.0, 0.0);      
      when  6 => Clock.Bitmaps(1).Position := (280.0,   60.0, 0.0);
      when 12 => Clock.Bitmaps(1).Position := (120.0,  -50.0, 0.0);
      when 24 => Clock.Bitmaps(1).Position := (310.0,   30.0, 0.0);
      when others => null;
    end case;
    if Clock_N >= 6 then
      Clock.Bitmaps(1).Size_Scaling := Clock.Bitmaps(1).Size_Scaling / 2.0;
    end if;  

    return Timezone_Models'(Ada.Finalization.Controlled with
                            Size_Scaling => Scale,
                            Color        => Colors.White,
                            Position     => Position,
                            Attitude     => Attitude,
                            Clock        => Clock,
                            Timezones    => Timezones);                                   
 
  end Model_With;         
  
  --==========================================================================
  
end Opengl.Clock;
