--                    -*- Mode: Ada -*-
-- Filename        : opengl-clock.ads
-- Description     : Types and subprograms for an analog clock in OpenGL.
-- Author          : J. Schrer
-- Created On      : 9. Apr. 1998
-- Last Modified By:
-- Last Modified On: 
-- Update Count    : 
-- Status          :

with Ada.Calendar;

with Opengl.Models;
with Opengl.Models.Strings;
with Opengl.Models.Forms;
with Opengl.Models.Dialfaces;

with Opengl.Gl;
with Opengl.Screen;
with Opengl.Types;
with Opengl.Display_Lists;
with Opengl.Bitmaps;

with Utilities.Calendar_Utilities;

package Opengl.Clock is

  -- A simple clock ----------------------------------------------------------
  
  type Model is new Models.Model_3d with private;


  procedure Render(The_Model : in out Model);

  use type Gl.Glfloat;

  type Dialface_Type is (Hours_12, Hours_24);
  

  function Model_With
    (Position       : in Types.Vector_3d := (X => Screen.Width  / 2.0,
     	                                       Y => Screen.Height / 2.0,
                                             Z => 0.0);
     Attitude       : in Types.Vector_3d := (X => 0.0,
     	                                       Y => 0.0,
                                             Z => 0.0);
     Scale          : in Gl.Glfloat      := Screen.Height / 2.0;
     Dialface       : in Dialface_Type   := Hours_12;
     Time_Offset    : in Duration        := 0.0;
     Earth_Scale    : in Gl.Glfloat      := 0.0;
     Earth_Location : in String          := "";
     Latitude       : in Gl.Glfloat      := 0.0; -- of geographic location
     Longitude      : in Gl.Glfloat      := 0.0; -- of geographic location
     Sunlight       : in Boolean         := False)
     return Model;
     
  -- If Dialface = Hours_12 a normal 12 H dialface is drawn.   
  -- If Dialface = Hours_24 a dialface with 24 H is drawn.   
     
  -- An array of clocks with different timezone ------------------------------

  type Timezone_Models is new Models.Model_3d with private;

  procedure Render(The_Model : in out Timezone_Models);
  
  function Model_With
    (Position       : in Types.Vector_3d := (X => Screen.Width  / 2.0,
     	                                       Y => Screen.Height / 2.0,
                                             Z => 0.0);
     Attitude       : in Types.Vector_3d := (X => 0.0,
     	                                       Y => 0.0,
                                             Z => 0.0);
     Scale          : in Gl.Glfloat      := Screen.Height / 2.0;
     Dialface       : in Dialface_Type   := Hours_12;
     Earth_Scale    : in Gl.Glfloat      := 0.0;
     Number         : in Positive        := 4)
     return Timezone_Models;
     
  -- Generates Number clocks with different timezones.   

private ----------------------------------------------------------------------    

  -- Some time objects -------------------------------------------------------

  Ada_Now    : constant Ada.Calendar.Time := Ada.Calendar.Clock;
  Now        : constant Utilities.Calendar_Utilities.Time :=
                        Utilities.Calendar_Utilities.Time_Of(Ada_Now);
  Full_Image : constant String := Utilities.Calendar_Utilities.Time_Image_Of(Now);
  Date_Image : constant String := Utilities.Calendar_Utilities.Date_Image_Of(Now);
  Time_Image : constant String := Full_Image(Full_Image'First .. Full_Image'Last - 3);
  Time_First : constant Ada.Calendar.Time :=
                        Ada.Calendar.Time_Of(Year  => Ada.Calendar.Year_Number'First,
                                             Month => Ada.Calendar.Month_Number'First,
                                             Day   => Ada.Calendar.Day_Number'First);

  -- Bitmaps to fill parts of the empty space --------------------------------
                                             
  type Bitmaps_Array is array(Positive range <>) of Bitmaps.Model;
  
  -- Texts -------------------------------------------------------------------
    
  type Text_Type is record
    Background       : Models.Forms.Rectangle_And_Frame_2d;
    String           : Models.Strings.String_2d_Access;
  end record; 
  
  -- Alarm messages ----------------------------------------------------------
  
  type Color_Index_Type is mod 3;
  
  type Message is record
    Text             : Text_Type;                           -- to show
    Time             : Ada.Calendar.Time := Ada_Now; -- text is switched on
    Offset_Sec       : Gl.Glfloat        := Gl.Glfloat'Last; -- dt for repeat 
    Times            : Positive          := 1;               -- repeat 
    Count            : Natural           := 0;               -- for times
    On               : Boolean           := True;            -- for blinking
    Color_Index      : Color_Index_Type  := 0;        
  end record;
  
  type Message_Array is array(Natural range <>) of Message;
  type Message_Array_Access is access all Message_Array;  
  
  type Alarm_Type is record
    Scale            : Gl.Glfloat        := Screen.Height / 2.0;
    Time_Stamp       : Ada.Calendar.Time := Time_First;
    Messages         : Message_Array_Access;
    On               : Boolean           := True;
    Window_Width     : gl.glfloat        := Screen.Width;
  end record;
      
  -- Dialface and clock parts ------------------------------------------------
  
  type Render_Mode_Type is (Render_Both, Render_Clock, Render_Earth);
  
  type Clock_Dialface_Type(Tick_Number : Positive := 12) is record
    H_Digits         : Models.Dialfaces.Ticks_With_Digits(Number => Tick_Number);
    Big_Ticks        : Models.Dialfaces.Ticks            (Number => Tick_Number);
    Small_Ticks      : Models.Dialfaces.Ticks            (Number => 4);
    Outer_Disk       : Models.Forms.Disk_2d_Z;
    Name             : Text_Type;
    Render_Mode      : Render_Mode_Type := Render_Both;    
  end record;
  
  type Clock_Dialface_Access is access Clock_Dialface_Type;
  
  type Clock_Parts_Type is record
    Background       : Models.Forms.Disk_2d_Z;    
    Dialface         : Clock_Dialface_Access;
    Hour_Pointer     : Models.Forms.Radial_Arrow_2d_Z;
    Minute_Pointer   : Models.Forms.Radial_Arrow_2d_Z;
    Second_Pointer   : Models.Forms.Radial_Arrow_2d_Z;
    Center_Disk      : Models.Forms.Disk_2d_Z;
  end record;  

  type Clock_Part_Lists_Type is record
    Background       : Gl.Gluint := Display_Lists.Next_List_Id;
    Dialface         : Gl.Gluint := Display_Lists.Next_List_Id;
    Hour_Pointer     : Models.Display_List_Model_2d_Z;
    Minute_Pointer   : Models.Display_List_Model_2d_Z;
    Second_Pointer   : Models.Display_List_Model_2d_Z;
    Center_Disk      : Gl.Gluint := Display_Lists.Next_List_Id;
  end record;  

  -- Clock model -------------------------------------------------------------  
  
  type Model is new Models.Model_3d with record
    Clock_Parts      : Clock_Parts_Type;    
    Clock_Part_Lists : Clock_Part_Lists_Type;    
    
    Digital_Date     : Models.Strings.String_2d(Length => Date_Image'Length);
    Digital_Time     : Models.Strings.String_2d(Length => Time_Image'Length);
    
    Now              : Ada.Calendar.Time := Ada_Now;    
    Time_Offset      : Duration          := 0.0;
    Calculate_Time   : Boolean           := True;
    Index            : Positive          := 1;

    Location         : Text_Type;
    Location_List    : Gl.Gluint         := Display_Lists.Next_List_Id;
    Latitude         : Gl.Glfloat      := 0.0; -- of geographic location
    Longitude        : Gl.Glfloat      := 0.0; -- of geographic location
    Earth            : Models.Root_Model_Access;
    Bitmaps          : Bitmaps_Array(1 .. 4);

    Alarm            : Alarm_Type;
  end record;

  -- Timezone clocks ---------------------------------------------------------
    
  type Timezone_Info is record
    Latitude         : Gl.Glfloat      := 0.0; -- of geographic location
    Longitude        : Gl.Glfloat      := 0.0; -- of geographic location
    Location         : Text_Type;              -- Name of geographic location
    Location_List    : Gl.Gluint       := Display_Lists.Next_List_Id;
    Time_Offset      : Duration        := 0.0; -- of geographic location 
    Position         : Types.Vector_3d := (0.0, 0.0, 0.0); -- on screen
  end record;  
    
  type Timezone_Infos is array(Positive range <>, 
                               Positive range <>) of Timezone_Info;
                                
  type Timezone_Infos_Access is access all Timezone_Infos;
  
  type Timezone_Models is new Models.Model_3d with record
    Clock            : Model;
    Timezones        : Timezone_Infos_Access;
  end record;
  
end Opengl.Clock;
