--                    -*- Mode: Ada -*-
-- Filename        : opengl-earth.adb
-- Description     : 
-- Author          : J. Schrer
-- Created On      : 
-- Last Modified By: 
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with System;
with Ada.Numerics;
with Ada.Calendar;
with Ada.Exceptions;
with Ada.Strings.Fixed;
with Interfaces.C;
with Utilities.Calendar_Utilities;
with Opengl.Command_Line;
with Opengl.Display_Lists;
with Opengl.Glu;
with Opengl.Colors;
with Opengl.Fonts;
with Opengl.Models.Strings;
with Opengl.Utils;
with Opengl.Float_32.Elementary_Functions;
with Opengl.Float_32.Arrays.Types;
with Opengl.Float_32.Arrays.Vector_Operations;
with Opengl.Timer;
with Opengl.Logging;
with Opengl.Win;
with Windows.Utils;
with Win32.Winuser;
with Win32.Windef;

package body Opengl.Earth is

  use Opengl.Float_32.Elementary_Functions;
  use Opengl.Float_32.Arrays.Vector_Operations;
  use type Opengl.Float_32.Arrays.Types.Vector;
  use type Interfaces.Unsigned_8;
  use type Gl.Glint;
  use type Gl.Gluint;
  use type Gl.Gldouble;
  use type Types.Integer_Array;
  use type Ada.Calendar.Time;
  use type Models.Root_Model_Access;
  use type Win.Ada_Called_Windows_Callback;
  
  package Calendar_Utilities renames Utilities.Calendar_Utilities;
                          
  -- For debugging -----------------------------------------------------------
  Current_Axis               : Types.Axis_Type := X;                              
  Offset                     : Types.Vector_3d := (0.0, 0.0, 0.0);
  -- End for debugging -------------------------------------------------------
  
  Render_Lists               : Boolean          := True;    
  Font_Deviation             : Gl.Glfloat       := 1.0;
  Font_Extrusion             : Gl.Glfloat       := 0.0;
  
  Resize_Callback            : Win.Ada_Called_Windows_Callback;
  Keypressed_Callback        : Win.Ada_Called_Windows_Callback;
  Middle_Mouse_Down_Callback : Win.Ada_Called_Windows_Callback;
    
  The_Model                  : Models.Root_Model_Access;
  
  --==========================================================================             
  
  function Time_Offset_Of(The_Model : in Model) return Duration is
  begin
    return The_Model.Time_Offset;
  end Time_Offset_Of;
                          
  --==========================================================================             
  
  procedure Set_Time_Offset(The_Model : in out Model;
                            To        : in     Duration) is
  begin
    The_Model.Time_Offset := To;
  end Set_Time_Offset;
     
  --==========================================================================             
     
  procedure Set_Location(The_Model : in out Model;
                         Latitude, 
                         Longitude : in     Gl.Glfloat) is
  begin
    if not The_Model.Animated_Latlon then
      The_Model.Latitude  := Latitude;                           
      The_Model.Longitude := Longitude;    
    end if;                       
  end Set_Location;
    
  --==========================================================================             
     
  function Position_Of(Latitude  : in Gl.Glfloat;
                       Longitude : in Gl.Glfloat;
                       Radius    : in Gl.Glfloat) return Types.Vector_3d is

    Sin_Lat : constant Gl.Glfloat := Sin(Latitude);
    Cos_Lat : constant Gl.Glfloat := Cos(Latitude);
    Sin_Lon : constant Gl.Glfloat := Sin(Longitude);
    Cos_Lon : constant Gl.Glfloat := Cos(Longitude);
  begin
    return Radius * (X =>   Cos_Lat * Sin_Lon,
                     Y => - Cos_Lat * Cos_Lon,
                     Z =>   Sin_Lat);
  end Position_Of;
                          
  --==========================================================================             
  
  function  Airports(Earth_Radius : in Gl.Glfloat) return Object_Array_Access;
  
  procedure Smooth_Edge(Pixels : in out Bitmaps.Pixel_Array);
                          
  --==========================================================================             
  
  The_Filter : Bitmaps.Filter_Type := 
    (Bitmaps.Red   => (First => 0, Last => 25, Bias => 0),
     Bitmaps.Green => (First => 0, Last => 40, Bias => 0),
     Bitmaps.Blue  => (First => 0, Last => 80, Bias => 30));
     
  The_Factor : array(Bitmaps.Color_Index_Type) of Gl.Glfloat := 
    (Bitmaps.Red => 1.0, Bitmaps.Green => 1.0, Bitmaps.Blue => 2.0);

  type Parameter_Type is (Filter, Light_Model, Light_Source, Object, 
                          Shininess, Spot_Exponent, Spot_Cutoff, 
                          Constant_Attenuation, Linear_Attenuation, 
                          Quadratic_Attenuation, Sun_Distance,
                          Deviation, Extrusion);
                          
  type Filter_Parameter_Type is 
    (Color_First, Color_Last, Color_Bias, Color_Factor);
  
  type Light_Type is (Ambient, Diffuse, Specular, Emission);  
  
  subtype Color_Index_Type is Positive range 1 .. 4;                        
  
  type Texture_Env_Parameter_Type is (Decal, Replace, Modulate, Blend);
    
  The_Parameter         : Parameter_Type             := Filter;       
  The_Filter_Parameter  : Filter_Parameter_Type      := Color_Last;       
  The_Color_Component   : Color_Index_Type           := Bitmaps.Blue;
  The_Light             : Light_Type                 := Ambient;
  Texture_Env_Parameter : Texture_Env_Parameter_Type := Modulate;
  
  --==========================================================================             
  
  procedure Filter(The_Pixel : in out Bitmaps.Pixel) is
    use Bitmaps;
  begin
    if The_Pixel(Red)   in The_Filter(Red).  First .. 
                           The_Filter(Red).  Last and then
       The_Pixel(Green) in The_Filter(Green).First .. 
                           The_Filter(Green).Last and then
       The_Pixel(Blue)  in The_Filter(Blue). First .. 
                           The_Filter(Blue). Last then

      The_Pixel(Red)   := Byte(Gl.Glfloat(The_Pixel(Red))   * The_Factor(Red))   +
                          The_Filter(Red).Bias;
      The_Pixel(Green) := Byte(Gl.Glfloat(The_Pixel(Green)) * The_Factor(Green)) +
                          The_Filter(Green).Bias;
      The_Pixel(Blue)  := Byte(Gl.Glfloat(The_Pixel(Blue))  * The_Factor(Blue))  +
                          The_Filter(Blue).Bias;          
    end if;                                                          
  end Filter;  
                          
  --==========================================================================             
  
  function Earth_Bitmap(Filename  : in String;
                        The_Model : in Models.Root_Model_Access) 
           return Bitmaps.Bitmap_Type is
             
    Earth : Bitmaps.Bitmap_Type;      
  begin
    if The_Model = null then
      Bitmaps.Read(Earth, Filename);
      Bitmaps.Generate_Pixels(Earth);
      Bitmaps.Filter(The_Bitmap => Earth, 
                     The_Filter => Filter'access);
--      Smooth_Edge(Earth.Pixels.all);                   
    else
      Earth := Model(The_Model.all).Earth_Bitmap; 
    end if;    
    return Earth;
  end Earth_Bitmap;
                          
  --==========================================================================             
  
  type Earth_Angles_Type is record
    Phi_Since_21_6     : Gl.Glfloat := 0.0;
    Phi_Since_Midnight : Gl.Glfloat := 0.0;
    Time_Zone_Angle    : Gl.Glfloat := 0.0;    
  end record;
      
  function Earth_Angles(Now : in Ada.Calendar.Time) return Earth_Angles_Type is
                 
    Year               : Ada.Calendar.Year_Number;
    Month              : Ada.Calendar.Month_Number;
    Day                : Ada.Calendar.Day_Number;
    Sec                : Ada.Calendar.Day_Duration;
    Sec_Since_21_6     : Gl.Glfloat;
    Year_Day_Of_Now    : Integer;
    Year_Day_Of_21_6   : Integer;
    Days_In_This_Year  : Integer;
  begin
    Ada.Calendar.Split (Date    => Now,
                        Year    => Year,
                        Month   => Month,
                        Day     => Day,
                        Seconds => Sec);
    Year_Day_Of_Now    := Calendar_Utilities.Year_Day_Of
                          (The_Year => Year, The_Month => Month, The_Day => Day);
    Year_Day_Of_21_6   := Calendar_Utilities.Year_Day_Of
                          (The_Year => Year, The_Month => 6, The_Day => 21);
    Days_In_This_Year  := Calendar_Utilities.Days_In(Year);                                                                 
    Sec_Since_21_6     := Gl.Glfloat(Year_Day_Of_Now - Year_Day_Of_21_6) * 
                          86400.0 + Gl.Glfloat(Sec);                             
                          
    return (Phi_Since_21_6     => Sec_Since_21_6 / Gl.Glfloat(Days_In_This_Year) / 
                                  86400.0 * 360.0,
            Phi_Since_Midnight => Gl.Glfloat(Sec) / 86400.0 * 360.0,
            Time_Zone_Angle    => 360.0 / 24.0 / 60.0 * Gl.Glfloat
                                  (Windows.Utils.Time_Zone_Bias_Minutes));
  end Earth_Angles;                               
                          
  --==========================================================================             
  
  Axis_Inclination : constant := 23.0 + 27.0 / 60.0;  
  
  function Earth_Attitude(Angles : in Earth_Angles_Type) return Types.Vector_3d is
  begin  
    return (X => - 90.0 + Axis_Inclination,                                               
            Y => - Angles.Phi_Since_21_6,
            Z => Angles.Phi_Since_Midnight + 
                 Angles.Phi_Since_21_6     + 
                 Angles.Time_Zone_Angle);                          
  end Earth_Attitude;
                          
  --==========================================================================             

  function Earth_Attitude(Now : in Ada.Calendar.Time) return Types.Vector_3d is
  begin  
    return Earth_Attitude(Earth_Angles(Now));
  end Earth_Attitude;
                          
  --==========================================================================

  procedure Render(The_Model : in out Model) is
    --------------------------------------------------------------------------
    procedure Render_Objects is
    begin  
      if The_Model.Objects /= null then
  --      Models.Motion_Sequence := Models.Rotate_First;
  --      Models.Rotation_1 := Z;
  --      Models.Rotation_2 := Y;
  --      Models.Rotation_3 := X;      
        for I in The_Model.Objects'range loop
          Models.Render(The_Model.Objects(I).Symbol);
        end loop;
  
        for I in The_Model.Objects'range loop
  	      if The_Model.Objects(I).Selected then
  	        Gl.Glpushmatrix;
  	--        Models.Rotate(The_Model.Objects(i).Name);  
  	        Models.Translate(The_Model.Objects(I).Name);
  	        Models.Rotate_Z(- The_Model.Attitude(Z));
  	        Models.Rotate_Y(- The_Model.Attitude(Y));
  	        Models.Rotate_X(- The_Model.Attitude(X));
  	        Gl.Glcalllist(The_Model.Objects(I).Name.Display_List);
  	        Gl.Glpopmatrix;
          end if;
        end loop;
  	--      Models.Rotation_1 := X;
  	--      Models.Rotation_2 := Y;
  	--      Models.Rotation_3 := Z;      
  	--      Models.Motion_Sequence := Models.Translate_First;
      end if;
    end Render_Objects;
    --------------------------------------------------------------------------
    procedure Render_Net is
    
      Number_Color : constant Types.Color_Type := Colors.Yellow;
      Number_Font  : constant Fonts.Font_Type  :=
                     Fonts.Font(Deviation => Font_Deviation,
                                Extrusion => Font_Extrusion);
      ------------------------------------------------------------------------
      procedure Generate(Numbers     : in Types.Integer_Array;
                         Axis        : in Types.Axis_Type;
                         Axis_Bias   : in Gl.Glfloat;
                         Start_Angle : in Gl.Glfloat) is
      begin
        for I in Numbers'range loop
          declare
            Image : constant String := 
                    Ada.Strings.Fixed.Trim(Integer'Image(Numbers(I)), 
                                           Side => Ada.Strings.Both);

            R     : Gl.Glfloat renames The_Model.Net_Sphere.Radius;                         
            Phi   : constant Gl.Glfloat := Start_Angle + Gl.Glfloat(I - 1) * 10.0;
            
            The_String : Models.Strings.String_3d :=
              (Models.Root_Model with
         		   Length       => Image'Length,
         		   Size_Scaling => 1.0,
         		   Color        => Number_Color,
         		   Position     => (X => 0.0, Y => 0.0, Z => R),
         		   Attitude     => (X => 0.0, Y => Phi, Z => 0.0),
         		   Text         => Image,
               Font         => Number_Font,
         		   Font_Size    => R * 0.05,
         		   Alignment    => (Horizontal => Models.Strings.Center,
         		                    Vertical   => Models.Strings.Middle),
         		   Hidden_Data  => Models.Strings.Default_Hidden_Data);
          begin
            The_String.Attitude(Axis) := Axis_Bias;
            Models.Strings.Initialize(The_String);            
            Models.Strings.Render(The_String);                                   
          end;
        end loop;
      end Generate;  
      ------------------------------------------------------------------------
      Longitude_Numbers : Types.Integer_Array :=
                          Utils.Number_Array_Of(First     => 180, 
                                                Increment => -10, 
                                                Number    => 36);
      Latitude_Numbers  : constant Types.Integer_Array :=
                          Utils.Number_Array_Of(First     => -90, 
                                                Increment => 10, 
                                                Number    => 19);
      Lon_Start_Angle   : Gl.Glfloat := 0.0;                                                   
    begin
      Glu.Gluquadricdrawstyle
        (Quadobject => The_Model.Net_Sphere.Glu_Quadric_Object,
     	   Drawstyle  => Glu.Glu_Line);
      Models.Forms.Render(The_Model.Net_Sphere);
      
      if The_Model.Sunlight then
        Lon_Start_Angle            := The_Model.Net_Sphere.Attitude(Z);
        Longitude_Numbers(2 .. 36) := - Longitude_Numbers(2 .. 36);
      else
        Lon_Start_Angle            := - The_Model.Net_Sphere.Attitude(Z);
      end if;
      
      Models.Rotate_X(-90.0);      
      Generate(Numbers     => Longitude_Numbers,
               Axis        => X,
               Axis_Bias   => 180.0,
               Start_Angle => Lon_Start_Angle);
               
      Models.Rotate_X(90.0);      
      Models.Rotate_Z(-90.0 + The_Model.Net_Sphere.Attitude(Z));      
      Generate(Numbers     => Latitude_Numbers(1 .. 9),
               Axis        => Z,
               Axis_Bias   => -90.0,
               Start_Angle => -180.0);
      Generate(Numbers     => Latitude_Numbers(11 .. 19),
               Axis        => Z,
               Axis_Bias   => -90.0,
               Start_Angle => -180.0);
               
      Models.Rotate_Z(180.0);         
      Generate(Numbers     => Latitude_Numbers(2 .. 9),
               Axis        => Z,
               Axis_Bias   => -90.0,
               Start_Angle => -180.0);
      Generate(Numbers     => Latitude_Numbers(11 .. 18),
               Axis        => Z,
               Axis_Bias   => -90.0,
               Start_Angle => -180.0);
    end Render_Net;
    --------------------------------------------------------------------------
    procedure Render_Solid is
      
      Tex_Env_Params : constant array(Texture_Env_Parameter_Type) of Gl.Glint :=
                        (Decal    => Gl.Gl_Decal,
                         Replace  => Gl.Gl_Replace, 
                         Modulate => Gl.Gl_Modulate, 
                         Blend    => Gl.Gl_Blend);
    begin
      Gl.Gltexparameteri(Target => Gl.Gl_Texture_2d,
                         Pname  => Gl.Gl_Texture_Mag_Filter,
                         Param  => Gl.Gl_Linear);
      Gl.Gltexparameteri(Target => Gl.Gl_Texture_2d,
                         Pname  => Gl.Gl_Texture_Min_Filter,
                         Param  => Gl.Gl_Linear);
                         
      Gl.Gltexparameteri(Target => Gl.Gl_Texture_2d,
                         Pname  => Gl.Gl_Texture_Wrap_S,
                         Param  => Gl.Gl_Repeat);
      Gl.Gltexparameteri(Target => Gl.Gl_Texture_2d,
                         Pname  => Gl.Gl_Texture_Wrap_T,
                         Param  => Gl.Gl_Repeat);
  
      Gl.Glteximage2d
          (Target     => Gl.Gl_Texture_2d,
           Level      => 0,
           Components => 3,
           Width      => Gl.Glsizei(The_Model.Earth_Bitmap.Pixels'Length(1)),
           Height     => Gl.Glsizei(The_Model.Earth_Bitmap.Pixels'Length(2)),
           Border     => 0,
           Format     => Gl.Gl_Rgb,
           C_Type     => Gl.Gl_Unsigned_Byte,
           Pixels     => The_Model.Earth_Bitmap.Pixels.all'Address);
  
      Gl.Glenable(Gl.Gl_Texture_2d);
        
--      if The_Model.Sunlight then
--        Texture_Env_Parameter := Modulate;
--      else               
--        Texture_Env_Parameter := Decal;
--      end if;
                               
      Gl.Gltexenvi(Target => Gl.Gl_Texture_Env,
                   Pname  => Gl.Gl_Texture_Env_Mode,
                   Param  => Tex_Env_Params(Texture_Env_Parameter));
             
      Glu.Gluquadrictexture
         (Quadobject    => The_Model.Solid_Sphere.Glu_Quadric_Object,
          Texturecoords => Gl.Gl_True);
           
      Glu.Gluquadricnormals
         (Quadobject    => The_Model.Solid_Sphere.Glu_Quadric_Object,
          Normals       => Glu.Glu_Smooth);
             
      Models.Forms.Render(The_Model.Solid_Sphere);
        
      Gl.Gldisable(Gl.Gl_Texture_2d);
    end Render_Solid;
    --------------------------------------------------------------------------
    procedure Render_Coordinate_System is
    begin  
      if The_Model.Render_C_sys then
        The_Model.Coord_System.Position := The_Model.Position;
        The_Model.Coord_System.Attitude := The_Model.Attitude;
        Coordinate_System.Render(The_Model.Coord_System);
      end if;        
    end Render_Coordinate_System;  
    --------------------------------------------------------------------------
    procedure Render_Net_Info is
    begin
      if The_Model.Render_Net then
        if The_Model.Sunlight then      
          Models.Lighting.Disable_Lighting;
        end if;
--        Clip_Planes.Disable;

        if Render_Lists then    
          Gl.Glcalllist(The_Model.Net_List);  
        else
          Render_Net;
        end if;  
        
--        Clip_Planes.Enable(The_Model.Z_Plane);    
        if The_Model.Sunlight then      
          Models.Lighting.Enable_Lighting;
        end if;  
      end if;
    end Render_Net_Info;  
    --------------------------------------------------------------------------
    procedure Render_Earth is
    begin  
      if Render_Lists then      
        if The_Model.Net_List = Invalid_List then
          Display_Lists.Generate(The_Model.Net_List);
          Gl.Glnewlist(List => The_Model.Net_List, Mode => Gl.Gl_Compile);
          Render_Net;
          Gl.Glendlist;
        end if;
            
        if The_Model.Solid_List = Invalid_List then
          Display_Lists.Generate(The_Model.Solid_List);
          Gl.Glnewlist(List => The_Model.Solid_List, Mode => Gl.Gl_Compile);
          Render_Solid; 
          Gl.Glendlist;
        end if;
      end if;

      Clip_Planes.Enable(The_Model.Z_Plane);    
      Models.Initialize_Render(The_Model);

      if not The_Model.Render_Bitmap then
        Models.Forms.Render(The_Model.Solid_Sphere);
      elsif Render_Lists then
        Gl.Glcalllist(The_Model.Solid_List);  
      else  
        Render_Solid;
      end if;  
      
      Render_Net_Info;
      Render_Objects;  
    
      Models.Finalize_Render(The_Model);      
      Clip_Planes.Disable;      
    end Render_Earth;  
    --------------------------------------------------------------------------
    Angles : Earth_Angles_Type;
    Angle  : Gl.Glfloat := 0.0;
    Now    : constant Ada.Calendar.Time := Ada.Calendar.Clock;
  begin
    if not The_Model.Animated_Time then
      The_Model.Time := Now;
    elsif The_Model.Time_Factor /= 1.0 then  
      The_Model.Time := The_Model.Time + Duration(Gl.Glfloat
                        (Now - The_Model.Render_Time) * The_Model.Time_Factor);
    end if;
    
    The_Model.Render_Time := Now;
    Angles                := Earth_Angles(The_Model.Time);
    The_Model.Time        := The_Model.Time + The_Model.Time_Offset;

    if The_Model.Sunlight then      
      Angle := Axis_Inclination * Cos(Angles.Phi_Since_21_6, Cycle => 360.0); 
      
      The_Model.Sun.Position(Y) := The_Model.Sun_Distance * 
                                   Sin(Angle, Cycle => 360.0);       
      The_Model.Sun.Position(Z) := The_Model.Sun_Distance * 
                                   Cos(Angle, Cycle => 360.0);       
      
      The_Model.Sun.Attitude := 
        (X => The_Model.Latitude, 
         Y => 180.0 + 30.0
              - The_Model.Longitude
              - Angles.Phi_Since_Midnight, -- Phi since midday
         Z => 0.0);

      The_Model.Attitude := (X => -90.0 + The_Model.Latitude,
                             Y => 0.0,
                             Z => 184.0 - The_Model.Longitude);
    
      Models.Lighting.Init(The_Model.Lighting_Model);
      Models.Lighting.Init(The_Model.Sun);
      Models.Lighting.Init(The_Model);
      
      Models.Lighting.Enable_Lighting;
      Models.Lighting.Enable(The_Model.Sun);    
                
      Models.Lighting.Render(The_Model.Sun);                                 
      Render_Earth;    
      
      Models.Lighting.Disable(The_Model.Sun);
      Models.Lighting.Disable_Lighting;
      
      Render_Coordinate_System;
    else
      The_Model.Attitude := Earth_Attitude(Angles);  

      Models.Rotation_1 := Y;
      Models.Rotation_2 := X;
      Models.Rotation_3 := Z;      
    
      Render_Earth;    
      Render_Coordinate_System;
      
      Models.Rotation_1 := X;
    	Models.Rotation_2 := Y;
    	Models.Rotation_3 := Z;          
    end if;
          
    if Command_Line.Clock in 0 .. 1 then
      if The_Model.Sunlight then      
        Timer.Set(Calendar_Utilities.Image_Of(The_Model.Time) & ", PHI(Sun):" & 
                  Gl.Glfloat'Image(The_Model.Sun.Attitude(X))&
                  Gl.Glfloat'Image(The_Model.Sun.Attitude(Y))&
                  Gl.Glfloat'Image(The_Model.Sun.Attitude(Z)), 2, Append => False);
      else            
        Timer.Set(Calendar_Utilities.Image_Of(The_Model.Time) & ", PHI(Earth):" & 
                  Gl.Glfloat'Image(The_Model.Attitude(X))&
                  Gl.Glfloat'Image(The_Model.Attitude(Y))&
                  Gl.Glfloat'Image(The_Model.Attitude(Z)), 2, Append => False);
      end if;            
--      Timer.Set("Phi_21_6:" & Gl.Glfloat'Image(Angles.Phi_Since_21_6) &
--               ", Phi_0_0:" & Gl.Glfloat'Image(Angles.Phi_Since_Midnight), 
--               1, Append => False);
    end if;                                           
  end Render;
                          
  --==========================================================================             
  
  Step   : constant Gl.Glfloat := Screen.Height / 10.0;
  Factor : constant Gl.Glfloat := 1.1;
                            
  --==========================================================================             
  
  function Keypressed
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    The : Model renames Model(The_Model.all);         
             
    subtype Sign_Type is Integer range -1 .. 1;
    --------------------------------------------------------------------------
    procedure Change_Filter(Sign : in Sign_Type) is
      --------------------------------------------- 
      procedure Add(The : in out Bitmaps.Byte) is
      begin
        The := The + 1;
      end Add;
      --------------------------------------------- 
      procedure Subtract(The : in out Bitmaps.Byte) is
      begin
        The := The - 1;
      end Subtract;
      ---------------------------------------------              
    begin  
      case The_Filter_Parameter is                      
        when Color_First  => 
          case Sign is
            when -1 => Subtract(The_Filter(The_Color_Component).First);
            when  1 => Add(The_Filter(The_Color_Component).First);
            when  0 => null;           
          end case;  
        when Color_Last   =>
          case Sign is
            when -1 => Subtract(The_Filter(The_Color_Component).Last);
            when  1 => Add(The_Filter(The_Color_Component).Last);
            when  0 => null;           
          end case;  
        when Color_Bias   =>
          case Sign is
            when -1 => Subtract(The_Filter(The_Color_Component).Bias);
            when  1 => Add(The_Filter(The_Color_Component).Bias);
            when  0 => null;           
          end case;  
        when Color_Factor => 
          The_Factor(The_Color_Component) := The_Factor(The_Color_Component) + 
                                             Gl.Glfloat(Sign) / 100.0;
      end case;            
      
      Timer.Set(Win32.Wparam'Image(Wparam) & ", '" & Character'Val(Wparam) & "'", 
                3, Append => False);    
      Timer.Set("Red:" & Bitmaps.Byte'Image(The_Filter(Bitmaps.Red).First) & " .." & 
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Red).Last) & ", Green:" &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Green).First) & " .." &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Green).Last) & ", Blue:" &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Blue).First) & " .." &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Blue).Last) & ", Red*" &
                Gl.Glfloat'Image(The_Factor(Bitmaps.Red)) & ", Green*" &              
                Gl.Glfloat'Image(The_Factor(Bitmaps.Green)) & ", Blue*" &              
                Gl.Glfloat'Image(The_Factor(Bitmaps.Blue)) & ", Red+" &              
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Red).Bias) & ", Green+" &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Green).Bias) & ", Blue+" &
                Bitmaps.Byte'Image(The_Filter(Bitmaps.Blue).Bias), 1, 
                Append => False);
      Timer.Set(Filter_Parameter_Type'Image(The_Filter_Parameter) & " " &
                Bitmaps.Color_Index_Type'Image(The_Color_Component), 2, 
                Append => False);              
    end Change_Filter;  
    --------------------------------------------------------------------------
    function New_Value(Value : in Gl.Glfloat;
                       Sign  : in Sign_Type  := 1;
                       Inc   : in Gl.Glfloat := 0.1;
                       First : in Gl.Glfloat := 0.0;
                       Last  : in Gl.Glfloat := 1.0) return Gl.Glfloat is
    begin
      case Sign is
       when -1 => return Gl.Glfloat'Max(Value - Inc, First);                   
       when  1 => return Gl.Glfloat'Min(Value + Inc, Last);                   
       when  0 => return Value;
      end case;
    end New_Value;                         
    --------------------------------------------------------------------------
    procedure Change_Color(The_Color : in out Types.Color_Type;
                           Sign      : in     Sign_Type;
                           Name      : in     String) is
    begin
      case The_Color_Component is
        when Bitmaps.Red   => The_Color.Red   := New_Value(The_Color.Red,   Sign);
        when Bitmaps.Green => The_Color.Green := New_Value(The_Color.Green, Sign);
        when Bitmaps.Blue  => The_Color.Blue  := New_Value(The_Color.Blue,  Sign);
        when Bitmaps.Alpha => The_Color.Alpha := New_Value(The_Color.Alpha, Sign);
      end case;
                                
      Timer.Set(Name & ": Red:" & Gl.Glfloat'Image(The_Color.Red)   &
                     ", Green:" & Gl.Glfloat'Image(The_Color.Green) &
                     ", Blue:"  & Gl.Glfloat'Image(The_Color.Blue)  &
                     ", Alpha:" & Gl.Glfloat'Image(The_Color.Alpha), 1,
                   Append => False);                                  
    end Change_Color;        
    --------------------------------------------------------------------------
    procedure Change_Sun(Sign : in Sign_Type) is
    begin                                      
      case The_Light is              
        when Ambient  => Change_Color(The.Sun.Ambient_Color,  Sign, "Sun.Ambient");
        when Diffuse  => Change_Color(The.Sun.Diffuse_Color,  Sign, "Sun.Diffuse");
        when Specular => Change_Color(The.Sun.Specular_Color, Sign, "Sun.Specular");
        when Emission => null;                            
      end case;  
      The.Sun.Initialized := False;
    end Change_Sun;  
    --------------------------------------------------------------------------
    procedure Change_Earth(Sign : in Sign_Type) is
    begin                                      
      case The_Light is              
        when Ambient  => Change_Color(The.Ambient_Color,  Sign, "Earth.Ambient");
        when Diffuse  => Change_Color(The.Diffuse_Color,  Sign, "Earth.Diffuse");
        when Specular => Change_Color(The.Specular_Color, Sign, "Earth.Specular");
        when Emission => Change_Color(The.Emission_Color, Sign, "Earth.Emission");                            
      end case;  
      The.Initialized := False;
    end Change_Earth;  
    --------------------------------------------------------------------------
    procedure Set(Value : in Gl.Glfloat;
                  Name  : in String) is
    begin                
      Timer.Set(Name & ":" & Gl.Glfloat'Image(Value), 2, Append => False);
    end Set;  
    --------------------------------------------------------------------------
    procedure Change(Sign : in Sign_Type) is
      -----------------------------------------------------------------
      procedure Change(Value : in out Gl.Glfloat;
                       Name  : in     String;
                       Inc   : in     Gl.Glfloat := 1.0;
                       First : in     Gl.Glfloat := Gl.Glfloat'First;
                       Last  : in     Gl.Glfloat := Gl.Glfloat'Last) is
      begin
        Value := New_Value(Value, Sign, Inc, First, Last);
        Set(Value, Name);
      end Change;    
      -----------------------------------------------------------------
    begin  
      case The_Parameter is                      
        when Filter                => 
          Change_Filter(Sign);
        when Light_Model           => 
          Change_Color(The.Lighting_Model.Ambient_Color, Sign, "Light.Ambient");
          The.Lighting_Model.Initialized := False;
        when Light_Source          => 
          Change_Sun(Sign);                       
        when Object                =>
          Change_Earth(Sign);
        when Shininess             => 
          Change(The.Shininess, "Shininess", 0.2, 0.0, 128.0); 
          The.Initialized := False;
        when Spot_Exponent         => 
          Change(The.Sun.Spot_Exponent, "Spot_Exponent", 1.0, 0.0); 
          The.Sun.Initialized := False;
        when Spot_Cutoff           => 
          Change(The.Sun.Spot_Cutoff, "Spot_Cutoff", 1.0, 0.0, 180.0); 
          The.Sun.Initialized := False;
        when Constant_Attenuation  => 
          Change(The.Sun.Constant_Attenuation,  "Constant_Attenuation"); 
          The.Sun.Initialized := False;
        when Linear_Attenuation    => 
          Change(The.Sun.Linear_Attenuation,    "Linear_Attenuation"); 
          The.Sun.Initialized := False;
        when Quadratic_Attenuation => 
          Change(The.Sun.Quadratic_Attenuation, "Quadratic_Attenuation");                 
          The.Sun.Initialized := False;
        when Sun_Distance          =>
          Change(The.Sun.Position(Z), "Sun.Position(Z)", 
                 The.Sun.Position(Z) / 4.0, 0.0);                                       
        when Deviation             =>
          Change(Font_Deviation, "Deviation", 0.1, 0.0);         
          The.Net_List := Invalid_List;
          Render_Lists := False; -- To save memory
        when Extrusion             =>
          Change(Font_Extrusion, "Extrusion", 0.01, 0.0);         
          The.Net_List := Invalid_List;
          Render_Lists := False; -- To save memory
      end case;
    end Change;  
    --------------------------------------------------------------------------
    procedure Change_Time(Offset : in Duration;
                          Times  : in Positive := 1) is
    begin  
      The.Animated_Time := True; 
      if The.Great_Change then
        The.Time := The.Time + Offset;
      else
        for I in 1 .. Times loop
          The.Time := The.Time + Offset;
        end loop;  
      end if;                                               
    end Change_Time;  
    --------------------------------------------------------------------------    
    procedure Translate_Earth(Step : in Gl.Glfloat;
                              Axis : in Types.Axis_Type) is
    begin                                
      The.Position(Axis) := The.Position(Axis) + Step;           
      Timer.Set("Screen.Width:" & Gl.Glfloat'Image(Screen.Width) &
                ", Screen.Height:" & Gl.Glfloat'Image(Screen.Height) &
                ", Position(" & Positive'Image(Axis) & "):" & 
                Gl.Glfloat'Image(The.Position(Axis)), 4, Append => False);
    end Translate_Earth;              
    --------------------------------------------------------------------------
    type Position_Type is (Lat, Lon);
        
    procedure Rotate_Earth(Angle : in Gl.Glfloat;
                           Which : in Position_Type) is
    begin                                
      The.Animated_Latlon := True;
      case Which is
        when Lat => The.Latitude  := The.Latitude  + Angle;
        when Lon => The.Longitude := The.Longitude + Angle;
      end case;  
      Timer.Set("lat:" & Gl.Glfloat'Image(The.Latitude) & 
               ",lon:" & Gl.Glfloat'Image(The.Longitude), 4, Append => False);                 
    end Rotate_Earth;              
    --------------------------------------------------------------------------    
    procedure Zoom_Earth(Factor : in Gl.Glfloat) is
    begin                                
      The.Size_Scaling        := The.Size_Scaling        * Factor;
      The.Solid_Sphere.Radius := The.Solid_Sphere.Radius * Factor;
      The.Net_Sphere.Radius   := The.Solid_Sphere.Radius;
      The.Solid_List          := Invalid_List;
      The.Net_List            := Invalid_List;
      Render_Lists            := False; -- To save memory
      
      if The.Objects /= null then
        for I in The.Objects'range loop
          The.Objects(I).Symbol.Position :=
            Position_Of(Latitude  => The.Objects(I).Latitude,
                        Longitude => The.Objects(I).Longitude,
                        Radius    => The.Solid_Sphere.Radius);
          The.Objects(I).Name.Position := The.Objects(I).Symbol.Position;
        end loop;
      end if;  
      Timer.Set("Screen.Width:" & Gl.Glfloat'Image(Screen.Width) &
                ", Screen.Height:" & Gl.Glfloat'Image(Screen.Height) &
                ", Earth-Radius:" & Gl.Glfloat'Image(The.Solid_Sphere.Radius), 
                4, Append => False);
    end Zoom_Earth;              
    --------------------------------------------------------------------------  
    procedure Change_Time_Factor(Factor : in Gl.Glfloat) is
    begin
      if Factor = 1.0 then
        The.Animated_Time := False;
        The.Time_Factor   := 1.0;
      else
        The.Animated_Time := True;
        The.Time_Factor   := The.Time_Factor * Factor;
      end if;  
      Timer.Set("Time-factor:" & Gl.Glfloat'Image(The.Time_Factor), 
                4, Append => False);      
    exception
      when others => null;                          
    end Change_Time_Factor;
    --------------------------------------------------------------------------               
    procedure Set_Axis(Axis : in Types.Axis_Type) is
    begin
      Current_Axis := Axis;
      Timer.Set("Axis:" & Positive'Image(Axis), 4, Append => False);
      Offset(Axis) := Offset(Axis) + 10.0;      
    end Set_Axis;            
    --------------------------------------------------------------------------               
    procedure Invert(Switch : in out Boolean;
                     Name   : in     String) is
    begin                 
      Switch := not Switch;
      Timer.Set("Render " & Name & ": " & Boolean'Image(Switch), 
                4, Append => False);
    end Invert;            
    --------------------------------------------------------------------------                                    
  begin
    Timer.Set("Earth.Keypressed: " & Win32.Wparam'Image(Wparam), 4, 
              Append => False);
            
    case Wparam is      
      when  72 => The.Sunlight := not The.Sunlight;             -- h
                  if The.Sunlight then
                    Texture_Env_Parameter := Modulate;
                  else
                    Texture_Env_Parameter := Decal;
                  end if;                      
                  The.Animated_Latlon := False;
                  The.Solid_List      := Invalid_List;
                  Render_Lists        := False; -- To save memory
      when  70 => The_Parameter       := Filter;                -- f  
      when  77 => The_Parameter       := Light_Model;           -- m  
      when  76 => The_Parameter       := Light_Source;          -- l  
      when  79 => The_Parameter       := Object;                -- o  
      when  75 => The_Parameter       := Shininess;             -- k 
      when 220 => The_Parameter       := Spot_Exponent;         -- ^ 
      when 226 => The_Parameter       := Spot_Cutoff;           -- < 
      when  49 => The_Parameter       := Constant_Attenuation;  -- 1 
      when  50 => The_Parameter       := Linear_Attenuation;    -- 2 
      when  51 => The_Parameter       := Quadratic_Attenuation; -- 3 
      when 191 => The_Parameter       := Sun_Distance;          -- #      
      when  65 => The_Light           := Ambient;               -- a
      when  68 => The_Light           := Diffuse;               -- d
      when  83 => The_Light           := Specular;              -- s
      when  73 => The_Light           := Emission;              -- i
      when  82 => The_Color_Component := Bitmaps.Red;           -- r
      when  71 => The_Color_Component := Bitmaps.Green;         -- g
      when  66 => The_Color_Component := Bitmaps.Blue;          -- b      
      when  52 => The_Color_Component := Bitmaps.Alpha;         -- 4      
      when 122 => The_Parameter       := Deviation;             -- F11
      when 123 => The_Parameter       := Extrusion;             -- F12
      when  53 .. 56 =>                                         -- 5 .. 8
        Texture_Env_Parameter := Texture_Env_Parameter_Type'Val
                                 (Natural(Wparam) - 53);
        The.Solid_List := Invalid_List;
        Render_Lists   := False; -- To save memory
        Timer.Set("Texture_Env_Parameter: " & Texture_Env_Parameter_Type'Image
                   (Texture_Env_Parameter), 1, Append => False);
      ------------------------------------------------------------------------
      -- Manipulate color of earth bitmap or lighting parameters.
      ------------------------------------------------------------------------      
      when 192 => The_Filter_Parameter := Color_First;  -- 
      when 222 => The_Filter_Parameter := Color_Last;   -- 
      when 107 => The_Filter_Parameter := Color_Bias;   -- +
      when 106 => The_Filter_Parameter := Color_Factor; -- *
      when  48 => The_Filter(Bitmaps.Red).Bias   := 0;  -- 0
                  The_Filter(Bitmaps.Green).Bias := 0;
                  The_Filter(Bitmaps.Blue).Bias  := 0;
                  The_Factor := (others => 1.0);
      when  80 => if The.Sun.Position_W = 0.0 then      -- p
                    The.Sun.Position_W := 1.0;
                  else
                    The.Sun.Position_W := 0.0;
                  end if;          
                  Set(The.Sun.Position_W, "Position(W)");
      when 187 => Change( 1);                           -- +
      when 189 => Change(-1);                           -- -
      ------------------------------------------------------------------------
      -- Manipulate lat and lon
      ------------------------------------------------------------------------
      when  38 => Rotate_Earth( 10.0, Lat);                      -- up
      when  40 => Rotate_Earth(-10.0, Lat);                      -- down
      when  37 => Rotate_Earth(-10.0, Lon);                      -- left
      when  39 => Rotate_Earth( 10.0, Lon);                      -- right
      ------------------------------------------------------------------------
      -- Manipulate time
      ------------------------------------------------------------------------
      when 112 => Change_Time( 60.0,     1);                    -- F1
      when 113 => Change_Time(-60.0,     1);                    -- F2
      when 114 => Change_Time( 86400.0, 30);                    -- F3
      when 115 => Change_Time(-86400.0, 30);                    -- F4
      when  16 => The.Great_Change := not The.Great_Change;     -- shift         
      when  87 => The.Time_Offset  := The.Time_Offset - 3600.0; -- w - west                
      when  69 => The.Time_Offset  := The.Time_Offset + 3600.0; -- e - east                
      ------------------------------------------------------------------------
      -- Translate Earth
      ------------------------------------------------------------------------
      when 102 => Translate_Earth( Step, Y);                    -- Keypad keyup
      when 100 => Translate_Earth(-Step, Y);                    -- Keypad keydown
      when 104 => Translate_Earth( Step, X);                    -- Keypad keyleft
      when  98 => Translate_Earth(-Step, X);                    -- Keypad keyright
      ------------------------------------------------------------------------            
      -- Zoom Earth            
      ------------------------------------------------------------------------            
      when  33 => Zoom_Earth(Factor);                           -- +
      when  34 => Zoom_Earth(1.0 / Factor);                     -- -
      ------------------------------------------------------------------------
      -- Manipulate time-factor
      ------------------------------------------------------------------------                        
      when 116 => Change_Time_Factor(2.0);                      -- F5
      when 117 => Change_Time_Factor(0.5);                      -- F6        
      when 118 => Change_Time_Factor(1.0);                      -- F7        
      ------------------------------------------------------------------------
      -- Some other manipulations
      ------------------------------------------------------------------------                        
      when  46 => Render_Lists := not Render_Lists;              -- del
      when  78 => Invert(The.Render_Net, "Net-Shpere");          -- n
      when  67 => Invert(The.Render_C_sys, "Coordinate-System"); -- c 
      when  32 => Invert(The.Render_Bitmap, "Bitmap");           -- blank
      when  88 => Set_Axis(X); -- x 
      when  89 => Set_Axis(Y); -- y 
      when  90 => Set_Axis(Z); -- z 
      when   8 => -- backspace
        if The.Objects /= null then                  
          for I in The.Objects'range loop
            The.Objects(I).Selected := False;
          end loop;
        end if;                  
        Timer.Set("Objects selection reset", 4);
      when others => null;
    end case;
      
    Timer.Set("Shineness: " & Gl.Glfloat'Image(The.Shininess) &
              ", Earth.Lists: " & Boolean'Image(Render_Lists) & 
              ", Param: " & Texture_Env_Parameter_Type'Image(Texture_Env_Parameter) &
              ", Parameter: " & Parameter_Type'Image(The_Parameter) &       
              ", Filter: "    & Filter_Parameter_Type'Image(The_Filter_Parameter) &       
              ", Component: " & Positive'Image(The_Color_Component) &
              ", Light: "     & Light_Type'Image(The_Light), 4, Append => False);
    
    if Keypressed_Callback /= null then
      return Keypressed_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Keypressed;
                          
  --==========================================================================             
                            
  function Middle_Mouse_Down
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    The : Model renames Model(The_Model.all);
    ---------------------------
    procedure Render_Objects is
    begin
      if The.Objects /= null then
        Gl.Glinitnames;
        Gl.Glpushname(0);
--        Models.Motion_Sequence := Models.Rotate_First;
--        Models.Rotation_1 := Z;
--        Models.Rotation_2 := Y;
--        Models.Rotation_3 := X;      
        The.Attitude := Gl.Glfloat(The.Time_Factor) * 
                        Earth_Attitude(Ada.Calendar.Clock);
	      for I in The.Objects'range loop
          Models.Model_Name := I;
          Models.Initialize_Render(The);
          Models.Model_Name := 0;
	        Models.Render(The.Objects(I).Symbol);
          Models.Finalize_Render(The);
	      end loop;
--        Models.Rotation_1 := X;
--        Models.Rotation_2 := Y;
--        Models.Rotation_3 := Z;      
--        Models.Motion_Sequence := Models.Translate_First;
        Gl.Glflush;
      end if;
    end Render_Objects;
    -------------------              
    Mouse_X : constant Win32.Word := Win32.Windef.Loword
                                    (Interfaces.C.Unsigned_Long(Lparam));
    Mouse_Y : constant Win32.Word := Win32.Windef.Hiword
                                    (Interfaces.C.Unsigned_Long(Lparam));

    Selection_Buffer : array(1 .. 64) of aliased Gl.Gluint := (others => 0);
    Hits             : Gl.Glint;
    Viewport         : array(0 .. 3) of aliased Gl.Glint;
    The_Viewport     : Glu.Glint_Array;
    Index            : Positive := 4;
  begin
    Gl.Glselectbuffer(Size   => Selection_Buffer'Length,
                      Buffer => Selection_Buffer
                               (Selection_Buffer'First)'Unchecked_Access);
    Gl.Glgetintegerv(Pname  => Gl.Gl_Viewport,
                     Params => Viewport(Viewport'First)'Unchecked_Access);

    for I in Viewport'range loop
      The_Viewport(I) := Viewport(I);
    end loop;

    Gl.Glmatrixmode(Gl.Gl_Projection);
    Gl.Glpushmatrix;

    if Gl.Glrendermode(Gl.Gl_Select) /= 0 then
      Ada.Exceptions.Raise_Exception(Opengl_Error'Identity,
        "Opengl.Maps.Left_Mouse_Down, Gl.GlRenderMode returns not 0");
    end if;

    Gl.Glloadidentity;
    Glu.Glupickmatrix(X        => Gl.Gldouble(Mouse_X),
                      Y        => Gl.Gldouble
                                 (Screen.Pixel_Height - Integer(Mouse_Y)),
                      Width    => 25.0,
                      Height   => 25.0,
                      Viewport => The_Viewport);
    Gl.Glortho(Left   => 0.0,    Right => Gl.Gldouble(Screen.Width),
               Bottom => 0.0,    Top   => Gl.Gldouble(Screen.Height),
               Znear  => 1000.0, Zfar  => - 1000.0);

    Render_Objects;

    Hits := Gl.Glrendermode(Gl.Gl_Render);
    if The.Objects /= null then
      for I in 1 .. Hits loop
        The.Objects(Integer(Selection_Buffer(Index))).Selected := True;
        Index := Index + 4;
      end loop;
    end if;  

    Gl.Glmatrixmode(Gl.Gl_Projection);
    Gl.Glpopmatrix;
    Gl.Glmatrixmode(Gl.Gl_Modelview);

    if Command_Line.Clock = 0 then
      Timer.Set("X:" & Win32.Word'Image(Mouse_X) &
                ", Y:" & Win32.Word'Image(Mouse_Y) &
                ", Hits:" & Gl.Glint'Image(Hits), 3, Append => False);
    end if;            

    if Middle_Mouse_Down_Callback /= null then
      return Middle_Mouse_Down_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Middle_Mouse_Down;
                          
  --==========================================================================             
  
  function Resize
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    The        : Model renames Model(The_Model.all);
    New_Width       : constant Natural := 
                      Natural(Win32.Windef.Loword(Win32.Dword(Lparam)));
    New_Height      : constant Natural := 
                      Natural(Win32.Windef.Hiword(Win32.Dword(Lparam)));                   
    Height_To_Width : constant Gl.Glfloat := 
                      Gl.Glfloat(New_Height) / Gl.Glfloat(New_Width);
    Screen_Height,
    Screen_Width    : Gl.Glfloat;
  begin  
    if New_Width <= New_Height then
      Screen_Height := Screen.Width  * Height_To_Width;
      Screen_Width  := Screen.Width;
    else
      Screen_Height := Screen.Height;
      Screen_Width  := Screen.Height / Height_To_Width;
    end if;    
         
    if The.Position(X .. Y) /= (0.0, 0.0) then
      The.Position(X) := Screen_Width  / 2.0; 
      The.Position(Y) := Screen_Height / 2.0; 
    end if;                                     
    
    if Resize_Callback /= null then
      return Resize_Callback(Window_Handle, Message, Wparam, Lparam);
    else    
      return Win32.Winuser.Defwindowproc(Window_Handle, Message, Wparam, Lparam);
    end if;                                
  end Resize;
                          
  --==========================================================================             

  function Model_With
    (Bitmap_File     : in String;
     Position        : in Types.Vector_3d :=
                         (X => Screen.Width  / 2.0,
          	              Y => Screen.Height / 2.0,
                          Z => 0.0);
     Attitude        : in Types.Vector_3d :=
                         (X => 0.0,
          	              Y => 0.0,
                          Z => 0.0);
     Scale           : in Gl.Glfloat := Screen.Height / 2.0;
     Longitude_Lines : in Natural    := 360;
     Latitude_Lines  : in Natural    := 180;
     Time_Factor     : in Gl.Glfloat := 1.0;
     Time_Offset     : in Duration   := 0.0;
     Sunlight        : in Boolean    := False;
     Latitude       : in Gl.Glfloat  := 0.0;
     Longitude      : in Gl.Glfloat  := 0.0) return Models.Root_Model_Access is

    Solid_Sphere : Models.Forms.Sphere_3d :=
      (Models.Root_Model with
       Position           => (0.0, 0.0, 0.0),
       Attitude           => (0.0, 0.0, -5.0 + 180.0),
       Size_Scaling       => 1.0,
       Color              => Colors.White,
       Radius             => Scale,
       Slices             => 360,
       Stacks             => 180,
       Glu_Quadric_Object => Glu.Glunewquadric);

    Net_Sphere : Models.Forms.Sphere_3d :=
      (Models.Root_Model with
       Position           => (0.0, 0.0, 0.0),
       Attitude           => (0.0, 0.0, -4.0),
       Size_Scaling       => 1.0,
       Color              => Colors.White,
       Radius             => Scale,
       Slices             => Longitude_Lines,
       Stacks             => Latitude_Lines,
       Glu_Quadric_Object => Glu.Glunewquadric);

    Earth_Coord_System : Coordinate_System.Model :=
      (Models.Root_Model with
       Position           => Position,
       Attitude           => Attitude,
       Size_Scaling       => Coordinate_System.Max_Length,
       Color              => Colors.Red,
       Axis_Width         => 1.0,
       Name               => (" Earth          "),
       Transformation     => 0);
       
    Sun_Distance : constant Gl.Glfloat := Scale * 150_000_000.0 / 6000.0;   
    Shininess    : Gl.Glfloat := 3.0;
  begin
    Keypressed_Callback := 
      Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Keydown);
    Middle_Mouse_Down_Callback := 
      Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Mbuttondown);
    Resize_Callback := Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Size);
      
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Keydown)     :=
      Keypressed'access;
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Mbuttondown) :=
      Middle_Mouse_Down'access;  
    Win.Standard_Window.Callbacks(Win32.Winuser.Wm_Size) := Resize'access;  
    
    if System.Name'Image(System.System_Name) = "SYSTEM_NAME_GNAT" then
      Shininess := 0.6;
    end if; -- 3.0 looks good with ObjectAda and corresponding lib.
      
    The_Model := new Model'
      (Models.Root_Model with
       Position        => Position,
	     Attitude        => Attitude,
    	 Size_Scaling    => 1.0,
	     Color           => Colors.White,
       
       Initialized     => False,
       Face            => Gl.Gl_Front,
       Ambient_Color   => Models.Lighting.Default_Ambient_Light,
       Diffuse_Color   => Models.Lighting.Default_Diffuse_Light,
       Specular_Color  => Colors.White,
       Emission_Color  => Colors.Black,
       Shininess       => Shininess,
       Color_Indexes   => (0, 1, 1),                         
       
       Solid_Sphere    => Solid_Sphere,
       Net_Sphere      => Net_Sphere,
       Earth_Bitmap    => Earth_Bitmap(Bitmap_File, The_Model),

       Solid_List      => Invalid_List,
       Net_List        => Invalid_List,
       
       Latitude        => Latitude,
       Longitude       => Longitude,

       Lighting_Model  => (Initialized   => False,
                           Ambient_Color => Colors.White,--Models.Lighting.Default_Ambient_Light,
                           Local_Viewer  => Gl.Gl_False,                                   
                           Two_Side      => Gl.Gl_False),
                           
       Sun             => Models.Lighting.Light_Source'(Models.Root_Model with       
                          Position              => (0.0, 0.0, Sun_Distance),
                          Attitude              => (0.0, 0.0, 0.0),
                      	  Size_Scaling          => 1.0,
                      	  Color                 => Colors.White,
                          
                          Id                    => Gl.Glenum'Last,
                          Initialized           => False,        
                          Ambient_Color         => Colors.White,
                          Diffuse_Color         => Colors.White,
                          Specular_Color        => Colors.White,
                          Spot_Direction        => (0.0, 0.0, -1.0),
                          Position_W            => 0.0,
                          Spot_Exponent         => 0.0,
                          Spot_Cutoff           => 180.0,
                          Constant_Attenuation  => 1.0,
                          Linear_Attenuation    => 0.0,
                          Quadratic_Attenuation => 0.0),
                          
       Sun_Distance    => Sun_Distance,
       
       Z_Plane         => Clip_Planes.Clip_Plane
                         (Axis    => Z,
                          Visible => Clip_Planes.Positive_Direction,
                          Cutoff  => 0.0),
                          
       Time            => Ada.Calendar.Clock,
       Render_Time     => Ada.Calendar.Clock,
       Time_Factor     => Time_Factor,
       Time_Offset     => Time_Offset,
       
       Sunlight        => Sunlight,
       Animated_Latlon => False,
       Animated_Time   => Time_Factor /= 1.0,
       Great_Change    => False,
       Render_Net      => False,
       Render_Bitmap   => True,
       Render_C_sys    => False,

       Coord_System    => Earth_Coord_System,
       Objects         => Airports(Earth_Radius => Scale));
                                                                              
    return The_Model;
  end Model_With;
                          
  --==========================================================================             
  
  function Airports(Earth_Radius : in Gl.Glfloat) return Object_Array_Access is
    ---------------------------------------------------------------------
    function Object_Of(Latitude,
                       Longitude : in Gl.Glfloat;
                       Name      : in String := "") return Object_Type is
      Latitude_Text,
      Longitude_Text : String(1 .. 32);
    begin
      Utils.Set_Float_Image(Image => Latitude_Text,
                            Value => Latitude  * 180.0 / Ada.Numerics.Pi,
                            Aft   => 3, 
                            Exp   => 0);
      Utils.Set_Float_Image(Image => Longitude_Text,
                            Value => Longitude * 180.0 / Ada.Numerics.Pi,
                            Aft   => 3, 
                            Exp   => 0);
      declare
  	    The_Position : constant Types.Vector_3d :=
                       Position_Of(Latitude  => Latitude,
                                   Longitude => Longitude,
                                   Radius    => Earth_Radius);

        The_Text   : constant String := Name &
          " (lat: " &
          Ada.Strings.Fixed.Trim(Latitude_Text,  Ada.Strings.Both) &
          ",lon: " &
          Ada.Strings.Fixed.Trim(Longitude_Text, Ada.Strings.Both) &
--          ", x:" & Gl.Glfloat'Image(The_Position(X)) &
--          ", y:" & Gl.Glfloat'Image(The_Position(Y)) &
--          ", z:" & Gl.Glfloat'Image(The_Position(Z)) & 
          ")";

        The_Name : Models.Strings.String_3d :=
  	      (Models.Root_Model with
  	       Size_Scaling => 1.0,
  	       Color        => Colors.Red,
  	       Position     => The_Position,
  	       Attitude     => (0.0, 0.0, 0.0),
--  	       Position     => (Earth_Radius, 0.0, 0.0), 
--  	       Attitude     => (180.0 / Ada.Numerics.Pi) * (0.0, Latitude, Longitude),
  	       Length       => The_Text'Length,
  	       Text         => The_Text,
           Alignment    => (Horizontal => Models.Strings.Left,
                            Vertical   => Models.Strings.Bottom),
  	       Font         => Fonts.Font,
  	       Font_Size    => 0.025 * Screen.Height / 2.0,
           Hidden_Data  => Models.Strings.Default_Hidden_Data);

        The_Symbol : Models.Forms.Sphere_3d :=
          (Models.Root_Model with
           Position           => The_Name.Position,
           Attitude           => The_Name.Attitude,
           Size_Scaling       => 1.0,
           Color              => Colors.Red,
           Radius             => Earth_Radius / 100.0,
           Slices             => 36,
           Stacks             => 18,
           Glu_Quadric_Object => Glu.Glunewquadric);
      begin
        return Object_Type'(Selected  => False,
                            Latitude  => Latitude,
                            Longitude => Longitude,
                            Symbol    => Models.Display_List_Model_Of(The_Symbol),
                            Name      => Models.Display_List_Model_Of(The_Name));
      end;
    end Object_Of;
    --------------------------------------------------------------------------
    The     : Object_Array_Access;
  begin
--    The := new Object_Array(1 .. 4);
--    The(The'Last - 3) := Object_Of(0.0, 0.0);
--    The(The'Last - 2) := Object_Of(0.0, Ada.Numerics.Pi);
--    The(The'Last - 1) := Object_Of(  Ada.Numerics.Pi / 2.0, 0.0, "North-Pole");
--    The(The'Last)     := Object_Of(- Ada.Numerics.Pi / 2.0, 0.0, "South-Pole");
    return The;
  exception when others => return null;
  end Airports;  
                          
  --==========================================================================             
  
  procedure Smooth_Edge(Pixels : in out Bitmaps.Pixel_Array) is
    -------------------------------------------------
    procedure Smooth(Col1, Col2 : in Positive) is
      -------------------------------------------------
      function Mean (Left, Right : in Bitmaps.Rgb_Color) 
                    return Bitmaps.Rgb_Color is          
        The : Bitmaps.Rgb_Color;
      begin
        for I in The'range loop
          The(I) := Bitmaps.Byte((Integer(Left(I)) + Integer(Right(I))) / 2);
        end loop;  
        return The;
      end Mean;     
      -------------------------------------------------
    begin
      for Row in Pixels'range(1) loop
        Pixels(Row, Col1) := Mean(Pixels(Row, Col1), Pixels(Row, Col2));
        Pixels(Row, Col2) := Pixels(Row, Col1);
      end loop;      
    end Smooth;          
    -------------------------------------------------
    Max : constant Integer := 1;--Utilities.Command_Line.Argument("-s",0);
  begin
    Timer.Set(Positive'Image(Pixels'First(1))&" .."&Positive'Image(Pixels'Last(1)) & "," &
              Positive'Image(Pixels'First(2))&" .."&Positive'Image(Pixels'Last(2)),3);
              
    Smooth(Col1 => Pixels'First(2), Col2 => Pixels'Last(2));
    for Col in 1 .. Max loop
      Smooth(Col1 => Col, Col2 => Col + 1);
      Smooth(Col1 => Pixels'Last(2) - Col + 1, Col2 => Pixels'Last(2) - Col);
    end loop;    
  end Smooth_Edge;  

  --==========================================================================             

end Opengl.Earth;
