--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-dialfaces.adb
-- Description     : Package for the definition of dialface-models. 
-- Author          : H. Altmann, A. Gaisbauer, J. Schrer
-- Created On      : 09. Feb. 1998
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Strings.Fixed;
with Opengl.Float_32.Elementary_Functions;
with Opengl.Float_32.Arrays.Vector_Operations;
with Opengl.Float_32.Arrays.Types;
with Opengl.Models.Forms;
with Opengl.Types;

package body Opengl.Models.Dialfaces is

  use Opengl.Float_32.Elementary_Functions;
      -- to make sin, cos, etc directly visible.
  use Opengl.Float_32.Arrays.Vector_Operations;
      -- to make vector-operators directly visible.
  use type Opengl.Float_32.Arrays.Types.Vector;
  use type Ada.Strings.Unbounded.String_Access;
  use type Models.Strings.String_2d_Access;
  ---------------------------------------------------------------------------
  procedure Render(The_Model  : in out Ticks) is

    Phi  : Gl.Glfloat := The_Model.Start_Angle;
    Tick : constant Models.Forms.Rectangle :=
      (Lower_Left  => The_Model.Size_Scaling * 
                      (X =>   The_Model.Min_Radius,
                       Y => - The_Model.Width / 2.0),
       Upper_Right => The_Model.Size_Scaling * 
                      (X =>   The_Model.Max_Radius,
                       Y =>   The_Model.Width / 2.0));
  begin
    Initialize_Render(The_Model);
    for I in 1 .. The_Model.Number loop
      Models.Forms.Render(The_Rectangle => Tick, Angle => Phi);
      Phi := Phi + The_Model.Angle_Increment;
    end loop;
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model  : in out Abstract_Ticks_With_Strings) is
  begin
    if The_Model.Width > 0.0 then
      Render(The_Model => Ticks(The_Model));
    end if;
      
    Initialize_Render(The_Model);
    for I in The_Model.Strings'range loop
      if The_Model.Strings(I) /= null then
        Models.Strings.Render(The_Model.Strings(I).all);
      end if;
    end loop;
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  function String_Of(The_Model : in Abstract_Ticks_With_Strings'Class;
                     Image     : in String;
                     Phi       : in Gl.Glfloat)
           return Models.Strings.String_2d_Access is

    R : constant Gl.Glfloat := The_Model.String_Center * The_Model.Min_Radius;

    The_String : Models.Strings.String_2d :=
      (Ada.Finalization.Controlled with
       Length       => Image'Length,
       Size_Scaling => The_Model.Size_Scaling,
       Color        => The_Model.Color,
       Position     => The_Model.Size_Scaling *
                       (X => R * Cos(Phi, Cycle => 360.0),
                        Y => R * Sin(Phi, Cycle => 360.0),
                        Z => The_Model.Position(Z)),
       Alignment    => (Horizontal => Models.Strings.Center,
                        Vertical   => Models.Strings.Middle),
       Attitude     => 0.0,
       Text         => Image,
       Font         => The_Model.Font,
       Font_Size    => The_Model.Font_Size,
       Hidden_Data  => Models.Strings.Default_Hidden_Data);
  begin
    case The_Model.String_Orientation is
      when Vertical      => null;
      when Top_Inside    => The_String.Attitude := Phi + 90.0;
      when Top_Outside   => The_String.Attitude := Phi - 90.0;
      when Left_Outside  => The_String.Attitude := Phi + 180.0;
      when Right_Outside => The_String.Attitude := Phi;
    end case;

    return new Models.Strings.String_2d'(The_String);
  end String_Of;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out Ticks_With_Digits) is
    Phi : Gl.Glfloat := The_Model.Start_Angle;
  begin
    for I in The_Model.Strings'range loop
      if The_Model.Digit_Value(I) /= Invalid_Number and then
         The_Model.Strings(I) = null then
        The_Model.Strings(I) := String_Of
            (The_Model => The_Model,
             Image     => Ada.Strings.Fixed.Trim
                         (Integer'Image(The_Model.Digit_Value(I)),
                          Side => Ada.Strings.Both),
             Phi       => Phi);
      end if;
      Phi := Phi + The_Model.Angle_Increment;
    end loop;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust(The_Model : in out Ticks_With_Digits) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out Ticks_With_Strings) is
    Phi : Gl.Glfloat := The_Model.Start_Angle;
  begin
    for I in The_Model.Strings'range loop
      if The_Model.Raw_Strings(I) /= null and then
         The_Model.Strings(I) = null then
        The_Model.Strings(I) := String_Of
            (The_Model => The_Model,
             Image     => The_Model.Raw_Strings(I).all,
             Phi       => Phi);
      end if;
      Phi := Phi + The_Model.Angle_Increment;
    end loop;
  end Initialize;
  ----------------------------------------------------------------------------  
  procedure Adjust(The_Model : in out Ticks_With_Strings) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Render(The_Model  : in out Tick_List) is

    Tick : constant Models.Forms.Rectangle :=
      (Lower_Left  => The_Model.Size_Scaling *
                      (X =>   The_Model.Min_Radius,
                       Y => - The_Model.Width / 2.0),
       Upper_Right => The_Model.Size_Scaling *
                      (X =>   The_Model.Max_Radius,
                       Y =>   The_Model.Width / 2.0));
  begin
    Initialize_Render(The_Model);
    for I in The_Model.Angles'range loop
      Models.Forms.Render(The_Rectangle => Tick, Angle => The_Model.Angles(I));
    end loop;
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model  : in out Abstract_Tick_List_With_Strings) is
  begin
    if The_Model.Width > 0.0 then
      Render(The_Model => Tick_List(The_Model));
    end if;
      
    Initialize_Render(The_Model);
    for I in The_Model.Strings'range loop
      if The_Model.Strings(I) /= null then
        Models.Strings.Render(The_Model.Strings(I).all);
      end if;
    end loop;
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  function String_Of(The_Model : in Abstract_Tick_List_With_Strings'Class;
                     Image     : in String;
                     Phi       : in Gl.Glfloat)
           return Models.Strings.String_2d_Access is

    R : constant Gl.Glfloat := The_Model.String_Center * The_Model.Min_Radius;

    The_String : Models.Strings.String_2d :=
      (Ada.Finalization.Controlled with
       Length       => Image'Length,
       Size_Scaling => The_Model.Size_Scaling,
       Color        => The_Model.Color,
       Position     => The_Model.Size_Scaling *
                       (X => R * Cos(Phi, Cycle => 360.0),
                        Y => R * Sin(Phi, Cycle => 360.0),
                        Z => The_Model.Position(Z)),
       Alignment    => (Horizontal => Models.Strings.Center,
                        Vertical   => Models.Strings.Middle),
       Attitude     => 0.0,
       Text         => Image,
       Font         => The_Model.Font,
       Font_Size    => The_Model.Font_Size,
       Hidden_Data  => Models.Strings.Default_Hidden_Data);
  begin
    case The_Model.String_Orientation is
      when Vertical      => null;
      when Top_Inside    => The_String.Attitude := Phi + 90.0;
      when Top_Outside   => The_String.Attitude := Phi - 90.0;
      when Left_Outside  => The_String.Attitude := Phi + 180.0;
      when Right_Outside => The_String.Attitude := Phi;
    end case;

    return new Models.Strings.String_2d'(The_String);
  end String_Of;
  ----------------------------------------------------------------------------             
  procedure Initialize(The_Model : in out Tick_List_With_Digits) is
  begin
    for I in The_Model.Strings'range loop
      if The_Model.Digit_Value(I) /= Invalid_Number and then
         The_Model.Strings(I) = null then
        The_Model.Strings(I) := String_Of
            (The_Model => The_Model,
             Image     => Ada.Strings.Fixed.Trim
                         (Integer'Image(The_Model.Digit_Value(I)),
                          Side => Ada.Strings.Both),
             Phi       => The_Model.Angles(I) );
      end if;
    end loop;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust(The_Model : in out Tick_List_With_Digits) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out Tick_List_With_Strings) is
  begin
    for I in The_Model.Strings'range loop
      if The_Model.Raw_Strings(I) /= null and then
         The_Model.Strings(I) = null then
        The_Model.Strings(I) := String_Of
            (The_Model => The_Model,
             Image     => The_Model.Raw_Strings(I).all,
             Phi       => The_Model.Angles(I) );
      end if;
    end loop;
  end Initialize;
  ----------------------------------------------------------------------------  
  procedure Adjust(The_Model : in out Tick_List_With_Strings) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
end Opengl.Models.Dialfaces;
