--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-dialfaces.ads
-- Description     : Package for the definition of dialface-models. 
-- Author          : H. Altmann, A. Gaisbauer, J. Schrer
-- Created On      : 09. Feb. 1998
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Strings.Unbounded;
with Opengl.Fonts;
with Opengl.Models.Strings;
with Opengl.Types;

package Opengl.Models.Dialfaces is

  use type Gl.Glfloat;

  ----------------------------------------------------------------------------
  -- Dialfaces

  subtype Digit_Number_Type is Positive range 1 .. 4;

  subtype Tick_Number_Type is Natural    range       0 .. 360;
  subtype Degree_Type      is Gl.Glfloat range - 360.0 .. 360.0;


  type Ticks(Number  : Tick_Number_Type) is new Model_2d_Z with record
    Start_Angle    : Degree_Type := 0.0;
    Angle_Increment: Degree_Type := 360.0 / 12.0;
    Min_Radius     : Gl.Glfloat  := 0.825;
    Max_Radius     : Gl.Glfloat  := 1.0;
    Width          : Gl.Glfloat  := 0.2;
--    Tick_Rectangle : Rectangle;
  end record;

  -- The ticks of a circular dialface. .
  -- All angles are measured in degree with reference to the positive x-axis
  -- 0 deg is direction x (points to right end of screen).

  --  Number          : is the number of ticks on the dialface.
  --  Start_Angle     : the angle of the first tick.
  --  Angle_Increment : angular difference between ticks
  --  Min_Radius      : radial distance from center to start of ticks.
  --  Max_Radius      : radial distance from center to end   of ticks.
  --  Width           : width of ticks

  procedure Render(The_Model  : in out Ticks);

  ----------------------------------------------------------------------------     

  type String_Orientation_Type is
    (Vertical,
     Top_Inside,
     Top_Outside,
     Left_Outside,
     Right_Outside);

  type Abstract_Ticks_With_Strings(Number : Tick_Number_Type)
       is abstract new Ticks(Number) with record

    String_Orientation : String_Orientation_Type      := Vertical;
    Strings            : Models.Strings.String_2d_Array
                                        (1 .. Number) := (others => null);
    String_Center      : Gl.Glfloat range 0.0 .. 3.0  := 0.8;
    Font               : Fonts.Font_Type              := Fonts.Font;
    Font_Size          : Fonts.Font_Size_Type         := 10.0;
  end record;

  -- Ticks with strings, abstract, because no "initialize" procedure is given
  -- that specifies how the strings are built.
  --
  -- String_Orientation : specifies the attitude of the strings.
  -- Strings            : Theese models are assigned by initialize.
  -- String_center      : Specifies a circle for the center of the strings 
  --                      and refers to min_radius. With string_center > 1.0 
  --                      the strings may be placed outside the max_radius.

--  procedure Initialize(The_Model : in out Abstract_Ticks_With_Strings'Class) is abstract;
--  procedure Adjust    (The_Model : in out Abstract_Ticks_With_Strings'Class) is abstract;

  procedure Render    (The_Model  : in out Abstract_Ticks_With_Strings);

  ----------------------------------------------------------------------------     

  Invalid_Number : constant := Integer'First;

  type Ticks_With_Digits(Number  : Tick_Number_Type)
       is new Abstract_Ticks_With_Strings(Number) with record

    Digit_Value : Types.Integer_Array(1 .. Number) := (others => Invalid_Number);
  end record;

  -- For a dialface where some but not necassary all ticks have digits.
  --
  -- The strings-component (derived from the ancestor type 
  -- Abstract_Ticks_With_Strings) is initialized in the following way:
  --
  -- The digits are specified by the digit_value number_array.
  -- If digit_value(i) /= Invalid_Number, then natural'image(digit_value(i))
  -- is taken for the string of the i'th tick.
  -- If digit_value(i)  = Invalid_Number, then the i'th tick has no string.


  procedure Initialize(The_Model : in out Ticks_With_Digits);
  procedure Adjust    (The_Model : in out Ticks_With_Digits);

  ----------------------------------------------------------------------------       

  type String_Vector is array (Positive range <>) of
       Ada.Strings.Unbounded.String_Access;

--  type String_Vector_Access is access all String_Vector;     

  type Ticks_With_Strings(Number : Tick_Number_Type)
       is new Abstract_Ticks_With_Strings(Number) with record

    Raw_Strings : String_Vector(1 .. Number);
  end record;

  -- For a dialface where some but not necassary all ticks have strings,
  -- which not all have to be digits.
  -- The strings-component (derived from the ancestor type 
  -- Abstract_Ticks_With_Strings)
  -- is initialized by the strings given in component raw_strings.
  -- If raw_strings(i) = null then the i'th tick has no string.


  procedure Initialize(The_Model : in out Ticks_With_Strings);
  procedure Adjust    (The_Model : in out Ticks_With_Strings);

  ----------------------------------------------------------------------------          
  -- For dialfaces with an irregular tick spacing. 

  type Tick_List(Number : Tick_Number_Type) is new Model_2d_Z with record
    Angles         : Types.Float_Vector(1 .. Number) := (others => 0.0);
    Min_Radius     : Gl.Glfloat                      := 0.825;
    Max_Radius     : Gl.Glfloat                      := 1.0;
    Width          : Gl.Glfloat                      := 0.2;
  end record;

  -- The ticks of a circular dialface. .
  -- All angles are measured in degree with reference to the positive x-axis
  -- 0 deg is direction x (points to right end of screen).

  --  Number     : is the number of ticks on the dialface.
  --  Angles     : angle-array of the ticks
  --  Min_Radius : radial distance from center to start of ticks.
  --  Max_Radius : radial distance from center to end   of ticks.
  --  width      : width of ticks

  procedure Render(The_Model  : in out Tick_List);

  ----------------------------------------------------------------------------   

  type Abstract_Tick_List_With_Strings(Number : Tick_Number_Type)
       is abstract new Tick_List(Number) with record

    String_Orientation : String_Orientation_Type      := Vertical;
    Strings            : Models.Strings.String_2d_Array
                                        (1 .. Number) := (others => null);
    String_Center      : Gl.Glfloat range 0.0 .. 3.0  := 0.8;
    Font               : Fonts.Font_Type              := Fonts.Font;
    Font_Size          : Fonts.Font_Size_Type         := 10.0;
  end record;

  -- Ticks with strings, abstract, because no "initialize" procedure is given
  -- that specifies how the strings are built.
  --
  -- String_Orientation : specifies the attitude of the strings.
  -- Strings            : The strings are assigned by initialize.
  -- String_center      : Specifies a circle for the center of the strings 
  --                      and refers to min_radius. With string_center > 1.0 
  --                      the strings may be placed outside the max_radius.

  procedure Render    (The_Model  : in out Abstract_Tick_List_With_Strings);

  ----------------------------------------------------------------------------     

  type Tick_List_With_Digits(Number  : Tick_Number_Type)
       is new Abstract_Tick_List_With_Strings(Number) with record

    Digit_Value : Types.Integer_Array(1 .. Number) := (others => Invalid_Number);
  end record;

  -- For a dialface where some but not necassary all ticks have digits.
  --
  -- The strings-component (derived from the ancestor type 
  -- Abstract_Ticks_With_Strings) is initialized in the following way:
  --
  -- The digits are specified by the digit_value number_array.
  -- If digit_value(i) /= Invalid_Number, then natural'image(digit_value(i))
  -- is taken for the string of the i'th tick.
  -- If digit_value(i)  = Invalid_Number, then the i'th tick has no string.


  procedure Initialize(The_Model : in out Tick_List_With_Digits);
  procedure Adjust    (The_Model : in out Tick_List_With_Digits);

  ----------------------------------------------------------------------------       

  type Tick_List_With_Strings(Number : Tick_Number_Type)
       is new Abstract_Tick_List_With_Strings(Number) with record

    Raw_Strings : String_Vector(1 .. Number);
  end record;

  -- For a dialface where some but not necassary all ticks have strings,
  -- which not all have to be digits.
  -- The strings-component (derived from the ancestor type 
  -- Abstract_Ticks_With_Strings)
  -- is initialized by the strings given in component raw_strings.
  -- If raw_strings(i) = null then the i'th tick has no string.


  procedure Initialize(The_Model : in out Tick_List_With_Strings);
  procedure Adjust    (The_Model : in out Tick_List_With_Strings);

end Opengl.Models.Dialfaces;
