--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-strings.adb
-- Description     : Package for the definition of strings opengl-models. 
-- Author          : H. Altmann, A. Gaisbauer, J. Schrer
-- Created On      : 09. Feb. 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Ada.Unchecked_Deallocation;
with Opengl.Display_Lists;
with Opengl.Float_32.Arrays.Vector_Operations;

package body Opengl.Models.Strings is

  use type Gl.Glfloat;
  use Opengl.Float_32.Arrays.Vector_Operations;
  ---------------------------------------------------------------------------
  -- Internal subprograms
  ---------------------------------------------------------------------------
  procedure Initialize(Data      : in out Hidden_Data_Type;
                       Text      : in     String;
                       Scale     : in     Gl.Glfloat;
                       Fontsize  : in     Gl.Glfloat;
                       Font      : in     Fonts.Font_Type;
                       Alignment : in     String_Alignment) is
  begin
    Data.Scale_Factor := Scale * Fontsize / Fonts.Max_Character_Height_Of(Font);

    case Alignment.Horizontal is
      when Left   => Data.Reference_Point(X) := 0.0;
      when Center => Data.Reference_Point(X) := - Data.Scale_Factor * 
                          Fonts.Width_Of(Text, Font) / 2.0;
      when Right  => Data.Reference_Point(X) := - Data.Scale_Factor * 
                          Fonts.Width_Of(Text, Font);
    end case;

    case Alignment.Vertical   is
      when Bottom => Data.Reference_Point(Y) := 0.0;
      when Middle => Data.Reference_Point(Y) := - Data.Scale_Factor * 
                          Fonts.Height_Of(Text, Font) / 2.0;
      when Top    => Data.Reference_Point(Y) := - Data.Scale_Factor * 
                          Fonts.Height_Of(Text, Font);
    end case;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Initialize(Data      : in out Hidden_Data_Type;
                       Text      : in     String;
                       Chars     : in     Positive;
                       Scale     : in     Gl.Glfloat;
                       Fontsize  : in     Gl.Glfloat;
                       Font      : in     Fonts.Font_Type;
                       Alignment : in     String_Alignment) is
  begin
    Data.Scale_Factor := Scale * Fontsize / Fonts.Max_Character_Height_Of(Font);

    case Alignment.Horizontal is
      when Left   => Data.Reference_Point(X) := 0.0;
      when Center => Data.Reference_Point(X) := 
                          - Data.Scale_Factor * Gl.Glfloat(Chars) *
                          Fonts.Mean_Character_Width_Of(Font) / 2.0;
      when Right  => Data.Reference_Point(X) := 
                          - Data.Scale_Factor * Gl.Glfloat(Chars) *
                          Fonts.Mean_Character_Width_Of(Font);
    end case;

    case Alignment.Vertical   is
      when Bottom => Data.Reference_Point(Y) := 0.0;
      when Middle => Data.Reference_Point(Y) := - Data.Scale_Factor * 
                          Fonts.Height_Of(Text, Font) / 2.0;
      when Top    => Data.Reference_Point(Y) := - Data.Scale_Factor * 
                          Fonts.Height_Of(Text, Font);
    end case;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Render(Text : in String;
                   Font : in Fonts.Font_Type;
                   Data : in Hidden_Data_Type) is
    pragma Inline(Render);               
  begin                 
    Gl.Gllistbase(Gl.Gluint(Fonts.List_Index_Of(Font)));
    Gl.Glpushmatrix;
    Gl.Gltranslatef(X => Data.Reference_Point(X), 
                    Y => Data.Reference_Point(Y),                       
                    Z => 0.0);
    Gl.Glscalef(X => Data.Scale_Factor, 
                Y => Data.Scale_Factor,
                Z => 1.0);
    for I in 1 .. Render_Count loop
      Gl.Glpushmatrix;                          
    	Gl.Glcalllists(N      => Text'Length,
                     C_Type => Gl.Gl_Unsigned_Byte,
                     Lists  => Text'Address);
      Gl.Glpopmatrix;
    end loop;
    Gl.Glpopmatrix;            
  end Render;
  ----------------------------------------------------------------------------  
  procedure Render(Text      : in     String;
                   Font      : in     Fonts.Font_Type;
                   Data      : in out Hidden_Data_Type;
                   Chars     : in     Positive;
                   Increment : in     Positive;
                   Scale     : in     Gl.Glfloat;
                   Fontsize  : in     Gl.Glfloat;
                   Alignment : in     String_Alignment) is                   
    --------------------------------------------------------------------------               
    procedure Render(The_Text : in String) is               
    begin
      Initialize(Data      => Data,
                 Text      => The_Text,
                 Scale     => Scale,
                 Fontsize  => Fontsize,
                 Font      => Font,
                 Alignment => Alignment);
      Render(Text => The_Text,
             Font => Font,
             Data => Data);
    end Render;
    --------------------------------------------------------------------------                        
    Last : Positive;
  begin
    if Chars < Text'Length then
      Last := Data.First + Chars - 1;
      if Last > Text'Length then
        Last := Text'Length;        
        Render(Text(Data.First .. Last) & " " & 
               Text(1 .. Chars + Data.First - Last - 1));
      else         
        Render(Text(Data.First .. Last));
      end if;        
      Data.First := Increment + Data.First rem Text'Length;
    else
      Render(Text);
    end if;
  end Render;
  ---------------------------------------------------------------------------
  -- End internal subprograms
  ---------------------------------------------------------------------------
  procedure Free(Item : in out String_2d_Access) is
    procedure Deallocate is new Ada.Unchecked_Deallocation
             (Name => String_2d_Access, Object => String_2d);
  begin
    Deallocate(Item);
  end Free;
  ---------------------------------------------------------------------------
  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;                       
                   Item   : in     String_2d_Access) is                     
  begin
    if Item = null then
      Boolean'Write(Stream, False);
    else
      Boolean'Write(Stream, True);
      String_2d'Output(Stream, Item.all);    
    end if;
  end Write;
  ----------------------------------------------------------------------------
  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out String_2d_Access) is
                    
    Is_Not_Null : Boolean;                
  begin
    Boolean'Read(Stream, Is_Not_Null);
    if Is_Not_Null then 
      Item := new String_2d'(String_2d'Input(Stream));
    else
      Item := null;
    end if;
  end Read;
  ----------------------------------------------------------------------------
  procedure Free(Item : in out String_3d_Access) is
    procedure Deallocate is new Ada.Unchecked_Deallocation
             (Name => String_3d_Access, Object => String_3d);
  begin
    Deallocate(Item);
  end Free;
  ---------------------------------------------------------------------------
  procedure Write(Stream : access Ada.Streams.Root_Stream_Type'Class;                       
                  Item   : in     String_3d_Access) is                     
  begin
    if Item = null then
      Boolean'Write(Stream, False);
    else
      Boolean'Write(Stream, True);
      String_3d'Output(Stream, Item.all);    
    end if;
  end Write;
  ----------------------------------------------------------------------------
  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out String_3d_Access) is
                    
    Is_Not_Null : Boolean;                
  begin
    Boolean'Read(Stream, Is_Not_Null);
    if Is_Not_Null then 
      Item := new String_3d'(String_3d'Input(Stream));
    else
      Item := null;
    end if;
  end Read;
  ----------------------------------------------------------------------------                       
  procedure Initialize(The_Model : in out String_2d) is
  begin
    Initialize(Data      => The_Model.Hidden_Data,
               Text      => The_Model.Text,
               Scale     => The_Model.Size_Scaling,
               Fontsize  => The_Model.Font_Size,
               Font      => The_Model.Font,
               Alignment => The_Model.Alignment);
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust(The_Model : in out String_2d) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out String_2d;
                       Chars     : in     Positive) is
  begin                         
    Initialize(Data      => The_Model.Hidden_Data,
               Text      => The_Model.Text,
               Chars     => Chars,
               Scale     => The_Model.Size_Scaling,
               Fontsize  => The_Model.Font_Size,
               Font      => The_Model.Font,
               Alignment => The_Model.Alignment);
  end initialize;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_2d) is
  begin
    Initialize_Render(The_Model);
    Render(Text => The_Model.Text,
           Font => The_Model.Font,
           Data => The_Model.Hidden_Data);
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_2d;
                   Chars     : in     Positive;
                   Increment : in     Positive := 1) is
  begin
    Initialize_Render(The_Model);
    Render(Text      => The_Model.Text,
           Font      => The_Model.Font,
           Data      => The_Model.Hidden_Data,
           Chars     => Chars,
           Increment => Increment,
           Scale     => The_Model.Size_Scaling,
           Fontsize  => The_Model.Font_Size,
           Alignment => The_Model.Alignment);                    
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_2d_Length_1_array) is
  begin
    for I in The_Model'range loop
      Render(The_Model(I));  
    end loop;
  end Render;
  ----------------------------------------------------------------------------  
  procedure Render (The_Model : in out String_2d_Access) is
  begin
    Render(The_Model.all);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_2d_Array) is
  begin
    for I in The_Model'range loop
      Render(The_Model(I));
    end loop;
  end Render;
  ----------------------------------------------------------------------------  
  procedure Initialize(The_Model : in out String_3d) is
  begin
    Initialize(Data      => The_Model.Hidden_Data,
               Text      => The_Model.Text,
               Scale     => The_Model.Size_Scaling,
               Fontsize  => The_Model.Font_Size,
               Font      => The_Model.Font,
               Alignment => The_Model.Alignment);
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust    (The_Model : in out String_3d) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out String_3d;
                       Chars     : in     Positive) is
  begin                         
    Initialize(Data      => The_Model.Hidden_Data,
               Text      => The_Model.Text,
               Chars     => Chars,
               Scale     => The_Model.Size_Scaling,
               Fontsize  => The_Model.Font_Size,
               Font      => The_Model.Font,
               Alignment => The_Model.Alignment);
  end initialize;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_3d) is
  begin
    Motion_Sequence := Rotate_First;
    Initialize_Render(The_Model);
    Render(Text => The_Model.Text,
           Font => The_Model.Font,
           Data => The_Model.Hidden_Data);
    Finalize_Render(The_Model);
    Motion_Sequence := Translate_First;
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_3d;
                   Chars     : in     Positive;
                   Increment : in     Positive := 1) is
  begin
    Motion_Sequence := Rotate_First;
    Initialize_Render(The_Model);
    Render(Text      => The_Model.Text,
           Font      => The_Model.Font,
           Data      => The_Model.Hidden_Data,
           Chars     => Chars,
           Increment => Increment,
           Scale     => The_Model.Size_Scaling,
           Fontsize  => The_Model.Font_Size,
           Alignment => The_Model.Alignment);                    
    Finalize_Render(The_Model);
    Motion_Sequence := Translate_First;
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_3d_Length_1_array) is
  begin
    for I in The_Model'range loop
      Render(The_Model(I));  
    end loop;
  end Render;
  ----------------------------------------------------------------------------
  procedure Render (The_Model : in out String_3d_Access) is
  begin
    Render(The_Model.all);
  end Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out String_3d_Array) is
  begin
    for I in The_Model'range loop
      Render(The_Model(I));
    end loop;
  end Render;
  ----------------------------------------------------------------------------  
  function Background_Of
          (The_String       : in String_2d;
           Background_Color : in Types.Color_Type := Default_Background_Color;
           Frame            : in Boolean          := True;
           Frame_Color      : in Types.Color_Type := Default_Frame_Color;
           Blanks           : in boolean          := true) 
    return models.forms.rectangle_and_frame_2d is
    
    Dimension  : Types.Vector_2d := Types.Vector_2d
                (Fonts.Outer_Rectangle_Of
                   (The_String.Text, The_String.Font, Blanks));   
    Line_Width : Gl.Glfloat := 0.0;            
  begin
    Dimension(X) := Gl.Glfloat'Max
      (Dimension(X), Fonts.Max_Character_Width_Of(The_String.Font));                      
    -- Does only work for the_string.alignment = (center, middle).  
    Dimension := 2.0 * The_String.Size_Scaling * The_String.Font_Size *
                      (X => 0.67 * Dimension(X), Y => Dimension(Y));                
    if Frame then
      Line_Width := 0.1 * Dimension(Y);
    end if;                    
    
    return Forms.Rectangle_And_Frame
          (Position         => The_String.Position,
           Width            => Dimension(X),
           Height           => Dimension(Y), 
           Frame_Line_Width => Line_Width,
           Background_Color => Background_Color,
           Frame_Color      => Frame_Color);                                                                                         
  end Background_Of;                             
  ----------------------------------------------------------------------------                                     
  function Background_Of
          (The_String       : in String_2d;
           Background_Color : in Types.Color_Type := Default_Background_Color;
           Frame            : in Boolean          := True;
           Frame_Color      : in Types.Color_Type := Default_Frame_Color;
           Blanks           : in boolean          := true) 
    return Gl.Gluint is
      
    obj : forms.rectangle_and_frame_2d := Background_Of
          (the_string, Background_Color, frame, Frame_Color, blanks);          
    The : constant Gl.Gluint := Display_Lists.Next_List_Id;                              
  begin
    Gl.Glnewlist(The, Mode => Gl.Gl_Compile);    
    forms.render(obj);
    Gl.Glendlist;      
    return The;
  end Background_Of;                             
  ----------------------------------------------------------------------------                                 
end Opengl.Models.Strings;
