--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-strings.ads
-- Description     : Package for the definition of string-models. 
-- Author          : H. Altmann, A. Gaisbauer, J. Schrer
-- Created On      : 09. Feb. 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Opengl.Fonts;
with opengl.models.forms;

package Opengl.Models.Strings is

  type Horizontal_Alignment_Type is (Left,   Center, Right);
  type Vertical_Alignment_Type   is (Bottom, Middle, Top);

  type String_Alignment is record
    Horizontal : Horizontal_Alignment_Type := Center;
    Vertical   : Vertical_Alignment_Type   := Middle;
  end record;

  -- In order to spezify a string-location flexibly, a reference point for 
  -- the position-component has to be given using the above enum-types.

  type Hidden_Data_Type is private;

  Default_Hidden_Data : constant Hidden_Data_Type;

  -- Data that is calculated once during initialization, 
  -- and used by the render procedure.

  -- String 2D

  type String_2d(Length : Natural) is new Model_2d_Z with record
    Text        : String(1 .. Length)  := (others => ' ');
    Font        : Fonts.Font_Type      := Fonts.Font;
                  -- available after windows initialization
    Font_Size   : Fonts.Font_Size_Type := 10.0;
    Alignment   : String_Alignment;
    Hidden_Data : Hidden_Data_Type;
  end record;

  -- Text       : is the text to be displayed.
  -- Font       : is the first list index of the opengl-calllists for the font.
  -- Font_size  : and component size_scaling are used for the size of the string.
  -- Alignment  : Is used to calculate the Reference_Point.
  -- Hidden_Data: see in private part.


  procedure Initialize(The_Model : in out String_2d);
  procedure Adjust    (The_Model : in out String_2d);

  procedure Render(The_Model : in out String_2d);

  procedure Initialize(The_Model : in out String_2d;
                       Chars     : in     Positive);

  procedure Render(The_Model : in out String_2d;
                   Chars     : in     Positive;
                   Increment : in     Positive := 1);

  -- String array with strings all having length 1

  subtype String_2d_Length_1 is String_2d(Length => 1);

  type String_2d_Length_1_array is array (Natural range <>) of String_2d_Length_1;

  procedure Render(The_Model : in out String_2d_Length_1_array);

  type String_2d_Access is access all String_2d;
  
  procedure Free(Item : in out String_2d_Access);

  procedure Write(Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   : in     String_2d_Access);

  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out String_2d_Access);

  for String_2d_Access'Write use Write;
  for String_2d_Access'Read  use Read;

  procedure Render (The_Model : in out String_2d_Access);

  -- String array for strings of different length.

  type String_2d_Array is array (Natural range <>) of String_2d_Access;

  procedure Render(The_Model : in out String_2d_Array);

  
  type String_3d(Length : Natural) is new Model_3d with record
    Text        : String(1 .. Length)  := (others => ' ');
    Font        : Fonts.Font_Type      := Fonts.Font;
                  -- available after windows initialization
    Font_Size   : Fonts.Font_Size_Type := 10.0;
    Alignment   : String_Alignment;
    Hidden_Data : Hidden_Data_Type;
  end record;

  -- Text       : is the text to be displayed.
  -- Font       : is the first list index of the opengl-calllists for the font.
  -- Font_size  : and component size_scaling are used for the size of the string.
  -- Alignment  : Is used to calculate the Reference_Point.
  -- Hidden_Data: see in private part.


  procedure Initialize(The_Model : in out String_3d);
  procedure Adjust    (The_Model : in out String_3d);

  procedure Render(The_Model : in out String_3d);

  procedure Initialize(The_Model : in out String_3d;
                       Chars     : in     Positive);

  procedure Render(The_Model : in out String_3d;
                   Chars     : in     Positive;
                   Increment : in     Positive := 1);

  -- String array with strings all having length 1

  subtype String_3d_Length_1 is String_3d(Length => 1);

  type String_3d_Length_1_array is array (Natural range <>) of String_3d_Length_1;

  procedure Render(The_Model : in out String_3d_Length_1_array);


  type String_3d_Access is access all String_3d;

  procedure Free(Item : in out String_3d_Access);

  procedure Write(Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   : in     String_3d_Access);

  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out String_3d_Access);

  for String_3d_Access'Write use Write;
  for String_3d_Access'Read  use Read;

  procedure Render (The_Model : in out String_3d_Access);

  -- String array for strings of different length.

  type String_3d_Array is array (Natural range <>) of String_3d_Access;

  procedure Render(The_Model : in out String_3d_Array);
  
  ----------------------------------------------------------------------------

  Default_Frame_Color      : constant Types.Color_Type := Colors.White;
  Default_Background_Color : constant Types.Color_Type := Colors.Black;


  function Background_Of
          (The_String       : in String_2d;
           Background_Color : in Types.Color_Type := Default_Background_Color;
           Frame            : in Boolean          := True;
           Frame_Color      : in Types.Color_Type := Default_Frame_Color;
           Blanks           : in boolean          := true) 
    return models.forms.rectangle_and_frame_2d;
    
                               
  function Background_Of
          (The_String       : in String_2d;
           Background_Color : in Types.Color_Type := Default_Background_Color;
           Frame            : in Boolean          := True;
           Frame_Color      : in Types.Color_Type := Default_Frame_Color;
           Blanks           : in boolean          := true) 
    return Gl.Gluint;
    
  -- Returns (a display-list for) a rectangle around a given string.
  -- If Frame then a frame with different color is drawn around the rectangle.    
    

private ----------------------------------------------------------------------                               

  type Hidden_Data_Type is record
    Reference_Point : Types.Vector_2d := (0.0, 0.0);
    Scale_Factor    : Gl.Glfloat      := 0.0;
    First           : positive        := 1;
  end record;

  -- Reference_Point: Initialize computes the Reference_Point using alignment.
  --                  Render translates to The_Model.Hidden_Data.Reference_Point
  --                  before rendering.             
  --                  E.g. for Alignment := (horizontal => center, vertical => middle),
  --                  the Reference_Point is calculated by initialize automatically via
  --
  --                  The_model.Hidden_Data.Reference_Point := - Center_Of(The_Model); 
  --
  --                  (The rendering of characters/strings always starts at the
  --                   lower-left point).
  --
  -- Scale_Factor   : is calculated via
  --
  --                    The_Model.Size_Scaling * The_Model.Font_Size / 
  --                    Fonts.Max_Character_Height_Of(The_Model.Font)
  --
  -- First          : is used for the scrolling Render.

  Default_Hidden_Data : constant Hidden_Data_Type :=
    (Reference_Point => (0.0, 0.0),
     Scale_Factor    => 0.0,
     First           => 1);

end Opengl.Models.Strings;
