--                    -*- Mode: Ada -*-
-- Filename        : opengl-timer.ads
-- Description     : Package for measuring opengl graphics performance. 
-- Author          : J. Schrer
-- Created On      : 23. Feb. 1998
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Calendar;
with Ada.Integer_Text_Io, Ada.Float_Text_Io;
with Opengl.Command_Line;
with Opengl.Fonts;
with Opengl.Gl;
with Opengl.Colors;
with Opengl.Display_Lists;
with Opengl.Logging;
with Opengl.Float_32.Arrays.Vector_Operations;
with Opengl.Float_32.Arrays.Types;

package body Opengl.Timer is

  use type Ada.Strings.Unbounded.String_Access,
           Ada.Calendar.Time,
           Ada.Real_Time.Time,
           Ada.Real_Time.Time_Span,
           Gl.Glfloat,
           Opengl.Float_32.Arrays.Types.Vector,
           Models.Strings.String_2d_Access;
  use Opengl.Float_32.Arrays.Vector_Operations;
  
  Last_Time     : Ada.Calendar.Time := Ada.Calendar.Clock;
  Run_Time      : Float             := 0.0; 

  subtype String_12 is String(1 .. 12);

  Count_Image   : constant String_12 := "Count:      ";
  Dt_Image      : constant String_12 := "Dt [ms]:    ";
  Mean_Dt_Image : constant String_12 := "E[Dt] [ms]: ";
  Max_Dt_Image  : constant String_12 := "Max[Dt][ms]:";
  
  Line_Count    : Line_Number_Range := 0;
  Last_Chars    : array(Line'range) of Natural := (others => 0);
  ----------------------------------------------------------------------------
  procedure Set(The_Item : in String;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True) is
                  
    First_Char : Positive := 1;                  
  begin
    if To_Line = 0 then
      Line_Count := Line_Number_Range'Max((Line_Count + 1) rem 5, 1);      
    else
      Line_Count := To_Line;
    end if;
    if Append then  
      First_Char := Last_Chars(Line_Count) + 1;        
      Last_Chars(Line_Count) := Last_Chars(Line_Count) + The_Item'Length;
    else  
      Last_Chars(Line_Count) := The_Item'Length;
    end if;      
    if Last_Chars(Line_Count) <= Line(Line_Count)'Length then
      Line(Line_Count)(First_Char .. Last_Chars(Line_Count)) := The_Item;
      if Clear then
        Line(Line_Count)(Last_Chars(Line_Count) + 1 .. Line(Line_Count)'Last) := 
                        (others => ' ');
      end if;
    else
      Line(Line_Count)(First_Char .. Line(Line_Count)'Length) := 
        The_Item(The_Item'First .. The_Item'First + 
                                   Line(Line_Count)'Length - First_Char);
    end if;
  end Set;
  ----------------------------------------------------------------------------
  procedure Set(The_Item : in Integer;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True) is
  begin
    Set(Integer'Image(The_Item), To_Line, Clear, Append);
  end Set;
  ----------------------------------------------------------------------------
  procedure Set(The_Item : in Gl.Glfloat;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True) is
  begin
    Set(Gl.Glfloat'Image(The_Item), To_Line, Clear, Append);
  end Set;
  ----------------------------------------------------------------------------
  procedure Set(The_Item : in Boolean;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True) is
  begin
    Set(Boolean'Image(The_Item), To_Line, Clear, Append);
  end Set;
  ----------------------------------------------------------------------------
  function Sec_Since_Start return Float is
    Now : constant Ada.Calendar.Time := Ada.Calendar.Clock;
  begin
    Run_Time  := Run_Time + Float(Now - Last_Time);
    Last_Time := Now;   
    return Run_Time;
  exception
    when Constraint_Error => return 0.0;      
  end Sec_Since_Start;  
  ----------------------------------------------------------------------------
  procedure Print_Statistics(The_Model : in Model) is
  begin
    Logging.put("---------------------------------------------------------------------");
    Logging.put("Timer Statistics:");
    Logging.put(Count_Image   & The_Model.Counter_Image.Text);
    Logging.put(Dt_Image      & The_Model.Dt_Image.Text);
    Logging.put(Mean_Dt_Image & The_Model.Mean_Dt_Image.Text);
    Logging.put(Max_Dt_Image  & The_Model.Max_Dt_Image.Text);
  end Print_Statistics;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out Model) is
    ----------------------
    procedure Do_Render is
      pragma Inline(Do_Render);
      
	    Time : constant Ada.Real_Time.Time := Ada.Real_Time.Clock;
	    Dt   : constant Float := Float(Ada.Real_Time.To_Duration
                              (Time - The_Model.Last_Time));      
    begin
	    The_Model.Mean_Dt := (The_Model.Mean_Dt * Float(The_Model.Counter - 1) + 
                            Dt) / Float(The_Model.Counter);
      The_Model.Max_Dt  := Float'Max(Dt, The_Model.Max_Dt);
        
      if Output then
  	    if The_Model.Counter <= The_Model.Max_Iterations and then
                                Time - The_Model.Last_Screen_Update_Time >= 
                                The_Model.Screen_Update_Dt then                           
          Ada.Integer_Text_Io.Put(To   => The_Model.Counter_Image.Text,
                                  Item => The_Model.Counter);
          Ada.Float_Text_Io.Put(To   => The_Model.Dt_Image.Text,
                                Item => Dt * 1000.0,
                                Aft  => 3,
                                Exp  => 0);
          Ada.Float_Text_Io.Put(To   => The_Model.Mean_Dt_Image.Text,
                                Item => The_Model.Mean_Dt * 1000.0,
                                Aft  => 3,
                                Exp  => 0);
          Ada.Float_Text_Io.Put(To   => The_Model.Max_Dt_Image.Text,
                                Item => The_Model.Max_Dt * 1000.0,
                                Aft  => 3,
                                Exp  => 0);
  	      The_Model.Last_Screen_Update_Time := Time;
        end if;
  
        Models.Initialize_Render(The_Model);
  
        if The_Model.Text_Image /= null then
  	      Models.Strings.Render(The_Model.Text_Image.all);
        end if;
  
        Gl.Glcalllist(The_Model.Display_List_Index);
  
  	    Models.Strings.Render(The_Model.Counter_Image);
  	    Models.Strings.Render(The_Model.Dt_Image);
  	    Models.Strings.Render(The_Model.Mean_Dt_Image);
  	    Models.Strings.Render(The_Model.Max_Dt_Image);
  
        for I in Line'range loop
          The_Model.Line_Images(I).Text := Line(I);
  	      Models.Strings.Render(The_Model.Line_Images(I));
        end loop;
  
  	    Models.Finalize_Render(The_Model);
      end if;  

      if The_Model.Include_Render_Time then
        The_Model.Last_Time := Time;
      else
        The_Model.Last_Time := Ada.Real_Time.Clock;
      end if;
    end Do_Render;
    --------------
  begin
    begin
      The_Model.Counter := The_Model.Counter + 1;
    exception
      when Constraint_Error =>
        The_Model.Counter := 1;
    end;
    if The_Model.Counter rem The_Model.Increment = 0 then
      Do_Render;
    end if;
  exception when others => null;
  end Render;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out Model) is

    Font_Size    : constant Fonts.Font_Size_Type :=
                   Fonts.Font_Size_Type(Opengl.Command_Line.Timer_Fontsize);              
    The_Font     : constant Fonts.Font_Type := Fonts.Font;--(Name   => "Courier New",
                                                          -- Weight => Fonts.Font_Weight_Normal);
    Offset       : constant Gl.Glfloat :=
                   The_Model.Size_Scaling * Font_Size /
                   Fonts.Max_Character_Height_Of(The_Font);
    Max_Width    : constant Gl.Glfloat := Offset *
                   Fonts.Width_Of(Max_Dt_Image, The_Font);
                    
    Counter_Text : Models.Strings.String_2d(String_12'Length);
    Dt_Text      : Models.Strings.String_2d(String_12'Length);
    Mean_Dt_Text : Models.Strings.String_2d(String_12'Length);
    Max_Dt_Text  : Models.Strings.String_2d(String_12'Length);
  begin
    The_Model.Screen_Update_Dt :=
      Ada.Real_Time.Milliseconds(The_Model.Screen_Update_Dt_Sec * 1000);
      
    The_Model.Color       := Colors.Red;
    The_Model.Position(Y) := 1.0;
    
    Counter_Text  :=
         (Models.Root_Model with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Attitude     => The_Model.Attitude,
          Position     => The_Model.Position + (0.0, 3.0 * Offset, 0.0),
          Font         => The_Font,
          Font_Size    => Font_Size,
          Length       => Count_Image'Length,
          Text         => Count_Image,
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
                    
    Dt_Text       :=
         (Models.Root_Model with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Attitude     => The_Model.Attitude,
          Position     => The_Model.Position + (0.0, 2.0 * Offset, 0.0),
          Font         => The_Font,
          Font_Size    => Font_Size,
          Length       => Dt_Image'Length,
          Text         => Dt_Image,
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
                    
    Mean_Dt_Text  :=
         (Models.Root_Model with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Attitude     => The_Model.Attitude,
          Position     => The_Model.Position + (0.0, Offset, 0.0),
          Font         => The_Font,
          Font_Size    => Font_Size,
          Length       => Mean_Dt_Image'Length,
          Text         => Mean_Dt_Image,
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
                    
    Max_Dt_Text   :=
 	       (Models.Root_Model with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Attitude     => The_Model.Attitude,
          Position     => The_Model.Position,
          Font         => The_Font,
          Font_Size    => Font_Size,
          Length       => Max_Dt_Image'Length,
          Text         => Max_Dt_Image,
          Alignment    => (Horizontal => Models.Strings.Left,
                             Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
          
	  The_Model.Counter_Image :=
         (Models.Root_Model with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Attitude     => The_Model.Attitude,
          Position     => The_Model.Position + (Max_Width, 3.0 * Offset, 0.0),
          Font         => The_Font,
          Font_Size    => Font_Size,
          Length       => 10,
          Text         => (others => '0'),
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);

	  The_Model.Dt_Image :=
	       (Models.Root_Model with
	        Size_Scaling => The_Model.Size_Scaling,
	        Color        => The_Model.Color,
	        Attitude     => The_Model.Attitude,
	        Position     => The_Model.Position + (Max_Width, 2.0 * Offset, 0.0),
	        Font         => The_Font,
	        Font_Size    => Font_Size,
	        Length       => 10,
	        Text         => (others => '0'),
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);

	  The_Model.Mean_Dt_Image :=
	       (Models.Root_Model with
	        Size_Scaling => The_Model.Size_Scaling,
	        Color        => The_Model.Color,
	        Attitude     => The_Model.Attitude,
	        Position     => The_Model.Position + (Max_Width, Offset, 0.0),
	        Font         => The_Font,
	        Font_Size    => Font_Size,
	        Length       => 10,
	        Text         => (others => '0'),
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
          
	  The_Model.Max_Dt_Image :=
	       (Models.Root_Model with
	        Size_Scaling => The_Model.Size_Scaling,
	        Color        => The_Model.Color,
	        Attitude     => The_Model.Attitude,
	        Position     => The_Model.Position + (Max_Width, 0.0, 0.0),
	        Font         => The_Font,
	        Font_Size    => Font_Size,
	        Length       => 10,
	        Text         => (others => '0'),
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);

    if The_Model.Text /= null then
      The_Model.Text_Image := new Models.Strings.String_2d'
	       (Models.Root_Model with
	        Size_Scaling => The_Model.Size_Scaling,
	        Color        => The_Model.Color,
	        Attitude     => The_Model.Attitude,
	        Position     => The_Model.Position + (0.0, 3.0 * Offset, 0.0),
	        Font         => The_Font,
	        Font_Size    => Font_Size,
	        Length       => The_Model.Text'Length,
	        Text         => The_Model.Text.all,
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
    end if;

    for I in The_Model.Line_Images'range loop
      The_Model.Line_Images(I) :=
	       (Models.Root_Model with
	        Size_Scaling => The_Model.Size_Scaling,
	        Color        => Colors.Green,
	        Attitude     => The_Model.Attitude,
	        Position     => The_Model.Position +
                         (X => 2.5 * Max_Width,
                          Y => Gl.Glfloat(I-1) * Offset,
                          Z => 0.0),
	        Font         => The_Font,
	        Font_Size    => Font_Size,
	        Length       => Info_Text'Length,
	        Text         => (others => '0'),
          Alignment    => (Horizontal => Models.Strings.Left,
                           Vertical   => Models.Strings.Bottom),
          Hidden_Data  => Models.Strings.Default_Hidden_Data);
    end loop;

    -- Create display list for constant strings.
    The_Model.Display_List_Index := Display_Lists.Next_List_Id;
    Gl.Glnewlist(List => The_Model.Display_List_Index,
                 Mode => Gl.Gl_Compile);
    Models.Strings.Render(Counter_Text);
    Models.Strings.Render(Dt_Text);
	  Models.Strings.Render(Mean_Dt_Text);
	  Models.Strings.Render(Max_Dt_Text);
    Gl.Glendlist;
  end Initialize;
  ----------------------------------------------------------------------------
end Opengl.Timer;
