--                    -*- Mode: Ada -*-
-- Filename        : opengl-fonts.adb
-- Description     : Package that holds an object for the actual font for
--                   displaying strings in Opengl.
-- Author          : J. Schrer, A. Gaisbauer.
-- Created On      : 27. Jan. 1998
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Exceptions;
with Ada.Characters.Handling;
with Ada.Strings.Unbounded;
with Ada.Text_Io;
with Ada.Float_Text_Io;
with Ada.Integer_Text_Io;
with Ada.Unchecked_Deallocation;
with Opengl.Messages;
with Opengl.Display_Lists;
with Opengl.Command_Line;
with Utilities.Generics.Hash;
with Utilities.Prime_Numbers;
--with Utilities.Command_Line;
with Utilities.Code_Generator;

package body Opengl.Fonts is

  use type Ada.Strings.Unbounded.String_Access;
  use type Interfaces.C.Int;
  use type Interfaces.C.Unsigned_Long;
  use type Gl.Glfloat;
  use type Win32.Bool;
  use type Win32.Windef.Hdc;
  use type Win32.Windef.Hfont;
  use type Win32.Windef.Hgdiobj;

  ----------------------------------------------------------------------------  
  procedure Write(File               : in Ada.Text_Io.File_Type;
                  The_Font           : in Font_Type;
                  Glyphs_Metric_Data : in Boolean := True) is
    -----------------------------------------------------------------
    procedure Put(Char : in Character;
                  Data : in Win32_Addons.Wingdi.Glyphmetricsfloat) is
      ------------------------------------
      procedure Put(F : in Win32.Float) is
      begin
        Ada.Float_Text_Io.Put(File, Float(F), Fore => 1, Aft => 9, Exp => 0);
        Ada.Text_Io.Put(File, "  ");
      end Put;
      ------------------------------------                                      
    begin
      Ada.Integer_Text_Io.Put(File, Item => Character'Pos(Char), Width => 3);
      Ada.Text_Io.Put(File, " '" & Char & "' ");
      Put(Data.Gmfblackboxx);
      Put(Data.Gmfblackboxy);
      Put(Data.Gmfptglyphorigin.X);
      Put(Data.Gmfptglyphorigin.Y);
      Put(Data.Gmfcellincx);
      Put(Data.Gmfcellincy);
      Ada.Text_Io.New_Line(File);
    end Put;
    -----------------------------------------------------------------
  begin
    Utilities.Code_Generator.Put_Lines(File,
      "--FONT created-------------------------------------------------------",
      "Name            : " & The_Font.Name,
      "Size            : " & Font_Size_Type'Image(The_Font.Size),
      "Weight          : " & Interfaces.C.Int'Image(The_Font.Weight),
      "Italic          : " & Boolean'Image(The_Font.Italic),
      "Underline       : " & Boolean'Image(The_Font.Underline),
      "Strikeout       : " & Boolean'Image(The_Font.Strikeout),
      "Charset         : " & Interfaces.C.Unsigned_Long'Image(The_Font.Charset),
      "Output-Precision: " & Interfaces.C.Unsigned_Long'Image(The_Font.Output_Precision),
      "Clip-Precision  : " & Interfaces.C.Unsigned_Long'Image(The_Font.Clip_Precision),
      "Quality         : " & Interfaces.C.Unsigned_Long'Image(The_Font.Quality),
      "Pitch-and-Family: " & Interfaces.C.Unsigned_Long'Image(The_Font.Pitch_And_Family),
      "List-Index      : " & Interfaces.C.Unsigned_Long'Image(The_Font.List_Index),
      "Deviation       : " & Interfaces.C.C_Float'Image(The_Font.Deviation),
      "Extrusion       : " & Interfaces.C.C_Float'Image(The_Font.Extrusion),
      "Max_Char_Width  : " & Gl.GlFloat'Image(The_Font.Max_Character_Width),
      "Max_Char_Height : " & Gl.GlFloat'Image(The_Font.Max_Character_Height),
      "Mean_Char_Width : " & Gl.GlFloat'Image(The_Font.Mean_Character_Width),
      "Mean_Char_Height: " & Gl.GlFloat'Image(The_Font.Mean_Character_Height));

    Ada.Text_Io.Put(File, "Format          : ");
    if    The_Font.Format = Win32_Addons.Wingdi.Wgl_Font_Polygons then
      Ada.Text_Io.Put_Line(File, "Wgl_Font_Polygons");
    elsif The_Font.Format = Win32_Addons.Wingdi.Wgl_Font_Lines    then
      Ada.Text_Io.Put_Line(File, "Wgl_Font_Lines");
    end if;

    Ada.Text_Io.Put_Line(File,
      "First_Character = '" & The_Font.First_Character & "' = " &
      Natural'Image(Character'Pos(The_Font.First_Character)) &
      " .. Last_Character = '" & The_Font.Last_Character & "' = " &
      Natural'Image(Character'Pos(The_Font.Last_Character)));

    if Glyphs_Metric_Data then
      Ada.Text_Io.Put_Line(File,
  "  N  C  GmfBlackBoxX GmfBlackBoxY Origin.X     Origin.Y     GmfCellIncX  GmfCellIncY");
--          1         2         3         4         5         6         7         8
--"12345678901234567890123456789012345678901234567890123456789012345678901234567890"      
      for Char in The_Font.Glyphs_Metric_Data'range loop
        Put(Char, The_Font.Glyphs_Metric_Data(Char));
      end loop;
    end if;

    Ada.Text_Io.Put_Line(File,
      "---------------------------------------------------------------------");
  end Write;
  ----------------------------------------------------------------------------    
  function Is_Equal(Left, Right : in Ada.Strings.Unbounded.String_Access)
           return Boolean is
  begin
    if     Left = null then
      return Right = null;
    elsif Right = null then
      return Left = null;
    else
      return Left = Right or else Left.all = Right.all;
    end if;
  end Is_Equal;
  ----------------------------------------------------------------------------  
  procedure Free (Item : in out Font_Type) is
    procedure Deallocate is new Ada.Unchecked_Deallocation
             (Object => Opengl_Outline_Font_Type, Name => Font_Type);
  begin
    null;
--    Deallocate(Item);                 
  end Free;
  ----------------------------------------------------------------------------  
  package Font_Hash is new Utilities.Generics.Hash
         (Element            => Font_Type,
          Key                => Ada.Strings.Unbounded.String_Access,
          Index              => Natural,
          Is_Equal           => Is_Equal,
          Hash_Function      => Utilities.Prime_Numbers.Standard_Hash_Function,
          Deallocate_Element => Free,
          Deallocate_Key     => Ada.Strings.Unbounded.Free);
  ----------------------------------------------------------------------------
  Font_Table : Font_Hash.Table(Length => Utilities.Prime_Numbers.
                                         Give_Next_Prime(Near => 100) - 1);

  The_Device_Context : Win32.Windef.Hdc := System.Null_Address;
  ----------------------------------------------------------------------------
  function Is_Initialized(The_Font : in Font_Type) return Boolean is 
  begin
    return The_Font /= null;
  end Is_Initialized;
  ----------------------------------------------------------------------------
  function Font(Name            : in Font_Name_Type   := Default_Font_Name;
                First_Character : in Character        := Default_First_Character;
                Last_Character  : in Character        := Default_Last_Character;
                Size            : in Font_Size_Type   := Default_Font_Size;
                Weight          : in Font_Weight_Type := Default_Font_Weight;
                Italic          : in Boolean          := False;
                Underline       : in Boolean          := False;
                Strikeout       : in Boolean          := False;
                Deviation       : in Gl.Glfloat       := Default_Deviation;
                Extrusion       : in Gl.Glfloat       := Default_Extrusion)
           return Font_Type is

    Font_Name : constant String :=
                Ada.Characters.Handling.To_Upper
                  (Ada.Strings.Fixed.Trim(Source => Name,
                                          Side   => Ada.Strings.Both));
                                          
    Hash_Name : aliased String :=
                Font_Name & First_Character & Last_Character &
                Ada.Strings.Fixed.Trim(Font_Size_Type'Image(Size),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(Font_Weight_Type'Image(Weight),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(Natural'Image(Boolean'Pos(Italic)),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(Natural'Image(Boolean'Pos(Underline)),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(Natural'Image(Boolean'Pos(Strikeout)),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(gl.glfloat'Image(Deviation),
                                       Side => Ada.Strings.Both) &
                Ada.Strings.Fixed.Trim(gl.glfloat'Image(Extrusion),
                                       Side => Ada.Strings.Both);
                                       
    The_Font  : Font_Type;
  begin
    if Font_Hash.Is_In_Table(The_Table => Font_Table,
                             The_Key   => Hash_Name'Unchecked_Access) then
      The_Font := Font_Hash.Get_Element
                  (The_Table => Font_Table,
                   The_Key   => Hash_Name'Unchecked_Access);
    else
      if The_Device_Context = System.Null_Address then
        Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
          (Subprogram => "OpenGL.Fonts.Font",
           Message    => "The_Device_Context = null"));
      end if;

      Create_New_Font:
      declare
        Font_Name_C : constant Interfaces.C.Char_Array := 
                               Interfaces.C.To_C(Font_Name);
        Width, 
        Height      : gl.glfloat;
        count       : natural := 0;                       
      begin
        The_Font := new Opengl_Outline_Font_Type
                       (First_Character => First_Character,
                        Last_Character  => Last_Character);

        The_Font.Name       := Font_Name &
                              (Font_Name_Length - Font_Name'Length) * ' ';
        The_Font.Size       := Size;
        The_Font.Weight     := Interfaces.C.Int(Weight);
        The_Font.Italic     := Italic;
        The_Font.Underline  := Underline;
        The_Font.Strikeout  := Strikeout;
        The_Font.List_Index :=
          Interfaces.C.Unsigned_Long
            (Display_Lists.Next_List_Id
               (Required_Number => The_Font.Glyphs_Metric_Data'Length));

        --The_Font.Output_Precision := Win32.Wingdi.Out_Tt_Only_Precis;
        --The_Font.Quality          := Win32.Wingdi.Proof_Quality;
        
        The_Font.Deviation := Interfaces.C.C_Float(Deviation);
        The_Font.Extrusion := Interfaces.C.C_Float(Extrusion);                          

        The_Font.Handle := Win32.Wingdi.Createfont
  	      (Nheight            => Interfaces.C.Int(- The_Font.Size),
  	       Nwidth             => 0,
  	       Nescapement        => 0,
  	       Norientation       => 0,
  	       Fnweight           => The_Font.Weight,
  	       Fdwitalic          => Boolean'Pos(The_Font.Italic),
  	       Fdwunderline       => Boolean'Pos(The_Font.Underline),
  	       Fdwstrikeout       => Boolean'Pos(The_Font.Strikeout),
  	       Fdwcharset         => The_Font.Charset,
  	       Fdwoutputprecision => The_Font.Output_Precision,
  	       Fdwclipprecision   => The_Font.Clip_Precision,
  	       Fdwquality         => The_Font.Quality,
  	       Fdwpitchandfamily  => The_Font.Pitch_And_Family,
  	       Lpszface           => Font_Name_C(Font_Name_C'First)'Unchecked_Access);

        if The_Font.Handle = System.Null_Address then
          Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
            (Subprogram => "OpenGL.Fonts.Font",
             Message    => "Win32.Windgi.CreateFont failed"));
        end if;

        if Win32.Wingdi.Selectobject
               (Hdc     => The_Device_Context,
                Hgdiobj => The_Font.Handle) = System.Null_Address then
          Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
            (Subprogram => "OpenGL.Fonts.Font",
             Message    => "Win32.Windgi.SelectObject failed (return value is null"));
        end if;

        if Win32_Addons.Wingdi.Wglusefontoutlines
               (Hdc        => The_Device_Context,
                First      => Character'Pos(The_Font.First_Character),
     		    Count      => The_Font.Glyphs_Metric_Data'Length,
                Listbase   => The_Font.List_Index,
                Deviation  => The_Font.Deviation,
                Extrusion  => The_Font.Extrusion,
                Format     => The_Font.Format,
                Lpgmf      => The_Font.Glyphs_Metric_Data
                             (The_Font.Glyphs_Metric_Data'First)'Unchecked_Access)
            = Win32.False then

    	    Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
            (Subprogram => "OpenGL.Fonts.Font",
             Message    => "Win32_Addons.Windgi.wglUseFontOutlines failed"));
        end if;

        The_Font.List_Index := The_Font.List_Index -
                               Character'Pos(The_Font.First_Character);

        for Char in The_Font.Glyphs_Metric_Data'range loop
          if Ada.Characters.Handling.Is_Graphic(Char) then
            Count  := Count + 1;
            Width  := Gl.Glfloat(The_Font.Glyphs_Metric_Data(Char).Gmfcellincx);
            Height := Gl.Glfloat(The_Font.Glyphs_Metric_Data(Char).Gmfblackboxy);
            The_Font.Max_Character_Width   := 
               Gl.Glfloat'Max(The_Font.Max_Character_Width, Width);
            The_Font.Max_Character_Height  := 
               Gl.Glfloat'Max(The_Font.Max_Character_Height, Height);
            The_Font.Mean_Character_Width  := The_Font.Mean_Character_Width  +
                                              Width;
            The_Font.Mean_Character_Height := The_Font.Mean_Character_Height +
                                              Height;                
          end if;
        end loop;
        
        The_Font.Mean_Character_Width  := The_Font.Mean_Character_Width  / 
                                          gl.glfloat(Count);
        The_Font.Mean_Character_Height := The_Font.Mean_Character_Height / 
                                          gl.glfloat(Count);        
      end Create_New_Font;

      if Ada.Text_Io.Is_Open(Log_File) then
        Write(Log_File, The_Font, Glyphs_Metric_Data => Command_Line.Font_Info);
      end if;

      Font_Hash.Insert(The_Table   => Font_Table,
                       The_Key     => new String'(Hash_Name),
                       The_Content => The_Font);
    end if;

    return The_Font;
  end Font;
  ----------------------------------------------------------------------------
  procedure Write(Filename           : in String;
                  The_Font           : in Font_Type;
                  Glyphs_Metric_Data : in Boolean := True) is
    File : Ada.Text_Io.File_Type;
  begin
    Ada.Text_Io.Create(File, Ada.Text_Io.Append_File, Filename);
    Write(File, The_Font, Glyphs_Metric_Data);
    Ada.Text_Io.Close(File);
  end Write;
  ----------------------------------------------------------------------------     
  function All_Created_Fonts return Font_Type_Array_Access is

    Font_Number : constant Natural :=
                  Font_Hash.Number_Of_Inserted_Elements(Font_Table);
    The_Strings : Font_Hash.Key_Array    (1 .. Font_Number);
    The_Fonts   : Font_Hash.Element_Array(1 .. Font_Number);
    Fonts       : Font_Type_Array_Access;
  begin
    Font_Hash.Save_Content(The_Table    => Font_Table,
                           The_Keys     => The_Strings,
                           The_Elements => The_Fonts);
    Fonts := new Font_Type_Array(The_Fonts'range);
    for I in Fonts'range loop
      Fonts(I) := The_Fonts(I);
    end loop;
    return Fonts;
  end All_Created_Fonts;
  ----------------------------------------------------------------------------     
  procedure Set(Device_Context : in Win32.Windef.Hdc) is
    Standard_Font : Font_Type;
  begin
--    Font_Hash.Deallocate(Font_Table);
    The_Device_Context := Device_Context;
    Standard_Font      := Font;
  end Set;
  ----------------------------------------------------------------------------     
  procedure Get(Device_Context : out Win32.Windef.Hdc) is
  begin
    Device_Context := The_Device_Context;
  end Get;
  ----------------------------------------------------------------------------     
  function List_Index_Of(The_Font : in Font_Type)
           return Interfaces.C.Unsigned_Long is
  begin
    return The_Font.List_Index;
  end List_Index_Of;
  ----------------------------------------------------------------------------                
  function Max_Character_Width_Of(The_Font : in Font_Type)
           return Gl.Glfloat is
  begin
    return The_Font.Max_Character_Width;
  end Max_Character_Width_Of;
  ----------------------------------------------------------------------------
  function Max_Character_Height_Of(The_Font : in Font_Type)
           return Gl.Glfloat is
  begin
    return The_Font.Max_Character_Height;
  end Max_Character_Height_Of;
  ----------------------------------------------------------------------------
  function Mean_Character_Width_Of(The_Font : in Font_Type)
           return Gl.Glfloat is
  begin
    return The_Font.Mean_Character_Width;
  end Mean_Character_Width_Of;
  ----------------------------------------------------------------------------
  function Mean_Character_Height_Of(The_Font : in Font_Type)
           return Gl.Glfloat is
  begin
    return The_Font.Mean_Character_Height;
  end Mean_Character_Height_Of;
  ----------------------------------------------------------------------------
  function Width_Of (The_Character : in Character;
                     The_Font      : in Font_Type) return Gl.Glfloat is
  begin
    return Gl.Glfloat(The_Font.Glyphs_Metric_Data(The_Character).Gmfcellincx);
  end Width_Of;
  ----------------------------------------------------------------------------
  function Height_Of(The_Character : in Character;
                     The_Font      : in Font_Type) return Gl.Glfloat is
  begin
    return Gl.Glfloat(The_Font.Glyphs_Metric_Data(The_Character).Gmfblackboxy);
  end Height_Of;
  ----------------------------------------------------------------------------
  function Width_Of (The_String    : in String;
                     The_Font      : in Font_Type;
                     Blanks        : in boolean := true) return Gl.Glfloat is
    Width : Gl.Glfloat := 0.0;
  begin
    for I in The_String'range loop
      if Blanks or else The_String(I) /= ' ' then
        Width := Width + Gl.Glfloat
          (The_Font.Glyphs_Metric_Data(The_String(I)).Gmfcellincx);
      end if;    
    end loop;    
    return Width;
  end Width_Of;
  ----------------------------------------------------------------------------
  function Height_Of(The_String    : in String;
                     The_Font      : in Font_Type)    return Gl.Glfloat is
    Height : Gl.Glfloat := 0.0;
  begin
    for I in The_String'range loop
      Height := Gl.Glfloat'Max(Height, Gl.Glfloat
        (The_Font.Glyphs_Metric_Data(The_String(I)).Gmfblackboxy));
    end loop;
    return Height;
  end Height_Of;
  ----------------------------------------------------------------------------
  function Center_Of(The_Character : in Character;
                     The_Font      : in Font_Type) return Types.Vector_2d is
  begin
    return (X => Width_Of (The_Character, The_Font) / 2.0,
            Y => Height_Of(The_Character, The_Font) / 2.0);
  end Center_Of;
  ----------------------------------------------------------------------------
  function Center_Of(The_String    : in String;
                     The_Font      : in Font_Type;
                     Blanks        : in boolean := true) 
           return Types.Vector_2d is
  begin
    return (X => Width_Of (The_String, The_Font, Blanks) / 2.0,
            Y => Height_Of(The_String, The_Font)         / 2.0);
  end Center_Of;
  ----------------------------------------------------------------------------
  function Outer_Rectangle_Of(The_Character : in Character;
                              The_Font      : in Font_Type)
           return Types.Rectangle is
  begin
    return Types.Rectangle'(X => Width_Of (The_Character, The_Font),
                            Y => Height_Of(The_Character, The_Font));
  end Outer_Rectangle_Of;
  ----------------------------------------------------------------------------
  function Outer_Rectangle_Of(The_String    : in String;
                              The_Font      : in Font_Type;
                              Blanks        : in boolean := true)
           return Types.Rectangle is
  begin
    return Types.Rectangle'(X => Width_Of (The_String, The_Font, Blanks),
                            Y => Height_Of(The_String, The_Font));
  end Outer_Rectangle_Of;
  ----------------------------------------------------------------------------
end Opengl.Fonts;
