--                    -*- Mode: Ada -*-
-- Filename        : opengl-fonts.ads
-- Description     : Package that holds an object for the actual font for
--                   displaying strings in Opengl.
-- Author          : J. Schrer, A. Gaisbauer.
-- Created On      : 27. Jan. 1998
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Strings.Fixed;
with Interfaces.C;
with Opengl.Gl;
with Opengl.Types;
with Opengl.Win_Stuff;
with Win32.Windef;
with Win32.Wingdi;
with Win32_Addons.Wingdi;
with System;

package Opengl.Fonts is

  Error : exception;

  -- Is raised in the following subprograms if any error occurs, e.g. if a
  -- subprogram is called with a font_type parameter, that is not created
  -- by a call of function font before.

--  type Font_Type is 
--    (Arial,
--     Times_Roman,
--     Courier_New);
  -- Add fonds as needed.

  subtype Font_Name_Type is String;--(1 .. 31);
  -- Font_name can be maximum 32 characters with the terminating nul in C.
  -- (see description of parameter lpszFace for function 
  -- Win32.WinGDI.CreateFont)

  subtype Font_Size_Type   is Gl.Glfloat range 0.0 .. 100.0;
  subtype Font_Weight_Type is Win_Stuff.Font_Weight_Type;

  Default_Font_Name   : constant Font_Name_Type   := Win_Stuff.Default_Font_Name;
  Default_Font_Size   : constant Font_Size_Type   := 10.0;
  
  Font_Weight_Bold    : constant Font_Weight_Type := 
                                 Win_Stuff.Font_Weight_Bold;
  Font_Weight_Normal  : constant Font_Weight_Type := 
                                 Win_Stuff.Font_Weight_Normal;
  Font_Weight_Thin    : constant Font_Weight_Type := 
                                 Win_Stuff.Font_Weight_Thin;
  
  Default_Font_Weight : constant Win_Stuff.Font_Weight_Type := Font_Weight_Bold;

  Default_First_Character : constant Character := ' '; -- 32
  Default_Last_Character  : constant Character := Character'Val(255);
  
  Default_Deviation       : constant Gl.Glfloat := 1.0;
  Default_Extrusion       : constant Gl.Glfloat := 0.0;

  ----------------------------------------------------------------------------
  type Font_Type is private;
  ----------------------------------------------------------------------------

  function Is_Initialized(The_Font : in Font_Type) return Boolean;
  
  -- Returns true if the_font is created by function font below or assigned
  --   a value of a font_type-variable created by the function.
  -- Returns false if the_font is only declared.

  function Font(Name            : in Font_Name_Type   := Default_Font_Name;
                First_Character : in Character        := Default_First_Character;
                Last_Character  : in Character        := Default_Last_Character;
                Size            : in Font_Size_Type   := Default_Font_Size;
                Weight          : in Font_Weight_Type := Default_Font_Weight;
                Italic          : in Boolean          := False;
                Underline       : in Boolean          := False;
                Strikeout       : in Boolean          := False;
                Deviation       : in Gl.Glfloat       := Default_Deviation;
                Extrusion       : in Gl.Glfloat       := Default_Extrusion)
           return Font_Type;

  -- Returns a font_type with the given parameters.
  -- All other parameters are fixed (see private part).

  procedure Write(Filename           : in String;
                  The_Font           : in Font_Type;
                  Glyphs_Metric_Data : in Boolean := True);

  -- Writes the attributes of the_font in readable format to file filename.    
  -- If Glyphs_Metric_Data = true writes the Glyphs_Metric_Data of the_font, 
  -- else writes only the definition parameters.

  type Font_Type_Array is array(Positive range <>) of Font_Type;

  type Font_Type_Array_Access is access Font_Type_Array;


  function All_Created_Fonts return Font_Type_Array_Access;

  -- Returns all fonts, that are created during the model creation phase.

  procedure Set(Device_Context : in Win32.Windef.Hdc);

  -- Is called from the window-creation callback after the Device_Context
  -- is created. The Device_Context is needed before function font may be
  -- called by any model-creation procedure.                    
  --
  -- Set also calls function Font, to create a font_type object with the 
  -- default parameters. Reason: When a font is created, after gl.glnewlist
  -- was called but before a call to gl.glendlist, wglusefontoutlines failes
  -- in function font. Between the calls of gl.glnewlist and gl.glendlist no
  -- new font may be created, but an existing one may be retrieved by a call
  -- of function font.

  procedure Get(Device_Context : out Win32.Windef.Hdc);

  -- Is called at the end of the window-creation callback after all models
  -- are created. 

  function List_Index_Of(The_Font : in Font_Type)
           return Interfaces.C.Unsigned_Long;

  -- Gives the list index of the_font, that is used by Gllistbase 
  -- to render a string.           

  ----------------------------------------------------------------------------                    

  function Max_Character_Width_Of  (The_Font : in Font_Type) return Gl.Glfloat;
  function Max_Character_Height_Of (The_Font : in Font_Type) return Gl.Glfloat;

  function Mean_Character_Width_Of (The_Font : in Font_Type) return Gl.Glfloat;
  function Mean_Character_Height_Of(The_Font : in Font_Type) return Gl.Glfloat;

  pragma Inline(Max_Character_Height_Of);

  -- Return the width or heigth of the_font.

  function Width_Of (The_Character : in Character;
                     The_Font      : in Font_Type)          return Gl.Glfloat;

  function Height_Of(The_Character : in Character;
                     The_Font      : in Font_Type)          return Gl.Glfloat;

  -- Return the width or heigth of The_Character.

  function Width_Of (The_String    : in String;
                     The_Font      : in Font_Type;
                     Blanks        : in boolean := true)    return Gl.Glfloat;

  -- Returns the width of the_string (the sum of the widths of all characters).
  -- If Blanks = true adds the width of all blanks else omits them.

  function Height_Of(The_String    : in String;
                     The_Font      : in Font_Type)          return Gl.Glfloat;

  -- Returns the height of the highest character in the_string.

  function Center_Of(The_Character : in Character;
                     The_Font      : in Font_Type) return Types.Vector_2d;

  function Center_Of(The_String    : in String;
                     The_Font      : in Font_Type;
                     Blanks        : in boolean := true) 
           return Types.Vector_2d;

  -- Returns the center of the_character/the_string 
  -- in relation to the lower left corner.
  -- If Blanks = true adds the width of all blanks else omits them.

  function Outer_Rectangle_Of(The_Character : in Character;
                              The_Font      : in Font_Type)
           return Types.Rectangle;

  function Outer_Rectangle_Of(The_String    : in String;
                              The_Font      : in Font_Type;
                              Blanks        : in boolean := true)
           return Types.Rectangle;

  -- Return (widht_of(), height_of()).
  -- If Blanks = true adds the width of all blanks else omits them.

private ----------------------------------------------------------------------

  Font_Name_Length : constant := 31;

  use Ada.Strings.Fixed; -- to make natural '*' character directly visible.

  type Logical_Font_Type is tagged record
    Name            : Font_Name_Type(1 .. Font_Name_Length) :=
                      Default_Font_Name &
                     (Font_Name_Length - Default_Font_Name'Length) * ' ';
    Size            : Font_Size_Type             := Default_Font_Size;
    Weight          : Interfaces.C.Int           := Interfaces.C.Int
                                                   (Default_Font_Weight);
    Italic          : Boolean                    := False;
    Underline       : Boolean                    := False;
    Strikeout       : Boolean                    := False;
    Charset         : Interfaces.C.Unsigned_Long := Win32.Wingdi.Ansi_Charset;
--    Output_Precision: Interfaces.C.Unsigned_Long := Win32.Wingdi.Out_Default_Precis;
    Output_Precision: Interfaces.C.Unsigned_Long := Win32.Wingdi.Out_Tt_Only_Precis;
    Clip_Precision  : Interfaces.C.Unsigned_Long := Win32.Wingdi.Clip_Default_Precis;
--    Quality         : Interfaces.C.Unsigned_Long := Win32.Wingdi.Default_Quality;
    Quality         : Interfaces.C.Unsigned_Long := Win32.Wingdi.Proof_Quality;
    Pitch_And_Family: Interfaces.C.Unsigned_Long := Win32.Wingdi.Default_Pitch;
  end record;

  -- "Logical font" attributes, see win32.wingdi.logfont


  type Glyphs_Metric_Data_Type is array(Character range <>) of
       aliased Win32_Addons.Wingdi.Glyphmetricsfloat;


  type Opengl_Outline_Font_Type(First_Character, Last_Character : Character)
       is new Logical_Font_Type with record

    Handle               : Win32.Windef.Hfont         := System.Null_Address;
    List_Index           : Interfaces.C.Unsigned_Long := 
                           Interfaces.C.Unsigned_Long'Last;
    Deviation            : Interfaces.C.C_Float       := 
                           Interfaces.C.C_Float(Default_Deviation);
    Extrusion            : Interfaces.C.C_Float       := 
                           Interfaces.C.C_Float(Default_Extrusion);
    Format               : Interfaces.C.Int           := 
                           Win32_Addons.Wingdi.Wgl_Font_Polygons;
    Glyphs_Metric_Data   : Glyphs_Metric_Data_Type
                          (First_Character .. Last_Character);

    Max_Character_Width  : Gl.Glfloat := 0.0;
    Max_Character_Height : Gl.Glfloat := 0.0;
    Mean_Character_Width : Gl.Glfloat := 0.0;
    Mean_Character_Height: Gl.Glfloat := 0.0;
  end record;

  -- See parameters of function win32_addons.Wglusefontoutlines

  type Font_Type is access Opengl_Outline_Font_Type;


  pragma Inline(Is_Initialized);
  pragma Inline(List_Index_Of);
  pragma Inline(Max_Character_Width_Of);
  pragma Inline(Max_Character_Height_Of);

end Opengl.Fonts;
