--                              -*- Mode: Ada -*-
-- Filename        : opengl-win.adb
-- Description     : Package with utilities to setup an opengl-window
--                   for MS-Windows using the Win32Ada-binding
-- Author          : J. Schrer
-- Created On      : 17 Nov. 1997
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Command_Line;
with Ada.Exceptions;
with Ada.Task_Identification;
with Ada.Unchecked_Deallocation;

with Utilities.Command_Line;
with Utilities.Duration_Text_Io;

with Opengl.Gl;
with Opengl.Types;
with Opengl.Bitmaps;
with Opengl.Logging;
with Opengl.Messages;
with Opengl.Utils;
with Opengl.Timer;
with Opengl.Test_Screen_Prototyping;
with Opengl.Win_Stuff;

with Win32.Winmain;
with Win32.Winbase;
with Win32.Winnt;
with Windows.Messages;

package body Opengl.Win is

  use type Ada.Real_Time.Time_Span;
  use type Interfaces.C.Int;
  use type Win32.Windef.Hwnd;
  use type Win32.Bool;
  use type Win32.Windef.Atom;
  use type Gl.Glfloat;
  use type Models.Model_Group_Access;
  use type Models.Window_Models_Access;
  ----------------------------------------------------------------------------
  procedure Free is new Ada.Unchecked_Deallocation
    (Object => Windows_Timer, Name => Windows_Timer_Access);
  ----------------------------------------------------------------------------
  function Window_Index(Handle : in Win32.Windef.Hwnd) return Positive is
  begin
    for I in Main_Window.Windows'range loop
      if Main_Window.Windows(I).Handle = System.Null_Address
         -- then in initialization part in function main below
         or else Main_Window.Windows(I).Handle = Handle then
        return I;
      end if;
    end loop;
    return Positive'Last;
  end Window_Index;
  ----------------------------------------------------------------------------
  task body Windows_Timer is

    Counter : Natural := 0;
    --------------------------------
	  procedure Send_Render_Message is

      pragma Inline(Send_Render_Message);

	    Dummy : Win32.Lresult;
	  begin
	    Dummy := Win32.Winuser.Sendmessage
        (Hwnd    => Main_Window.Windows(Window_Index).Handle,
         Msg     => Win32.Uint(Window_Message),
         Wparam  => Win32.Uint(Window_Index),
         Lparam  => 0);
      Counter := Counter + 1;
    exception
      when Constraint_Error =>
        Counter := 1;
        Logging.Put("Opengl.Win.Windows_Timer(Window_Index =>" &
                    Natural'Image(Window_Index) & "), Counter reset to 0!");
	  end Send_Render_Message;
    ------------------------    
--    Inc          : Ada.Real_Time.Time_Span := Ada.Real_Time.Milliseconds(1);
    Next_Time    : Ada.Real_Time.Time      := Ada.Real_Time.Clock;
    Send_Message : Boolean                 := True;    
  begin
    Logging.Put("Windows_Timer(Window_Index =>" & 
      Natural'Image(Window_Index) & "): " &
        Ada.Task_Identification.Image(Ada.Task_Identification.Current_Task));

--    loop
--      Send_Render_Message;      
--      exit when Inc > Time_Increment.all;
--      Next_Time := Next_Time + Inc;
--      select 
--        accept Render_Immediately;
--      or  
--        delay until Next_Time;
--      end select;  
--      Inc := 2 * Inc;
--    end loop;

    loop
      if Send_Message then
        Send_Render_Message;
      end if;

      select
        accept Shutdown;
        exit;
      or
        accept Render_Immediately;
        Next_Time := Ada.Real_Time.Clock;
      or
        accept Toggle_Run_Stop;
        Send_Message := not Send_Message;
      or
        delay until Next_Time;
        Next_Time := Next_Time + Time_Increment.all;
      end select;
    end loop;

    Opengl.Logging.Put("Counter =" & Natural'Image(Counter) &
      "  in task Opengl.Win.Windows_Timer(Window_Index =>" &
        Natural'Image(Window_Index) & ")");
  end Windows_Timer;
  ----------------------------------------------------------------------------
  procedure Main_Windows_Message_Loop
    (Main_Window : access Windows_Attributes'Class) is

    Bool_Result : Win32.Bool;
    Long_Result : Win32.Long;
  begin
    while Win32.Winuser.Getmessage
              (Lpmsg         => Main_Window.Message'access,
               Hwnd          => System.Null_Address,--Main_Window.Window_Handle, 
               Wmsgfiltermin => Main_Window.Message_Filter_Minimum,
               Wmsgfiltermax => Main_Window.Message_Filter_Maximum) /= 0 loop
   	  Bool_Result := Win32.Winuser.Translatemessage
	                  (Lpmsg => Main_Window.Message'access);
	  	Long_Result := Win32.Winuser.Dispatchmessage
	                  (Lpmsg => Main_Window.Message'access);
	  end loop;
  end Main_Windows_Message_Loop;
  ----------------------------------------------------------------------------
  -- The specification of function Main is taken from 
  -- ..\objectada\samples\W32PTZD\common\src\routines.ads.
  -- The body of function Main is taken (and adapted for "reentrance") from 
  -- ..\objectada\samples\W32PTZD\2\hellowin\src\routines.ads
  -- It initializes the window and starts the windows message loop.

  function Main(Instance          : Win32.Windef.Hinstance;
			          Previous_Instance : Win32.Windef.Hinstance;
			          Command_Line      : Win32.Pcstr;
			          Show_Command      : Win32.Int;
			          Main_Window       : access Windows_Attributes'Class)
           return Win32.Uint is
    ---------------------------------------------------------
    procedure Show_Mouse_Cursor is 
    begin
      while Win32.Winuser.Showcursor(Win32.False) >= 0 loop
        null;
      end loop;
    end Show_Mouse_Cursor;
    ---------------------------------------------------------
    procedure Hide_Mouse_Cursor is 
    begin
      if Win32.Winuser.Showcursor(Win32.True) < 0 then
 	      null;
      end if;
    end Hide_Mouse_Cursor;
    ---------------------------------------------------------    
    Bool_Result : Win32.Bool;
    Atom_Result : Win32.Windef.Atom;
  begin
    if Main_Window.Window_Class.Hinstance = System.Null_Address and
       Previous_Instance = System.Null_Address then
         
     	Main_Window.Window_Class :=
	          (Style         => Main_Window.Window_Class_Style,
	           Lpfnwndproc   => Win32.Winuser.Wndproc(Main_Window.Callback),
	           Cbclsextra    => Main_Window.Window_Class_Cbclsextra,
	           Cbwndextra    => Main_Window.Window_Class_Cbwndextra,
	           Hinstance     => Instance,
	           Hicon         => Main_Window.Window_Class_Icon,
	           Hcursor       => Main_Window.Window_Class_Cursor,
	           Hbrbackground => Main_Window.Window_Class_Background,
	           Lpszmenuname  => Win32.To_Pcstr(Main_Window.Window_Class_Menu_Name),
	           Lpszclassname => Win32.To_Pcstr(Main_Window.Class_Name));
             
	    Atom_Result := Win32.Winuser.Registerclass
	                  (Lpwndclass => Main_Window.Window_Class'access);
                    
      if Atom_Result = 0 then
        Ada.Exceptions.Raise_Exception(Error'Identity, Message =>
          Messages.Error_Message
            (Subprogram => "OpenGL.Win.main",
             Message    => "Result of Win32.Winuser.Registerclass is 0"));
      end if;
    end if;

    if Main_Window.Window_Parent /= System.Null_Address then
      Main_Window.Window_Style := Main_Window.Window_Style or
   	    Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Child);
    end if;

    Main_Window.Caption := Interfaces.C.Strings.New_String
      (Utilities.Command_Line.Total_Command_Line(Full_Path => False));

    if not Opengl.Command_Line.Full_Screen_Window then
      Main_Window.Window_Style :=
        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Border)           or 
        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Caption)          or 
        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Visible)          or 
        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Thickframe)       or 
        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Overlappedwindow) or
--        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Sizebox)          or 
    	  Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Popup);--            or
--        Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Maximize);
    end if;
           
    for I in Main_Window.Windows'range loop
      Screen.Set_Current_Window_Area(To => Main_Window.Windows(I).Area);
      
	    Main_Window.Windows(I).Handle := Win32.Winuser.Createwindow
	       (Lpclassname  => Win32.To_Pcstr(Main_Window.Class_Name),
	        Lpwindowname => Win32.To_Pcstr(Main_Window.Caption),
	        Dwstyle      => Main_Window.Window_Style,
	--        X            => Win32.Winuser.Cw_Usedefault, 
	--        Y            => Win32.Winuser.Cw_Usedefault,
	--        Nwidth       => Win32.Winuser.Cw_Usedefault,
	--        Nheight      => Win32.Winuser.Cw_Usedefault,
	        X            => Interfaces.C.Int(Main_Window.Windows(I).Area.X_Upper_Left),
	        Y            => Interfaces.C.Int(Main_Window.Windows(I).Area.Y_Upper_Left),
	        Nwidth       => Interfaces.C.Int(Main_Window.Windows(I).Area.Width),
	        Nheight      => Interfaces.C.Int(Main_Window.Windows(I).Area.Height),
	        Hwndparent   => Main_Window.Window_Parent,
	        Hmenu        => Main_Window.Window_Menu,
	        Hinstance    => Instance,
	        Lpparam      => Main_Window.Window_Parameters);

	    if Main_Window.Windows(I).Handle = System.Null_Address then
	      Ada.Exceptions.Raise_Exception(Error'Identity, Message =>
	        Messages.Error_Message
            (Subprogram => "OpenGL.Win.main",
	           Message    => "Result of Win32.Winuser.Createwindow is null"));
	    end if;

	    Bool_Result := Win32.Winuser.Showwindow
	                  (Hwnd     => Main_Window.Windows(I).Handle,
		                 Ncmdshow => Win32.Winmain.Get_Ncmdshow);
	    Bool_Result := Win32.Winuser.Updatewindow
	                  (Hwnd => Main_Window.Windows(I).Handle);

	    if Bool_Result = Win32.False then
	      Ada.Exceptions.Raise_Exception(Error'Identity,  Message =>
	        Messages.Error_Message
            (Subprogram => "OpenGL.Win.main",
	           Message    => "Result of Win32.Winuser.Updatewindow is false"));
	    end if;
    end loop;

    if not Opengl.Command_Line.Show_Mouse then
      Hide_Mouse_Cursor;
    end if;  

    Main_Windows_Message_Loop(Main_Window);

    if not Opengl.Command_Line.Show_Mouse then
      Show_Mouse_Cursor;    
    end if;  

--    if Win32.Winuser.Destroywindow(main_window.Window_Handle) = Win32.False then
--      Ada.Exceptions.Raise_Exception(Error'Identity,  Message =>
--          messages.error_message(subprogram => "OpenGL.Win.main",
--            Message => "Result of Win32.Winuser.Destroywindow is false"));
--    end if;                   
--    if Win32.Winuser.Unregisterclass
--           (Lpclassname => Win32.To_Pcstr(main_window.Class_Name), 
--            Hinstance   => Instance) = Win32.False then
--      Ada.Exceptions.Raise_Exception(Error'Identity,  Message =>
--          messages.error_message(subprogram => "OpenGL.Win.main",
--            Message => "Result of Win32.Winuser.Unregisterclass is false"));
--    end if; 

    return Main_Window.Message.Wparam;
  end Main;
  ----------------------------------------------------------------------------
  -- Some additional callbacks for some basic interaction with the 
  -- OpenGL program
  ----------------------------------------------------------------------------
  procedure Output_Info(Dt      : in Ada.Real_Time.Time_Span;
                        Index   : in Positive;                        
                        Info    : in String := "";
                        Text    : in String := "";
                        Line_No : in Positive := 1) is
  begin
    Timer.Line(Line_No)(1 .. Text'Length) := Text;
    Utilities.Duration_Text_Io.Put
      (To   => Timer.Line(Line_No)
              (Text'Length + (Index - 1) * 10 + 1 .. Text'Length + Index * 10),
       Item => Ada.Real_Time.To_Duration(Dt * 1000),
       Aft  => 3,
       Exp  => 0);
    Timer.Line(Line_No + 1)(1 .. Info'Length) := Info;
  end Output_Info;
  ---------------------------------------------------------------------------- 
  -- Pressing the Down-arrow divides  the current render interval by 2.           
  -- Pressing the Up-arrow multiplies the current render interval by 2.
  -- Pressing any other key has no effect.           

  function Keypressed_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin
    for I in Main_Window.Windows'range loop
      declare
		    The_Window : Window renames Main_Window.Windows(I);
		    Dt         : Ada.Real_Time.Time_Span renames The_Window.Models.Render_Intervall;
		    The_Timer  : Windows_Timer renames The_Window.Timer.all;
		  begin
		    case Wparam is
          when 107 => Dt := 2 * Dt;                     -- numpad +
                      Output_Info(Dt, I, Text    => "Render interval[ms] = ",
                                         Line_No => 1);
		      when 109 => Dt := Dt / 2;                     -- numpad -
                      Output_Info(Dt, I, Text    => "Render interval[ms] = ", 
                                         Line_No => 1);
          when  84 => Timer.Output := not Timer.Output; -- t              
		      when others => null;
        end case;
        
        select
		      The_Timer.Render_Immediately;
        or
          delay Ada.Real_Time.To_Duration(Dt);
        end select;
      end;
    end loop;
    return Win32.Winuser.Defwindowproc
	          (Hwnd   => Window_Handle,
	           Msg    => Message,
	           Wparam => Wparam,
	           Lparam => Lparam);
  end Keypressed_Opengl_Window;
  ----------------------------------------------------------------------------
  -- Pressing the left mouse button, holds rendering.
  -- Pressing the left mouse button again, resumes rendering.

  function Left_Mouse_Down_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin
    for I in Main_Window.Windows'range loop
      declare
		    The_Window : Window renames Main_Window.Windows(I);
      begin
        select
          The_Window.Timer.Toggle_Run_Stop;
        or
--          delay Ada.Real_Time.To_Duration(The_Window.Models.Render_Intervall);
          delay 1.0;
        end select;
      end;
    end loop;
    return Win32.Winuser.Defwindowproc
	          (Hwnd   => Window_Handle,
	           Msg    => Message,
	           Wparam => Wparam,
	           Lparam => Lparam);
  end Left_Mouse_Down_Opengl_Window;
  ----------------------------------------------------------------------------
  -- Pressing the right mouse button, stops the program.

  function Right_Mouse_Down_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
    Dummy : Win32.Lresult;
  begin
    for I in Main_Window.Windows'range loop
      declare
		    The_Window : Window renames Main_Window.Windows(I);
      begin
        Dummy := Destroy_Opengl_Window(Window_Handle => The_Window.Handle,
       	   		                         Message       => Message,
		                                   Wparam        => Wparam,
		                                   Lparam        => Lparam);
      end;
    end loop;
    return Dummy;
  end Right_Mouse_Down_Opengl_Window;
  ----------------------------------------------------------------------------
  procedure Set_Priority is
       
    type win_priority_class is (realtime, high, normal, idle);   
       
    priorities  : constant array(win_priority_class) of Win32.Dword :=
                  (realtime => Win32.Winbase.Realtime_Priority_Class,
                   High     => Win32.Winbase.High_Priority_Class,
                   Normal   => Win32.Winbase.Normal_Priority_Class,
                   Idle     => Win32.Winbase.Idle_Priority_Class);
                  
    win_priority: win_priority_class := idle;                    
    Prozess     : constant Win32.Winnt.Handle := Win32.Winbase.Getcurrentprocess;
    Priority    : constant Win32.Dword := Win32.Winbase.Getpriorityclass(Prozess);
    New_Priority: Win32.Dword;
    Success     : Win32.Bool; 
  begin
    begin
      win_Priority := win_priority_class'value(Command_Line.Priority);
    exception  
      when constraint_error => 
        win_Priority := idle;
        logging.put("Wrong priority on command-line: '" & Command_Line.Priority & 
                    "', set to idle, allowed: realtime, high, normal, idle");      
    end;  
    
    new_priority := priorities(win_Priority); 
    
    Logging.put("Prozess priority    : " & Win32.Dword'Image(Priority));
    
    if Priority /= New_Priority then
	Success := Win32.Winbase.Setpriorityclass
                 (Hprocess        => Prozess,  
                  Dwpriorityclass => New_Priority);
	if Success = Win32.False then
	  Ada.Exceptions.Raise_Exception(Error'Identity, Message => 
	    Messages.Error_Message
	         (Subprogram => "OpenGL.Win.Set_Priority",
	          Message    => "Win32.Winbase.Setpriorityclass failed"));
    	end if;
    	New_Priority := Win32.Winbase.Getpriorityclass(Prozess);
      Logging.put("New prozess priority: " & Win32.Dword'Image(New_Priority) & 
                  " = " & win_priority_class'image(win_priority));
    end if;
  end Set_Priority;
  ----------------------------------------------------------------------------  
  procedure Execute(Main_Window : access Windows_Attributes'Class :=
                                         Standard_Window'access) is
  begin
    Logging.Initialize_Log_File(Win_Stuff.Os_Info);
    
    Win.Main_Window := Opengl_Windows_Attributes_Access(Main_Window);

    if Win.Main_Window.Callbacks(Win32.Winuser.Wm_Keydown) = null then
      Win.Main_Window.Callbacks(Win32.Winuser.Wm_Keydown)     :=
        Keypressed_Opengl_Window'access;
    end if;  
    if Win.Main_Window.Callbacks(Win32.Winuser.Wm_Lbuttondown) = null then
      Win.Main_Window.Callbacks(Win32.Winuser.Wm_Lbuttondown) :=
        Left_Mouse_Down_Opengl_Window'access;
    end if;    
    if Win.Main_Window.Callbacks(Win32.Winuser.Wm_Rbuttondown) = null then
      Win.Main_Window.Callbacks(Win32.Winuser.Wm_Rbuttondown) :=
        Right_Mouse_Down_Opengl_Window'access;
    end if;    
    
    Ada.Command_Line.Set_Exit_Status
      (Code => Ada.Command_Line.Exit_Status
              (Main(Instance          => Win32.Winmain.Get_Hinstance,
                    Previous_Instance => Win32.Winmain.Get_Hprevinstance,
                    Command_Line      => Win32.Winmain.Get_Lpcmdline,
                    Show_Command      => Win32.Winmain.Get_Ncmdshow,
                    Main_Window       => Main_Window)));
               
    Logging.Finalize_Log_File;
  end Execute;
  ---------------------------------------------------------------------------- 
  procedure make_opengl_foreground_window is
  begin      
    if win32.winuser.SetForegroundWindow
                     (Main_Window.Windows(1).Handle) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Make_OpenGL_Foreground_Window",
         Message    => "WinUser.SetForegroundWindow failed"));
    end if;
  end make_opengl_foreground_window;
  ---------------------------------------------------------------------------- 
  procedure Send_Message(Message : in Win32.Uint;
                         Wparam  : in Win32.Wparam := 0;
                         Lparam  : in Win32.Lparam := 0) is
                         
    Dummy : Win32.Lresult;
  begin
    Dummy := Win32.Winuser.Sendmessage
                 (Hwnd    => Main_Window.Windows(1).Handle,
                  Msg     => Message,
                  Wparam  => Wparam,
                  Lparam  => Lparam);
  end Send_Message;                       
  ---------------------------------------------------------------------------- 
  procedure Create_Opengl_Window_Pre_Operation
    (The_Window    : in out Window;
     Window_Handle : in     System.Address) is
       
    Pixel_Format_Index : Interfaces.C.Int;
  begin  
    Logging.Write_Current_Task_Id;
    
    --------------------------------------------------------------------------
    -- Set up device and rendering context and create an outline-font
    --------------------------------------------------------------------------

    -- Set Devive context
    The_Window.Device_Context := Win32.Winuser.Getdc(Window_Handle);

    if The_Window.Device_Context = System.Null_Address then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Create_Opengl_Window",
         Message    => "Winuser.Getdc failed"));
    end if;

    -- Set pixel format     
    Pixel_Format_Index := Win32.Wingdi.Choosepixelformat
      (Hdc  => The_Window.Device_Context,
       Ppfd => Main_Window.Pixel_Format'access);

    if Pixel_Format_Index = 0 then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Create_Opengl_Window",
         Message    => "Windgi.Choseppixelformat failed (returns 0)"));
    end if;

    if Win32.Wingdi.Setpixelformat
           (Hdc          => The_Window.Device_Context,
            Ipixelformat => Pixel_Format_Index,
            Ppfd         => Main_Window.Pixel_Format'access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Create_Opengl_Window",
         Message    => "Windgi.Setpixelformat failed"));
    end if;

    -- Set Rendering_Context and make it current
    The_Window.Rendering_Context :=
      Win32.Wingdi.Wglcreatecontext(The_Window.Device_Context);

    if The_Window.Rendering_Context = System.Null_Address then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Create_Opengl_Window",
         Message    => "Windgi.Wglcreatecontext failed"));
    end if;

    if Win32.Wingdi.Wglmakecurrent
           (Hdc   => The_Window.Device_Context,
            Hglrc => The_Window.Rendering_Context) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Create_Opengl_Window",
         Message    => "Windgi.WglMakeCurrent failed"));
    end if;

    -- Antialiasing
    if Command_Line.Anti_Aliasing then
      Models.Antialiasing_On;
      Opengl.Logging.put("Anti-aliasing");
    end if;

    Fonts.Set(Device_Context => The_Window.Device_Context);
  end Create_Opengl_Window_Pre_Operation;  
  ----------------------------------------------------------------------------
  procedure Create_Opengl_Window_Post_Operation
    (The_Window   : in out Window;
     Window_Index : in     Positive) is
  begin  
    The_Window.Outline_Fonts := Fonts.All_Created_Fonts;
    Fonts.Get(Device_Context => The_Window.Device_Context);

    The_Window.Models.Render_Intervall := Opengl.Command_Line.Timer_Delta;

    The_Window.Timer := new Windows_Timer
      (Time_Increment => The_Window.Models.Render_Intervall'access,
       Window_Index   => Window_Index,
       Window_Message => Wm_Render);
    
    The_Window.Models.Background_Color := Utils.Background_Color;      
  end Create_Opengl_Window_Post_Operation;  
  ----------------------------------------------------------------------------                  
  function Create_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    Index      : constant Positive := Window_Index(Window_Handle);
    The_Window : Window renames Main_Window.Windows(Index);
  begin
    Create_Opengl_Window_Pre_Operation(The_Window, Window_Handle);
   
    The_Window.Models := new Models.Window_Models(Group_Number => 0);
    The_Window.Models.Permanent_Group := 
      Test_Screen_Prototyping.Models_Of(Index);

    Create_Opengl_Window_Post_Operation(The_Window, Index);
   
    return Win32.Winuser.Defwindowproc
          (Hwnd   => Window_Handle,
           Msg    => Message,
           Wparam => Wparam,
           Lparam => Lparam);
  end Create_Opengl_Window;
  ----------------------------------------------------------------------------                          
  procedure Destroy_Opengl_Window_Pre_Operation(The_Window : in out Window) is
  begin      
    if Win32.Wingdi.Wglmakecurrent
           (Hdc   => The_Window.Device_Context,
            Hglrc => System.Null_Address) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Destroy_Opengl_Window",
         Message    => "Windgi.WglMakeCurrent failed"));
    end if;

    if Win32.Wingdi.Wgldeletecontext
           (The_Window.Rendering_Context) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Destroy_Opengl_Window",
         Message    => "Windgi.WglDeleteContext failed"));
    end if;

    Logging.Write_Timing_Info(The_Window.Models);

    if The_Window.Timer /= null and then The_Window.Timer'Callable then
      select
        The_Window.Timer.Shutdown;
      or
        delay 0.1;
        abort The_Window.Timer.all;
      end select;
      Free(The_Window.Timer);
    end if;
  end Destroy_Opengl_Window_Pre_Operation;
  ----------------------------------------------------------------------------    
  procedure Destroy_Opengl_Window_Post_Operation(The_Window : in out Window) is
  begin
    Win32.Winuser.Postquitmessage(Nexitcode => 0);
    The_Window.Handle := System.Null_Address;
    The_Window.Models := null;
  end Destroy_Opengl_Window_Post_Operation;
  ----------------------------------------------------------------------------    
  function Destroy_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
             
    Index      : constant Positive := Window_Index(Window_Handle);
    The_Window : Window renames Main_Window.Windows(Index);
  begin
    if The_Window.Handle = System.Null_Address then 
      -- When test_inputs generated in the same exe Destroy_Opengl_Window
      -- was already called. 
      -- Now called from simulator.network_interface.Render_Opengl_Window after
      -- receiving a shutdown from the task 
      -- Simulator.Instruments.Input_Generator.Manager.
      return Win32.Winuser.Defwindowproc
            (Hwnd   => Window_Handle,
             Msg    => Message,
             Wparam => Wparam,
             Lparam => Lparam);
    end if;  
    
    Destroy_Opengl_Window_Pre_Operation (The_Window);
    Destroy_Opengl_Window_Post_Operation(The_Window);
    
    return Win32.Winuser.Defwindowproc
          (Hwnd   => Window_Handle,
           Msg    => Message,
           Wparam => Wparam,
           Lparam => Lparam);
  end Destroy_Opengl_Window;
  ----------------------------------------------------------------------------     
  function Resize_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
    --------------------------------------------------------
-- GetWindowRect does not work, Windows-Error "Function only available in win32-mode" is shown             
--    function Area_Of(The_Window : in Win32.Windef.Hwnd) 
--             return Screen.Window_Area_Type is
--               
--      Area : constant Win32.Windef.Rect := Win_Stuff.Area_Of(The_Window);
--    begin
--      return (X_Upper_Left => Natural(Area.Left),
--              Y_Upper_Left => Natural(Area.Top),
--              Width        => Natural(Area.Right) - Natural(Area.Left),
----              Height       => Natural(Area.Top)   - Natural(Area.Bottom));
--              Height       => Natural(Area.Top)   + Natural(Area.Bottom));
--    end Area_Of;
    ----------------------------------------------------------------                 
    procedure Set_Bitmap_Scale(The : in out Models.Window_Models;
                               To  : in     Types.Vector_2d) is
      --------------------------------------------------------------
      procedure Set_Bitmap_Scale(The : in out Models.Model_Group) is
      begin
        for I in The.Models'range loop
          if The.Models(I).all in Bitmaps.Model'Class then
            Bitmaps.Model(The.Models(I).all).Bitmap.Scale := To;
          end if;  
        end loop;  
      end Set_Bitmap_Scale;
      --------------------------------------------------------------                           
    begin  
      for I in The.Model_Groups'range loop
        if The.Model_Groups(I) /= null then
          Set_Bitmap_Scale(The.Model_Groups(I).all);
        end if;
      end loop;
      if The.Permanent_Group /= null then
        Set_Bitmap_Scale(The.Permanent_Group.all);
      end if;
    end Set_Bitmap_Scale;    
    -----------------------------------------------------------------                 
    Index        : constant Positive := Window_Index(Window_Handle);
    The_Window   : Window renames Main_Window.Windows(Index);
    Now          : constant Ada.Real_Time.Time := Ada.Real_Time.Clock;
    New_Width    : constant Natural    := 
                   Natural(Win32.Windef.Loword(Win32.Dword(Lparam)));
    New_Height   : constant Natural    := 
                   Natural(Win32.Windef.Hiword(Win32.Dword(Lparam)));                   
  begin
    if Win32.Wingdi.Wglmakecurrent
           (Hdc   => The_Window.Device_Context,
            Hglrc => The_Window.Rendering_Context) = Win32.False then
      Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
        (Subprogram => "OpenGL.Win.Resize_Opengl_Window",
         Message    => "Windgi.wglMakeCurrent failed"));
    end if;    
    
    Utils.Resize_Window(New_Width  => New_Width,
                        New_Height => New_Height);

    if New_Width /= 0 and then New_Height /= 0 then
      Set_Bitmap_Scale(The_Window.Models.all, 
                       To => (X => Gl.Glfloat(New_Width)  / 
                                   Gl.Glfloat(The_Window.Area.Width),
                              Y => Gl.Glfloat(New_Height) / 
                                   Gl.Glfloat(The_Window.Area.Height)));
                                 
      The_Window.Area.Width  := New_Width;                        
      The_Window.Area.Height := New_Height;               
         
      screen.set_current_window_area(to => The_Window.Area);
    end if;  
                                 
    if Now - The_Window.Last_Resize_Time > Ada.Real_Time.Milliseconds(100) then     
      select 
        The_Window.Timer.Render_Immediately;
        The_Window.Last_Resize_Time := Now;
      or delay 0.1;
      end select;  
    end if;   

    return Win32.Winuser.Defwindowproc
          (Hwnd   => Window_Handle,
           Msg    => Message,
           Wparam => Wparam,
           Lparam => Lparam);
  end Resize_Opengl_Window;
  ----------------------------------------------------------------------------    
  function Render_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is

    The_Window : Window renames Main_Window.Windows(Positive(Wparam));
  begin
    if The_Window.Handle /= System.Null_Address then
      if Win32.Wingdi.Wglmakecurrent
             (Hdc   => The_Window.Device_Context,
              Hglrc => The_Window.Rendering_Context) = Win32.False then
        Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
          (Subprogram => "OpenGL.Win.Render_Opengl_Window",
           Message    => "Windgi.wglMakeCurrent failed"));
      end if;

      Models.Render(The_Window.Models.all);
      Gl.Glflush;

      if Win32.Wingdi.Swapbuffers(The_Window.Device_Context) = Win32.False then
        Ada.Exceptions.Raise_Exception(Error'Identity, Messages.Error_Message
          (Subprogram => "OpenGL.Win.Render_Opengl_Window",
           Message    => "Wingdi.SwapBuffers failed"));
      end if;      
    end if;
    
    return Win32.Winuser.Defwindowproc
          (Hwnd   => Window_Handle,
           Msg    => Message,
           Wparam => Wparam,
           Lparam => Lparam);
  end Render_Opengl_Window;
  ----------------------------------------------------------------------------     
  function Timer_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin
    return Render_Opengl_Window(Window_Handle, Message, Wparam, Lparam);
  end Timer_Opengl_Window;
  ----------------------------------------------------------------------------
  function General_C_called_windows_callback
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult is
  begin
    if Natural(Message) in Main_Window.Callbacks'range and then
       Main_Window.Callbacks(Natural(Message)) /= null then
      return Main_Window.Callbacks(Natural(Message)).all
               (Window_Handle => Window_Handle,
                Message       => Message,
                Wparam        => Wparam,
                Lparam        => Lparam);
    else
      return Win32.Winuser.Defwindowproc
	          (Hwnd   => Window_Handle,
	           Msg    => Message,
	           Wparam => Wparam,
	           Lparam => Lparam);
    end if;
  exception
    when E : others =>
      Messages.Log(The_Exception => E);
      if Command_Line.Show_Errors then
        Logging.Write_Error_File(E);      
        Windows.Messages.Display_Information
	             (The_Window          => Window_Handle,
	              The_Exception       => E,
	              Close_Window        => False,
	              Log_File            => Log_File'access,
	              Add_Windows_Message => True,
                Beep                => 1);
        return Destroy_Opengl_Window
		          (Window_Handle => Window_Handle,
		           Message       => Message,
		           Wparam        => Wparam,
		           Lparam        => Lparam);
      else      
	  return Win32.Winuser.Defwindowproc
		          (Hwnd   => Window_Handle,
		           Msg    => Message,
		           Wparam => Wparam,
		           Lparam => Lparam);
      end if;
--      Ada.Exceptions.Reraise_Occurrence(E); 
  end General_C_called_windows_callback;
  ----------------------------------------------------------------------------     
end Opengl.Win;
