--                              -*- Mode: Ada -*-
-- Filename        : opengl-win.ads
-- Description     : Package with utilities to setup an opengl-window
--                   for MS-Windows using the Win32Ada-binding.
-- Author          : J. Schrer
-- Created On      : 17 Nov. 1997
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Calendar;
with Ada.Real_Time;
with Interfaces.C;
with Interfaces.C.Strings;
with Win32.Windef;
with Win32.Wingdi;
with Win32.Winuser;
with Opengl.Screen;
with Opengl.Fonts;
with Opengl.Models;
with Opengl.Command_Line;
with System;

package Opengl.Win is

  Error : exception;
  -- Is raised when errors are detected in the usage of functions from this
  -- package or when called win32 subprograms fail. 

  ----------------------------------------------------------------------------
  -- Types, subprograms for (opengl-)window

  type C_Called_Windows_Callback is access
       function (Window_Handle : Win32.Windef.Hwnd;
                 Message       : Win32.Uint;
                 Wparam        : Win32.Wparam;
                 Lparam        : Win32.Lparam) return Win32.Lresult;

  pragma Convention(Stdcall, C_Called_Windows_Callback);

  -- Is (intrinsic or) called from C because of convention pragma.


--   type Windows_Attributes; 
--  
--    type Ada_Called_Windows_Callback is access 
--         function (Window_Handle : Win32.Windef.Hwnd;
--                   Message       : Win32.Uint;
--                   Wparam        : Win32.Wparam;
--                   Lparam        : Win32.Lparam;
--                   Attributes    : access Windows_Attributes'Class) 
--         return Win32.Lresult;
--  
--   This Forward Declaration Construct Gives 
--   "Internal Fatal Internal Memory access Failure" in Oa 7.1 in Line
--                   Attributes    : access Windows_Attributes'Class) 


  type Ada_Called_Windows_Callback is access
       function (Window_Handle : Win32.Windef.Hwnd;
                 Message       : Win32.Uint;
                 Wparam        : Win32.Wparam;
                 Lparam        : Win32.Lparam) return Win32.Lresult;

  -- "Windows Ada Callback" (without convention pragma -> not called from C).

  Max_Messages : constant := Win32.Winuser.Wm_User + 100;

  Wm_Render    : constant := Win32.Winuser.Wm_User;

  subtype Callback_Number_Type is Natural range 0 .. Max_Messages - 1;

  type Windows_Callbacks is array(Callback_Number_Type range <>)
    of Ada_Called_Windows_Callback;

  -- And an array of "Windows Ada Callbacks"

  ----------------------------------------------------------------------------  
  -- "Standard" opengl-windows callbacks.

  function Create_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;

  function Destroy_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;
           
  -- Create_Opengl_Window and Destroy_Opengl_Window contain no special code for 
  -- rendering instruments.         

  function Resize_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;

  function Render_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;

  function Timer_Opengl_Window
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;

  -- End standard opengl-windows callbacks.
  ----------------------------------------------------------------------------  

  function General_C_called_windows_callback
          (Window_Handle : Win32.Windef.Hwnd;
           Message       : Win32.Uint;
           Wparam        : Win32.Wparam;
           Lparam        : Win32.Lparam) return Win32.Lresult;

  pragma Convention(Stdcall, General_C_called_windows_callback);

  -- Calls the function the_windows_callbacks(message).all(...),
  -- if the_windows_callbacks(message) /= null.
  -- The_windows_callbacks is an array of type Windows_Callbacks above and 
  -- part of the whole windows_attributes-datastructure below.

  -- This corresponds to the specification of the function main_windows taken from 
  -- ..\objectada\samples\W32PTZD\common\src\routines.ads.

  ----------------------------------------------------------------------------                             
  -- The following tagged type is the root of a class.
  -- Objects of that type contain all data that should be permanent 
  -- for a windows program. 
  -- The components of this root class type are taken from  
  --   ..\objectada\samples\W32PTZD\2\hellowin\src\data.ads

  -- The components callback and callbacks are added.                               
  -- If the default callback General_C_called_windows_callback'access is not 
  -- used, the specification of the function assigned to component callback
  -- has to be declared in the specification of an application specific package
  -- belonging to the windows main procedure.
  -- (see hellowin_pac in hellowin.adb as an example).
  -- the same holds for the functions assigned to the components of 
  -- array Callbacks.


  Default_Class_Name : aliased Interfaces.C.Char_Array :=
                               Interfaces.C.To_C("OpenGL main program");
  Default_Caption    : aliased Interfaces.C.Char_Array :=
                               Interfaces.C.To_C("OpenGL main window");

  use type Interfaces.C.Unsigned_Long,
           Interfaces.C.Unsigned;      -- for "or" below.                          


  -- A task that generates a Window_Message event each Time_Increment seconds
  -- and sends it to the window with index Window_Index. 
  -- The handle of the window is determined by Window_Index.           
  -- Here used to send a WM_Render message to an OpenGL window.
           
  task type Windows_Timer(Time_Increment : access Ada.Real_Time.Time_Span;
                          Window_Index   : Positive;
                          Window_Message : Natural) is

    pragma Priority(Command_Line.Winnt_Priority);

    entry Render_Immediately;
          -- To bypass the "delay until" under certain circumstances.
    entry Toggle_Run_Stop;
          -- When running stops, when stopped runs again.
    entry Shutdown;
          -- Terminates the task.

  end Windows_Timer;

  type Windows_Timer_Access is access Windows_Timer;


  type Window is record
    Models            : Opengl.Models.Window_Models_Access := null;
    Timer             : Windows_Timer_Access               := null;
    Area              : Screen.Window_Area_Type            := -- whole screen as default
                        (X_Upper_Left => Command_Line.Window_X,
                         Y_Upper_Left => Command_Line.Window_Y,
                         Width        => Command_Line.Window_Width,
                         Height       => Command_Line.Window_Height);
    Handle            : Win32.Windef.Hwnd                  := System.Null_Address;
    Device_Context    : Win32.Windef.Hdc                   := System.Null_Address;
    Rendering_Context : Win32.Windef.Hglrc                 := System.Null_Address;
    Outline_Fonts     : Fonts.Font_Type_Array_Access       := null;
    Last_Resize_Time  : Ada.Real_Time.Time                 := Ada.Real_Time.Clock;
  end record;

  -- Models.Window_Models contains all conceptual information for the models
  -- and the window.
  -- The type Window adds all MS-Win specific data necessary for
  -- the implementation.

  type Window_Array  is array(Positive range <>) of aliased Window;


  type Windows_Attributes(Callback_Number : Callback_Number_Type;
                          Window_Number   : Positive) is tagged record

    Windows                 : Window_Array(1 .. Window_Number);

    Callback                : C_Called_Windows_Callback :=
                              General_C_called_windows_callback'access;

    -- The initialization with this functions is Opengl specific -------------
    Callbacks               : Windows_Callbacks(0 .. Callback_Number) :=
      (Win32.Winuser.Wm_Create  => Create_Opengl_Window'access,
       Win32.Winuser.Wm_Destroy => Destroy_Opengl_Window'access,
       Win32.Winuser.Wm_Size    => Resize_Opengl_Window'access,
       Wm_Render                => Render_Opengl_Window'access,
--       Win32.Winuser.Wm_Timer   => Timer_Opengl_Window'access,
       others                   => null);

    Class_Name              : Interfaces.C.Strings.Chars_Ptr :=
                              Interfaces.C.Strings.To_Chars_Ptr
                                (Default_Class_Name'access);
    Caption                 : Interfaces.C.Strings.Chars_Ptr :=
                              Interfaces.C.Strings.To_Chars_Ptr
                                (Default_Caption'access);

    Window_Class            : aliased Win32.Winuser.Wndclass;

    Message                 : aliased Win32.Winuser.Msg;

    -- Window_style is Opengl specific --------------------------------------
	  Window_Style            : Interfaces.C.Unsigned_Long :=
--      Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Overlappedwindow);
--      Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Sizebox)          or 
--      Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Overlapped)       or 
--      Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Border)           or 
	    Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Popup);
--      Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Clipchildren)     or
--	    Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Clipsiblings)     or
-- 	    Interfaces.C.Unsigned_Long(Win32.Winuser.Ws_Maximize);

	  Window_Class_Style      : Interfaces.C.Unsigned :=
	                            Win32.Winuser.Cs_Hredraw or
	                            Win32.Winuser.Cs_Vredraw;

	  Window_Class_Cbclsextra : Interfaces.C.Int := 0;
	  Window_Class_Cbwndextra : Interfaces.C.Int := 0;

	  Window_Class_Icon       : Win32.Windef.Hicon :=
                        	    Win32.Winuser.Loadicon
	                           (Hinstance  => System.Null_Address,
	                            Lpiconname => Win32.Lpcstr
                                           (Win32.Winuser.Idi_Application));

	  Window_Class_Cursor     : Win32.Windef.Hcursor :=
	                            Win32.Winuser.Loadcursor
	                           (Hinstance    => System.Null_Address,
	                            Lpcursorname => Win32.Lpcstr
                                             (Win32.Winuser.Idc_Arrow));

	  Window_Class_Background : Win32.Windef.Hbrush :=
	                            Win32.Wingdi.Getstockobject
                             (Fnobject => Win32.Wingdi.White_Brush);

	  Window_Class_Menu_Name  : Interfaces.C.Strings.Chars_Ptr :=
                              Interfaces.C.Strings.Null_Ptr;

	  Window_Parent           : Win32.Windef.Hwnd     := System.Null_Address;
	  Window_Menu             : Win32.Windef.Hmenu    := System.Null_Address;
	  Window_Parameters       : Win32.Lpvoid          := System.Null_Address;
	  Message_Filter_Minimum  : Interfaces.C.Unsigned := 0;
	  Message_Filter_Maximum  : Interfaces.C.Unsigned := 0;

    Creation_Time           : Ada.Calendar.Time     := Ada.Calendar.Clock;
  end record;

  -- The implementation is a first proposal. Possibly more data has to be
  -- added to this root. The user can add data in derived classes. 

  ----------------------------------------------------------------------------
  -- Types, constants for opengl outline-fonts:

  Default_Pixel_Format : constant Win32.Wingdi.Pixelformatdescriptor :=

      (Nsize           => Win32.Word(Win32.Wingdi.Pixelformatdescriptor'Size / 8),
                             -- Size of this structure
       Nversion        => 1, -- Version of this structure    
       Dwflags         =>
         Win32.Dword(Win32.Wingdi.Pfd_Draw_To_Window) or -- Draw to Window (not to bitmap)
         Win32.Dword(Win32.Wingdi.Pfd_Support_Opengl) or -- Support OpenGL calls in window
         Win32.Dword(Win32.Wingdi.Pfd_Doublebuffer),     -- doublebuffering
       Ipixeltype      => Win32.Wingdi.Pfd_Type_Rgba,    -- RGBA Color mode
       Ccolorbits      => 24,                            -- Want 24bit color 

       Credbits        => 0,                             -- Not used to select mode
       Credshift       => 0,
       Cgreenbits      => 0,
       Cgreenshift     => 0,
       Cbluebits       => 0,
       Cblueshift      => 0,
       Calphabits      => 0,

       Calphashift     => 0,
       Caccumbits      => 0,
       Caccumredbits   => 0,
       Caccumgreenbits => 0,
       Caccumbluebits  => 0,
       Caccumalphabits => 0,
       Cdepthbits      => 16,                            -- Size of depth buffer

       Cstencilbits    => 0,
       Cauxbuffers     => 0,
       Ilayertype      => Win32.Wingdi.Pfd_Main_Plane,   -- Draw in main plane
       Breserved       => 0,
       Dwlayermask     => 0,
       Dwvisiblemask   => 0,
       Dwdamagemask    => 0);

  ----------------------------------------------------------------------------         

  type Opengl_Windows_Attributes(Callback_Number : Callback_Number_Type;
                                 Window_Number   : Positive)

     is new Windows_Attributes(Callback_Number => Callback_Number,
                               Window_Number   => Window_Number) with record

    Pixel_Format : aliased Win32.Wingdi.Pixelformatdescriptor :=
                           Default_Pixel_Format;
  end record;

  -- A complete opengl-window including a font.

  ----------------------------------------------------------------------------         

  Standard_Window : aliased Opengl_Windows_Attributes
                   (Callback_Number => Callback_Number_Type'Last,
                    Window_Number   => 1);

  -- This is a global object to work on. 
  -- All subprograms and types shall be private in the final version
  -- of the package.
  ----------------------------------------------------------------------------         

  procedure Set_Priority;
  
  -- Shall be called at start of the main program.
  
  -- Reads the command-line and if it contains one of the strings
  -- "-realtime", "-high", "-normal", "-idle" sets the priority class
  -- of the process correspondingly, default is idle.
  
  procedure Execute(Main_Window : access Windows_Attributes'Class :=
                                         Standard_Window'access);

  -- The Main program has to call this routine.
  -- This is the routine from               
  -- ..\objectada\samples\W32PTZD\common\src\main.adb.

  -- Other objects than Standard_Window have to be given to execute when
  -- some of the functions of Windows_Attributes.callbacks are overwritten by
  -- ones that use other components in derived types.
  -- (derived from Windows_Attributes).                                                                            

  -- This is an Ada-version of the opengl-setup for MS-windows given in
  -- listings 4.1, 4.2 of 
  -- "OpenGL superbible, R.S. Wright, M. Sweet, ISBN 1-57169-073-5"                                         

  procedure make_opengl_foreground_window;
  
  -- Sets the OpenGL-window to the foreground.

  procedure Send_Message(Message : in Win32.Uint;
                         Wparam  : in Win32.Wparam := 0;
                         Lparam  : in Win32.Lparam := 0);
                         
  -- Sends parameters to the OpenGL-window.                       
           
private ----------------------------------------------------------------------
    
  -- Subprograms and data used in package opengl.win.callbacks.
  
  function Window_Index(Handle : in Win32.Windef.Hwnd) return Positive;
  
  procedure Output_Info(Dt      : in Ada.Real_Time.Time_Span;
                        Index   : in Positive;                        
                        Info    : in String := "";
                        Text    : in String := "";
                        Line_No : in Positive := 1);
                        
  -- Procedures defining common code for simple rendering and instrument
  -- rendering implemented in package opengl.win.callbacks.
  
  procedure Create_Opengl_Window_Pre_Operation
    (The_Window    : in out Window;
     Window_Handle : in     System.Address);
                                         
  procedure Create_Opengl_Window_Post_Operation
    (The_Window    : in out Window;
     Window_Index  : in     Positive);
    
  procedure Destroy_Opengl_Window_Pre_Operation (The_Window : in out Window);
    
  procedure Destroy_Opengl_Window_Post_Operation(The_Window : in out Window);
    
  type Opengl_Windows_Attributes_Access is access all Opengl_Windows_Attributes;

  ----------------------------------------------------------------------------
  Main_Window : Opengl_Windows_Attributes_Access;

  -- Global object, set by procedure execute, used by callbacks.
  
end Opengl.Win;
