--                              -*- Mode: Ada -*-
-- Filename        : math-generics-arrays-matrix_operations.adb
-- Description     : Matrix operations
-- Author          : J. Schrer
-- Created On      : Wed May 21 19:03:29 1997
-- Last Modified By: 
-- Last Modified On: Wed May 21 19:04:11 1997
-- Update Count    : 1
-- Status          :

package body Math.Generics.Arrays.Matrix_Operations is
  -----------------------------------------------------------------------------
  function "+" (Left, Right : in Matrix) return Matrix is
    
    R : Matrix (Left'range(1), Left'range(2));
  begin
    if    Left'Length(1) /= Right'Length(1) or else 
          Left'Length(2) /= Right'Length(2) then
      raise Matrix_Dimension_Error;
    elsif Left'First(1) = Right'First(1) and then 
          Left'First(2) = Right'First(2) then
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) + Right(I,J);
        end loop;
      end loop;
    else
      R := Right;
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) + R(I,J);
        end loop;
      end loop;
    end if;
    return R;
  end "+";
  -----------------------------------------------------------------------------
  function "-" (Right : in Matrix) return Matrix is
    
    R : Matrix (Right'range(1), Right'range(2));
  begin
    for I in R'range(1) loop
      for J in R'range(2) loop
        R(I,J) := - Right(I,J);
      end loop;
    end loop;
    return R;
  end "-";
  -----------------------------------------------------------------------------
  function "-" (Left, Right : in Matrix) return Matrix is
    
    R : Matrix (Left'range(1), Left'range(2));
  begin
    if    Left'Length(1) /= Right'Length(1) or else 
          Left'Length(2) /= Right'Length(2) then
      raise Matrix_Dimension_Error;
    elsif Left'First(1) = Right'First(1) and then 
          Left'First(2) = Right'First(2) then
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) - Right(I,J);
        end loop;
      end loop;
    else
      R := Right;
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) - R(I,J);
        end loop;
      end loop;
    end if;
    return R;
  end "-";
  -----------------------------------------------------------------------------
  function "*" (Left, Right : in Matrix) return Matrix is
    
    R      : Matrix (Left'range(1), Right'range(2));
    E      : Element;
    Offset : Integer; -- to avoid constraint_error in case index is
                      -- positive subrange and left'first(2) > right'first(1)
  begin
    if    Left'Length(2) /= Right'Length(1) then
      raise Matrix_Dimension_Error;
    elsif Left'First(2) = Right'First(1) then
      for I in R'range(1) loop
        for J in R'range(2) loop
          E := Zero;
          for K in Left'range(2) loop -- or right'range(1)
            E := E + Left(I,K) * Right(K,J);
          end loop;
          R(I,J) := E;
        end loop;
      end loop;
    else
      Offset := Integer(Right'First(1)) - Integer(Left'First(2));
      for I in R'range(1) loop
        for J in R'range(2) loop
          E := Zero;
          for K in Left'range(2) loop
            E := E + Left(I,K) * Right(Index(Integer(K)+Offset),J);
          end loop;
          R(I,J) := E;
        end loop;
      end loop;
    end if;
    return R;
  end "*";
  -----------------------------------------------------------------------------
  function "*" (Left : in Element; Right : in Matrix) return Matrix is
    
    R : Matrix (Right'range(1), Right'range(2));
  begin
    for I in R'range(1) loop
      for J in R'range(2) loop
        R(I,J) := Left * Right(I,J);
      end loop;
    end loop;
    return R;
  end "*";
  -----------------------------------------------------------------------------
  function "*" (Left : in Matrix; Right : in Element) return Matrix is
    
    R : Matrix (Left'range(1), Left'range(2));
  begin
    for I in R'range(1) loop
      for J in R'range(2) loop
        R(I,J) := Left(I,J) * Right;
      end loop;
    end loop;
    return R;
  end "*";
  -----------------------------------------------------------------------------
  function "/" (Left : in Matrix; Right : in Element) return Matrix is
    
    R : Matrix (Left'range(1), Left'range(2));
  begin
    for I in R'range(1) loop
      for J in R'range(2) loop
        R(I,J) := Left(I,J) / Right;
      end loop;
    end loop;
    return R;
  end "/";
  -----------------------------------------------------------------------------
  function Transpose (The_Matrix : in Matrix) return Matrix is
    
    R : Matrix (The_Matrix'range(2), The_Matrix'range(1));
  begin
    for I in R'range(1) loop
      for J in R'range(2) loop
        R(I,J) := The_Matrix(J,I);
      end loop;
    end loop;
    return R;
  end Transpose;
  ----------------------------------------------------------------------------
  function Submatrix(The_Matrix : in Matrix;
                     Row, Col   : in Index) return Matrix is

    R         : Matrix(1 .. The_Matrix'Length(1) - 1,
                       1 .. The_Matrix'Length(2) - 1);
    Row_Count,
    Col_Count : Index := 0;
  begin
    if The_Matrix'Length(1) /= The_Matrix'Length(2) then
      raise Matrix_Dimension_Error;
    else
      for I in The_Matrix'range(1) loop
        if I /= Row then
          Row_Count := Row_Count + 1;
          Col_Count := 0;
          for J in The_Matrix'range(2) loop
            if J /= Col then
              Col_Count := Col_Count + 1;
              R(Row_Count, Col_Count) := The_Matrix(I,J);
            end if;
          end loop;
        end if;
      end loop;
      return R;
    end if;
  end Submatrix;
  ---------------------------------------------------------------------------
  type Sign_Type is range -1 .. 1;
  ---------------------------------------------------------------------------
  function Sign_Of(Row, Col : in Index) return Sign_Type is
  begin
    if (Row + Col) rem 2 = 0 then
      return  1;
    else
      return -1;
    end if;
  end Sign_Of;
  ---------------------------------------------------------------------------
  function Adjoint(The_Matrix : in Matrix;
                   Row, Col   : in Index) return Element is

    Sign : constant Sign_Type := Sign_Of(Row, Col);
    Det  : constant Element   := 
           abs Submatrix(The_Matrix, Row => Row, Col => Col);
  begin
    if Sign = 1 then
      return   Det;
    else
      return - Det;
    end if;
  end Adjoint;
  ---------------------------------------------------------------------------
  function "abs" (Right : in Matrix) return Element is
    
    -----------------------------------
    procedure Row_Col_With_Min_Elements
             (M       : in     Matrix;
              Min_Row,
              Min_Col,
              Row_Min,
              Col_Min : in out Index) is

      Row_Count : array(M'range(1)) of Index := (others => 0);
      Col_Count : array(M'range(2)) of Index := (others => 0);
    begin
      for I in M'range(1) loop
        for J in M'range(2) loop
          if M(I,J) /= Zero then
            Row_Count(I) := Row_Count(I) + 1;
            Col_Count(J) := Col_Count(J) + 1;
          end if;
        end loop;
      end loop;
      Row_Min := M'Length(1);
      Col_Min := M'Length(2);
      Min_Row := 1;
      Min_Col := 1;
      for I in Row_Count'range loop
        if Row_Count(I) < Row_Min then
          Row_Min := Row_Count(I);
          Min_Row := I;
        end if;
        if Col_Count(I) < Col_Min then
          Col_Min := Col_Count(I);
          Min_Col := I;
        end if;
      end loop;
    end Row_Col_With_Min_Elements;
    ------------------------------------------------
    procedure Invert(The_Sign : in out Sign_Type) is
    begin
      if The_Sign = 1 then
        The_Sign := -1;
      else
        The_Sign :=  1;
      end if;
    end Invert;
    -----------------------------------------------------
    function Determinant(M : in Matrix) return Element is
    begin
      if M'Length(1) = 1 then
        return M(1,1);
      else
        declare
          Min_Row, Min_Col,
          Row_Min, Col_Min : Index;
          Det              : Element;
          Sign             : Sign_Type;
        begin
          Row_Col_With_Min_Elements(M,Min_Row => Min_Row, Min_Col => Min_Col,
                                      Row_Min => Row_Min, Col_Min => Col_Min);
          Det := Zero;
          if Row_Min <= Col_Min then
            Sign := Sign_Of(Row => Min_Row, Col => 1);
            for J in M'range(2) loop
              if M(Min_Row, J) /= Zero then
                if Sign = 1 then
                  Det := Det + M(Min_Row, J) * -- Recursion
                         Determinant(Submatrix(M, Row => Min_Row, Col => J));
                else -- sign = -1
                  Det := Det - M(Min_Row, J) * -- Recursion
                         Determinant(Submatrix(M, Row => Min_Row, Col => J));
                end if;
              end if;
              Invert(Sign);
            end loop;
          else
            Sign := Sign_Of(Row => 1, Col => Min_Col);
            for I in M'range(1) loop
              if M(I, Min_Col) /= Zero then
                if Sign = 1 then
                  Det := Det + M(I, Min_Col) * -- Recursion
                         Determinant(Submatrix(M, Row => I, Col => Min_Col));
                else -- sign = -1
                  Det := Det - M(I, Min_Col) * -- Recursion
                         Determinant(Submatrix(M, Row => I, Col => Min_Col));
                end if;
              end if;
              Invert(Sign);
            end loop;
          end if;
          return Det;
        end;
      end if;
    end Determinant;
    ----------------
  begin
    if    Right'Length(1) /= Right'Length(2) then
      raise Matrix_Dimension_Error;
    elsif Right'Length(1) = 1 then
      return Right(Right'First(1),Right'First(2));
    elsif Right'Length(1) = 0 then
      return Zero;
    elsif Right'First(1) = 1 and then Right'First(2) = 1 then
      return Determinant(Right);
    else
      declare
        X : constant Matrix(1 .. Right'Length(1), 
                            1 .. Right'Length(2)) := Right;
      begin
        return Determinant(X);
      end;
    end if;
  end "abs";
  ---------------------------------------------------------------------------
  function Inverse(The_Matrix : in Matrix) return Matrix is
    
    R   : Matrix(The_Matrix'range(1), The_Matrix'range(2));
    Det : Element;
  begin
    if    The_Matrix'Length(1) /= The_Matrix'Length(2) then
      raise Matrix_Dimension_Error;
    elsif The_Matrix'Length(1) = 1 then
      R(R'First(1), R'First(2)) := One / The_Matrix(The_Matrix'First(1), 
                                                    The_Matrix'First(2));
    else
      Det := abs The_Matrix;
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Adjoint(The_Matrix, Row => J, Col => I) / Det;
        end loop;
      end loop;
    end if;
    return R;
  exception
    when Constraint_Error => raise Singular_Matrix;
  end Inverse;
  ----------------------------------------------------------------------------
  function Elementwise_Multiplication 
          (Left, Right : in Matrix) return Matrix is
    
    R : Matrix (Left'range(1), Left'range(2));
  begin
    if    Left'Length(1) /= Right'Length(1) or else 
          Left'Length(2) /= Right'Length(2) then
      raise Matrix_Dimension_Error;
    elsif Left'First(1) = Right'First(1) and then 
          Left'First(2) = Right'First(2) then
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) * Right(I,J);
        end loop;
      end loop;
    else
      R := Right;
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Left(I,J) * R(I,J);
        end loop;
      end loop;
    end if;
    return R;
  end Elementwise_Multiplication;
  -----------------------------------------------------------------------------
  function  Copy (Outoff          : in Matrix;
                  From_First_Row,
                  To_Last_Row,
                  From_First_Col,
                  To_Last_Col     : in Index;
                  Multiply_With   : in Element := One) return Matrix is

    R : Matrix (From_First_Row .. To_Last_Row, From_First_Col .. To_Last_Col);
  begin
    if Multiply_With = One then
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Outoff(I,J);
        end loop;
      end loop;
    else
      for I in R'range(1) loop
        for J in R'range(2) loop
          R(I,J) := Multiply_With * Outoff(I,J);
        end loop;
      end loop;
    end if;
    return R;
  end Copy;
  -----------------------------------------------------------------------------
  procedure Copy (Into           : in out Matrix;
                  Source         : in     Matrix;
                  First_Row,
                  First_Col      : in     Index;
                  Multiply_With  : in     Element := One) is

    Col : Index;
    Row : Index := First_Row - 1;
  begin
    if Multiply_With = One then
      for I in Source'range(1) loop
        Row := Row + 1;
        Col := First_Col - 1;
        for J in Source'range(2) loop
          Col := Col + 1;
          Into(Row,Col) := Source(I,J);
        end loop;
      end loop;
    else
      for I in Source'range(1) loop
        Row := Row + 1;
        Col := First_Col - 1;
        for J in Source'range(2) loop
          Col := Col + 1;
          Into(Row,Col) := Multiply_With * Source(I,J);
        end loop;
      end loop;
    end if;
  end Copy;
  -----------------------------------------------------------------------------
  procedure Copy (Into           : in out Matrix;
                  Outoff         : in     Matrix;
                  From_First_Row,
                  To_Last_Row,
                  From_First_Col,
                  To_Last_Col,
                  First_Row,
                  First_Col      : in     Index;
                  Multiply_With  : in     Element := One) is

    Col : Index;
    Row : Index := First_Row - 1;
  begin
    if Multiply_With = One then
      for I in From_First_Row .. To_Last_Row loop
        Row := Row + 1;
        Col := First_Col - 1;
        for J in From_First_Col .. To_Last_Col loop
          Col := Col + 1;
          Into(Row, Col) := Outoff(I,J);
        end loop;
      end loop;
    else
      for I in From_First_Row .. To_Last_Row loop
        Row := Row + 1;
        Col := First_Col - 1;
        for J in From_First_Col .. To_Last_Col loop
          Col := Col + 1;
          Into(Row, Col) := Multiply_With * Outoff(I,J);
        end loop;
      end loop;
    end if;
  end Copy;
  -----------------------------------------------------------------------------
  procedure Accumulate (In_Matrix      : in out Matrix;
                        Source         : in     Matrix;
                        First_Row,
                        First_Col      : in     Index;
                        Positive_Accu  : in     Boolean := True;
                        Multiply_With  : in     Element := One) is

    Col : Index;
    Row : Index := First_Row - 1;
  begin
    if Positive_Accu then
      if Multiply_With = One then
        for I in Source'range(1) loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in Source'range(2) loop
            Col := Col + 1;
            In_Matrix(Row,Col) := In_Matrix(Row,Col) + Source(I,J);
          end loop;
        end loop;
      else
        for I in Source'range(1) loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in Source'range(2) loop
            Col := Col + 1;
            In_Matrix(Row,Col) := In_Matrix(Row,Col) +
                                  Multiply_With * Source(I,J);
          end loop;
        end loop;
      end if;
    else
      if Multiply_With = One then
        for I in Source'range(1) loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in Source'range(2) loop
            Col := Col + 1;
            In_Matrix(Row,Col) := In_Matrix(Row,Col) - Source(I,J);
          end loop;
        end loop;
      else
        for I in Source'range(1) loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in Source'range(2) loop
            Col := Col + 1;
            In_Matrix(Row,Col) := In_Matrix(Row,Col) -
                                  Multiply_With * Source(I,J);
          end loop;
        end loop;
      end if;
    end if;
  end Accumulate;
  -----------------------------------------------------------------------------
  procedure Accumulate (In_Matrix      : in out Matrix;
                        Outoff         : in     Matrix;
                        From_First_Row,
                        To_Last_Row,
                        From_First_Col,
                        To_Last_Col,
                        First_Row,
                        First_Col      : in     Index;
                        Positive_Accu  : in     Boolean := True;
                        Multiply_With  : in     Element := One) is

    Col : Index;
    Row : Index := First_Row - 1;
  begin
    if Positive_Accu then
      if Multiply_With = One then
        for I in From_First_Row .. To_Last_Row loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in From_First_Col .. To_Last_Col loop
            Col := Col + 1;
            In_Matrix(Row, Col) := In_Matrix(Row, Col) + Outoff(I,J);
          end loop;
        end loop;
      else
        for I in From_First_Row .. To_Last_Row loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in From_First_Col .. To_Last_Col loop
            Col := Col + 1;
            In_Matrix(Row, Col) := In_Matrix(Row, Col) +
                                   Multiply_With * Outoff(I,J);
          end loop;
        end loop;
      end if;
    else
      if Multiply_With = One then
        for I in From_First_Row .. To_Last_Row loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in From_First_Col .. To_Last_Col loop
            Col := Col + 1;
            In_Matrix(Row, Col) := In_Matrix(Row, Col) - Outoff(I,J);
          end loop;
        end loop;
      else
        for I in From_First_Row .. To_Last_Row loop
          Row := Row + 1;
          Col := First_Col - 1;
          for J in From_First_Col .. To_Last_Col loop
            Col := Col + 1;
            In_Matrix(Row, Col) := In_Matrix(Row, Col) -
                                   Multiply_With * Outoff(I,J);
          end loop;
        end loop;
      end if;
    end if;
  end Accumulate;
  -----------------------------------------------------------------------------
  procedure Zero_Matrix (The_Matrix : out Matrix) is
  begin
    The_Matrix := Matrix'(The_Matrix'range(1) => 
                         (The_Matrix'range(2) => Zero));
  end Zero_Matrix;
  -----------------------------------------------------------------------------
  function  Zero_Matrix (Of_Order : in Index) return Matrix is
    
    R : constant Matrix(1 .. Of_Order, 1 .. Of_Order) := 
                       (others => (others => Zero));
  begin
    return R;
  end Zero_Matrix;
  -----------------------------------------------------------------------------
  procedure Identity_Matrix (The_Matrix : out Matrix) is
    
    Offset: Integer;
  begin
    if    The_Matrix'Length(1) /= The_Matrix'Length(2) then
      raise Matrix_Dimension_Error;
    elsif The_Matrix'First(1) = The_Matrix'First(2) then
      The_Matrix := Matrix'(The_Matrix'range(1) => 
                           (The_Matrix'range(2) => Zero));
      for I in The_Matrix'range(1) loop
        The_Matrix(I,I) := One;
      end loop;
    else
      Offset := Integer(The_Matrix'First(2)) - Integer(The_Matrix'First(1));
      The_Matrix := Matrix'(The_Matrix'range(1) => 
                           (The_Matrix'range(2) => Zero));
      for I in The_Matrix'range(1) loop
        The_Matrix(I, Index(Integer(I) + Offset)) := One;
      end loop;
     end if;
  end Identity_Matrix;
  -----------------------------------------------------------------------------
  function  Identity_Matrix (Of_Order : in Index) return Matrix is
    
    R : Matrix (1 .. Of_Order, 1 .. Of_Order);
  begin
    Identity_Matrix(R);
    return R;
  end Identity_Matrix;
  -----------------------------------------------------------------------------
end Math.Generics.Arrays.Matrix_Operations;
