--                              -*- Mode: Ada -*-
-- Filename        : math-generics-interpolation.ads
-- Description     : Package for generic interpolation types and algorithms
-- Author          : J. Schrer
-- Created On      : 22. April 1998
-- Last Modified By: 
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Finalization;

generic
  type Float_Type is digits <>;
  
package Math.Generics.Interpolation is
  
  -- general array types
  
  type Array_1d is array(Integer range <>) of Float_Type;

  type Array_2d is array(Integer range <>,
                         Integer range <>) of Float_Type;

  type Array_3d is array(Integer range <>,
                         Integer range <>,
                         Integer range <>) of Float_Type;
  
  -- type for value ranges                                                                          
                         
  type Search_Result is record
    In_Range : Boolean    := False; -- See 'binary_search'-explanation
    Index    : Integer    := 0;     -- actual index
    Factor   : Float_Type := 0.0;   -- (the_value - allvalues(index)) /
  end record;                       -- (the_range(index + 1) - the_range(index))                        
         
  -- types for equidistant and non-equidistant range values
  
  type Range_Type is abstract new Ada.Finalization.Controlled with null record;
  
  function Search(The_Value  : in Float_Type;
                  The_Range : in Range_Type) return Search_Result is abstract;
                  
  function Is_In (The_Value  : in Float_Type;
                  The_Range  : in Range_Type) return Boolean is abstract;                  
                  
  type Range_Type_Access is access all Range_Type'Class;
  
  
  type Equidistant_Range is new Range_Type with record
    Min       : Float_Type := Float_Type'First;
    Max       : Float_Type := Float_Type'Last;
    Increment : Float_Type := 1.0;
    First     : Integer    := 1;
    Last      : Integer    := 0;
  end record;
  
  procedure Initialize(The : in out Equidistant_Range);
  
  -- Calculates the.last.
  
  -- and search functions
  
  function Search(The_Value : in Float_Type;
                  The_Range : in Equidistant_Range) return Search_Result;
    
  function Is_In (The_Value : in Float_Type;
                  The_Range : in Equidistant_Range) return Boolean;
                                                  
  type Non_Equidistant_Range(First, Last : Integer) is new Range_Type with record
    Values : Array_1d(First .. Last);
  end record;
                                                        
  function Search(The_Value : in Float_Type;
                  The_Range : in Non_Equidistant_Range) return Search_Result;
                         
  -- Perfomrs a binary search for The_Value in array The_Range and
  -- returns an Index, an In_Range and a Factor as follows:
                         
	--  if    The_Value < the_range(the_range'First) then
	--    return (In_Range => False,
  --            Index    => the_range'First,
	--            Factor   => 0.0);                                                  
	--  elsif The_Value > the_range(the_range'Last) then
	--    return (In_Range => False,
  --            Index    => the_range'Last,
	--            Factor   => 0.0);                                                  
	--  else -- I in the_range'First .. the_range'Last - 1
	--    return (In_Range => True,
  --            Index    => I,
	--            Factor   => (The_Value - the_range(I)) / 
	--                        (the_range(I + 1) - the_range(I)));                                                  
	                         
  -- Table types for "regular" tables
                             
  function Is_In (The_Value : in Float_Type;
                  The_Range : in Non_Equidistant_Range) return Boolean;
                  
	type Table_1d(Last_1  : Integer := 0;
	              First_1 : Integer := 1) is record
	  Values_1     : Range_Type_Access;
	  Table_Values : Array_1d(First_1 .. Last_1);
  end record;
                
	type Table_2d(Last_1,  Last_2  : Integer := 0;
	              First_1, First_2 : Integer := 1) is record
	  Values_1     : Range_Type_Access;
	  Values_2     : Range_Type_Access;
	  Table_Values : Array_2d(First_1 .. Last_1, 
                            First_2 .. Last_2);
	end record;
                  
	type Table_3d(Last_1,  Last_2,  Last_3  : Integer := 0;
	              First_1, First_2, First_3 : Integer := 1) is record
	  Values_1     : Range_Type_Access;
	  Values_2     : Range_Type_Access;
	  Values_3     : Range_Type_Access;
	  Table_Values : Array_3d(First_1 .. Last_1, 
                            First_2 .. Last_2, 
                            First_3 .. Last_3);
	end record;
                  
  function Interpolated_Value(The_Table : in Array_1d;
                              Value_1   : in Search_Result) return Float_Type;
                           
  function Interpolated_Value(The_Table : in Array_2d;
                              Value_1   : in Search_Result;
                              Value_2   : in Search_Result) return Float_Type;
                           
  function Interpolated_Value(The_Table : in Array_3d;
                              Value_1   : in Search_Result;
                              Value_2   : in Search_Result;
                              Value_3   : in Search_Result) return Float_Type;
                              
  -- "irregular" table 2D
                              
  type Array_1d_Access is access Array_1d;
  
  type Array_2d_Irregular is array(Integer range <>) of Array_1d_Access;
  
  type Table_2d_Irregular(Last_1  : Integer := 0;
	                        First_1 : Integer := 1) is record
	  Values_1     : Range_Type_Access;
	  Values_2     : Range_Type_Access;
	  Table_Values : Array_2d_Irregular(First_1 .. Last_1);
	end record;
                           
  function Interpolated_Value(The_Table : in Array_2d_Irregular;
                              Value_1   : in Search_Result;
                              Value_2   : in Search_Result) return Float_Type;
                              
end Math.Generics.Interpolation;
