--                              -*- Mode: Ada -*-
-- Filename        : portable-Operating_System.adb
-- Description     : Version for Windows.
-- Author          : J. Schroeer, AMST
-- Created On      : Wed Feb  5 14:41:56 1997
-- Last Modified By:
-- Last Modified On: Thu Jul  3 22:41:44 1997
-- Update Count    : 23
-- Status          : ok

with Ada.Text_Io, Ada.Direct_Io, Ada.Exceptions, Ada.Strings.Unbounded;
with Interfaces.C;
with Windows.Utils;
with Utilities.String_Utilities, Utilities.String_Stack, Utilities.Files;
with Win32, Win32.Winbase, Win32.Winnt;
with System;

package body Portable.Operating_System is

  Max_Path_Length : constant Interfaces.C.Size_T := 260;

--  the_Newline             : constant String    := "" & Ascii.Lf;       -- Unix
--  the_Directory_Separator : constant Character := '/';                 -- Unix
--  the_Path_Separator      : constant Character := ':';                 -- Unix
--  Executable_Suffix       : constant String    := "";                  -- Unix

  The_New_Line            : constant String    := Ascii.Cr & Ascii.Lf; -- Windows
  The_Directory_Separator : constant Character := '\';                 -- Windows
  The_Path_Separator      : constant Character := ';';                 -- Windows
  Executable_Suffix       : constant String    := ".exe";              -- Windows

  use type Win32.Bool;
  use type System.Address; -- For handle types (are subtypes of system.address)
  use type Interfaces.C.Size_T, Interfaces.C.Unsigned_Long;
  use type Ada.Strings.Unbounded.String_Access;


  package String_Utilities renames Utilities.String_Utilities;
  -------------------------------------------------------------------------
  function New_Line return String is
  begin
    return The_New_Line;
  end New_Line;
  -------------------------------------------------------------------------
  function Directory_Separator return Character is
  begin
    return The_Directory_Separator;
  end Directory_Separator;
  -------------------------------------------------------------------------
  function Path_Separator return Character is
  begin
    return The_Path_Separator;
  end Path_Separator;
  -------------------------------------------------------------------------
  function Environment_Value_Of (Name : in String) return String is

    C_Name : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Name);
    Buffer : Interfaces.C.Char_Array(0 .. 1023);
    Length : constant Interfaces.C.Size_T := Interfaces.C.Size_T
                     (Win32.Winbase.Getenvironmentvariable
                          (Lpname   => C_Name(C_Name'First)'Unchecked_Access,
                           Lpbuffer => Buffer(Buffer'First)'Unchecked_Access,
                           Nsize    => Buffer'Length));
  begin
    if    Length = 0 then
      return "";
    elsif Length in Buffer'range then
      return Interfaces.C.To_Ada(Buffer(Buffer'First .. Length));
    else
      declare
        New_Buffer : Interfaces.C.Char_Array(0 .. Length);
        New_Length : constant Interfaces.C.Size_T := Interfaces.C.Size_T
                    (Win32.Winbase.Getenvironmentvariable
                    (Lpname   => C_Name(C_Name'First)'Unchecked_Access,
                     Lpbuffer => New_Buffer(New_Buffer'First)'Unchecked_Access,
                     Nsize    => New_Buffer'Length));
      begin
        return Interfaces.C.To_Ada(New_Buffer(New_Buffer'First .. New_Length));
      end;
    end if;
  end Environment_Value_Of;
  --------------------------------------------------------------------------
  procedure Change_Working_Directory (Directory_Name : in String) is

    C_Name : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Directory_Name);
  begin
    if Win32.Winbase.Setcurrentdirectory
           (C_Name(C_Name'First)'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.Change_Working_Directory" &
         New_Line & "  ("  & '"' & Directory_Name & '"' & ") failed.");
    end if;
  end Change_Working_Directory;
  ----------------------------------------------------------------------------
  function Get_Working_Directory return String is

    Buffer : Interfaces.C.Char_Array(0 .. Max_Path_Length);
    Length : constant Interfaces.C.Size_T := Interfaces.C.Size_T
            (Win32.Winbase.Getcurrentdirectory
               (Nbufferlength => Buffer'Length,
                Lpbuffer      => Buffer(Buffer'First)'Unchecked_Access));
  begin
    if Length in Buffer'range then
      return Interfaces.C.To_Ada(Buffer(Buffer'First .. Length));
    else
      declare
        New_Buffer : Interfaces.C.Char_Array(0 .. Length);
        New_Length : constant Interfaces.C.Size_T := Interfaces.C.Size_T
                    (Win32.Winbase.Getcurrentdirectory
                    (Nbufferlength => New_Buffer'Length,
                     Lpbuffer      => New_Buffer(New_Buffer'First)'Unchecked_Access));
      begin
        return Interfaces.C.To_Ada(New_Buffer(New_Buffer'First .. New_Length));
      end;
    end if;
  end Get_Working_Directory;
  ----------------------------------------------------------------------------
  procedure Create_Directory (Pathname : in String) is

    C_Name : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Pathname);
  begin
    if Win32.Winbase.Createdirectory
          (Lppathname           => C_Name(C_Name'First)'Unchecked_Access,
           Lpsecurityattributes => null) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.Create_Directory" &
         New_Line & "  (" & '"' & Pathname & '"' & ") failed.");
    end if;
  end Create_Directory;
  ----------------------------------------------------------------------------
  procedure Remove_Directory (Pathname : in String) is

    C_Name : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Pathname);
  begin
    if Win32.Winbase.Removedirectory
           (C_Name(C_Name'First)'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.Remove_Directory" &
         New_Line & "  (" & '"' & Pathname & '"' & ") failed.");
    end if;
  end Remove_Directory;
  ----------------------------------------------------------------------------
  function Is_File     (Pathname: in String) return Boolean is

    C_Name    : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Pathname);
    File_Attr : constant Interfaces.C.Unsigned_Long :=
                Win32.Winbase.Getfileattributes
               (C_Name(C_Name'First)'Unchecked_Access);
  begin
    return File_Attr = Win32.Winnt.File_Attribute_Normal   or else
           File_Attr = Win32.Winnt.File_Attribute_Readonly or else
           File_Attr = Win32.Winnt.File_Attribute_Hidden   or else
           File_Attr = Win32.Winnt.File_Attribute_Archive  or else
           File_Attr = Win32.Winnt.File_Attribute_System   or else
           File_Attr = Win32.Winnt.File_Attribute_Temporary;
  end Is_File;
  ----------------------------------------------------------------------------
  function Is_Directory(Pathname: in String) return Boolean is

    C_Name    : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Pathname);
    File_Attr : constant Interfaces.C.Unsigned_Long :=
                Win32.Winbase.Getfileattributes
               (C_Name(C_Name'First)'Unchecked_Access);
  begin
    return File_Attr = Win32.Winnt.File_Attribute_Directory;
  end Is_Directory;
  ----------------------------------------------------------------------------
  procedure For_Every_Directory_Entry(Pathname : in String) is

    Quit        : Boolean := False;
    Current_Dir : constant String := Get_Working_Directory;
    Pattern     : aliased Interfaces.C.Char_Array := Interfaces.C.To_C("*");
    Find_Data   : aliased Win32.Winbase.Win32_Find_Dataa;
    Handle      : System.Address;
  begin
    Change_Working_Directory(Pathname);
    Handle := Win32.Winbase.Findfirstfile
	                (Lpfilename     => Pattern(Pattern'First)'Unchecked_Access,
	                 Lpfindfiledata => Find_Data'Unchecked_Access);
		if Handle = Win32.Winbase.Invalid_Handle_Value then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.For_Every_Directory_Entry" &
                    New_Line & "  (" & '"' & Pathname & '"' &
                    ") failed. (Win32.Winbase.Findfirstfile)");
    end if;
		Action(D_Entry => Windows.Utils.To_Ada(Find_Data.Cfilename), Quit => Quit);
 	  while not Quit and then
		      Win32.Winbase.Findnextfile
		          (Hfindfile      => Handle,
		           Lpfindfiledata => Find_Data'Unchecked_Access) = Win32.True loop
		  Action(D_Entry => Windows.Utils.To_Ada(Find_Data.Cfilename), Quit => Quit);
    end loop;
    Change_Working_Directory(Current_Dir);
  end For_Every_Directory_Entry;
  ----------------------------------------------------------------------------                    
  function User return String is

    Buffer : Interfaces.C.Char_Array(0 .. 127);
    Length : aliased Interfaces.C.Unsigned_Long :=
                     Interfaces.C.Unsigned_Long(Buffer'Length);
  begin
    if Win32.Winbase.Getusername
           (Lpbuffer => Buffer(Buffer'First)'Unchecked_Access,
            Nsize    => Length'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.user failed.");
    end if;
    return Interfaces.C.To_Ada
          (Buffer(Buffer'First .. Interfaces.C.Size_T(Length)));
  end User;
  ----------------------------------------------------------------------------                      
  function Owner(Pathname : in String) return String is

    C_Name            : aliased Interfaces.C.Char_Array :=
                        Interfaces.C.To_C(Pathname, Append_Nul => True);
--    C_User            : aliased Interfaces.C.Char_Array :=
--                        Interfaces.C.To_C(User, Append_Nul => True);
    Security          : aliased Win32.Winnt.Security_Descriptor;
    Length            : aliased Interfaces.C.Unsigned_Long;
    Owner_Defaulted   : aliased Win32.Bool;
    Psid_Address      : aliased System.Address;
    Name              : aliased Interfaces.C.Char_Array(0 .. 127);
    Domain_Name       : aliased Interfaces.C.Char_Array(0 .. 127);
    Name_Length,
    Domain_Name_Length: aliased Interfaces.C.Unsigned_Long                                                                ;
    Sid_Name_Use      : aliased Win32.Winnt.Sid_Name_Use;
  begin
    if Win32.Winbase.Getfilesecurity
           (Lpfilename           => C_Name(C_Name'First)'Unchecked_Access,
            Requestedinformation => Win32.Winnt.Owner_Security_Information,
            Psecuritydescriptor  => Security'Address,
            Nlength              => Interfaces.C.Unsigned_Long(Security'Size / 8),
            Lpnlengthneeded      => Length'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.owner(" & Pathname &
                    ") failed (Win32.Winbase.Getfilesecurity).");
    end if;
    if Length /= 0 then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.owner(" & Pathname &
                    ") failed (Win32.Winbase.Getfilesecurity, Lpnlengthneeded /= 0).");
    end if;
    if Win32.Winbase.Getsecuritydescriptorowner
           (Psecuritydescriptor => Security'Address,
            Powner              => Psid_Address'Unchecked_Access,
            Lpbownerdefaulted   => Owner_Defaulted'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.owner(" & Pathname &
                    ") failed (Win32.Winbase.Getsecuritydescriptorowner).");
    end if;
    if Win32.Winbase.Lookupaccountsid
           (Lpsystemname           => null,
            Sid                    => Psid_Address,
            Name                   => Name(Name'First)'Unchecked_Access,
            Cbname                 => Name_Length'Unchecked_Access,
            Referenceddomainname   => Domain_Name(Domain_Name'First)'Unchecked_Access,
            Cbreferenceddomainname => Domain_Name_Length'Unchecked_Access,
            Peuse                  => Sid_Name_Use'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.owner(" & Pathname &
                    ") failed (Win32.Winbase.Lookupaccountsid).");
    end if;
    return Interfaces.C.To_Ada
          (Name(Name'First .. Interfaces.C.Size_T(Name_Length)));
  end Owner;
  ----------------------------------------------------------------------------                    
  procedure Null_Operation(The_Pathname : in     String;
                           Continue     : in out Boolean) is
  begin
    null;
  end Null_Operation;
  ----------------------------------------------------------------------------
  procedure Process_Files(The_Directory : in String  := Get_Working_Directory;
                          Recursive     : in Boolean := True) is

    Continue : Boolean := True;
    ---------------------------------------------------------
    procedure Process_Directory(Directory : in String) is
      -------------------------------------------
      procedure Process(The_Entry : in String) is
      begin
        if    Is_File(The_Entry) then
          File_Operation(The_Pathname => The_Entry,
                         Continue     => Continue);
        elsif Recursive and then Is_Directory(The_Entry) then
          Process_Directory(Directory => The_Entry); -- Recursion
        end if;
      end Process;
      ------------
      Pattern     : aliased Interfaces.C.Char_Array := Interfaces.C.To_C("*");
      Find_Data   : aliased Win32.Winbase.Win32_Find_Dataa;
      Handle      : System.Address;
      Current_Dir : constant String := Get_Working_Directory;
    begin
      if not Continue or else Directory = "." or else Directory = ".." then
        return;
      end if;
      Change_Working_Directory(Directory);
      Directory_Pre_Operation(Directory, Continue);
      if Continue then
		    Handle := Win32.Winbase.Findfirstfile
		                  (Lpfilename     => Pattern(Pattern'First)'Unchecked_Access,
		                   Lpfindfiledata => Find_Data'Unchecked_Access);
		    if Handle = Win32.Winbase.Invalid_Handle_Value then
          Ada.Exceptions.Raise_Exception(Os_Error'Identity,
             Message => "portable.operating_system.Process_Files" &
                        New_Line & "  (" & '"' & Directory & '"' &
                        ") failed. (Win32.Winbase.Findfirstfile)");
		    end if;
		    Process(The_Entry => Windows.Utils.To_Ada(Find_Data.Cfilename));
		    while Continue and then
		          Win32.Winbase.Findnextfile
		              (Hfindfile      => Handle,
		               Lpfindfiledata => Find_Data'Unchecked_Access) = Win32.True loop
		      Process(The_Entry => Windows.Utils.To_Ada(Find_Data.Cfilename));
		    end loop;
      end if;
      Directory_Post_Operation(Directory, Continue);
      Change_Working_Directory(Current_Dir);
    end Process_Directory;
    ----------------------
  begin
    Process_Directory(The_Directory);
  end Process_Files;
  ----------------------------------------------------------------------------   
  function All_Entries(Directory         : in String  := Get_Working_Directory;
                       Recursive         : in Boolean := True;
                       Relative_Pathnames: in Boolean := True;
                       Directories,
                       Files             : in Boolean := True)
           return Utilities.String_Vectors.String_Vector is
    ------------------------------------------------------------------------
    function Absolute_Dirname_Of(The_Directory : in String) return String is
      Current_Dir : constant String := Get_Working_Directory;
    begin
      Change_Working_Directory(The_Directory);
      declare
        Dir : constant String := Get_Working_Directory;
      begin
        Change_Working_Directory(Current_Dir);
        return Dir;
      end;
    end Absolute_Dirname_Of;
    ---------------------------------------------------------------       
    Root_Dir   : constant String := Absolute_Dirname_Of(Directory);
    Name_Stack : Utilities.String_Stack.Stack;
    ---------------------------------------------------------------
    procedure Process_Directory(Directory_Name : in String) is
      ------------------------------------------------
      Current_Dir : constant String := Absolute_Dirname_Of(Directory_Name);
	    ------------------------------------------------
	    procedure Save_Name(The_Pathname : in String) is
        Name : Ada.Strings.Unbounded.String_Access;
      begin
        if Relative_Pathnames then
          if Current_Dir = Root_Dir then
            Name := new String'(The_Pathname);
          else
            Name := new String'(Current_Dir(Root_Dir'Last + 2 ..
                    Current_Dir'Last) & The_Directory_Separator & The_Pathname);
          end if;
        else
          Name := new String'(Current_Dir &
                  The_Directory_Separator & The_Pathname);
        end if;
	      Utilities.String_Stack.Push(Item => Name, Onto => Name_Stack);
	    end Save_Name;
      ------------------------------------------------------- 
	    procedure Process_Entry(Entry_Name : in     String;
	                            Quit       : in out Boolean) is
	    begin
        if Files and then Is_File(Entry_Name) then
          Save_Name(Entry_Name);
	      elsif Is_Directory(Entry_Name) then
	        if Entry_Name = "." or else Entry_Name = ".." then
	          return;
	        end if;
	        if Directories then
            Save_Name(Entry_Name);
	        end if;
	        if Recursive then
            Process_Directory(Directory_Name => Entry_Name); -- RECURSION
          end if;
	      end if;
	    end Process_Entry;
      -------------------------------------------------------------
      procedure Traverse_Directory is new For_Every_Directory_Entry
               (Action => Process_Entry);
    begin
      Traverse_Directory(Directory_Name);
    end Process_Directory;
    ----------------------
  begin
    Process_Directory(Directory);
    declare
      Names : Utilities.String_Vectors.String_Vector
             (1 .. Utilities.String_Stack.Depth_Of(Name_Stack));
    begin
      Utilities.String_Stack.Convert(The_Stack => Name_Stack,
                                     To_Vector => Names);
      return Names;
    end;
  end All_Entries;
  ----------------------------------------------------------------------------
  function File_Time_Stamp      (Name       : in String;
                                 Local_Time : in Boolean := True)
           return Ada.Calendar.Time is

    C_Name           : aliased Interfaces.C.Char_Array := Interfaces.C.To_C(Name);
    File_Handle      : Win32.Winnt.Handle;
    Creation_Time,
    Last_Access_Time,
    Last_Write_Time,
    Last_Time        : aliased Win32.Winbase.Filetime;
    System_Time      : aliased Win32.Winbase.Systemtime;
  begin
    if Is_File(Name) then
      File_Handle := Win32.Winbase.Createfile
                    (Lpfilename           => C_Name(C_Name'First)'Unchecked_Access,
--                     Dwdesiredaccess      => Win32.Winnt.Generic_Read, 
                     Dwdesiredaccess      => 0,
                     Dwsharemode          => Win32.Winnt.File_Share_Read or
                                             Win32.Winnt.File_Share_Write,
                     Lpsecurityattributes => null,
                     Dwcreationdisposition=> Win32.Winbase.Open_Existing,
                     Dwflagsandattributes => Win32.Winnt.File_Attribute_Normal,
                     Htemplatefile        => System.Null_Address);
    else
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.file_time_stamp" & New_Line &
         "  (" & '"' & Name & '"' & ") failed (not existing).");
--      return Ada.Calendar.Time_Of(Year  => Ada.Calendar.Year_Number'First,
--                                  month => Ada.Calendar.month_Number'First,
--                                  Day   => Ada.Calendar.day_Number'First);
    end if;
    if File_Handle = Win32.Winbase.Invalid_Handle_Value then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.file_time_stamp" & New_Line &
         "  (" & '"' & Name & '"' & ") failed. (Win32.Winbase.Createfile)");
    end if;
    if Win32.Winbase.Getfiletime
           (Hfile           => File_Handle,
            Lpcreationtime  => Creation_Time'Unchecked_Access,
            Lplastaccesstime=> Last_Access_Time'Unchecked_Access,
            Lplastwritetime => Last_Write_Time'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.file_time_stamp" & New_Line &
         "  (" & '"' & Name & '"' & ") failed. (Win32.Winbase.Getfiletime)");
    end if;
    if Local_Time then
      if Win32.Winbase.Filetimetolocalfiletime
             (Lpfiletime     => Last_Write_Time'Unchecked_Access,
              Lplocalfiletime=> Last_Time'Unchecked_Access) = Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
           Message => "portable.operating_system.file_time_stamp" & New_Line &
           "  (" & '"' & Name & '"' & ") failed. (Win32.Winbase.Filetimetolocalfiletime)");
      end if;
    else
      Last_Time := Last_Write_Time;
    end if;
    if Win32.Winbase.Filetimetosystemtime
           (Lpfiletime  => Last_Time'Unchecked_Access,
            Lpsystemtime=> System_Time'Unchecked_Access) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.file_time_stamp" & New_Line &
         "  (" & '"' & Name & '"' & ") failed. (Win32.Winbase.Filetimetosystemtime)");
    end if;
    if Win32.Winbase.Closehandle(File_Handle) = Win32.False then
      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
         Message => "portable.operating_system.file_time_stamp" & New_Line &
         "  (" & '"' & Name & '"' & ") failed. (Win32.Winbase.Closehandle)");
    end if;
    return Ada.Calendar.Time_Of
          (Year    => Ada.Calendar.Year_Number (System_Time.Wyear),
           Month   => Ada.Calendar.Month_Number(System_Time.Wmonth),
           Day     => Ada.Calendar.Day_Number  (System_Time.Wday),
           Seconds => Duration(Float(System_Time.Whour)   * 3600.0 +
                               Float(System_Time.Wminute) * 60.0   +
                               Float(System_Time.Wsecond)          +
                               Float(System_Time.Wmilliseconds) / 1000.0));
  end File_Time_Stamp;
  --------------------------------------------------------------------------
  -- local functions 
  --------------------------------------------------------------------------
  function Full_Pathname_Of(Filename  : in String;
                            Directory : in String;
                            Recursive : in Boolean) return String is

    Is_Pathname : constant Boolean := String_Utilities.Has
                                  (The_String    => Filename,
                                   The_Character => The_Directory_Separator);
    Found_Name  : Ada.Strings.Unbounded.String_Access;
    --------------------------------------------------------
--    procedure Save_Filename(The_Pathname : in     String; 
--                            Continue     : in out Boolean) is
--    begin
--      if Is_Pathname then 
--        declare
--          Full_Name : constant String := Get_Working_Directory &
--                                         The_Directory_Separator & The_Pathname;
--        begin
--          if String_Utilities.Right_Part_Slice
--                  (From_The_String => Full_Name,
--                   Number          => Filename'Length) = Filename then
--            Found_Name := new String'(Full_Name);
--            Continue   := False;
--          end if;
--        end;          
--      elsif The_Pathname = Filename then
--        Found_Name := new String'(Get_Working_Directory &
--                          The_Directory_Separator & The_Pathname);
--        Continue   := False;
--      end if;
--    end Save_Filename;                              
--    ------------------------------------------------------------------------
--    procedure Process is new Process_Files(File_Operation => Save_Filename);
    ------------------------------------------------------------------------
    procedure Search_Filename(The_Pathname : in     String;
                              Continue     : in out Boolean) is
    begin
      if Is_Pathname then
        declare
          Full_Name : constant String := Get_Working_Directory &
                                         The_Directory_Separator & Filename;
        begin
          if Is_File(Full_Name) then
            Found_Name := new String'(Full_Name);
            Continue   := False;
          end if;
        end;
      elsif Is_File(Filename) then
        Found_Name := new String'(Get_Working_Directory &
                          The_Directory_Separator & Filename);
        Continue   := False;
      end if;
    end Search_Filename;
    ------------------------------------------------------
    procedure Process is new Process_Files
             (Directory_Pre_Operation => Search_Filename);
    ------------------------------------------------------
  begin
    begin
      Process(The_Directory => Directory, Recursive => Recursive);
    exception
      when Os_Error => null;
    end;
    if Found_Name = null then
      return "";
    else
      declare
        Name : constant String := Found_Name.all;
      begin
        Ada.Strings.Unbounded.Free(Found_Name);
        return Name;
      end;
    end if;
  end Full_Pathname_Of;
  --------------------------------------------------------------------------
  function Locate(Filename : in String;
                  Path     : in String) return String is
  begin
    if String_Utilities.Has(The_String    => Path,
                            The_Character => The_Path_Separator) then
      declare
        Dir_Names : Utilities.String_Vectors.String_Vector :=
                    Utilities.String_Vectors.Split
                            (The_String => Path,
                             Separator  => The_Path_Separator);
      begin
        for I in Dir_Names'range loop
          declare
            Name : constant String :=
                   Full_Pathname_Of(Filename  => Filename,
                                    Directory => Dir_Names(I).all,
                                    Recursive => False);
          begin
            if Name /= "" then
              Utilities.String_Vectors.Deallocate(Dir_Names);
              return Name;
            end if;
          end;
        end loop;
        Utilities.String_Vectors.Deallocate(Dir_Names);
        return "";
      end;
    else
      return Full_Pathname_Of(Filename  => Filename,
                              Directory => Path,
                              Recursive => True);
    end if;
  end Locate;
  --------------------------------------------------------------------------
  -- End local functions
  -------------------------------------------------------------------------
  function Locate_Exec_On_Path(Exec_Name : in String) return String is
  begin
    if Executable_Suffix = Utilities.Files.Suffix_Of(Exec_Name) then
      return Locate(Filename => Exec_Name,
                    Path     => Environment_Value_Of("PATH"));
    else
      return Locate(Filename => Exec_Name & Executable_Suffix,
                    Path     => Environment_Value_Of("PATH"));
    end if;
  end Locate_Exec_On_Path;
  --------------------------------------------------------------------------
  function Locate_Regular_File(File_Name : in String;
                               Path      : in String :=
                                              Environment_Value_Of("PATH"))
           return String is
  begin
    return Locate(Filename => File_Name, Path => Path);
  end Locate_Regular_File;
  --------------------------------------------------------------------------
  function Size_Of(Name : in String) return Natural is

    type Byte is mod 255;
    for  Byte'Size use 8;

    package Byte_Direct_Io is new Ada.Direct_Io(Byte);

    File : Byte_Direct_Io.File_Type;
    Size : Natural;
  begin
    Byte_Direct_Io.Open(File, Mode => Byte_Direct_Io.In_File,
                              Name => Name,
                              Form => "shared=no");

    -- The Form parameter with shared = yes|no is GNAT specific, see
    -- GNAT reference manual, chapter 6.8, shared files.
    -- In order to use open for already opened external files, shared = no
    -- is required. If an external file with name Name is already open and
    -- "shared=yes" is assigned to Form, USE_ERROR is raised.

    Size := Natural(Byte_Direct_Io.Size(File));
    Byte_Direct_Io.Close(File);
    return Size;
  end Size_Of;
  ---------------------------------------------------------------------------
  procedure Rename(Old_Pathname,
                   New_Pathname        : in String;
                   Overwrite_If_Exists : in Boolean := True) is
    -----------------------------------------------------------
    procedure Rename_File is

      C_Oldname : aliased Interfaces.C.Char_Array :=
                          Interfaces.C.To_C(Old_Pathname);
      C_Newname : aliased Interfaces.C.Char_Array :=
                          Interfaces.C.To_C(New_Pathname);
    begin
      if Win32.Winbase.Movefile
             (Lpexistingfilename => C_Oldname(C_Oldname'First)'Unchecked_Access,
              Lpnewfilename      => C_Newname(C_Newname'First)'Unchecked_Access) =
         Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
           Message => "portable.operating_system.rename" & New_Line &
            "  (" & '"' & Old_Pathname & '"' & ", " & '"' &
                          New_Pathname & '"' & ") failed.");
      end if;
--      Execute("ren " & Old_Pathname & " " & New_Pathname);
    end Rename_File;
    ----------------
  begin
    if Is_File(New_Pathname) then
      if Overwrite_If_Exists then
        Delete(New_Pathname);
        Rename_File;
      end if;
    else
      Rename_File;
    end if;
  end Rename;
  --------------------------------------------------------------------------
  procedure Copy  (From_Pathname,
                   To_Pathname         : in String;
                   Overwrite_If_Exists : in Boolean := True) is
    ----------------------
    procedure Copy_File is

      C_Fromname : aliased Interfaces.C.Char_Array :=
                           Interfaces.C.To_C(From_Pathname);
      C_Toname   : aliased Interfaces.C.Char_Array :=
                           Interfaces.C.To_C(To_Pathname);
    begin
	    if Win32.Winbase.Copyfile
             (Lpexistingfilename => C_Fromname(C_Fromname'First)'Unchecked_Access,
              Lpnewfilename      => C_Toname(C_Toname'First)'Unchecked_Access,
	            Bfailifexists      => Win32.False) = Win32.False then
	      Ada.Exceptions.Raise_Exception(Os_Error'Identity,
	         Message => "portable.operating_system.copy" & New_Line &
	            "  (" & '"' & From_Pathname & '"' & ", " & '"' & To_Pathname & '"'
	             & ", " & New_Line & "  overwrite_if_exists => " &
	             Boolean'Image(Overwrite_If_Exists) & ") failed." -- );
                 & New_Line & "  last error : " & Win32.Dword'Image(Win32.Winbase.Getlasterror));
      end if;
--      Execute("copy " & From_Pathname & " " & To_Pathname); 
    end Copy_File;
    --------------
  begin
    if Is_File(To_Pathname) then
      if Overwrite_If_Exists then
        Delete(To_Pathname);
        Copy_File;
      end if;
    else
      Copy_File;
    end if;
  end Copy;
  --------------------------------------------------------------------------
  procedure Delete(Pathname : in String) is
    File : Ada.Text_Io.File_Type;
  begin
    if Is_File(Pathname) then
      Ada.Text_Io.Open(File, Ada.Text_Io.In_File, Pathname);
      Ada.Text_Io.Delete(File);
    end if;
  end Delete;
  --------------------------------------------------------------------------
  procedure Execute(Command : in String) is
    --------------------------------------------------
    procedure Exec(Program, Command : in String) is

      Full_Command : aliased Interfaces.C.Char_Array :=
                             Interfaces.C.To_C(Program & Command);
	    Startup_Info : aliased Win32.Winbase.Startupinfo;
	    Process_Info : aliased Win32.Winbase.Process_Information;
	      -- Cannot use WINBASE.INFINITE since it will raise constraint error.  So
	      -- use the following INFINITE until the error in WINBASE is corrected.
	    Infinite     : constant := 16#FFFF_FFFF#; -- 16#7FFFFFFF# - 16#80000000#;  -- what about '-1'?
	    Exit_Code    : aliased Interfaces.C.Unsigned_Long := 0;
	  begin
	    Win32.Winbase.Getstartupinfo(Startup_Info'Unchecked_Access);
	    if Win32.Winbase.Createprocess
	           (Lpapplicationname   => null,
	            Lpcommandline       => Full_Command(Full_Command'First)'Unchecked_Access,
	            Lpprocessattributes => null,
	            Lpthreadattributes  => null,
	            Binherithandles     => Win32.False,
	            Dwcreationflags     => 0,
	            Lpenvironment       => System.Null_Address,
	            Lpcurrentdirectory  => null,
	            Lpstartupinfo       => Startup_Info'Unchecked_Access,
	            Lpprocessinformation=> Process_Info'Unchecked_Access) =
         Win32.False then
	        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
	           Message => "portable.operating_system.execute" & New_Line &
	                      "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                        ") failed. (Win32.Winbase.Createprocess)");
	    end if;
      if Win32.Winbase.Waitforsingleobject
             (Hhandle       => Process_Info.Hprocess,
              Dwmilliseconds=> Infinite) = -1 then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
            Message => "portable.operating_system.execute" & New_Line &
                       "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                       ") failed. (Win32.Winbase.Waitforsingleobject)");
      end if;
      if Win32.Winbase.Getexitcodeprocess
             (Hprocess  => Process_Info.Hprocess,
              Lpexitcode=> Exit_Code'Unchecked_Access) = Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
          Message => "portable.operating_system.execute" & New_Line &
                     "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                     ") failed. (Win32.Winbase.Getexitcodeprocess)");
    	end if;
      if Program /= "command.com" then
        if Win32.Winbase.Closehandle(Process_Info.Hthread) = Win32.False then
           Ada.Exceptions.Raise_Exception(Os_Error'Identity,
            Message => "portable.operating_system.execute" & New_Line &
                       "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                       ") failed. (Win32.Winbase.Closehandle(hthread))");
        end if;
      end if;
	    if Win32.Winbase.Closehandle(Process_Info.Hprocess) = Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
          Message => "portable.operating_system.execute" & New_Line &
                     "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                     ") failed. (Win32.Winbase.Closehandle(hprocess))");
      end if;
    end Exec;
    ---------
  begin
    Exec(Program => "command.com", Command => " /c " & Command);
  end Execute;
  ---------------------------------------------------------------------------
end Portable.Operating_System;