--                              -*- Mode: Ada -*-
-- Filename        : portable-Operating_System.adb
-- Description     : Version for Windows.
-- Author          : J. Schroeer, AMST
-- Created On      : Wed Feb  5 14:41:56 1997
-- Last Modified By:
-- Last Modified On: Thu Jul  3 22:41:44 1997
-- Update Count    : 23
-- Status          : ok

with Ada.Text_Io, Ada.Direct_Io, Ada.Exceptions, Ada.Strings.Unbounded;     
with Interfaces.C;
with Utilities.String_Utilities, Utilities.String_Vectors,
     Utilities.String_Stack,     Utilities.Files;
with System;

with Win32;
with Win32.Winbase;
--with Windows.Utils;

with Posix;
with Posix.Calendar;
with Posix.Files;
with Posix.Io;
with Posix.File_Status;
with Posix.Process_Environment;
--with Posix.Process_Primitives;
with Posix.Process_Identification;
with Posix.Permissions;

package body Portable.Operating_System is

  Max_Path_Length : constant Interfaces.C.Size_T := 260;

--  the_Newline             : constant String    := "" & Ascii.Lf;       -- Unix
--  the_Directory_Separator : constant Character := '/';                 -- Unix
--  the_Path_Separator      : constant Character := ':';                 -- Unix
--  Executable_Suffix       : constant String    := "";                  -- Unix

  The_New_Line            : constant String    := Ascii.Cr & Ascii.Lf; -- Windows
  The_Directory_Separator : constant Character := '\';                 -- Windows
  The_Path_Separator      : constant Character := ';';                 -- Windows
  Executable_Suffix       : constant String    := ".exe";              -- Windows

  use type Win32.Bool;
  use type System.Address; -- For handle types (are subtypes of system.address)
  use type Interfaces.C.Size_T, Interfaces.C.Unsigned_Long;
  use type Ada.Strings.Unbounded.String_Access;


  package String_Utilities renames Utilities.String_Utilities;
  -------------------------------------------------------------------------
  function New_Line return String is
  begin
    return The_New_Line;
  end New_Line;
  -------------------------------------------------------------------------
  function Directory_Separator return Character is
  begin
    return The_Directory_Separator;
  end Directory_Separator;
  -------------------------------------------------------------------------
  function Path_Separator return Character is
  begin
    return The_Path_Separator;
  end Path_Separator;
  -------------------------------------------------------------------------
  function Environment_Value_Of (Name : in String) return String is
  begin  
    return Posix.To_String(Posix.Process_Environment.Environment_Value_Of
             (Posix.To_Posix_String(Name)));
  end Environment_Value_Of;
  --------------------------------------------------------------------------
  procedure Change_Working_Directory (Directory_Name : in String) is
  begin
    Posix.Process_Environment.Change_Working_Directory
      (Posix.To_Posix_String(Directory_Name));
  end Change_Working_Directory;
  ----------------------------------------------------------------------------
  function Get_Working_Directory return String is
  begin
    return Posix.To_String(Posix.Process_Environment.Get_Working_Directory);
  end Get_Working_Directory;
  ----------------------------------------------------------------------------
  procedure Create_Directory (Pathname : in String) is
  begin
    Posix.Files.Create_Directory
     (Pathname   => Posix.To_Posix_String(Pathname),
      Permission => Posix.Permissions.Group_Permission_Set);
  end Create_Directory;
  ----------------------------------------------------------------------------
  procedure Remove_Directory (Pathname : in String) is
  begin
    Posix.Files.Remove_Directory(Posix.To_Posix_String(Pathname));
  end Remove_Directory;
  ----------------------------------------------------------------------------
  function Is_File     (Pathname: in String) return Boolean is
  begin
    return Posix.Files.Is_File(Posix.To_Posix_String(Pathname));
  end Is_File;
  ----------------------------------------------------------------------------
  function Is_Directory(Pathname: in String) return Boolean is
  begin
    return Posix.Files.Is_Directory(Posix.To_Posix_String(Pathname));
  end Is_Directory;
  ----------------------------------------------------------------------------
  procedure For_Every_Directory_Entry(Pathname : in String) is
    --------------------------------------------------------------
    procedure Action(D_Entry : in     Posix.Files.Directory_Entry;
                     Quit    : in out Boolean) is
    begin
      Action(D_Entry => Posix.To_String(Posix.Files.Filename_Of(D_Entry)),
             Quit    => Quit);
    end Action;     
    --------------------------------------------------------------
    procedure Traverse is new Posix.Files.For_Every_Directory_Entry(Action);
  begin
    Traverse(Posix.To_Posix_String(Pathname));
  end For_Every_Directory_Entry;
  ----------------------------------------------------------------------------                    
  function User return String is
  begin
    return Posix.Process_Identification.Image
          (Posix.Process_Identification.Get_Real_User_Id);
  end User;
  ----------------------------------------------------------------------------                      
  function Owner(Pathname : in String) return String is
  begin
    return Posix.Process_Identification.Image(Posix.File_Status.Owner_Of 
          (File_Status => Posix.File_Status.Get_File_Status
                         (Posix.To_Posix_String(Pathname))));
  end Owner;
  ----------------------------------------------------------------------------                    
  procedure Null_Operation(The_Pathname : in     String;
                           Continue     : in out Boolean) is
  begin
    null;
  end Null_Operation;
  ----------------------------------------------------------------------------
  procedure Process_Files(The_Directory : in String  := Get_Working_Directory;
                          Recursive     : in Boolean := True) is

    Continue : Boolean := True;
    ---------------------------------------------------------
    procedure Process_Directory(Directory : in String) is
      -------------------------------------------
      procedure Process(The_Entry : in     String;
                        Quit      : in out Boolean) is
      begin
        if    Is_File(The_Entry) then
          File_Operation(The_Pathname => The_Entry,
                         Continue     => Continue);
        elsif Recursive and then Is_Directory(The_Entry) then
          Process_Directory(Directory => The_Entry); -- Recursion          
        end if;
        Quit := not Continue;               
      end Process;
      -------------------------------------------
      procedure Traverse is new For_Every_Directory_Entry(Action => Process);
      -------------------------------------------
      Current_Dir : constant String := Get_Working_Directory;
    begin
      if not Continue or else Directory = "." or else Directory = ".." then
        return;
      end if;
      Change_Working_Directory(Directory);
      Directory_Pre_Operation(Directory, Continue);
      if Continue then
        Traverse(Directory);
      end if;
      Directory_Post_Operation(Directory, Continue);
      Change_Working_Directory(Current_Dir);
    end Process_Directory;
    ----------------------
  begin
    Process_Directory(The_Directory);
  end Process_Files;
  ----------------------------------------------------------------------------   
  function All_Entries(Directory         : in String  := Get_Working_Directory;
                       Recursive         : in Boolean := True;
                       Relative_Pathnames: in Boolean := True;
                       Directories,
                       Files             : in Boolean := True)
           return Utilities.String_Vectors.String_Vector is
    ------------------------------------------------------------------------
    function Absolute_Dirname_Of(The_Directory : in String) return String is
      Current_Dir : constant String := Get_Working_Directory;
    begin
      Change_Working_Directory(The_Directory);
      declare
        Dir : constant String := Get_Working_Directory;
      begin
        Change_Working_Directory(Current_Dir);
        return Dir;
      end;
    end Absolute_Dirname_Of;
    ---------------------------------------------------------------       
    Root_Dir   : constant String := Absolute_Dirname_Of(Directory);
    Name_Stack : Utilities.String_Stack.Stack;
    ---------------------------------------------------------------
    procedure Process_Directory(Directory_Name : in String) is
      ------------------------------------------------
      Current_Dir : constant String := Absolute_Dirname_Of(Directory_Name);
     	------------------------------------------------
    	procedure Save_Name(The_Pathname : in String) is
        Name : Ada.Strings.Unbounded.String_Access;
      begin
        if Relative_Pathnames then
          if Current_Dir = Root_Dir then
            Name := new String'(The_Pathname);
          else
            Name := new String'(Current_Dir(Root_Dir'Last + 2 ..
                    Current_Dir'Last) & The_Directory_Separator & The_Pathname);
          end if;
        else
          Name := new String'(Current_Dir &
                  The_Directory_Separator & The_Pathname);
        end if;
    	  Utilities.String_Stack.Push(Item => Name, Onto => Name_Stack);
    	end Save_Name;
      ------------------------------------------------------- 
    	procedure Process_Entry(Entry_Name : in     String;
	                            Quit       : in out Boolean) is
     	begin
        if Files and then Is_File(Entry_Name) then
          Save_Name(Entry_Name);
    	  elsif Is_Directory(Entry_Name) then
  	      if Entry_Name = "." or else Entry_Name = ".." then
	          return;
  	      end if;
	        if Directories then
            Save_Name(Entry_Name);
	        end if;
	        if Recursive then
            Process_Directory(Directory_Name => Entry_Name); -- RECURSION
          end if;
    	  end if;
   	  end Process_Entry;
      -------------------------------------------------------------
      procedure Traverse_Directory is new For_Every_Directory_Entry
               (Action => Process_Entry);
    begin
      Traverse_Directory(Directory_Name);
    end Process_Directory;
    ----------------------
  begin
    Process_Directory(Directory);
    declare
      Names : Utilities.String_Vectors.String_Vector
              (1 .. Utilities.String_Stack.Depth_Of(Name_Stack));
    begin
      Utilities.String_Stack.Convert(The_Stack => Name_Stack,
                                     To_Vector => Names);
      return Names;
    end;
  end All_Entries;
  ----------------------------------------------------------------------------
  function File_Time_Stamp      (Name       : in String;
                                 Local_Time : in Boolean := True)
           return Ada.Calendar.Time is
  begin
    return Posix.Calendar.To_Time
          (Posix.File_Status.Last_Modification_Time_Of 
          (File_Status => Posix.File_Status.Get_File_Status
                         (Posix.To_Posix_String(Name))));
  end File_Time_Stamp;
  --------------------------------------------------------------------------
  -- local functions 
  --------------------------------------------------------------------------
  function Full_Pathname_Of(Filename  : in String;
                            Directory : in String;
                            Recursive : in Boolean) return String is

    Is_Pathname : constant Boolean := String_Utilities.Has
                                  (The_String    => Filename,
                                   The_Character => The_Directory_Separator);
    Found_Name  : Ada.Strings.Unbounded.String_Access;
    -----------------------------------------------------------
    procedure Search_Filename(The_Pathname : in     String;
                              Continue     : in out Boolean) is
    begin
      if Is_Pathname then
        declare
          Full_Name : constant String := Get_Working_Directory &
                                         The_Directory_Separator & Filename;
        begin
          if Is_File(Full_Name) then
            Found_Name := new String'(Full_Name);
            Continue   := False;
          end if;
        end;
      elsif Is_File(Filename) then
        Found_Name := new String'(Get_Working_Directory &
                          The_Directory_Separator & Filename);
        Continue   := False;
      end if;
    end Search_Filename;
    ------------------------------------------------------
    procedure Process is new Process_Files
             (Directory_Pre_Operation => Search_Filename);
    ------------------------------------------------------
  begin
    begin
      Process(The_Directory => Directory, Recursive => Recursive);
    exception
      when Os_Error => null;
    end;
    if Found_Name = null then
      return "";
    else
      declare
        Name : constant String := Found_Name.all;
      begin
        Ada.Strings.Unbounded.Free(Found_Name);
        return Name;
      end;
    end if;
  end Full_Pathname_Of;
  --------------------------------------------------------------------------
  function Locate(Filename : in String;
                  Path     : in String) return String is
  begin
    if String_Utilities.Has(The_String    => Path,
                            The_Character => The_Path_Separator) then
      declare
        Dir_Names : Utilities.String_Vectors.String_Vector :=
                    Utilities.String_Vectors.Split
                            (The_String => Path,
                             Separator  => The_Path_Separator);
      begin
        for I in Dir_Names'range loop
          declare
            Name : constant String :=
                   Full_Pathname_Of(Filename  => Filename,
                                    Directory => Dir_Names(I).all,
                                    Recursive => False);
          begin
            if Name /= "" then
              Utilities.String_Vectors.Deallocate(Dir_Names);
              return Name;
            end if;
          end;
        end loop;
        Utilities.String_Vectors.Deallocate(Dir_Names);
        return "";
      end;
    else
      return Full_Pathname_Of(Filename  => Filename,
                              Directory => Path,
                              Recursive => True);
    end if;
  end Locate;
  --------------------------------------------------------------------------
  -- End local functions
  -------------------------------------------------------------------------
  function Locate_Exec_On_Path(Exec_Name : in String) return String is
  begin
    if Executable_Suffix = Utilities.Files.Suffix_Of(Exec_Name) then
      return Locate(Filename => Exec_Name,
                    Path     => Environment_Value_Of("PATH"));
    else
      return Locate(Filename => Exec_Name & Executable_Suffix,
                    Path     => Environment_Value_Of("PATH"));
    end if;
  end Locate_Exec_On_Path;
  --------------------------------------------------------------------------
  function Locate_Regular_File(File_Name : in String;
                               Path      : in String :=
                                              Environment_Value_Of("PATH"))
           return String is
  begin
    return Locate(Filename => File_Name, Path => Path);
  end Locate_Regular_File;
  --------------------------------------------------------------------------
  function Size_Of(Name : in String) return Natural is

    type Byte is mod 255;
    for  Byte'Size use 8;

    package Byte_Direct_Io is new Ada.Direct_Io(Byte);

    File : Byte_Direct_Io.File_Type;
    Size : Natural;
  begin
    Byte_Direct_Io.Open(File, Mode => Byte_Direct_Io.In_File,
                              Name => Name,
                              Form => "shared=no");

    -- The Form parameter with shared = yes|no is GNAT specific, see
    -- GNAT reference manual, chapter 6.8, shared files.
    -- In order to use open for already opened external files, shared = no
    -- is required. If an external file with name Name is already open and
    -- "shared=yes" is assigned to Form, USE_ERROR is raised.

    Size := Natural(Byte_Direct_Io.Size(File));
    Byte_Direct_Io.Close(File);
    return Size;
  end Size_Of;
  ---------------------------------------------------------------------------
  procedure Rename(Old_Pathname,
                   New_Pathname        : in String;
                   Overwrite_If_Exists : in Boolean := True) is
  begin
    if Is_File(New_Pathname) then
      if Overwrite_If_Exists then
        Delete(New_Pathname);
        Posix.Files.Rename(Old_Pathname => Posix.To_Posix_String(Old_Pathname),
                           New_Pathname => Posix.To_Posix_String(New_Pathname));
      end if;
    else
      Posix.Files.Rename(Old_Pathname => Posix.To_Posix_String(Old_Pathname),
                         New_Pathname => Posix.To_Posix_String(New_Pathname));
    end if;
  end Rename;
  --------------------------------------------------------------------------
  procedure Copy  (From_Pathname,
                   To_Pathname         : in String;
                   Overwrite_If_Exists : in Boolean := True) is 
                   
    Source, Target : Posix.Io.File_Descriptor;               
  begin
    if Overwrite_If_Exists or else not Utilities.Files.Exists(To_Pathname) then
      Source := Posix.Io.Open(Name => Posix.To_Posix_String(From_Pathname),
                              Mode => Posix.Io.Read_Write);   
      Target := Posix.Io.Open_Or_Create
               (Name        => Posix.To_Posix_String(To_Pathname),
                Mode        => Posix.Io.Read_Write,
                Permissions => Posix.File_Status.Permission_Set_Of
                              (Posix.File_Status.Get_File_Status(Source)));   
      Target := Posix.Io.Duplicate(File => Source);
      Posix.Io.Close(Source);
      Posix.Io.Close(Target);
    end if;                         
  end Copy;
  --------------------------------------------------------------------------
  procedure Delete(Pathname : in String) is
    File : Ada.Text_Io.File_Type;
  begin
    if Is_File(Pathname) then
      Ada.Text_Io.Open(File, Ada.Text_Io.In_File, Pathname);
      Ada.Text_Io.Delete(File);
    end if;
  end Delete;
  --------------------------------------------------------------------------
--  procedure Execute(Command : in String) is  
--  begin
--    Windows.Utils.Start_Application("command.com", Command);
--  end Execute;
  ---------------------------------------------------------------------------
  procedure Execute(Command : in String) is
  
    use type Win32.Bool;
    --------------------------------------------------
    procedure Exec(Program, Command : in String) is

      Full_Command : aliased Interfaces.C.Char_Array :=
                             Interfaces.C.To_C(Program & Command);
    	Startup_Info : aliased Win32.Winbase.Startupinfo;
    	Process_Info : aliased Win32.Winbase.Process_Information;
    	-- Cannot use WINBASE.INFINITE since it will raise constraint error.  So
	    -- use the following INFINITE until the error in WINBASE is corrected.
	    Infinite     : constant := 16#FFFF_FFFF#; -- 16#7FFFFFFF# - 16#80000000#;  -- what about '-1'?
	    Exit_Code    : aliased Interfaces.C.Unsigned_Long := 0;
    begin
    	Win32.Winbase.Getstartupinfo(Startup_Info'Unchecked_Access);
	    if Win32.Winbase.Createprocess
	           (Lpapplicationname   => null,
	            Lpcommandline       => Full_Command(Full_Command'First)'Unchecked_Access,
	            Lpprocessattributes => null,
	            Lpthreadattributes  => null,
	            Binherithandles     => Win32.False,
	            Dwcreationflags     => 0,
	            Lpenvironment       => System.Null_Address,
	            Lpcurrentdirectory  => null,
	            Lpstartupinfo       => Startup_Info'Unchecked_Access,
	            Lpprocessinformation=> Process_Info'Unchecked_Access) =
         Win32.False then
    	  Ada.Exceptions.Raise_Exception(Os_Error'Identity,
	           Message => "portable.operating_system.execute" & New_Line &
	                      "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                        ") failed. (Win32.Winbase.Createprocess)");
	    end if;
      if Win32.Winbase.Waitforsingleobject
             (Hhandle       => Process_Info.Hprocess,
              Dwmilliseconds=> Infinite) = -1 then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
            Message => "portable.operating_system.execute" & New_Line &
                       "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                       ") failed. (Win32.Winbase.Waitforsingleobject)");
      end if;
      if Win32.Winbase.Getexitcodeprocess
             (Hprocess  => Process_Info.Hprocess,
              Lpexitcode=> Exit_Code'Unchecked_Access) = Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
          Message => "portable.operating_system.execute" & New_Line &
                     "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                     ") failed. (Win32.Winbase.Getexitcodeprocess)");
    	end if;
      if Program /= "command.com" then
        if Win32.Winbase.Closehandle(Process_Info.Hthread) = Win32.False then
           Ada.Exceptions.Raise_Exception(Os_Error'Identity,
            Message => "portable.operating_system.execute" & New_Line &
                       "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                       ") failed. (Win32.Winbase.Closehandle(hthread))");
        end if;
      end if;
    	if Win32.Winbase.Closehandle(Process_Info.Hprocess) = Win32.False then
        Ada.Exceptions.Raise_Exception(Os_Error'Identity,
          Message => "portable.operating_system.execute" & New_Line &
                     "  (" & '"' & Interfaces.C.To_Ada(Full_Command) & '"' &
                     ") failed. (Win32.Winbase.Closehandle(hprocess))");
      end if;
    end Exec;
    ---------
  begin
    Exec(Program => "command.com", Command => " /c " & Command);
  end Execute;
  ---------------------------------------------------------------------------
end Portable.Operating_System;
