--                              -*- Mode: Ada -*-
-- Filename        : portable-operating_system.ads
-- Description     : Portable specification of operating system calls
--
--                   For this specification more than one file for 
--                   the body may exist.
--
--                   Currently one version for GNAT and one for Windows
--                   The GNAT version works on all platforms GNAT 
--                   is available for.
--             
--                   The windows version is built on top of the win32ada
--                   binding and should work with GNAT and objectada.      
--                   
-- Author          : J. Schrer
-- Created On      : Thu Mar 27 10:23:07 1997
-- Last Modified By:
-- Last Modified On: 4.12.1997
-- Update Count    : 10
-- Status          :

-- PURPOSE: When changing
--          + the plattform, the operating system
--          + the Ada compiler or a foreign language compiler when
--            interfacing to C, FTN, Cobol, C++ or whatever.
--
--    eventually necessary adaptions shall be isolated to the package body.
--
--
-- REMARKS:
--
-- The package has an abstract interface for all subprograms whose
-- implementation eventually needs adaption because they fall into one of two
-- categories
--
-- 1) the use of Ada-compiler resources (i.e. package gnat.os_lib) or
--    other specialities of the Ada compiler implementation as the
--    form-string in open- or create-procedures
--
-- 2) the use of package Interfaces     (or children i.e. Interfaces.C)
--
-- All eventually necessary adaptions are isolated to the implementation
-- of package adaption (to the package body).
--
-- 1) When changing from GNAT to another Ada compiler all uses of features from
--    category 1) need adaption in the body.
--
-- 2) When porting to another operating system and or changing a
--    foreign language (C-) compiler,
--    possibly the interfacing pragmas (see Ada95 LRM, Annex B, page 343 ff.)
--
--     pragma Import(
--            [Convention    =>] convention_identifier,
--            [Entity        =>] local_name
--         [, [External_Name =>] string_expression]
--         [, [Link_Name     =>] string_expression]);
--
--     pragma Export(
--            [Convention    =>] convention_identifier,
--            [Entity        =>] local_name
--         [, [External_Name =>] string_expression]
--         [, [Link_Name     =>] string_expression]);
--
--     pragma Convention(
--            [Convention    =>] convention_identifier,
--            [Entity        =>] local_name);
--
--     pragma Linker_Options(string_expression);
--
--    need adaption in the implementation.
--
--    Possibly the optional parameters External_Name, Link_Name have to be
--    added or altered in the pragmas Import and Export.
--    Possibly the parameter Convention has to be altered in
--    pragma Convention when changing a foreign language compiler
--    with different calling conventions.

with Ada.Calendar;
with Utilities.String_Vectors;

package Portable.Operating_System is

  Os_Error                       : exception;
  Non_Existing_File_Or_Directory : exception;

  ----------------------------------------------------------------------------
    
  function Directory_Separator return Character;

   --  The character that is used to separate parts of a pathname.
   --  '/' on unix
   --  '\' on windows
  
  function Path_Separator      return Character;
  
   --  The character to separate paths in an environment variable value.
   --  '%' on unix (depending on shell)
   --  ';' on windows


  function New_Line            return String;

  -- The string that produces a new_line
  -- ascii.lf on Unix
  -- ascii.cr & ascii.lf on DOS / Windows
  
  ------------------------------------------------------------------
  -- The following subprograms are called as the ones from different 
  -- posix Ada API packages.

  function Environment_Value_Of (Name : in String) return String;

  --  Get the value of the environment variable. Returns the empty
  --  string if the environment variable does not exist.

  procedure Change_Working_Directory (Directory_Name : in String);
  
	-- Set the current directory path to directory_name.

  function  Get_Working_Directory  return String;
  
	-- Get the current directory path.

  procedure Create_Directory (Pathname : in String);
  
	-- Create a new directory with path New_Directory.
	-- Use the default security permissions.

  procedure Remove_Directory (Pathname : in String);
  
	-- Remove the directory indicated by path Directory.
	-- The directory must be empty.

      
  function Is_File      (Pathname: in String) return Boolean;
  
	-- Returns true if pathname refers to a regular file.

  function Is_Directory (Pathname: in String) return Boolean;
  
	-- Returns true if pathname refers to a directory.

  
  generic
    with procedure Action
                  (D_Entry :	in     String;
                   Quit    :	in out Boolean);
  procedure For_Every_Directory_Entry(Pathname : in String);
	
	-- Call the action procedure for each entry in the directory.
	-- Returns when all the entries have been processed or Action
	-- sets quit to true.

  -- A few additional subprograms -------------------------------

  function User return String;

  -- The user name of the current process.
  
  function Owner(Pathname : in String) return String;
  
  -- Runs only with winnt (raises exception with win95).
      
  procedure Null_Operation(The_Pathname : in     String; 
                           Continue     : in out Boolean); 
  
  -- Serves as default procedure for generic formal procedures of 
  -- following procedure process_files.                          
 
  generic
    with procedure File_Operation         
        (The_Pathname : in     String; 
         Continue     : in out Boolean) is Null_Operation;
                               
    with procedure Directory_Pre_Operation
        (The_Pathname : in     String; 
         Continue     : in out Boolean) is Null_Operation;

    with procedure Directory_Post_Operation
        (The_Pathname : in     String; 
         Continue     : in out Boolean) is Null_Operation;
                               
  procedure Process_Files(The_Directory : in String  := Get_Working_Directory;
                          Recursive     : in Boolean := True);
  
  -- This directory and file iterator performs the following operations:
  --
  -- Directory_Pre_Operation     is called after  entering a directory
  --   File_Operation            is called for every file in the directory
  -- Directory_Post_Operation    is called before leaving the directory
  --                                                 
  -- Processing starts at The_Directory.
  -- If recursive, all subdirectories are traversed.
  -- If Directory_Pre_Operation gives continue = false, 
  -- Directory_Post_Operation for that directory is called, but no
  -- files in that directory are processed.
  -- If recursive, in each case after execution of the program, the
  -- current directory is the same as before the call of the procedure.
  
 
  function All_Entries(Directory         : in String  := Get_Working_Directory;
                       Recursive         : in Boolean := True;
                       Relative_Pathnames: in Boolean := True;
                       Directories, 
                       Files             : in Boolean := True)
           return Utilities.String_Vectors.String_Vector;
  
  -- Returns all entries in the directory.
  -- If recursive, then all entries in all subdirectories are searched.
  -- If Relative_Pathnames then relative names are given, else 
  -- absolute pathnames are generated.
  -- The switches Directories and Files mark, if directory names / 
  -- filenames are returned.
    

  function File_Time_Stamp      (Name       : in String;
                                 Local_Time : in Boolean := True) 
           return Ada.Calendar.Time;

  --  Get time stamp of file from name (use for unopened file)
  --  os_error is raised if Name cannot be found.
  --  If local_time then the local time is returned else the GMT is given.

  function Locate_Exec_On_Path  (Exec_Name : in String) return String;

  --  Try to locate an executable whose name is given by Exec_Name in the
  --  directories listed in the environment Path.  If the Exec_Name doesn't
  --  have the executable suffix, it will be appended before the search.
  --  Otherwise works like Locate_Regular_File below.

  function Locate_Regular_File  (File_Name : in String;
                                 Path      : in String := 
                                                Environment_Value_Of("PATH"))
           return String;

  --  Try to locate a regular file whose name is given by File_Name in the
  --  directories listed in  Path.  If a file is found, its full pathname is
  --  returned; otherwise, an empty string is returned. If the File_Name given
  --  is an absolute pathname, then Locate_Regular_File just checks that the
  --  file exists and is a regular file.  Otherwise, the Path argument is
  --  parsed according to OS conventions, and for each directory in the Path
  --  a check is made if File_Name is a relative pathname of a regular file
  --  from that directory.
          
  --  Default value for Path is added compared to
  --  Gnat.Os_lib.Locate_Regular_File.

  -- If path is not a path (contains no path_separator character), it is
  -- treated as a directory, that is searched recursively. 
  -- The directories in a normal path variable are not searched recursively.
  
  function Size_Of(Name : in String) return Natural;

  -- Filesize in bytes.

  procedure Rename(Old_Pathname, 
                   New_Pathname        : in String;
                   Overwrite_If_Exists : in Boolean := True);

  -- Renames Old_Filename in New_Filename.

  procedure Copy  (From_Pathname, 
                   To_Pathname         : in String;
                   Overwrite_If_Exists : in Boolean := True);

  -- Copies From_Filename to To_Filename.
                   
  procedure Delete(Pathname : in String);
  
  -- Deletes file pathname.

  procedure Execute(Command : in String);

  -- Executes command by sending it to the operating system.

end Portable.Operating_System;

