--                              -*- Mode: Ada -*-
-- Filename        : portable-standard.ads
-- Description     : Declare portable numeric scalar types
-- Author          : J. Schrer
-- Created On      : Thu Mar 27 09:30:18 1997
-- Last Modified By:
-- Last Modified On: Thu Mar 27 11:02:39 1997
-- Update Count    : 2
-- Status          :

-------------------------------------------------------------------------------
-- Purpose: Define own numeric types (integer and real) corresponding to the
--          ones of package standard. The package is used to enhance
--          portability.
--
-- Remarks: For all own software, these types shall be used instead of the
--          ones from package standard.
--
--          Direct visibility of types and operations:
--          ------------------------------------------
--          The predefined package standard is directly visible by the implicit
--          "with standard; use standard;" context clause in front of each
--          compilation unit.
--
--          The following operators are implicitly declared for each
--          numeric type (predefined types in package standard or 'own' types,
--          see Ada95 LRM, chapter A.1, page 260 ff.)
--
--           integer types (integer_8, integer_16, integer_32, see below)
--
--            relational: =, /=, <, <=, >, >=,
--            unary     : +, -, abs
--            binary    : +, -, *, /, rem, mod, **
--
--           floating point types (float_32, float_64, see below)
--
--            relational: =, /=, <, <=, >, >=,
--            unary     : +, -, abs
--            binary    : +, -, *, /, ** (float ** integer)
--
--           To gain direct visibility of these operations for the 'own'
--           types declared in this package in order to use them in
--           infix operator notation as
--
--           'a+b'
--
--           rather than in function call notation as
--
--           'portable.standard."+"(a, b)'
--
--           in Ada83
--           --------
--           a with- and a use clause may be written in each
--           compilation unit, that uses one of the types.
--
--           "with portable.standard; use portable.standard;"
--
--           in Ada95
--           --------
--           a 'with clause' shall be written as a context clause and a
--           'use type clause' shall be written for each type, operators
--           shall be made direcly visible for in order to use them in infix
--           notation, e.g.
--
--           with portable.standard;
--           package body models.controllers is
--
--             procedure calculate(the_model : in out simulink.model) is
--               use type portable.standard.float_64;
--             begin
--               the_model.t := the_model.t + the_model.delta_t;
--
--           the_model.t and the_model.delta_t are of type 
--           portable.standard.float_64
--
-- Effects:
--   None

package Portable.Standard is

  pragma Pure(Portable.Standard);

  type Integer_8  is range - 2 ** 7 .. 2 ** 7 - 1;
  for  Integer_8'Size use 8;

  type Integer_16 is range - 2 ** 15 .. 2 ** 15 - 1;
  for  Integer_16'Size use 16;

  type Integer_32 is range - 2 ** 31 .. 2 ** 31 - 1;
  for  Integer_32'Size use 32;

  subtype Natural_8   is Integer_8  range 0 .. Integer_8'Last;
  subtype Positive_8  is Integer_8  range 1 .. Integer_8'Last;
  subtype Natural_16  is Integer_16 range 0 .. Integer_16'Last;
  subtype Positive_16 is Integer_16 range 1 .. Integer_16'Last;
  subtype Natural_32  is Integer_32 range 0 .. Integer_32'Last;
  subtype Positive_32 is Integer_32 range 1 .. Integer_32'Last;

 
  type Float_32 is digits 6;
  for  Float_32'Size use 32;

  type Float_64 is digits 15;
  for  Float_64'Size use 64;

 
  type Byte is mod 2 ** 8;
  for  Byte'Size use 8;

  type Word is mod 2 ** 16;
  for  Word'Size use 16;

end Portable.Standard;
