
--  $Id: posix-process_primitives.adb,v 1.1 2000/02/16 19:31:53 obry Exp $
--  Author : Pascal Obry
--  p.obry@wanadoo.fr

with Ada.Unchecked_Deallocation;
with Ada.Unchecked_Conversion;
with System;

with Posix_Win32;
with Win32;

package body Posix.Process_Primitives is

   use Posix;
   use type Win32.Ulong;

   ---------------------------------------
   -- Process_ID_To_PROCESS_INFORMATION --
   ---------------------------------------

   function Process_Id_To_Process_Information is
      new Ada.Unchecked_Conversion (Posix.Process_Identification.Process_Id,
                                    Win32.Winbase.Process_Information);

   ---------------------------------------
   -- PROCESS_INFORMATION_To_Process_ID --
   ---------------------------------------

   function Process_Information_To_Process_Id is
      new Ada.Unchecked_Conversion (Win32.Winbase.Process_Information,
                                    Posix.Process_Identification.Process_Id);

   Security_Handles_Inherited : aliased Win32.Winbase.Security_Attributes
     := (Win32.Winbase.Security_Attributes'Size / 8,
         System.Null_Address,
         Win32.True);

   --  Process Template

   -------------------
   -- Open_Template --
   -------------------

   procedure Open_Template (Template : in out Process_Template) is
   begin
      if Template.Is_Open then
         Close_Template (Template);
      end if;
      Template.Keep_Effective_Ids := False;
      --  POSIX.Signals.Delete_All_Signals (Template.Signal_Mask);
      Template.Is_Open := True;
      Template.Process_Informations := new
        Win32.Winbase.Process_Information;
      Template.File_Request_List := null;
      Template.Last_File_Request := null;
   end Open_Template;


   ----------------
   -- Check_Open --
   ----------------

   procedure Check_Open (Template : in Process_Template;
                         Message  : in String) is
   begin
      if not Template.Is_Open then
         Posix_Win32.Raise_Error (Message, Invalid_Argument);
      end if;
   end Check_Open;


   ----------------
   -- Check_File --
   ----------------

   procedure Check_File (File : in Posix.Io.File_Descriptor) is
   begin
      case File is
         when Posix.Io.Standard_Input |
           Posix.Io.Standard_Output |
           Posix.Io.Standard_Error =>
            null;
         when others =>
            Posix_Win32.Raise_Error
              ("(POSIX implementation restriction) Check_File",
               Invalid_Argument);
      end case;
   end Check_File;


   --------------------
   -- Close_Template --
   --------------------

   procedure Close_Template (Template : in out Process_Template) is

      P      : File_Request_Access;
      Result : Win32.Bool;

      procedure Free is
        new Ada.Unchecked_Deallocation (File_Request, File_Request_Access);

      procedure Free is
        new Ada.Unchecked_Deallocation (Win32.Winbase.Process_Information,
                                        Win32.Winbase.Lpprocess_Information);


   begin
      Check_Open (Template, "Close_Template");

      while Template.File_Request_List /= null loop
         P := Template.File_Request_List;

         case P.Action is

            when Open =>
               case P.File is
                  when
                    Posix.Io.Standard_Input |
                    Posix.Io.Standard_Output |
                    Posix.Io.Standard_Error =>
                     Result := Win32.Winbase.Closehandle (P.Ohandle);
                  when others =>
                     null; -- not handled
               end case;

            when Close =>
               case P.File is
                  when
                    Posix.Io.Standard_Input |
                    Posix.Io.Standard_Output |
                    Posix.Io.Standard_Error =>
                     Result := Win32.Winbase.Closehandle (P.Chandle);
                  when others =>
                     null; -- not handled
               end case;

            when Duplicate =>
               null;
         end case;

         Template.File_Request_List := P.Next;
         Free (P);
      end loop;

      Template.Is_Open := False;

      --  close the handle of the Process and Primary Thread.
      Result := Win32.Winbase.Closehandle
        (Template.Process_Informations.Hprocess);

      Result := Win32.Winbase.Closehandle
        (Template.Process_Informations.Hthread);

      Free (Template.Process_Informations);
   end Close_Template;


   ----------------------------
   -- Set_Keep_Effective_IDs --
   ----------------------------

   procedure Set_Keep_Effective_Ids (Template : in out Process_Template) is
   begin
      Check_Open (Template, "Set_Keep_Effective_IDs");
      Template.Keep_Effective_Ids := True;
   end Set_Keep_Effective_Ids;


   ---------------------
   -- Set_Signal_Mask --
   ---------------------

   procedure Set_Signal_Mask
     (Template : in out Process_Template;
      Mask     : in     Posix.Signals.Signal_Set) is
   begin
      Check_Open (Template, "Set_Signal_Mask");
      Template.Signal_Mask := Mask;
   end Set_Signal_Mask;


   ---------------------------------
   -- Set_Creation_Signal_Masking --
   ---------------------------------

   procedure Set_Creation_Signal_Masking
     (Template       : in out Process_Template;
      Masked_Signals : in     Posix.Signal_Masking := Posix.Rts_Signals) is
   begin
      Check_Open (Template, "Set_Creation_Signal_Masking");
      Template.Signal_Creation_Masking := Masked_Signals;
   end Set_Creation_Signal_Masking;


   ------------
   -- Insert --
   ------------

   procedure Insert (Request_Access : in     File_Request_Access;
                     Into           : in out Process_Template)
   is
   begin
      if Into.File_Request_List = null then
         Into.File_Request_List := Request_Access;
         Into.Last_File_Request := Request_Access;
      else
         Into.Last_File_Request.Next := Request_Access;
         Into.Last_File_Request      := Request_Access;
      end if;
   end Insert;

   -----------------------------
   -- Set_File_Action_To_Open --
   -----------------------------

   procedure Set_File_Action_To_Open
     (Template : in out Process_Template;
      File     : in     Posix.Io.File_Descriptor;
      Name     : in     Posix.Pathname;
      Mode     : in     Posix.Io.File_Mode       := Posix.Io.Read_Only;
      Options  : in     Posix.Io.Open_Option_Set := Posix.Io.Empty_Set)
   is
      New_Action : File_Request_Access;
   begin
      Check_Open (Template, "Set_File_Action_To_Open");
      Check_File (File);
      New_Action :=
       new File_Request'(Next    => null,
                         File    => File,
                         Action  => Open,
                         Name    => To_Unbounded_String
                                     (Posix.To_String (Name)),
                         Mode    => Mode,
                         Options => Options,
                         Ohandle => Posix_Win32.Null_Handle);
      Insert (New_Action, Into => Template);
   end Set_File_Action_To_Open;


   ------------------------------
   -- Set_File_Action_To_Close --
   ------------------------------

   procedure Set_File_Action_To_Close
     (Template : in out Process_Template;
      File     : in     Posix.Io.File_Descriptor)
   is
      New_Action : File_Request_Access;
   begin
      Check_Open (Template, "Set_File_Action_To_Close");
      New_Action := new File_Request'(Next    => null,
                                      File    => File,
                                      Action  => Close,
                                      Chandle => Posix_Win32.Null_Handle);
      Insert (New_Action, Into => Template);
   end Set_File_Action_To_Close;


   ----------------------------------
   -- Set_File_Action_To_Duplicate --
   ----------------------------------

   procedure Set_File_Action_To_Duplicate
     (Template  : in out Process_Template;
      File      : in     Posix.Io.File_Descriptor;
      From_File : in     Posix.Io.File_Descriptor)
   is
      New_Action : File_Request_Access;
   begin
      Check_Open (Template, "Set_File_Action_To_Duplicate");
      New_Action :=
       new File_Request'(Next      => null,
                         File      => File,
                         Action    => Duplicate,
                         From_File => From_File);
      Insert (New_Action, Into => Template);
   end Set_File_Action_To_Duplicate;


   ----------------------
   -- Execute_Template --
   ----------------------

   procedure Execute_Template (Template : in     Process_Template;
                               Si       : in out Win32.Winbase.Startupinfo)
   is

      Null_Filename : constant String := "nul";

      P             : File_Request_Access;

      -------------------------
      -- Mode_To_File_Access --
      -------------------------

      function Mode_To_File_Access (Mode : in Posix.Io.File_Mode)
                                    return Win32.Dword
      is
         use type Win32.Ulong;
      begin
         case Mode is
            when Posix.Io.Read_Only =>
               return Win32.Winnt.Generic_Read;
            when  Posix.Io.Write_Only | Posix.Io.Read_Write =>
               return Win32.Winnt.Generic_Read + Win32.Winnt.Generic_Write;
         end case;
      end Mode_To_File_Access;

      ---------------
      -- Open_File --
      ---------------

      function Open_File (Name : in String;
                          Mode : in Posix.Io.File_Mode)
                          return Win32.Winnt.Handle
      is
         use type Win32.Int;
         use type System.Address;
         L_Name      : constant String := Name & Ascii.Nul;
         H           : Win32.Winnt.Handle;
      begin
         H := Win32.Winbase.Createfile
           (Win32.Addr (L_Name),
            Mode_To_File_Access (Mode),
            Win32.Winnt.File_Share_Read,
            Security_Handles_Inherited'access,
            Win32.Winbase.Open_Existing,
            Win32.Winnt.File_Attribute_Normal,
            System.Null_Address);
         if H = Win32.Winbase.Invalid_Handle_Value then
            Posix_Win32.Check_Retcode (Posix_Win32.Retcode_Error,
                                       "Template : Open_File " & Name);
         end if;
         return H;
      end Open_File;

      -----------------
      -- Create_File --
      -----------------

      function Create_File (Name : in String;
                            Mode : in Posix.Io.File_Mode)
                            return Win32.Winnt.Handle
      is
         use type Win32.Int;
         use type System.Address;
         L_Name : constant String := Name & Ascii.Nul;
         H : Win32.Winnt.Handle;
      begin
         H := Win32.Winbase.Createfile
           (Win32.Addr (L_Name),
            Mode_To_File_Access (Mode),
            Win32.Winnt.File_Share_Write,
            Security_Handles_Inherited'access,
            Win32.Winbase.Create_Always,
            Win32.Winnt.File_Attribute_Normal,
            System.Null_Address);
         if H = Win32.Winbase.Invalid_Handle_Value then
            Posix_Win32.Check_Retcode (Posix_Win32.Retcode_Error,
                                       "Template : Create_File " & Name);
         end if;
         return H;
      end Create_File;

   begin
      if not Template.Keep_Effective_Ids then
         Set_User_Id (Get_Real_User_Id);
         Set_Group_Id (Get_Real_Group_Id);
      end if;
      --  POSIX.Signals.Set_Blocked_Signals (Template.Signal_Mask, Junk1);

      P := Template.File_Request_List;

      while P /= null loop

         case P.Action is

            when Open =>
               case P.File is
                  when Posix.Io.Standard_Input =>
                     Si.Hstdinput := Open_File (To_String (P.Name),
                                                P.Mode);
                     P.Ohandle := Si.Hstdinput;
                  when Posix.Io.Standard_Output =>
                     Si.Hstdoutput := Create_File (To_String (P.Name),
                                                   P.Mode);
                     P.Ohandle := Si.Hstdoutput;
                  when Posix.Io.Standard_Error =>
                     Si.Hstderror := Create_File (To_String (P.Name),
                                                  P.Mode);
                     P.Ohandle := Si.Hstderror;
                  when others =>
                     Posix_Win32.Raise_Not_Yet_Implemented
                       ("Execute Template Open for non Std file");
               end case;

            when Close =>
               case P.File is
                  when Posix.Io.Standard_Input =>
                     Si.Hstdinput := Open_File (Null_Filename,
                                                Posix.Io.Read_Only);
                     P.Chandle := Si.Hstdinput;
                  when Posix.Io.Standard_Output =>
                     Si.Hstdoutput := Create_File (Null_Filename,
                                                   Posix.Io.Write_Only);
                     P.Chandle := Si.Hstdoutput;
                  when Posix.Io.Standard_Error =>
                     Si.Hstderror := Create_File (Null_Filename,
                                                  Posix.Io.Write_Only);
                     P.Chandle := Si.Hstderror;
                  when others =>
                     Posix_Win32.Raise_Not_Yet_Implemented
                       ("Execute Template Close for non Std file");
               end case;

            when Duplicate =>
               Posix_Win32.Raise_Not_Yet_Implemented
                 ("Execute Template Duplicate for non Std file");
         end case;
         P := P.Next;
      end loop;
   end Execute_Template;


   -------------------
   -- Start_Process --
   -------------------

   procedure Start_Process
     (Child       :    out Posix.Process_Identification.Process_Id;
      Pathname    : in     Posix.Pathname;
      Template    : in     Process_Template;
      Env_List    : in     Posix.Process_Environment.Environment;
      Arg_List    : in     Posix.Posix_String_List := Posix.Empty_String_List;
      Environment : in Boolean)
   is
      use type Win32.Bool;
      use type Win32.Ulong;

      --  add .exe suffix if it is not set
      function Add_Exe (S : in String) return String is
      begin
         if S'Length > 4 and then
           S (S'Last - 3 .. S'Last) /= ".exe"
         then
            return S & ".exe";
         else
            return S;
         end if;
      end Add_Exe;

      Startup_Informations : aliased Win32.Winbase.Startupinfo;
      Process_Informations : aliased Win32.Winbase.Process_Information;

      L_Pathname : constant String
        := Add_Exe (Posix.To_String (Pathname)) & Ascii.Nul;

      Result : Win32.Bool;

      Arguments      : Unbounded_String;
      Argument_Count : Natural := 0;

      ------------
      -- Concat --
      ------------

      procedure Concat (Item : in Posix_String;
                        Quit : in out Boolean) is
      begin
         --  the first argument is by convention the program name and we don't
         --  want to add it to the command line.
         if Argument_Count /= 0 then
            Arguments := Arguments & Posix.To_String (Item) & ' ';
         end if;
         Argument_Count := Argument_Count + 1;
         Quit := False;
      end Concat;

      --------------------------
      -- Concat_All_Arguments --
      --------------------------

      procedure Concat_All_Arguments is new Posix.For_Every_Item (Concat);

      ---------------
      -- To_LPVOID --
      ---------------

      function To_Lpvoid is
        new Ada.Unchecked_Conversion (Posix.Process_Environment.Environment,
                                      Win32.Lpvoid);

      Env_Pointer : Win32.Lpvoid;

   begin
      Check_Open (Template, "Start_Process");

      Startup_Informations.Cb          := Win32.Winbase.Startupinfo'Size / 8;
      Startup_Informations.Lpreserved  := null;
      Startup_Informations.Lpdesktop   := null;
      Startup_Informations.Lptitle     := null;
      Startup_Informations.Dwflags     := Win32.Winbase.Startf_Usestdhandles;
      Startup_Informations.Cbreserved2 := 0;
      Startup_Informations.Lpreserved2 := null;

      Execute_Template (Template, Startup_Informations);

      Concat_All_Arguments (Arg_List);

      if Environment then
         Env_Pointer := To_Lpvoid (Env_List);
      else
         Env_Pointer := System.Null_Address;
      end if;

      Run_Process :
      declare
         use type Win32.Ulong;
         use Win32.Winbase;
         L_Command : constant String :=
           Posix.To_String (Pathname) & ' ' &
           To_String (Arguments) &
           Ascii.Nul;
      begin
         Result := Createprocess
           (Lpapplicationname    => Win32.Addr (L_Pathname),
            Lpcommandline        => Win32.Addr (L_Command),
            Lpprocessattributes  => Security_Handles_Inherited'access,
            Lpthreadattributes   => null,
            Binherithandles      => Win32.True,
            Dwcreationflags      => Normal_Priority_Class,
            Lpenvironment        => Env_Pointer,
            Lpcurrentdirectory   => null,
            Lpstartupinfo        => Startup_Informations'Unchecked_Access,
            Lpprocessinformation => Process_Informations'Unchecked_Access
            );
      end Run_Process;

      if Result = Win32.False then
         Posix_Win32.Raise_Error ("Create_Process", Posix.Not_Enough_Space);
      end if;

      Template.Process_Informations.all := Process_Informations;
      Child := Process_Information_To_Process_Id (Process_Informations);

      --  record all child process for Wait_For_Child_Process functions
      Posix_Win32.Add_Child (Child);
   end Start_Process;


   -------------------
   -- Start_Process --
   -------------------

   procedure Start_Process
     (Child    :    out Posix.Process_Identification.Process_Id;
      Pathname : in     Posix.Pathname;
      Template : in     Process_Template;
      Arg_List : in     Posix.Posix_String_List :=
        Posix.Empty_String_List)
   is
      Null_Environment : Posix.Process_Environment.Environment;
   begin
      Start_Process (Child, Pathname, Template,
                     Null_Environment, Arg_List,
                     Environment => False);
   end Start_Process;


   -------------------
   -- Start_Process --
   -------------------

   procedure Start_Process
     (Child    :    out Posix.Process_Identification.Process_Id;
      Pathname : in     Posix.Pathname;
      Template : in     Process_Template;
      Env_List : in     Posix.Process_Environment.Environment;
      Arg_List : in     Posix.Posix_String_List := Posix.Empty_String_List)
   is
   begin
      Start_Process (Child, Pathname, Template, Env_List, Arg_List,
                     Environment => True);
   end Start_Process;


   Lpfilepart : aliased Win32.Lpstr;

   --------------------------
   -- Start_Process_Search --
   --------------------------

   procedure Start_Process_Search
     (Child    :    out Posix.Process_Identification.Process_Id;
      Filename : in     Posix.Filename;
      Template : in     Process_Template;
      Arg_List : in     Posix.Posix_String_List := Posix.Empty_String_List)
   is
      Null_Environment : Posix.Process_Environment.Environment;
      Max_Len    : constant := 500;
      Pathname   : String (1 .. Max_Len) := (others => '.');
      pragma Warnings (Off, Pathname);
      L_Filename : constant String := Posix.To_String (Filename) & Ascii.Nul;
      Ext_Var    : constant String := ".exe" & Ascii.Nul;

      Result     : Win32.Dword;
   begin
      Check_Open (Template, "Start_Process_Search");

      Result := Win32.Winbase.Searchpath (null,
                                          Win32.Addr (L_Filename),
                                          Win32.Addr (Ext_Var),
                                          Win32.Dword (Pathname'Length),
                                          Win32.Addr (Pathname),
                                          Lpfilepart'access);

      if Result = 0 then
         Start_Process (Child, Filename,
                        Template, Null_Environment, Arg_List,
                        Environment => False);
      else
         Start_Process
           (Child,
            Posix.To_Posix_String (Pathname (1 .. Positive (Result))),
            Template, Null_Environment, Arg_List,
            Environment => False);
      end if;
   end Start_Process_Search;


   --------------------------
   -- Start_Process_Search --
   --------------------------

   procedure Start_Process_Search
     (Child    :    out Posix.Process_Identification.Process_Id;
      Filename : in     Posix.Filename;
      Template : in     Process_Template;
      Env_List : in     Posix.Process_Environment.Environment;
      Arg_List : in     Posix.Posix_String_List := Posix.Empty_String_List)
   is
      Max_Len    : constant := 500;
      Pathname   : String (1 .. Max_Len);
      pragma Warnings (Off, Pathname);
      L_Filename : constant String := Posix.To_String (Filename) & Ascii.Nul;
      Ext_Var    : constant String := ".exe" & Ascii.Nul;
      Result     : Win32.Dword;
   begin
      Check_Open (Template, "Start_Process_Search");

      Search_Filename_In_Env :
      declare
         Old_Environment : Posix.Process_Environment.Environment;
      begin
         Posix.Process_Environment.Copy_From_Current_Environment
           (Old_Environment);

         Posix.Process_Environment.Copy_To_Current_Environment
           (Env_List);

         Result := Win32.Winbase.Searchpath (null,
                                             Win32.Addr (L_Filename),
                                             Win32.Addr (Ext_Var),
                                             Win32.Dword (Pathname'Length),
                                             Win32.Addr (Pathname),
                                             Lpfilepart'access);

         Posix.Process_Environment.Copy_To_Current_Environment
           (Old_Environment);
      end Search_Filename_In_Env;

      if Result = 0 then
         Start_Process (Child, Filename,
                        Template, Env_List, Arg_List,
                        Environment => True);
      else
         Start_Process
           (Child,
            Posix.To_Posix_String (Pathname (1 .. Positive (Result))),
            Template, Env_List, Arg_List,
            Environment => True);
      end if;
   end Start_Process_Search;


   ------------------
   -- Exit_Process --
   ------------------

   procedure Exit_Process (Status : in Exit_Status := Normal_Exit) is
   begin
      Win32.Winbase.Exitprocess (Win32.Uint (Status));
   end Exit_Process;


   ----------------------
   -- Status_Available --
   ----------------------

   function Status_Available (Status : in Termination_Status)
                              return Boolean is
   begin
      return Status.Pid /= Null_Process_Id;
   end Status_Available;


   -------------------
   -- Process_ID_Of --
   -------------------

   function Process_Id_Of (Status : in Termination_Status)
                           return Process_Id is
   begin
      if not Status_Available (Status) then
         Posix.Set_Error_Code (Invalid_Argument);
         raise Posix_Error;
      end if;
      return Status.Pid;
   end  Process_Id_Of;


   --------------------------
   -- Termination_Cause_Of --
   --------------------------

   function Termination_Cause_Of (Status : in Termination_Status)
                                  return Termination_Cause is
   begin
      if not Status_Available (Status) then
         Posix.Set_Error_Code (Invalid_Argument);
         raise Posix_Error;
      end if;
      return Exited;
   end Termination_Cause_Of;


   --------------------
   -- Exit_Status_Of --
   --------------------

   function Exit_Status_Of (Status : in Termination_Status)
                            return Exit_Status is
   begin
      if not Status_Available (Status)
        or Termination_Cause_Of (Status) /= Exited then
         Posix.Set_Error_Code (Invalid_Argument);
         raise Posix_Error;
      end if;
      return Exit_Status (Status.Exit_Status);
   end Exit_Status_Of;


   ---------------------------
   -- Termination_Signal_Of --
   ---------------------------

   function Termination_Signal_Of (Status : Termination_Status)
                                   return Posix.Signals.Signal is
   begin
      if not Status_Available (Status)
        or Termination_Cause_Of (Status) /= Terminated_By_Signal then
         Posix.Set_Error_Code (Invalid_Argument);
         raise Posix_Error;
      end if;
      return Posix.Signals.Signal_Terminate;
   end Termination_Signal_Of;


   ------------------------
   -- Stopping_Signal_Of --
   ------------------------

   function Stopping_Signal_Of (Status : Termination_Status)
                                return Posix.Signals.Signal is
   begin
      if not Status_Available (Status)
        or Termination_Cause_Of (Status) /= Stopped_By_Signal then
         Posix.Set_Error_Code (Invalid_Argument);
         raise Posix_Error;
      end if;
      return Posix.Signals.Signal_Abort;
   end Stopping_Signal_Of;


   Process_Status : aliased Win32.Dword;

   ----------------------------
   -- Wait_For_Child_Process --
   ----------------------------

   procedure Wait_For_Child_Process
     (Status         :    out Termination_Status;
      Child          : in     POSIX.Process_Identification.Process_ID;
      Block          : in     Boolean := True;
      Trace_Stopped  : in     Boolean := True;
      Masked_Signals : in     POSIX.Signal_Masking := POSIX.RTS_Signals)
   is
      use type Win32.ULONG;
      Retcode  : Win32.DWORD;
      Result   : Win32.BOOL;
      Process_Informations : Win32.Winbase.Process_Information;
   begin
      if not Posix_Win32.Exist (Child) then
         Posix_Win32.Raise_Error ("No Child Process", Posix.No_Child_Process);
      end if;

      Process_Informations := Process_Id_To_Process_Information (Child);

      if Child = Null_Process_Id then
         Posix.Set_Error_Code (No_Child_Process);
         raise Posix_Error;
      elsif Process_Informations.Dwprocessid = -1 then
         Status.Exit_Status := Exit_Stat (Failed_Creation_Exit);
         return;
      end if;

      if Block then
         Retcode := Win32.Winbase.Waitforsingleobject
           (Process_Informations.Hprocess, Win32.Winbase.Infinite);
      else
         Retcode := Win32.Winbase.Waitforsingleobject
           (Process_Informations.Hprocess, 0);
         if Retcode = Win32.Winbase.WAIT_TIMEOUT then
            Status := Termination_Status'(Null_Process_ID, 0);
            return;
         end if;
      end if;

      Status.Pid := Child;

      Result := Win32.Winbase.GetExitCodeProcess
        (Process_Informations.hProcess,
         Process_Status'Access);
      Status.Exit_Status := Exit_Stat (Process_Status);

      --  remove child from the list of child process
      POSIX_Win32.Remove_Child (Child);
   end Wait_For_Child_Process;


   ----------------------------
   -- Wait_For_Child_Process --
   ----------------------------

   procedure Wait_For_Child_Process
     (Status         :    out Termination_Status;
      Group          : in     POSIX.Process_Identification.Process_Group_ID;
      Block          : in     Boolean := True;
      Trace_Stopped  : in     Boolean := True;
      Masked_Signals : in     POSIX.Signal_Masking := POSIX.RTS_Signals) is
   begin
      POSIX_Win32.Wait (Status, Block);
   end Wait_For_Child_Process;


   ----------------------------
   -- Wait_For_Child_Process --
   ----------------------------

   procedure Wait_For_Child_Process
     (Status         :    out Termination_Status;
      Block          : in     Boolean := True;
      Trace_Stopped  : in     Boolean := True;
      Masked_Signals : in     POSIX.Signal_Masking := POSIX.RTS_Signals) is
   begin
      POSIX_Win32.Wait (Status, Block);
   end Wait_For_Child_Process;

end POSIX.Process_Primitives;
