--                              -*- Mode: Ada -*-
-- Filename        : windows-arrays.ads
-- Description     : Utilities for interfacing with c and win32.
--                   Contains instantiations of interfaces.c.pointers
--                   for all array types in declared in win32.
-- Author          : J. Schrer
-- Created On      : 19. Nov. 1997
-- Last Modified By:
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Win32;
with Interfaces.C,
     Interfaces.C.Strings, 
     Interfaces.C.Pointers;

package Windows.Arrays is
   
  ----------------------------------------------------------------------------              
  -- Conversion routines between ada wide_string's and win32  wide char pointers:
                        
  function To_Ada (Item       : in Win32.Pwch)
           return Wide_String;

  -- Use this function also for the subtypes of win32.pwch which are
  -- lpwch, pwchar, nwpstr, lpwstr, pwstr
           
  function To_Ada (Item       : in Win32.Pcwch)
           return Wide_String;       
           
  -- Use this function also for the subtypes of win32.pcwch which are
  -- lpcwch, pcwstr, lpcwstr

  ----------------------------------------------------------------------------                    
  -- An instantiation of interfaces.c.pointers is given for all 
  -- types for which an access-to-variable-type and an array are declared. 
           
  -- Utilities to get win32.byte_array and win32.uchar_array from
  -- win32.pbyte, win32.pcbyte, win32.pcuchar, win32.puchar.
                   
  package Uchar_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Unsigned_Char,
		      Element_Array      => Win32.Uchar_Array,
		      Default_Terminator => Interfaces.C.Unsigned_Char'Last);
          
  -- Pointer objects of type puchar         
  -- have to be converted to uchar_pointers.pointer for calls to subprograms of
  -- Uchar_Pointers. 
       
  function To_Var_Ptr(Item : in Win32.Pcuchar) return Uchar_Pointers.Pointer;
         
  -- A conversion routine from any access-to-constant-type to the type 
  -- pointer declared in the package is given.
  -- All functios are overloaded and called to_var_ptr. 
    
  
  --==========================================================================
  -- Example of use (see description of interfaces.c.pointers in the LRM):
  
  -- Example with an access-to-variable-type (e.g. win32.puchar)        
                    
  -- To fetch a win32.short_array of length N from a win32.puchar pointer.
  -- Note that N has to by of type interfaces.c.ptrdiff_t.
  
	--		N         : Natural := 10;        
	--		  
	--    Short_Ptr : Win32.Puchar := Any_Win32_Function_Returning_Puchar(...);
	--		Short_Arr : constant Win32.Uchar_Array := 
	--		            Windows.Utils.Uchar_Pointers.Value
	--		           (Ref    => Windows.Utils.Uchar_Pointers.Pointer(Short_Ptr), 
	--		            Length => Interfaces.C.Ptrdiff_T(N));
          
  -- win32.puchar and windows.utils.short_pointers.pointer are two different types.
  -- Therefore the type conversion is necessary.
                                 
  -- short_arr has length N after the call.

  ----------------------------------------------------------------------------
  -- Example with an access-to-constant-type (e.g. win32.pcuchar)        
                    
  -- To fetch a win32.short_array of length N from a win32.pcuchar pointer.
          
	--		Const_Short_Ptr : Win32.Pcuchar := Any_Win32_Function_Returning_Pcuchar(...);
	--    Const_Short_Arr : constant Win32.Uchar_Array := 
	--		                  Windows.Utils.Uchar_Pointers.Value
	--		                 (Ref    => Windows.Utils.To_Var_Ptr(Const_Short_Ptr), 
	--		                  Length => Interfaces.C.Ptrdiff_T(N));
          
  -- short_arr has length N after the call.
  --==========================================================================
  
  package Byte_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Win32.Byte, -- win32.lpbyte, 
		      Element_Array      => Win32.Byte_Array,
		      Default_Terminator => Win32.Byte'Last);
          
  -- Pointer objects of type pbyte and lpbyte (which is a subtype of pbyte)         
  -- have to be converted to byte_pointers.pointer for calls to subprograms of
  -- byte_Pointers. 
          
  function To_Var_Ptr(Item : in Win32.Pcbyte) return Byte_Pointers.Pointer;
                    
  ----------------------------------------------------------------------------         
  -- 16 bit wide chars
    
  package Wchar_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Wchar_T, 
                                -- win32.wchar_t, win32.wchar, 
		      Element_Array      => Win32.Wchar_Array,
		      Default_Terminator => Interfaces.C.Wide_Nul);
          
  -- Pointer objects of type pwch, lpwch, pwchar, nwpstr and lpwstr, pwstr
  -- (which are subtypes of pwch)         
  -- have to be converted to wchar_pointers.pointer for calls to subprograms of
  -- wchar_Pointers. 
          
  function To_Var_Ptr(Item : in Win32.Pcwch) return Wchar_Pointers.Pointer;
                    
  -- Use also for subtypes
  -- lpcwch, pcwstr, lpcwstr
  
 ----------------------------------------------------------------------------
  -- Utilities for all other types declared in win32.
  ----------------------------------------------------------------------------           
  -- 16 bit signed integers
                    
  package Short_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Short, -- win32.short
		      Element_Array      => Win32.Short_Array,
		      Default_Terminator => Interfaces.C.Short'Last);
          
  -- Pointer objects of type pshort         
  -- have to be converted to short_pointers.pointer for calls to subprograms of
  -- short_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pcshort) return Short_Pointers.Pointer;
  
  ----------------------------------------------------------------------------        
  -- 16 bit unsigned integers
                    
  package Ushort_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Unsigned_Short, -- win32.ushort
		      Element_Array      => Win32.Word_Array,
		      Default_Terminator => Interfaces.C.Unsigned_Short'Last);

  -- Pointer objects of type pushort and pword (which is derived from pushort)         
  -- have to be converted to ushort_pointers.pointer for calls to subprograms of
  -- Ushort_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pcword) return Ushort_Pointers.Pointer;
  
  -- Use also for subtype 
  -- lpword        
  ----------------------------------------------------------------------------        
  -- "int" types, signed
                    
  package Int_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Int, -- win32.int
		      Element_Array      => Win32.Int_Array,
		      Default_Terminator => Interfaces.C.Int'First);

  -- Pointer objects of type pint and lpint (which is a subtype of pint)         
  -- have to be converted to int_pointers.pointer for calls to subprograms of
  -- int_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pcint) return Int_Pointers.Pointer;
  
  function To_Var_Ptr(Item : in Win32.Pbool) return Int_Pointers.Pointer;
  
  -- Use also for subtype 
  -- lpbool        
  ----------------------------------------------------------------------------        
  -- "unsigned" or "unsigned int" unsigned
                    
  package Uint_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Unsigned, -- win32.uint, win32.wparam
		      Element_Array      => Win32.Uint_Array,
		      Default_Terminator => Interfaces.C.Unsigned'Last);

  -- Pointer objects of type puint          
  -- have to be converted to uint_pointers.pointer for calls to subprograms of
  -- uint_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pcuint) return Uint_Pointers.Pointer;
  ----------------------------------------------------------------------------        
  -- "long" 32 bit signed integers
                    
  package Long_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Long, 
                                -- win32.long, win32.lparam, win32.lresult
		      Element_Array      => Win32.Long_Array,
		      Default_Terminator => Interfaces.C.Long'First);

  -- Pointer objects of type plong and lplong (which is a subtype of lplong)         
  -- have to be converted to long_pointers.pointer for calls to subprograms of
  -- long_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pclong) return Long_Pointers.Pointer;

  ----------------------------------------------------------------------------        
  -- "unsigned long" 32 bit unsigned integers
                    
  package Ulong_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Unsigned_Long, -- win32.dword
		      Element_Array      => Win32.Ulong_Array,
		      Default_Terminator => Interfaces.C.Unsigned_Long'Last);

  -- Pointer objects of type pulong and pdword, lpdword (which are subtypes of pulong)         
  -- have to be converted to ulong_pointers.pointer for calls to subprograms of
  -- ulong_Pointers. 
                    
  function To_Var_Ptr(Item : in Win32.Pcdword) return Ulong_Pointers.Pointer;
  
  ----------------------------------------------------------------------------         
  -- Utilities to handle C float or C double arrays from C functions that
  -- deal with float* or double* parameters.
  
  -- win32.float und win32.double are subtypes of the corresponding types of
  -- interfaces.c. Array declarations and instantiations are given.
            
  type C_Float_Vector is array(Natural range <>) of aliased 
       Interfaces.C.C_Float;          
                      
  package C_Float_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.C_Float,
		      Element_Array      => C_Float_Vector,
		      Default_Terminator => Interfaces.C.C_Float'First);
          
  type Double_Vector is array(Natural range <>) of aliased 
       Interfaces.C.Double;          
                      
  -- Pointer objects of type pfloat          
  -- have to be converted to c_float_pointers.pointer for calls to subprograms of
  -- c_float_Pointers. 

  package Double_Pointers is new Interfaces.C.Pointers
	 	     (Index              => Natural,
		      Element            => Interfaces.C.Double,
		      Element_Array      => Double_Vector,
		      Default_Terminator => Interfaces.C.Double'First);
          
  -- Pointer objects of type pdouble          
  -- have to be converted to double_pointers.pointer for calls to subprograms of
  -- double_Pointers. 
                                  
end Windows.Arrays;           
 
