--                              -*- Mode: Ada -*-
-- Filename        : utilities-error_interface.adb
-- Description     : Error messages for coco generated parsers
-- Author          : J. Schroeer
-- Created On      : Fri Jul  4 00:16:41 1997
-- Last Modified By:
-- Last Modified On: Wed Jul  9 19:58:32 1997
-- Update Count    : 3
-- Status          :

with Ada.Strings.Unbounded,Utilities.Scanner, --Utilities.String_Utilities, 
Portable.Operating_System;

package body Utilities.Error_Interface is

  use type Ada.Strings.Unbounded.String_Access;
    
  Comment_Number,
  Warning_Number,
  Syntax_Error_Number,
  Semantic_Error_Number : Natural := 0;

  The_First_Error_Line_In_Listing,
  The_First_Error_Line              : Positive;
  The_First_Error_File              : Ada.Strings.Unbounded.String_Access;

  The_First_Warning_Line_In_Listing,
  The_First_Warning_Line            : Positive;
  The_First_Warning_File            : Ada.Strings.Unbounded.String_Access;

  No_Error        : Boolean := True;

  Max_Line_Length : constant Positive := 78;
  New_Line        : constant String   := Portable.Operating_System.New_Line;
  Offset          : constant Positive := 7;
  Blanks          : constant String   := "       ";

  -----------------------------------------------------------------------------
  procedure Initialize is
  begin
    Comment_Number                    := 0;
    Warning_Number                    := 0;
    Syntax_Error_Number               := 0;
    Semantic_Error_Number             := 0;
    The_First_Error_Line_In_Listing   := 1;
    The_First_Error_Line              := 1;
    The_First_Warning_Line_In_Listing := 1;
    The_First_Warning_Line            := 1;
    Ada.Strings.Unbounded.Free(The_First_Error_File);
    Ada.Strings.Unbounded.Free(The_First_Warning_File);
    No_Error := True;
  end Initialize;
  -----------------------------------------------------------------------------
  function Number_Of_Comments return Natural is
  begin
    return Comment_Number;
  end Number_Of_Comments;
  -----------------------------------------------------------------------------
  function Number_Of_Warnings return Natural is
  begin
    return Warning_Number;
  end Number_Of_Warnings;
  -----------------------------------------------------------------------------
  function Number_Of_Errors return Natural is
  begin
    return Syntax_Error_Number + Semantic_Error_Number;
  end Number_Of_Errors;
  -----------------------------------------------------------------------------
  function Number_Of_Syntax_Errors return Natural is
  begin
    return Syntax_Error_Number;
  end Number_Of_Syntax_Errors;
  -----------------------------------------------------------------------------
  function Number_Of_Semantic_Errors return Natural is
  begin
    return Semantic_Error_Number;
  end Number_Of_Semantic_Errors;
  -----------------------------------------------------------------------------
  procedure Set_Editor_Variables
           (Listing_Line,
            Line : out Positive;
            File : out Ada.Strings.Unbounded.String_Access) is
  begin
    Listing_Line := Utilities.Scanner.Listing_Line_Number;
    Line         := Utilities.Scanner.Line_Number;
    File         := new String'(Utilities.Scanner.File_Name);
  exception
    when Utilities.Scanner.No_Open_File_Error =>
      Line := 1;
      File := new String'("");
  end Set_Editor_Variables;
  -----------------------------------------------------------------------------
  procedure Print_Next_Line (Message : in String;
                             Indent  : in Boolean) is
  begin
    case Indent is
      when True  =>
        Utilities.Scanner.Print_Message
                ("       " & Message,
                 Treat_As       => Utilities.Scanner.Message,
                 Line_Increment => 1);
      when False =>
        Utilities.Scanner.Print_Message
                (Message,
                 Treat_As       => Utilities.Scanner.Message,
                 Line_Increment => 1);
    end case;
  end Print_Next_Line;
  -----------------------------------------------------------------------------
  procedure Print_Comment (Message    : in String;
                           First_Line,
                           Indent     : in Boolean := True) is
  begin
    case First_Line is
      when True  =>
        Comment_Number := Comment_Number + 1;
        Utilities.Scanner.Print_Message
                ("**COMM " & Message,
                 Treat_As       => Utilities.Scanner.Comment,
                 Line_Increment => 0);
      when False =>
        Print_Next_Line (Message, Indent);
    end case;
  end Print_Comment;
  -----------------------------------------------------------------------------
  procedure Print_Warning (Message    : in String;
                           First_Line,
                           Indent     : in Boolean := True) is
  begin
    case First_Line is
      when True  =>
        Warning_Number := Warning_Number + 1;
        Utilities.Scanner.Print_Message
                 ("**WARN " & Message,
                  Treat_As       => Utilities.Scanner.Warning,
                  Line_Increment => 0);
        if Number_Of_Warnings = 1 then
          Set_Editor_Variables
                   (Listing_Line => The_First_Warning_Line_In_Listing,
                    Line         => The_First_Warning_Line,
                    File         => The_First_Warning_File);
        end if;
      when False =>
        Print_Next_Line (Message, Indent);
    end case;
  end Print_Warning;
  -----------------------------------------------------------------------------
  procedure Print_Syntax_Error
   (Expected_Symbols : in out Utilities.String_Vectors.String_Vector_Access) is

    Index     : Natural := 0;
    Column    : Natural;
    New_Lines : Natural := 0;
    ---------------------------------------------------------------------------
    function Symbols_To_String return String is
      Symbol : Ada.Strings.Unbounded.String_Access;
    begin
      Index := Index + 1;
      if Index = Expected_Symbols'First then
        Symbol := Expected_Symbols(Index);
        Column := Offset + 11 + Symbol'Length;
        return Symbol.all & " read, but " & Symbols_To_String;
      elsif Index = Expected_Symbols'First + 1 and
            Index < Expected_Symbols'Last then
        Symbol := Expected_Symbols(Index);
        Column := Column + Symbol'Length;
        if Column > Max_Line_Length then
          Column    := Offset + Symbol'Length;
          New_Lines := New_Lines + 1;
          return New_Line & Blanks & Symbol.all & Symbols_To_String;
        else
          return Symbol.all & Symbols_To_String;
        end if;
      elsif Index < Expected_Symbols'Last then
        Symbol := Expected_Symbols(Index);
        Column := Column + Symbol'Length + 2;
        if Column > Max_Line_Length then
          Column    := Offset + Symbol'Length;
          New_Lines := New_Lines + 1;
          return ',' & New_Line & Blanks & Symbol.all & Symbols_To_String;
        else
          return ", " & Symbol.all & Symbols_To_String;
        end if;
      elsif Index = Expected_Symbols'Last then
        Symbol := Expected_Symbols(Index);
        if Index > 2 then
          Column := Column + Symbol'Length + 13;
          if Column > Max_Line_Length then
            New_Lines := New_Lines + 1;
            return ',' & New_Line & Blanks & Symbol.all & " expected";
          else
            return ", " & Symbol.all & " expected";
          end if;
        else
          return Symbol.all & " expected";
        end if;
        -- PROGRAM_ERROR not possible, expected_symbols'length >= 2,
        -- see coco-generated parsers
      end if;
      return ""; -- to avoid warning
    end Symbols_To_String;
    ----------------------
  begin
    Syntax_Error_Number := Syntax_Error_Number + 1;
    Utilities.Scanner.Print_Message ("**SYN  " & Symbols_To_String,
                                     Line_Increment => New_Lines);
    Utilities.String_Vectors.Deallocate (Expected_Symbols);
    if No_Error then
      Set_Editor_Variables (Listing_Line => The_First_Error_Line_In_Listing,
                            Line         => The_First_Error_Line,
                            File         => The_First_Error_File);
      No_Error := False;
    end if;
  end Print_Syntax_Error;
  -----------------------------------------------------------------------------
  procedure Print_Semantic_Error (Message    : in String;
                                  First_Line,
                                  Indent     : in Boolean := True) is
  begin
    case First_Line is
      when True  =>
        Semantic_Error_Number := Semantic_Error_Number + 1;
        Utilities.Scanner.Print_Message
                ("**SEM  " & Message, Line_Increment => 0);
        if No_Error then
          Set_Editor_Variables(Listing_Line => The_First_Error_Line_In_Listing,
                               Line         => The_First_Error_Line,
                               File         => The_First_Error_File);
          No_Error := False;
        end if;
      when False =>
        Print_Next_Line (Message, Indent);
    end case;
  end Print_Semantic_Error;
  -----------------------------------------------------------------------------
  function Message_String (Number : in Natural;
                           Class  : in String) return String is
  begin
    if Number = 0 then
      return " No " & Class & 's';
    elsif Number = 1 then
      return Integer'Image(Number) & ' ' & Class;
    else
      return Integer'Image(Number) & ' ' & Class & 's';
    end if;
  end Message_String;
  -----------------------------------------------------------------------------
  procedure Print_End_Of_Listing is
  begin
    Utilities.Scanner.Print_Message
            (New_Line & Message_String(Number => Syntax_Error_Number,
                                       Class  => " Syntax   error"),
             Treat_As => Utilities.Scanner.Message);
    Utilities.Scanner.Print_Message
            (Message_String(Number => Semantic_Error_Number,
                            Class  => " Semantic error"),
             Treat_As => Utilities.Scanner.Message);
    Utilities.Scanner.Print_Message
            (Message_String(Number => Warning_Number,
                           Class  => " Warning"),
             Treat_As => Utilities.Scanner.Message);
    Utilities.Scanner.Print_Message
            (Message_String(Number => Comment_Number,
                            Class  => " Comment") & New_Line,
             Treat_As => Utilities.Scanner.Message);
  end Print_End_Of_Listing;
  -----------------------------------------------------------------------------
  function Statistics return String is
  begin
    return Message_String(Number => Syntax_Error_Number +
                                    Semantic_Error_Number,
                          Class  => "Error") & " detected." &
           Message_String(Number => Warning_Number,
                          Class  => "Warning") & ',' &
           Message_String(Number => Comment_Number,
                          Class  => "Comment") & " issued.";
  end Statistics;
  -----------------------------------------------------------------------------
  function First_Error_Line_In_Listing   return Positive is
  begin
    return The_First_Error_Line_In_Listing;
  end First_Error_Line_In_Listing;
  -----------------------------------------------------------------------------
  function First_Error_Line   return Positive is
  begin
    return The_First_Error_Line;
  end First_Error_Line;
  -----------------------------------------------------------------------------
  function First_Error_File   return String is
  begin
    if The_First_Error_File /= null then
      return The_First_Error_File.all;
    else
      return Utilities.Scanner.File_Name;
    end if;
  end First_Error_File;
  -----------------------------------------------------------------------------
  function First_Warning_Line_In_Listing return Positive is
  begin
    return The_First_Warning_Line_In_Listing;
  end First_Warning_Line_In_Listing;
  -----------------------------------------------------------------------------
  function First_Warning_Line return Positive is
  begin
    return The_First_Warning_Line;
  end First_Warning_Line;
  -----------------------------------------------------------------------------
  function First_Warning_File return String is
  begin
    if The_First_Error_File /= null then
      return The_First_Warning_File.all;
    else
      return Utilities.Scanner.File_Name;
    end if;
  end First_Warning_File;
  -----------------------------------------------------------------------------
end Utilities.Error_Interface;
