--                              -*- Mode: Ada -*-
-- Filename        : utilities-generics-block_stack.adb
-- Description     : Efficient unbounded stack
-- Author          : A. Schwarte, J. Schroeer
-- Created On      : Wed Jul  2 12:44:40 1997
-- Last Modified By:
-- Last Modified On: Wed Jul  2 13:47:29 1997
-- Update Count    : 3
-- Status          :

-- stack for arbitrary element types for ALSYS ADA Vs 3.2
-- elements must not be of an access type or contain components thereof
-- to be used if many elements of small size must be pushed
--
-- Copyright (C) 1988 by dSPACE GmbH, FRG, Paderborn
--
-- author : Albert Schwarte
-- date   : 14-Mar-88   Vs 1.0  initial release
--          09-Jun-88   Vs 1.1  block_size_in byte now generic parameter
--                              new exception block_size_error

with Ada.Unchecked_Deallocation;

package body Utilities.Generics.Block_Stack is

  Elements_Per_Block : constant Index := Index
                      (Block_Size_In_Byte / (Element'Size / 8));

  subtype Block is Vector (0 .. Elements_Per_Block - 1);

  type Block_Cell is record
    Items    : Block;
    Next,
    Previous : Block_Link;
  end record;
  -----------------------------------------------------------------------------
  procedure Deallocate_Block is new Ada.Unchecked_Deallocation
           (Object => Block_Cell, Name => Block_Link);
  -----------------------------------------------------------------------------
  procedure Push (Item : in     Element;
                  Onto : in out Stack) is
  begin
    if Onto.Depth = 0 then
      Onto.First_Block       := new Block_Cell;
      Onto.Current_Block     := Onto.First_Block;
      Onto.Position_In_Block := 0;
    elsif Onto.Position_In_Block = Block'Last then
      Onto.Current_Block.Next          := new Block_Cell;
      Onto.Current_Block.Next.Previous := Onto.Current_Block;
      Onto.Current_Block               := Onto.Current_Block.Next;
      Onto.Position_In_Block           := 0;
    else
      Onto.Position_In_Block := Onto.Position_In_Block + 1;
    end if;
    Onto.Current_Block.Items(Onto.Position_In_Block) := Item;
    Onto.Depth := Onto.Depth + 1;
  exception
    when Constraint_Error | Storage_Error => raise Overflow;
  end Push;
  -----------------------------------------------------------------------------
  procedure Pop (Item :    out Element;
                 From : in out Stack) is
  begin
    if From.Depth = 0 then
      raise Underflow;
    else
      Item       := From.Current_Block.Items(From.Position_In_Block);
      From.Depth := From.Depth - 1;
      if From.Position_In_Block > 0 then
        From.Position_In_Block := From.Position_In_Block - 1;
      elsif From.Current_Block.Previous /= null then
        From.Current_Block := From.Current_Block.Previous;
        Deallocate_Block (From.Current_Block.Next);
        From.Position_In_Block := Elements_Per_Block - 1;
      else
        Deallocate_Block (From.First_Block);
        From.Current_Block := null;
      end if;
    end if;
  end Pop;
  -----------------------------------------------------------------------------
  function  Top_Of (The_Stack       : in Stack) return Element is
  begin
    return The_Stack.Current_Block.Items (The_Stack.Position_In_Block);
  exception
    when Constraint_Error => raise Underflow;
  end Top_Of;
  -----------------------------------------------------------------------------
  procedure Change_Top (Of_Stack    : in out Stack;
                        To_Element  : in     Element) is
  begin
    Of_Stack.Current_Block.Items (Of_Stack.Position_In_Block) := To_Element;
  exception
    when Constraint_Error => raise Underflow;
  end Change_Top;
  -----------------------------------------------------------------------------
  function Depth_Of (The_Stack : in Stack) return Index is
  begin
    return The_Stack.Depth;
  end Depth_Of;
  -----------------------------------------------------------------------------
  function Is_Empty (The_Stack : in Stack) return Boolean is
  begin
    return The_Stack.Depth = 0;
  end Is_Empty;
  -----------------------------------------------------------------------------
  procedure Convert (The_Stack     : in out Stack;
                     To_Vector     : in out Vector;
                     Reverse_Order : in     Boolean := False) is

    First, Last, N : Index;
  begin
    if The_Stack.Depth > 0 then
      if Reverse_Order then
--        for i in to_vector'range loop
--          pop (item => to_vector(i), from => the_stack);
--        end loop;
        First := To_Vector'First;
        Last  := First + The_Stack.Position_In_Block;
        N     := The_Stack.Position_In_Block + 1;
        for I in First .. Last loop
          N := N - 1;
          To_Vector(I) := The_Stack.Current_Block.Items(N);
        end loop;
        while The_Stack.Current_Block.Previous /= null loop
          First := Last + 1;
          Last  := Last + Elements_Per_Block;
          N     := Elements_Per_Block;
          The_Stack.Current_Block := The_Stack.Current_Block.Previous;
          Deallocate_Block (The_Stack.Current_Block.Next);
          for I in First .. Last loop
            N := N - 1;
            To_Vector(I) := The_Stack.Current_Block.Items(N);
          end loop;
        end loop;
      else
--        for i in reverse to_vector'range loop
--          pop (item => to_vector(i), from => the_stack);
--        end loop;
        Last  := To_Vector'Last;
        First := Last - The_Stack.Position_In_Block;
        To_Vector(First .. Last) :=
           The_Stack.Current_Block.Items(0 .. The_Stack.Position_In_Block);
        while The_Stack.Current_Block.Previous /= null loop
          Last  := First - 1;
          First := First - Elements_Per_Block;
          The_Stack.Current_Block := The_Stack.Current_Block.Previous;
          Deallocate_Block (The_Stack.Current_Block.Next);
          To_Vector(First .. Last) := The_Stack.Current_Block.Items;
        end loop;
      end if;
      Deallocate_Block (The_Stack.First_Block);
      The_Stack.Current_Block     := null;
      The_Stack.Depth             := 0;
      The_Stack.Position_In_Block := 0;
    end if;
  exception
    when Constraint_Error | Underflow => raise Vector_Conversion_Error;
  end Convert;
  -----------------------------------------------------------------------------
  procedure Convert (The_Stack     : in out Stack;
                     To_Vector     : in out Vector_Access;
                     Reverse_Order : in     Boolean := False) is
    ----------------------------------------------------
    procedure Allocate (Storage_For : in out Vector_Access;
                        Length      : in     Index) is
      procedure Deallocate is new Ada.Unchecked_Deallocation
               (Object => Vector, Name => Vector_Access);
    begin
      if Storage_For = null then
        Storage_For := new Vector (1 .. Length);
      elsif Storage_For.all'Length /= Length then
        Deallocate (Storage_For);
        Storage_For := new Vector (1 .. Length);
      end if;
    end Allocate;
    -------------
  begin
    Allocate (Storage_For => To_Vector, Length => The_Stack.Depth);
    Convert(The_Stack, To_Vector.all, Reverse_Order);
  exception
    when Storage_Error => raise Overflow;
  end Convert;
  -----------------------------------------------------------------------------
  procedure Deallocate (The_Stack : in out Stack) is
  begin
    if The_Stack.Depth > 0 then
      while The_Stack.Current_Block.Previous /= null loop
        The_Stack.Current_Block := The_Stack.Current_Block.Previous;
        Deallocate_Block (The_Stack.Current_Block.Next);
      end loop;
      Deallocate_Block (The_Stack.First_Block);
      The_Stack.Depth             := 0;
      The_Stack.Position_In_Block := 0;
      The_Stack.Current_Block     := null;
    end if;
  end Deallocate;
  -----------------------------------------------------------------------------
begin
  if Elements_Per_Block < 1 then
    raise Block_Size_Error;
  end if;
end Utilities.Generics.Block_Stack;
