--                              -*- Mode: Ada -*-
-- Filename        : utilities-generics-hash.adb
-- Description     : Generic hash table
-- Author          : J. Schroeer
-- Created On      : Wed Jul  2 12:45:32 1997
-- Last Modified By:
-- Last Modified On: Wed Jul  2 13:48:51 1997
-- Update Count    : 3
-- Status          :

with Ada.Unchecked_Deallocation;

package body Utilities.Generics.Hash is

  type Cell is record
    The_Key      : Key;
    The_Content  : Element;
    Next         : Link;
    Insert_Index : Index;
  end record;
  -----------------------------------------------------------------------------
  function Search (The_Key      : in Key;
                   Array_Element: in Link) return Link is
    Temp : Link := Array_Element;
  begin
    while Temp /= null and then not Is_Equal(Temp.The_Key, The_Key) loop
      Temp := Temp.Next;
    end loop;
    return Temp;
  end Search;
  -----------------------------------------------------------------------------
  procedure Insert(The_Table  : in out Table;
                   The_Key    : in     Key;
                   The_Content: in     Element;
                   If_Inserted: in     Action_If_Inserted:= Raise_Exception) is

    Hash_Index: constant Index := 
                Hash_Function(The_Key      => The_Key,
                              Table_Length => The_Table.Length + 1);
    Temp      : Link  := Search(The_Key      => The_Key,
                                Array_Element=> The_Table.Content(Hash_Index));
  begin
    if Temp = null then
      The_Table.Inserted_Entries := The_Table.Inserted_Entries + 1;
      if The_Table.Content(Hash_Index) /= null then
        The_Table.Collisions := The_Table.Collisions + 1;
      end if;
      begin
        The_Table.Content(Hash_Index) := new Cell'
                (The_Key          => The_Key,
                 The_Content      => The_Content,
                 Next             => The_Table.Content(Hash_Index),
                 Insert_Index     => The_Table.Inserted_Entries);
      exception
        when Storage_Error => raise Overflow;
      end;
    else                                     -- element with key found
      case If_Inserted is
        when Raise_Exception   => raise Key_Already_Inserted;
        when Overwrite_Content => Temp.The_Content := The_Content;
      end case;
    end if;
  end Insert;
  -----------------------------------------------------------------------------
  function  Is_In_Table (The_Key   : in Key;
                         The_Table : in Table) return Boolean is
  begin
    return Search(The_Key       => The_Key,
                  Array_Element => The_Table.Content
                   (Hash_Function(The_Key      => The_Key,
                                  Table_Length => The_Table.Length + 1))) /= null;
  end Is_In_Table;
  -----------------------------------------------------------------------------
  function  Get_Element (The_Key   : in Key;
                         The_Table : in Table) return Element is
  begin
    return Search(The_Key       => The_Key,
                  Array_Element => The_Table.Content
                   (Hash_Function(The_Key      => The_Key,
                                  Table_Length => The_Table.Length + 1))).
           The_Content;
  exception
    when Constraint_Error => raise Key_Not_Found;
  end Get_Element;
  -----------------------------------------------------------------------------
  function Index_Of     (The_Key   : in Key;
                         The_Table : in Table) return Index is
  begin
    return Search(The_Key       => The_Key,
                  Array_Element => The_Table.Content
                   (Hash_Function(The_Key      => The_Key,
                                  Table_Length => The_Table.Length + 1))).
           Insert_Index;
  exception
    when Constraint_Error => raise Key_Not_Found;
  end Index_Of;
  -----------------------------------------------------------------------------
  procedure Deallocate is new Ada.Unchecked_Deallocation
           (Object => Cell, Name => Link);
  -----------------------------------------------------------------------------
  procedure Erase (The_Key   : in     Key;
                   The_Table : in out Table) is

    Hash_Index: constant Index := 
                Hash_Function(The_Key      => The_Key,
                              Table_Length => The_Table.Length + 1);
    Temp      : Link := The_Table.Content(Hash_Index);
    Del       : Link;
  begin
    if Temp /= null and then Is_Equal(Temp.The_Key, The_Key) then
      The_Table.Content(Hash_Index) := The_Table.Content(Hash_Index).Next;
      Deallocate_Key    (Temp.The_Key);
      Deallocate_Element(Temp.The_Content);
      Deallocate        (Temp);
      The_Table.Inserted_Entries := The_Table.Inserted_Entries - 1;
    else
      while Temp.Next /= null loop
        if Is_Equal(Temp.Next.The_Key, The_Key) then
          Del       := Temp.Next;
          Temp.Next := Del.Next;
          Deallocate_Key    (Del.The_Key);
          Deallocate_Element(Del.The_Content);
          Deallocate        (Del);
          The_Table.Inserted_Entries := The_Table.Inserted_Entries - 1;
          exit;
        end if;
        Temp := Temp.Next;
      end loop;
    end if;
  end Erase;
  -----------------------------------------------------------------------------
  function  Number_Of_Inserted_Elements(The_Table : in Table) return Index is
  begin
    return The_Table.Inserted_Entries;
  end Number_Of_Inserted_Elements;
  -----------------------------------------------------------------------------
  function Is_Empty(The_Table : in Table) return Boolean is
  begin
    return The_Table.Inserted_Entries = 0;
  end Is_Empty;
  -----------------------------------------------------------------------------
  procedure Deallocate (The_Table : in out Table) is
    Temp : Link;
  begin
    for I in The_Table.Content'range loop
      Temp := The_Table.Content(I);
      while Temp /= null loop
        The_Table.Content(I) := Temp.Next;
        Deallocate_Key    (Temp.The_Key);
        Deallocate_Element(Temp.The_Content);
        Deallocate        (Temp);
        Temp := The_Table.Content(I);
      end loop;
    end loop;
    The_Table.Inserted_Entries := 0;
    The_Table.Collisions       := 0;
  end Deallocate;
  -----------------------------------------------------------------------------
  procedure Deallocate(The_Table : in out Table_Access) is
    procedure Deallocate_Table is new Ada.Unchecked_Deallocation
             (Object => Table, Name => Table_Access);
  begin
    if The_Table /= null then
      Deallocate      (The_Table.all);
      Deallocate_Table(The_Table);
    end if;
  end Deallocate;
  -----------------------------------------------------------------------------
  procedure Statistics (The_Table                  : in     Table;
                        Number_Of_Inserted_Entries :    out Index;
                        Number_Of_Collisions       :    out Index) is
  begin
    Number_Of_Inserted_Entries := The_Table.Inserted_Entries;
    Number_Of_Collisions       := The_Table.Collisions;
  end Statistics;
  -----------------------------------------------------------------------------
  procedure Save_Content (The_Table    : in     Table;
                          The_Keys     :    out Key_Array;
                          The_Elements :    out Element_Array) is
    Temp    : Link;
    Counter : Index := 0;
  begin
    for I in The_Table.Content'range loop
      Temp := The_Table.Content(I);
      while Temp /= null loop
        Counter := Counter + 1;
        The_Keys    (Counter) := Temp.The_Key;  
        The_Elements(Counter) := Temp.The_Content;
        Temp := Temp.Next;
      end loop;
    end loop;      
  end Save_Content;                 
  -----------------------------------------------------------------------------
end Utilities.Generics.Hash;
