--                              -*- Mode: Ada -*-
-- Filename        : utilities-generics-hash.ads
-- Description     : Generic hash table, implemented as an array of lists.
--                   Choose a prime-number - 1 for the array length.
-- Author          : J. Schrer
-- Created On      : Wed Jul  2 12:51:29 1997
-- Last Modified By: J. Schrer
-- Last Modified On: 4. dez. 1998, js, added Save_Content
-- Update Count    : 
-- Status          :

generic
  type Element is private;
  type Key     is private; -- f.e. string
  type Index   is range <>;

  with function Is_Equal           (Left, Right : in Key) return Boolean is <>;

  with function Hash_Function      (The_Key     : in Key;
                                    Table_Length: in Index) return Index is <>;

  with procedure Deallocate_Element(The_Element : in out Element) is <>;
  with procedure Deallocate_Key    (The_Key     : in out Key)     is <>;

package Utilities.Generics.Hash is

  type Action_If_Inserted is (Raise_Exception, Overwrite_Content);

  type Table(Length : Index := 996) is limited private;
  type Table_Access is access all Table;

  procedure Insert(The_Table   : in out Table;
                   The_Key     : in     Key;
                   The_Content : in     Element;
                   If_Inserted : in     Action_If_Inserted := Raise_Exception);
                   
  -- Raises Key_Already_Inserted if The_Key already in The_Table when
  -- If_Inserted = Raise_Exception. When If_Inserted = Overwrite_Content,
  -- overwrites old content by The_Content.                                 

  function  Is_In_Table(The_Key   : in Key;
                        The_Table : in Table) return Boolean;

  function  Get_Element(The_Key   : in Key;
                        The_Table : in Table) return Element;
                        
  -- Raises Key_Not_Found if The_Key not in The_Table.                      

  function  Index_Of   (The_Key   : in Key;
                        The_Table : in Table) return Index;

  -- Returns the index of The_Key. 
  -- (index is incremented internally in each call of insert.)                                              
  -- Raises Key_Not_Found if The_Key not in The_Table.                                            
                        
  procedure Erase      (The_Key   : in     Key;
                        The_Table : in out Table);
                        
  -- Deleted The_Key (and the corresponding content) from The_Table.                        

  function  Number_Of_Inserted_Elements(The_Table : in Table) return Index;

  function Is_Empty (The_Table : in Table) return Boolean;

  procedure Deallocate (The_Table : in out Table);
  procedure Deallocate (The_Table : in out Table_Access);

  procedure Statistics (The_Table                  : in     Table;
                        Number_Of_Inserted_Entries :    out Index;
                        Number_Of_Collisions       :    out Index);
                        
  type Key_Array     is array(Index range <>) of Key;                        
  type Element_Array is array(Index range <>) of Element;                        

  procedure Save_Content (The_Table    : in     Table;
                          The_Keys     :    out Key_Array;
                          The_Elements :    out Element_Array);
                          
  -- Returns actual content of The_Table.
                                                    
                          
  Key_Not_Found        : exception;
  Key_Already_Inserted : exception;
  Overflow             : exception;

private

  type Cell;
  type Link is access Cell;

  type Search_Array is array (Index range <>) of Link;

  type Table (Length : Index := 996) is record
    Inserted_Entries : Index := 0;
    Collisions       : Index := 0;
    Content          : Search_Array (0 .. Length);
  end record;

end Utilities.Generics.Hash;
