--                              -*- Mode: Ada -*-
-- Filename        : utilities-low_level.adb
-- Description     : Utilities for conversion to low_level types
-- Author          : J. Schrer
-- Created On      : 21. Jan. 1998
-- Last Modified By:
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Unchecked_Conversion,
     Ada.Exceptions,
     Ada.Strings.Fixed;
with Portable.Operating_System;

package body Utilities.Low_Level is
  ----------------------------------------------------------------------------   
  package body Conversions is
    --------------------------------------------------------------------------                   
    function To_Element_Array(Item : in Item_Type) return Element_Array is
    
      function Convert is new Ada.Unchecked_Conversion
        (Source => Item_Type, Target => Element_Array);
        
      pragma Inline(Convert); -- to eliminate additional call
    begin
      return Convert(Item);    
    end To_Element_Array;
    --------------------------------------------------------------------------                   
    function To_Bit_Array    (Item : in Item_Type) return Bit_Array is
    
      function Convert is new Ada.Unchecked_Conversion
        (Source => Item_Type, Target => Bit_Array);
        
      pragma Inline(Convert); -- to eliminate additional call
    begin
      return Convert(Item);    
    end To_Bit_Array;
    --------------------------------------------------------------------------                   
    function To_Boolean_Array(Item : in Item_Type) return Boolean_Array is
    
      function Convert is new Ada.Unchecked_Conversion
        (Source => Item_Type, Target => Boolean_Array);
        
      pragma Inline(Convert); -- to eliminate additional call
    begin
      return Convert(Item);    
    end To_Boolean_Array;
    --------------------------------------------------------------------------                   
    function Image_Of(Item                 : in Item_Type;
                      Base                 : in Ada.Text_Io.Number_Base := 16;
                      Separator            : in String := " ";
                      Leading_Zeros        : in Boolean := True;
                      Base_And_Number_Sign : in Boolean := True) return String is
      -------------------------------------      
      function Max_Digits return Natural is
        -- Maximum number of digits needed to convert an object of 
        -- element_type to string image for the given base
        Log_2_base  : Positive;
        Actual_Base : Ada.Text_Io.Number_Base := 16;
      begin
        for I in reverse 1 .. 4 loop
          if Base >= Actual_Base then
            Log_2_base := I;
            exit;
          end if;
          Actual_Base := Actual_Base / 2;
        end loop;
        -- Log_2_base is set in each case because base = 2 .. 16.
        -- Program_error not possible.
        if Element_Type'Size rem Log_2_base = 0 then
          return Element_Type'Size / Log_2_base;
        else
          return Element_Type'Size / Log_2_base + 1;
        end if;
      end Max_Digits;
      ---------------
      Width      : constant Natural       := Max_Digits;                        
      Elements   : constant Element_Array := To_Element_Array(Item);
      Base_Image : constant String        := Positive'Image(Base);
      Image      : String(1 .. Elements'Length * (Width + Separator'Length) +
                               Base_Image'Length); 
      First      : Positive := 1;
      Last       : Natural;
      Buffer     : String(1 .. Width + Base_Image'Length + 2);
      First_Digit,
      Last_Digit : Natural;
      Length     : Natural;
      use Ada.Strings.Fixed; -- to make "*" directly visible
    begin
      for I in Elements'range loop
        Modular_Text_Io.Put(To => Buffer, Item => Elements(I), Base => Base);
        if Base = 10 then -- no base and '#' in buffer
          First_Digit := Ada.Strings.Fixed.Index_Non_Blank(Buffer);
          Last_Digit  := Buffer'Last;
        else
          First_Digit :=  1 + Ada.Strings.Fixed.Index
            (Source => Buffer, Pattern => "#", Going => Ada.Strings.Forward);
          Last_Digit  := -1 + Ada.Strings.Fixed.Index
            (Source => Buffer, Pattern => "#", Going => Ada.Strings.Backward);
        end if;
        Length := Last_Digit - First_Digit + 1;  
        if Leading_Zeros then
          Last := First + Width - 1;
          Image(First .. Last) := (Width - Length) * '0' & 
                                  Buffer(First_Digit .. Last_Digit);                                
        else
          Last := First + Length - 1;
          Image(First .. Last) := Buffer(First_Digit .. Last_Digit);                                
        end if;                                    
        if I < Elements'Last then
          First := Last + 1;
          Last  := Last + Separator'Length; 
          Image(First .. Last) := Separator; 
        end if;
        First := Last + 1;
      end loop;
      if Base_And_Number_Sign then
        return Base_Image(Base_Image'First + 1 .. Base_Image'Last) &
               '#' & Image(1 .. Last) & '#';
      else
        return Image(1 .. Last);
      end if;          
    end Image_Of;
    --------------------------------------------------------------------------                    
  begin
    if Item_Type'Size < Element_Type'Size        or else 
       Item_Type'Size rem Element_Type'Size /= 0 then
      Ada.Exceptions.Raise_Exception(Illegal_Instantiation'Identity,
        Message => "Illegal instantiation of Utilities.C.Low_Level " & 
                   Portable.Operating_System.New_Line &
                   "Item_Type'Size = " & Natural'Image(Item_Type'Size) &
                   ", Element_Type'Size = " & Natural'Image(Element_Type'Size));
    end if;                   
  end Conversions;
  ----------------------------------------------------------------------------                     
end Utilities.Low_Level;
           
