--                              -*- Mode: Ada -*-
-- Filename        : utilities-low_level.ads
-- Description     : utilities for bits, bit arrays and 
--                   byte-, word-, longword-images of arbitrary types.
-- Author          : J. Schroeer
-- Created On      : 16. Jan 1998
-- Last Modified By:
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Text_Io;
with Interfaces, Interfaces.C;

package Utilities.Low_Level is
  
  type Bit is range 0 .. 1;
  for Bit'Size use 1;           -- a bit has a size of one bit now.
  
  type Bit_Array is array    (Natural range <>) of Bit;
  pragma Pack(Bit_Array);       -- one bit for each element in bit_array
  
  type Boolean_Array is array(Natural range <>) of Boolean;
  pragma Pack(Boolean_Array);   -- one bit for each element in boolean_array

  -- Preinstantiations for third generic parameter Modular_Text_Io of
  -- below given generic package.
  
  package Unsigned_8_text_io  is new Ada.Text_Io.Modular_Io(Interfaces.Unsigned_8);
  package Unsigned_16_Text_Io is new Ada.Text_Io.Modular_Io(Interfaces.Unsigned_16);
  package Unsigned_32_Text_Io is new Ada.Text_Io.Modular_Io(Interfaces.Unsigned_32);
  

  generic
    type Item_Type    is private;
    type Element_Type is mod <>;  
    
    with package Modular_Text_Io is new Ada.Text_Io.Modular_Io(Element_Type);
    
  package Conversions is
    
    -- For item type each constraint type is possible.
    
    -- For element_type a modular (unsigned) type,
    -- like Interfaces.Unsigned_8, Unsigned_16, Unsigned_32 may be used.
    
    -- For Modular_Text_Io, one of the predefined packages from above 
    -- should be taken (to save memory).
    
    Illegal_Instantiation : exception;
    
    type Element_Array is array 
      (Natural range 1 .. Item_Type'Size / Element_Type'Size) of Element_Type;
        
    -- The body eleboration raises Illegal_Instantiation if  
    -- Item_Type'Size < Element_Type'Size or if
    -- Item_Type'Size and Element_Type'Size are not divisible without rest.
    -- ( Item_Type'Size rem Element_Type'Size /= 0 )

    subtype Bit_Array is Low_Level.Bit_Array(Natural range 1 .. Item_Type'Size);
      
    -- From here on bit_array refers to the subtype.
      
    subtype Boolean_Array is Low_Level.Boolean_Array(Bit_Array'range);
    
    -- dito for Boolean_Array
    
    function To_Element_Array(Item : in Item_Type) return Element_Array;
    
    -- Uses ada.unchecked_conversion.
    
    function To_Bit_Array    (Item : in Item_Type) return Bit_Array;

    -- Uses ada.unchecked_conversion.
    
    function To_Boolean_Array(Item : in Item_Type) return Boolean_Array;
    
    -- Uses ada.unchecked_conversion.

    function Image_Of(Item                 : in Item_Type;
                      Base                 : in Ada.Text_Io.Number_Base := 16;
                      Separator            : in String := " ";
                      Leading_Zeros        : in Boolean := True;
                      Base_And_Number_Sign : in Boolean := True) return String;
                     
    -- Gives a string image of item:
    -- Base                : 2 .. 16 (see ada.text_io) 
    --                       the base for each element-image from the element_array.      
    -- Separator           : A string that separates the images of two elements 
    --                       from the element_array                     
    -- Leading_Zeros       : true  - the image of each element is filled with 
    --                               leading zeros when the most significant digits
    --                               of the elements are zero.
    --                       False - leading zero-digits are ommitted.
    -- Base_And_Number_Sign: true  - The base'image and the number_sign ('#')
    --                               are given.
    --                       false - they are ommited.
                                        
    -- Example of ues:                  
		-- Image_Of(X,Base => 16, Separator => " ", Leading_Zeros => True, Base_And_Number_Sign => True)
		-- 16#00 00 FF FF#
                      
		-- Image_Of(X,Base => 16, Separator => "|", Leading_Zeros => True, Base_And_Number_Sign => True)
		-- 16#00|00|FF|FF#
                      
		-- Image_Of(X,Base => 16, Separator => " ", Leading_Zeros =>False, Base_And_Number_Sign => True)
		-- 16#0 0 FF FF#
                      
		-- Image_Of(X,Base => 16, Separator => " ", Leading_Zeros => True, Base_And_Number_Sign =>False)
		-- 00 00 FF FF
                      
		-- Image_Of(X,Base =>  2, Separator => "  ", Leading_Zeros => True, Base_And_Number_Sign => True)
		-- 2#00000000  00000000  11111111  11111111#

  end Conversions;
  
	-- Example of use of generic package conversion:
	                      
	--    type Any_Type is record 
	--      X : Interfaces.Unsigned_16 := 0;
	--      Y : Interfaces.Unsigned_16 := Interfaces.Unsigned_16'Last;
	--    end record;
	--    
	--	  package Any_Type_Bytes is new Utilities.Low_Level.Conversions
	--	    (Item_Type       => Any_Type,               -- Assume any_type'size = 32
	--	     Element_Type    => Interfaces.Unsigned_8,  -- This is a byte of size 8.
	--       Modular_Text_Io => Utilities.Low_Level.Unsigned_8_text_io);                                   
	--	    
	--	  X       : Any_Type;
	--	  X_Bytes : constant Any_Type_Bytes.Element_Array :=     
	--	                     Any_Type_Bytes.To_Element_Array(X);
	--	  -- An array of 4 Interfaces.Unsigned_8 elements.    
                   
	--	  X_Bits      : constant Any_Type_Bytes.Bit_Array :=     
	--	                         Any_Type_Bytes.To_Bit_Array(X);
	--	  -- An array of 32 bits of type bit above.                                        
  
	--	  X_Booleans  : constant Any_Type_Bytes.Boolean_Array :=     
	--	                         Any_Type_Bytes.To_Boolean_Array(X);
	--	  -- An array of 32 boolean's.                                        
  
	--	  Bit_1_set   : constant Boolean := Boolean'Val(X_Bits(1));
	--	  Bit_32_Set  : constant Boolean := X_Booleans(32);                   
	
	--		Any_Type_Bytes.Element_Array'Size =  32
	--		Any_Type_Bytes.Bit_Array'Size     =  32
	--		Any_Type_Bytes.Boolean_Array'Size =  32
	--		Any_Type'Size                     =  32
	--		X_Bytes'Size                      =  32
	--		X_Bits'Size                       =  32
	--		X_Booleans'Size                   =  32
	--		Bit_1_set  = False
	--		Bit_32_Set = True
  
end Utilities.Low_Level;


