--                              -*- Mode: Ada -*-
-- Filename        : utilities-scanner.adb
-- Description     : Hand written scanner for syntax with Ada lexical elements
-- Author          : J. Schroeer
-- Created On      : Thu Jul  3 20:34:20 1997
-- Last Modified By:
-- Last Modified On: Fri Jul  4 11:16:16 1997
-- Update Count    : 25
-- Status          :

with Utilities.String_Utilities, Utilities.Calendar_Utilities,
     Ada.Strings.Unbounded, Ada.Calendar, Portable.Operating_System;

package body Utilities.Scanner is

  Eof        : constant Character := Ascii.Etx;  -- end of file
  Eol        : constant Character := Ascii.Cr;   -- end of line
  Start_Char : constant Character := Ascii.Nul;  -- start of file

  -- eof and eol are stored in component line of internal text by procedure
  -- get_character. start_char is stored in component last_character of
  -- internal text by procedure open

  -- scanns outoff an object of this type.
  -- package is a multi state-machine, because scanning outoff different
  -- sources in parallel should be possible.task types are not best
  -- appropriate for this work, because slowing task-switchs must occur
  -- for each symbol and program-space is multiplied.

  Tab_Length : constant Natural := 8;

  type Internal_Text (Source_Type : Source_Class := Ascii_File) is record
    Line_Number,
    Last_Line_Number : Natural  := 1; -- of source-text
    Last_Error_Line  : Positive := 1; -- that has been printed
    Last_Sign        : Natural  := 0; -- position of last read character
                                      -- in file
    Column_Number    : Natural  := 0; -- position of last read character
                                      -- with respect to tabs
    Read_Blanks      : Natural  := 0; -- to adjust column_number,
                                      -- if tab is read
    Last_Character   : Character := Start_Char; -- = line(last_sign);
    Dot_Saved        : Boolean := False; -- used in get_number
    Apostrophe_Saved : Boolean := False; -- used in get_string
    Line,
    Last_Line        : Ada.Strings.Unbounded.String_Access;
                                         -- content  of source-line
    Line_Length,
    Last_Line_Length : Natural := 0;     -- length   of source-line
    Symbol_Counter   : Natural := 0;     -- counts symbol in line
    File_Name        : Ada.Strings.Unbounded.String_Access;
  end record;


  Max_Open_Files : constant Natural := 10;  -- maximum number of files allowed
                                            -- to open at a time
  Source_Counter : Natural := 0;            -- total number of open files
  Source_Info    : array (1 .. Max_Open_Files) of Internal_Text;
  Source_File    : array (1 .. Max_Open_Files) of Ada.Text_Io.File_Type;

  The_Source      : Internal_Text; -- := source_info(source_counter)
                                   --    for faster access
  The_Listing_Line_Number : Positive := 1;

  Line_Message_Number     : Natural; -- used in get_symbol.get_character,
                                     --         open, message
  -----------------------------------------------------------------------------
  -- local error procedures

  procedure Trace_Error_Message (Program_Trace, Trace, Message : in String) is
  begin
    if Program_Trace = "" then
      Print_Message(Trace & Message);
    else
      Print_Message(Program_Trace & Message);
    end if;
  end Trace_Error_Message;
  -----------------------------------------------------------------------------
  function Image (Sym : in Symbol) return String is
  begin
    case Sym is
      when Not_Equal             => return "/=";
      when Greater_Than_Or_Equal => return ">=";
      when Less_Than_Or_Equal    => return "<=";
      when Becomes               => return ":=";
      when Arrow                 => return "=>";
      when Double_Dot            => return "..";
      when Exponentiate          => return "**";
      when End_Of_Source         => return "EOF";
      when End_Of_Line           => return "EOL";
      when others                => return Symbol'Image(Sym)(2..2);
    end case;
  end Image;
  -----------------------------------------------------------------------------
  procedure Open (Filename : in String       := "";
                  Source   : in Source_Class := Ascii_File;
                  Listing  : in Boolean      := True) is
  begin
    if Source_Counter >= Max_Open_Files then
      Print_Message("Restriction: Only " & Natural'Image(Max_Open_Files) &
                    " Files allowed to be opened at a time for scanner",
                    Treat_As => Message);
      Print_Message("The following FILES are open :", Message);
      if Source_Counter = Max_Open_Files then -- Schroer 24.2.1992
        Source_Info(Source_Counter) := The_Source;
      end if;
      for I in 1 .. Source_Counter loop
        Print_Message(Source_Class'Image(Source_Info(I).Source_Type) &
                      " : " & Source_Info(I).File_Name.all,
                      Treat_As => Message);
      end loop;
      Print_Message(
"If you have source-text, you can recompile body of package UTILITIES.SCANNER",
       Treat_As => Message);
      raise Too_Many_Sources_Error;
    end if;

    if Source_Counter > 0 then
      Source_Info(Source_Counter) := The_Source;
    end if;
    Source_Counter      := Source_Counter + 1;
    Line_Message_Number := 0;
    case Source is
      when Ascii_File =>
        The_Source := Internal_Text'
                     (Source_Type      => Ascii_File,
                      Line_Number      => 1,
                      Last_Line_Number => 1,
                      Last_Error_Line  => 1,
                      Last_Sign        => 0,
                      Column_Number    => 0,
                      Read_Blanks      => 0,
                      Last_Character   => Start_Char,
                      Dot_Saved        => False,
                      Apostrophe_Saved => False,
                      Line             => new String(1 .. Max_Line_Length),
                      Last_Line        => new String(1 .. Max_Line_Length),
                      Line_Length      => 0,
                      Last_Line_Length => 0,
                      Symbol_Counter   => 0,
                      File_Name        => new String'(Filename));
        begin
          Ada.Text_Io.Open(File => Source_File(Source_Counter),
                           Mode => Ada.Text_Io.In_File,
                           Name => Filename);
        exception
          when Ada.Text_Io.Name_Error =>
            Source_Counter := Source_Counter - 1;
            if Source_Counter > 0 then  -- Schrer 24.2.92
              The_Source := Source_Info(Source_Counter);
            end if;
            raise Name_Error;
          when Ada.Text_Io.Status_Error =>
            Source_Counter := Source_Counter - 1;
            if Source_Counter > 0 then  -- Schroer 24.2.92
              The_Source := Source_Info(Source_Counter);
            end if;
            raise Status_Error;
        end;
        if Listing and not Ada.Text_Io.Is_Open (Listing_File) then
          declare
            Last : Natural := Filename'Length;
            Time : Utilities.Calendar_Utilities.Time :=
                   Utilities.Calendar_Utilities.Time_Of(Ada.Calendar.Clock);
            Separator : constant Character :=
                        Portable.Operating_System.Directory_Separator;
          begin
            for I in reverse Filename'range loop
              exit when Filename(I) = Separator;
              if Filename(I) = '.' then
                Last := I - 1;
                exit;
              end if;
            end loop;
            begin
              Ada.Text_Io.Create(File => Listing_File,
                                 Mode => Ada.Text_Io.Out_File,
                                 Name => Filename(Filename'First..Last) &
                                         ".lst");
            exception
              when Ada.Text_Io.Status_Error |
                   Ada.Text_Io.Use_Error    =>
                Ada.Text_Io.Create(File => Listing_File,
                                   Mode => Ada.Text_Io.Out_File,
                                   Name => Filename(Filename'First..Last) &
                                           ".ls");
            end;
            The_Listing_Line_Number := 2;
            Print_Message(Utilities.Calendar_Utilities.Image_Of(Time),Message);
            Ada.Text_Io.New_Line(Listing_File);
          end;
        end if;
        Print_Message ("File: " & Filename, Message);
      when Normal_String =>
        The_Source := Internal_Text'
                     (Source_Type      => Normal_String,
                      Line_Number      => 1,
                      Last_Line_Number => 1,
                      Last_Error_Line  => 1,
                      Last_Sign        => 0,
                      Column_Number    => 0,
                      Read_Blanks      => 0,
                      Last_Character   => Start_Char,
                      Dot_Saved        => False,
                      Apostrophe_Saved => False,
                      Line             => new String'(Filename & Eof),
                      Last_Line        => new String'(""),
                      Line_Length      => Filename'Length + 1,
                      Last_Line_Length => 0,
                      Symbol_Counter   => 0,
                      File_Name        => new String'(""));
    end case;
  end Open;
  -----------------------------------------------------------------------------
  procedure Close is
  begin
    if Source_Counter = 0 then
      Print_Message("Semantic-error: No File is open", Message);
      raise No_Open_File_Error;
    end if;

    case The_Source.Source_Type is
      when Ascii_File    => Ada.Text_Io.Close(Source_File(Source_Counter));
      when Normal_String => null;
    end case;

    Ada.Strings.Unbounded.Free(The_Source.Line);
    Ada.Strings.Unbounded.Free(The_Source.Last_Line);
    Ada.Strings.Unbounded.Free(The_Source.File_Name);

    Source_Counter := Source_Counter - 1;
    if Source_Counter > 0 then
      The_Source := Source_Info(Source_Counter);
      if The_Source.Source_Type = Ascii_File and not End_Of_Source then
        Print_Message ("File: " & The_Source.File_Name.all,
                       Treat_As => Message);
      end if;
    end if;
  end Close;
  -----------------------------------------------------------------------------
  procedure Get_Symbol is separate;
  -----------------------------------------------------------------------------
  procedure Print_Message (Info           : in String;
                           Treat_As       : in Message_Type := Error;
                           Line_Increment : in Natural := 1) is

    Correct_Col_Number: Natural := Column_Number;
    ------------------------------------------------------
    procedure Put_Message (File : in Ada.Text_Io.File_Type) is
    begin
      if Treat_As /= Message then
        Ada.Text_Io.New_Line (File);
        if The_Source.Symbol_Counter = 1 and then
           The_Source.Last_Line_Number > 0 then
          Ada.Text_Io.Put(File, Utilities.String_Utilities.Right_Justified
            (The_String   => Utilities.String_Utilities.Stripped
                            (Integer'Image(The_Source.Last_Line_Number)),
             In_The_Width => 4));
          Ada.Text_Io.Put (File, ' ');
          if The_Source.Last_Line(The_Source.Last_Line_Length) = Eol then
            Ada.Text_Io.Put_Line (File,
                The_Source.Last_Line(1 .. The_Source.Last_Line_Length - 1));
          else
            Ada.Text_Io.Put_Line (File,
                The_Source.Last_Line(1 .. The_Source.Last_Line_Length));
          end if;
          The_Listing_Line_Number := The_Listing_Line_Number + 1;
        end if;
        Ada.Text_Io.Put (File, Utilities.String_Utilities.Right_Justified
                  (The_String   => Utilities.String_Utilities.Stripped
                                  (Integer'Image(The_Source.Line_Number)),
                   In_The_Width => 4));
        Ada.Text_Io.Put (File, ' ');
        if The_Source.Line(The_Source.Line_Length) = Eol then
          Ada.Text_Io.Put_Line(File,
              The_Source.Line(1 .. The_Source.Line_Length - 1));
        elsif not (The_Source.Line_Length = 1 and then -- Schroer 24.2.1992
                   The_Source.Line(1) = Eof) then
          Ada.Text_Io.Put_Line(File,
              The_Source.Line(1 .. The_Source.Line_Length));
        end if;
        The_Source.Last_Error_Line := The_Source.Line_Number;
        for I in 1 .. Correct_Col_Number + 4 loop
           Ada.Text_Io.Put (File, '-');
        end loop;
        Ada.Text_Io.Put_Line (File, "^");
        The_Listing_Line_Number := The_Listing_Line_Number + 3;
      end if;
      Ada.Text_Io.Put_Line (File, Info);
      if Line_Increment = 0 then
        The_Listing_Line_Number := The_Listing_Line_Number + 1 +
                                   Utilities.String_Utilities.Number_Of
                                  (The_Character => Ascii.Cr,
                                   In_The_String => Info);
      else
        The_Listing_Line_Number := The_Listing_Line_Number + Line_Increment;
      end if;
    end Put_Message;
    ----------------
  begin
    if Source_Counter in Source_Info'range then
      if The_Source.Column_Number > The_Source.Line'Last - 1 then
        -- line longer 255 read
        Correct_Col_Number := The_Source.Column_Number rem
                             (The_Source.Line'Last - 1);
      end if;
      if Ada.Text_Io.Is_Open(Listing_File) and then
             The_Source.Source_Type = Ascii_File then
        Put_Message (Listing_File);
--      else
--        Put_Message (Ada.Text_Io.Standard_Output);
      end if;
      Line_Message_Number := Line_Message_Number + 1;
      if Treat_As = Error then
        Current_Symbol.Is_Correct := False;
      end if;
    elsif Ada.Text_Io.Is_Open(Listing_File) then
      Ada.Text_Io.Put_Line (Listing_File, Info);
--    else
--      Ada.Text_Io.Put_Line (Info);
    end if;
  end Print_Message;
  -----------------------------------------------------------------------------
  function Source_Number return Natural is
  begin
    return Source_Counter;
  end Source_Number;
  -----------------------------------------------------------------------------
  function End_Of_Source return Boolean is
  begin
    return Current_Symbol.Class = Separator and then
           Current_Symbol.Key   = End_Of_Source;
  end End_Of_Source;
  -----------------------------------------------------------------------------
  function Source_Type (Source_Number : in Natural := 0) return Source_Class is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Source_Type;
    else
      return Source_Info(Source_Number).Source_Type;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error: No File is open");
        raise No_Open_File_Error;
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Source_Type;
  -----------------------------------------------------------------------------
  use type Ada.Strings.Unbounded.String_Access;
  -----------------------------------------------------------------------------
  function Line (Source_Number : in Natural := 0) return String is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Line(1 .. The_Source.Line_Length - 1);
    else
      return Source_Info(Source_Number).Line
            (1 .. Source_Info(Source_Number).Line_Length - 1);
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error: No File is open");
        raise No_Open_File_Error;
      elsif Source_Info(Source_Number).Line = null then
        return "";
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Line;
  -----------------------------------------------------------------------------
  function File_Name (Source_Number : in Natural := 0) return String is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.File_Name.all;
    else
      return Source_Info(Source_Number).File_Name.all;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error: No File is open");
        raise No_Open_File_Error;
      elsif Source_Info(Source_Number).File_Name = null then
        return "";
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end File_Name;
  -----------------------------------------------------------------------------
  function Line_Number (Source_Number : in Natural := 0) return Natural is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Line_Number;
    else
      return Source_Info(Source_Number).Line_Number;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error, No File is open");
        raise No_Open_File_Error;
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Line_Number;
  -----------------------------------------------------------------------------
  function Column_Number (Source_Number : in Natural := 0) return Natural is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Column_Number;
    else
      return Source_Info(Source_Number).Column_Number;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error: No File is open");
        raise No_Open_File_Error;
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Column_Number;
  -----------------------------------------------------------------------------
  function Last_Line (Source_Number : in Natural := 0) return String is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Last_Line(1 .. The_Source.Last_Line_Length - 1);
    else
      return Source_Info(Source_Number).Last_Line
            (1 .. Source_Info(Source_Number).Last_Line_Length - 1);
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error: No File is open");
        raise No_Open_File_Error;
      elsif Source_Info(Source_Number).Last_Line = null then
        return "";
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Last_Line;
  -----------------------------------------------------------------------------
  function Last_Line_Number (Source_Number : in Natural := 0) return Natural is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Last_Line_Number;
    else
      return Source_Info(Source_Number).Last_Line_Number;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error, No File is open");
        raise No_Open_File_Error;
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Last_Line_Number;
  -----------------------------------------------------------------------------
  function Symbol_Counter (Source_Number : in Natural := 0) return Natural is
  begin
    if Source_Number = 0 or else Source_Number = Source_Counter then
      return The_Source.Symbol_Counter;
    else
      return Source_Info(Source_Number).Symbol_Counter;
    end if;
  exception
    when Constraint_Error =>
      if Source_Number not in Source_Info'range then
        Print_Message ("Semantic-error, No File is open");
        raise No_Open_File_Error;
      else
        Print_Message ("Semantic-error: Unexpected CONSTRAINT_ERROR");
        raise;
      end if;
  end Symbol_Counter;
  -----------------------------------------------------------------------------
  function Listing_Line_Number return Positive is
  begin
    return The_Listing_Line_Number;
  end Listing_Line_Number;
  -----------------------------------------------------------------------------
  Line_Number_Difference  : Natural  := 0;
  The_Current_Line_Number : Positive := 1;
  -----------------------------------------------------------------------------
  procedure Save_Line_Number_Difference is
  begin
    if Ada.Text_Io.Is_Open (Listing_File) then
      Line_Number_Difference := Scanner.Listing_Line_Number -
                                Positive(Ada.Text_Io.Line(Listing_File));
    else
      Line_Number_Difference := 0;
    end if;
  end Save_Line_Number_Difference;
  -----------------------------------------------------------------------------
  procedure Save_Current_Line_Number is
  begin
    if Ada.Text_Io.Is_Open (Listing_File) then
      The_Current_Line_Number := Line_Number_Difference +
                                 Positive(Ada.Text_Io.Line(Listing_File)) + 20;
    else
      The_Current_Line_Number := 1;
    end if;
  end Save_Current_Line_Number;
  -----------------------------------------------------------------------------
  function Current_Line_Number return Positive is
  begin
    return The_Current_Line_Number;
  end Current_Line_Number;
  -----------------------------------------------------------------------------
end Utilities.Scanner;
