--                              -*- Mode: Ada -*-
-- Filename        : utilities-string_vectors.adb
-- Description     : String vector utilities
-- Author          : J. Schroeer
-- Created On      : Mon Apr  7 09:55:10 1997
-- Last Modified By:
-- Last Modified On: Fri Jul  4 01:21:54 1997
-- Update Count    : 8
-- Status          :

with Ada.Characters.Handling;
with Ada.Unchecked_Deallocation;
with Utilities.String_Utilities;

package body Utilities.String_Vectors is

  use type Ada.Strings.Unbounded.String_Access;
  -----------------------------------------------------------------------------
  function "=" (Left, Right : in String_Vector) return Boolean is
    Offset : constant Integer := Right'First - Left'First;
  begin
    if Left'Length /= Right'Length then
      return False;
    else
      for I in Left'range loop
        if Left(I) = null then
          if Right(I + Offset) /= null then
            return False;
          end if;
        elsif Right(I + Offset) = null then
          return False;
        elsif Left(I).all /= Right(I + Offset).all then
          return False;
        end if;
      end loop;
      return True;
    end if;
  end "=";
  -----------------------------------------------------------------------------
  function Copy (The_Vector : in String_Vector) return String_Vector is
    Vector : String_Vector (The_Vector'range);
  begin
    for I in The_Vector'range loop
      if The_Vector(I) /= null then
        Vector(I) := new String'(The_Vector(I).all);
      else
        Vector(I) := new String'("");
      end if;
    end loop;
    return Vector;
  end Copy;
  -----------------------------------------------------------------------------
  function Copy (The_Vector : in String_Vector_Access)
           return String_Vector_Access is
  begin
    if The_Vector /= null then
      return new String_Vector'(Copy(The_Vector.all));
    else
      return null;
    end if;
  end Copy;
  -----------------------------------------------------------------------------
  function First_Index (The_Element     : in String;
                        The_Vector      : in String_Vector;
                        Case_Sensitive,
                        Forward         : in Boolean := True) return Integer is
  begin
    if Case_Sensitive then
      if Forward then
        for I in The_Vector'range loop
          if The_Element = The_Vector(I).all then
            return I;
          end if;
        end loop;
      else
        for I in reverse The_Vector'range loop
          if The_Element = The_Vector(I).all then
            return I;
          end if;
        end loop;
      end if;
    else
      declare
        Local_String : constant String(The_Element'range) :=
                       Ada.Characters.Handling.To_Upper(The_Element);
      begin
        if Forward then
          for I in The_Vector'range loop
            if Local_String = Ada.Characters.Handling.To_Upper
                              (The_Vector(I).all) then
              return I;
            end if;
          end loop;
        else
          for I in reverse The_Vector'range loop
            if Local_String = Ada.Characters.Handling.To_Upper
                              (The_Vector(I).all) then
              return I;
            end if;
          end loop;
        end if;
      end;
    end if;
    return The_Vector'First - 1;
  end First_Index;
  -----------------------------------------------------------------------------
  function First_Index(The_Element    : in Ada.Strings.Unbounded.String_Access;
                       The_Vector     : in String_Vector;
                       Case_Sensitive,
                       Forward        : in Boolean := True) return Integer is
  begin
    return First_Index(The_Element.all, The_Vector, Case_Sensitive, Forward);
  end First_Index;
  -----------------------------------------------------------------------------
  function Split(The_String : in String;
                 Separator  : in Character := Ascii.Nul) return String_Vector is
             
    Separators : constant Utilities.String_Utilities.Index_Array :=
                          Utilities.String_Utilities.Indices_Of
                                  (The_Character => Separator,
                                   In_The_String => The_String);
  begin
    if Separators'Length = 0 then
      declare
        The_Strings : String_Vector(1 .. 1);
      begin
        The_Strings(1) := new String'(The_String);
        return The_Strings;                    
      end;
    elsif Separators(Separators'Last) = The_String'Last then
      declare
        The_Strings : String_Vector(Separators'range);
      begin
        The_Strings(The_Strings'First) := new String'
          (The_String(The_String'First .. Separators(Separators'First) - 1));
        for I in The_Strings'First + 1 .. The_Strings'Last loop
          The_Strings(I) := new String'
            (The_String(Separators(I - 1) + 1 .. Separators(I) - 1));
        end loop;
        return The_Strings;                    
      end;
    else
      declare
        The_Strings : String_Vector(Separators'First .. Separators'Last + 1);
      begin
        The_Strings(The_Strings'First) := new String'
          (The_String(The_String'First .. Separators(Separators'First) - 1));
        for I in The_Strings'First + 1 .. The_Strings'Last - 1 loop
          The_Strings(I) := new String'
            (The_String(Separators(I - 1) + 1 .. Separators(I) - 1));
        end loop;
        The_Strings(The_Strings'Last) := new String'
          (The_String(Separators(Separators'Last) + 1 .. The_String'Last));
        return The_Strings;                    
      end;
    end if;
  end Split;
  ---------------------------------------------------------------------------
  function Assemble(The_Strings : in String_Vector;
                    Separator   : in Character := Ascii.Nul) return String is
    
    Temp  : String(1 .. Total_Length(The_Strings) + The_Strings'Length);                  
    First : Natural := Temp'First;
    Last  : Natural := First;
  begin
    for I in The_Strings'range loop
      Last := First + The_Strings(I)'Length;
      Temp(First .. Last - 1) := The_Strings(I).all;
      Temp(Last) := Separator;
      First := Last + 1;
    end loop;
    return Temp;
  end Assemble;
  ---------------------------------------------------------------------------
  function Max_Length(The_Vector : in String_Vector) return Natural is

    Max : Natural := 0;
  begin
    for I in The_Vector'range loop
      declare
        Element : Ada.Strings.Unbounded.String_Access renames The_Vector(I);
      begin
        if Element /= null and then Element'Length > Max then
          Max := Element'Length;
        end if;
      end;
    end loop;
    return Max;
  end Max_Length;
  -----------------------------------------------------------------------------
  function Total_Length(The_Vector : in String_Vector) return Natural is

    Length : Natural := 0;
  begin
    for I in The_Vector'range loop
      declare
        Element : Ada.Strings.Unbounded.String_Access renames The_Vector(I);
      begin
        if Element /= null then
          Length := Length + Element'Length;
        end if;
      end;
    end loop;
    return Length;
  end Total_Length;
  -----------------------------------------------------------------------------
  procedure Allocate (Item  : in out String_Vector_Access;
                      Last  : in     Integer;
                      First : in     Integer := 1) is
  begin
    if Item = null then
      Item := new String_Vector (First .. Last);
    elsif Item'Last /= Last or Item'First /= First then
      Deallocate (Item);
      Item := new String_Vector (First .. Last);
    end if;
  end Allocate;
  -----------------------------------------------------------------------------
  procedure Deallocate (Item  : in out String_Vector) is
  begin
    for I in Item'range loop
      Ada.Strings.Unbounded.Free(Item(I));
    end loop;
  end Deallocate;
  -----------------------------------------------------------------------------
  procedure Deallocate (Item : in out String_Vector_Access;
                        Deep : in     Boolean := True) is

    procedure Deallocate_Vector is new Ada.Unchecked_Deallocation
             (Object => String_Vector, Name => String_Vector_Access);
  begin
    if Item /= null then
      if Deep then
        Deallocate(Item.all);
      end if;
      Deallocate_Vector(Item);
    end if;
  end Deallocate;
  -----------------------------------------------------------------------------
end Utilities.String_Vectors;
