--                              -*- Mode: Ada -*-
-- Filename        : utilities-string_vectors.ads
-- Description     : String vector utilities
-- Author          : J. Schroeer
-- Created On      : Mon Apr  7 09:47:43 1997
-- Last Modified By:
-- Last Modified On: Fri Jul  4 01:22:47 1997
-- Update Count    : 8
-- Status          :

with Ada.Strings.Unbounded;

package Utilities.String_Vectors is

  type String_Vector is array (Integer range <>) of
       Ada.Strings.Unbounded.String_Access;

  type String_Vector_Access is access all String_Vector;

  function "=" (Left, Right : in String_Vector) return Boolean;

  -- If left'length /= right'length, false is returned.
  -- Left'first may be /= right'first.
  -- Performs a deep comparison,
  -- i.e. left(i).all = right(i + offset).all are compared.

  function Copy (The_Vector : in String_Vector) return String_Vector;

  -- Performs a deep copy, the string_access-components of the_vector are
  -- copied. If a component of the_vector is null, a new empty string is
  -- allocated in the copy ( new string'("") ).

  function Copy (The_Vector : in String_Vector_Access)
           return String_Vector_Access;

  -- If the_vector /= null then calls copy above for a new allocated
  -- string_vector, else null is returned.

  function First_Index(The_Element    : in     String;
                       The_Vector     : in     String_Vector;
                       Case_Sensitive,
                       Forward        : in     Boolean := True) return Integer;

  -- Returns the index i of the first element in the_vector for which
  -- the_element = the_vector(i).all.
  -- If no component matches, the_vector'first - 1 is returned.
  --
  -- Case is not regarded if case_sensitive = false.
  -- If forward = false searches the components in the_vector in reverse order.
  -- If the_vector(i) = null, the_vector(i).all = "" is assumed,
  -- i.e. comparison with an empty string is performed.

  function First_Index(The_Element    : in Ada.Strings.Unbounded.String_Access;
                       The_Vector     : in String_Vector;
                       Case_Sensitive,
                       Forward        : in Boolean := True) return Integer;

  -- Calls first_index above.
  -- If the_element = null, the_element.all = "" is assumed,
  -- i.e. comparison with an empty string is performed.

  function Split(The_String : in String;
                 Separator  : in Character := Ascii.Nul) return String_Vector;
                 
  -- Returns a string_vector, where each element is a slice of the string  
  -- that ends with separator. The separator character is not included.
  -- If the_string does not contain separator the retuned string_vector has 
  -- the length 1 and the element contains the whole string.
  -- If the last character in the_string is not separator, that character is
  -- appended.                               
                 
  function Assemble(The_Strings : in String_Vector;
                    Separator   : in Character := Ascii.Nul) return String;
  
  -- Returns a string assembled from all elements in the_strings with character
  -- separator added between the slices formed by the elements.
                    
  function Max_Length  (The_Vector : in String_Vector) return Natural;

  -- Returns length of longest string in the_vector.

  function Total_Length(The_Vector : in String_Vector) return Natural;

  -- Returns length of all strings in the_vector.

  procedure Allocate   (Item  : in out String_Vector_Access;
                        Last  : in     Integer;
                        First : in     Integer := 1);

  -- If item = null it is allocated with the given range first .. last.
  -- If item /= null and its range does not match the given range,
  --   it is deallocated and then allocated with the given range.
  -- If item /= null and its range matches the given range, nothing is done.

  procedure Deallocate (Item  : in out String_Vector);

  -- Deallocates the elements of item.
                        
  procedure Deallocate (Item  : in out String_Vector_Access;
                        Deep  : in     Boolean := True);

  -- If deep = true, the components item(i) are deallocated too,
  -- else only item is deallocated.

end Utilities.String_Vectors;
