------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Rational_Arithmetics;
use  Rational_Arithmetics;

with Constrained_Checked_SI.Generic_Units    , Constrained_Checked_SI.Generic_Text_IO,
     Constrained_Unchecked_SI.Generic_Units  , Constrained_Unchecked_SI.Generic_Text_IO;
with Unconstrained_Checked_SI.Generic_Units  , Unconstrained_Checked_SI.Generic_Text_IO,
     Unconstrained_Unchecked_SI.Generic_Units, Unconstrained_Unchecked_SI.Generic_Text_IO;

with Ada.Calendar;
use  Ada.Calendar;
with Ada.Text_IO;
use  Ada.Text_IO;

procedure Measure is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      27 September 2002
  --====================================================================
  -- Measurements of relative speeds.
  --                             Gnat 3.14p Win 98
  -- Constrained   Checked  :  90.343699190   15.8
  -- Constrained   Unchecked:   6.849125865    1.2
  -- Unconstrained Checked  :  56.190207680    9.8
  -- Unconstrained Unchecked:   5.731462143    1.0
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  30.07.2002 with work-around for Gnat 3.14p bug
  --  C.G.    1.1  27.09.2002 Gnat work-around removed for 3.16w
  --                          (there is one work-around left)
  --====================================================================

  package Duration_Text_IO is new Ada.Text_IO.Fixed_IO (Duration);
  use Duration_Text_IO;

  package Constrained is

    package Checked_SI       is new Constrained_Checked_SI (Float);
    package Checked_SI_Units is new Checked_SI.Generic_Units;

    subtype Checked is Checked_SI.Item;

    function Schottky_Langmuir return Checked;

    package Unchecked_SI       is new Constrained_Unchecked_SI (Float);
    package Unchecked_SI_Units is new Unchecked_SI.Generic_Units;

    subtype Unchecked is Unchecked_SI.Item;

    function Schottky_Langmuir return Unchecked;

  end Constrained;

  package body Constrained is

    function Schottky_Langmuir return Checked is
      use Checked_SI, Checked_SI_Units;
      Eps0: constant Item := 8.854 * Coulomb / (Volt * Meter);
      Electron_Mass  : constant Item := 9.10814E-28 * Gram;
      Electron_Charge: constant Item := 1.6020E-19 * Coulomb;
      Voltage : constant Item := 50.0 * Volt;
      Distance: constant Item := 20.0 * centi * Meter;
    begin
      return (4.0/9.0) * Eps0 * Sqrt (2.0 * Electron_Charge / Electron_Mass) *
             Voltage**(3/2) / Distance**2;
    end Schottky_Langmuir;

    function Schottky_Langmuir return Unchecked is
      use Unchecked_SI, Unchecked_SI_Units;
      Eps0: constant Item := 8.854 * Coulomb / (Volt * Meter);
      Electron_Mass  : constant Item := 9.10814E-28 * Gram;
      Electron_Charge: constant Item := 1.6020E-19 * Coulomb;
      Voltage : constant Item := 50.0 * Volt;
      Distance: constant Item := 20.0 * centi * Meter;
    begin
      return (4.0/9.0) * Eps0 * Sqrt (2.0 * Electron_Charge / Electron_Mass) *
             Voltage**(3/2) / Distance**2;
    end Schottky_Langmuir;

  end Constrained;

  package Unconstrained is

    package Checked_SI is new Unconstrained_Checked_SI (Float);
    package Checked_SI_Units is new Checked_SI.Generic_Units;

    subtype Checked is Checked_SI.Item;

    function Schottky_Langmuir return Checked;

    package Unchecked_SI is new Unconstrained_Unchecked_SI (Float);
    package Unchecked_SI_Units is new Unchecked_SI.Generic_Units;

    subtype Unchecked is Unchecked_SI.Item;

    function Schottky_Langmuir return Unchecked;

  end Unconstrained;

  package body Unconstrained is

    function Schottky_Langmuir return Checked is
      use Checked_SI, Checked_SI_Units;
      Eps0: constant Item := 8.854 * Coulomb / (Volt * Meter);
      Electron_Mass  : constant Item := 9.10814E-28 * Gram;
      Electron_Charge: constant Item := 1.6020E-19 * Coulomb;
      Voltage : constant Item := 50.0 * Volt;
      Distance: constant Item := 20.0 * centi * Meter;
    begin
      return (4.0/9.0) * Eps0 * Sqrt (2.0 * Electron_Charge / Electron_Mass) *
             Voltage**(3/2) / Distance**2;
    end Schottky_Langmuir;

    function Schottky_Langmuir return Unchecked is
      use Unchecked_SI, Unchecked_SI_Units;
      Eps0: constant Item := 8.854 * Coulomb / (Volt * Meter);
      Electron_Mass  : constant Item := 9.10814E-28 * Gram;
      Electron_Charge: constant Item := 1.6020E-19 * Coulomb;
      Voltage : constant Item := 50.0 * Volt;
      Distance: constant Item := 20.0 * centi * Meter;
    begin
      return (4.0/9.0) * Eps0 * Sqrt (2.0 * Electron_Charge / Electron_Mass) *
             Voltage**(3/2) / Distance**2;
    end Schottky_Langmuir;

  end Unconstrained;

  type Repetitions is range 1 .. 1_000_000;

  Constrained_Checked_Start    , Constrained_Checked_End    ,
  Constrained_Unchecked_Start  , Constrained_Unchecked_End  ,
  Unconstrained_Checked_Start  , Unconstrained_Checked_End  ,
  Unconstrained_Unchecked_Start, Unconstrained_Unchecked_End: Time;

begin

  declare
    use Constrained;
    use Checked_SI, Checked_SI_Units;
    Current_Density: -- Checked  (Gnat bug work-around)
                     Checked_SI.Item := Ampere / Meter_2;
  begin
    Constrained_Checked_Start := Clock;
    for R in Repetitions loop
      Current_Density := Schottky_Langmuir;
    end loop;
    Constrained_Checked_End := Clock;
  end;

  declare
    use Constrained;
    use Unchecked_SI, Unchecked_SI_Units;
    Current_Density: Unchecked := Ampere / Meter_2;
  begin
    Constrained_Unchecked_Start := Clock;
    for R in Repetitions loop
      Current_Density := Schottky_Langmuir;
    end loop;
    Constrained_Unchecked_End := Clock;
  end;

  declare
    use Unconstrained;
    use Checked_SI, Checked_SI_Units;
    Current_Density: Checked;
  begin
    Unconstrained_Checked_Start := Clock;
    for R in Repetitions loop
      Current_Density := Schottky_Langmuir;
    end loop;
    Unconstrained_Checked_End := Clock;
  end;

  declare
    use Unconstrained;
    use Unchecked_SI, Unchecked_SI_Units;
    Current_Density: Unchecked;
  begin
    Unconstrained_Unchecked_Start := Clock;
    for R in Repetitions loop
      Current_Density := Schottky_Langmuir;
    end loop;
    Unconstrained_Unchecked_End := Clock;
  end;

  Put ("Constrained_Checked    :");  Put (Constrained_Checked_End     - Constrained_Checked_Start)    ;  New_Line;
  Put ("Constrained_Unchecked  :");  Put (Constrained_Unchecked_End   - Constrained_Unchecked_Start)  ;  New_Line;
  Put ("Unconstrained_Checked  :");  Put (Unconstrained_Checked_End   - Unconstrained_Checked_Start)  ;  New_Line;
  Put ("Unconstrained_Unchecked:");  Put (Unconstrained_Unchecked_End - Unconstrained_Unchecked_Start);  New_Line;

end Measure;
