------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Text_IO;
use  Ada.Text_IO;

with Rational_Arithmetics;
use  Rational_Arithmetics;

--#c with Unconstrained_Checked_SI.Generic_Units,
--#c      Unconstrained_Checked_SI.Generic_Text_IO;
--#u with Unconstrained_Unchecked_SI.Generic_Units,
--#u      Unconstrained_Unchecked_SI.Generic_Text_IO;

procedure Physical_Computations_Unconstrained is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      27 September 2002
  --====================================================================
  -- An example use.
  --
  --    To get the   checked version, uncomment the "--#c" lines.
  --    To get the unchecked version, uncomment the "--#u" lines.
  --
  -- Airspeed is measured by a Pitot tube. Since the pressure in the
  -- tube depends on the air density (p = rho*v**2/2) and is gauged to
  -- the norm atmosphere, the speed has to be corrected by
  --
  --    SQRT (Gauge_Air_Density/Surrounding_Air_Density)
  --
  -- Speed of sound in norm atmosphere
  -- Schottky-Langmuir equation for anode (emission) current
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  25.07.2002 with Gnat 3.14p bug work-around
  --                          (unnecesary qualification needed)
  --  C.G.    1.1  27.09.2002 Gnat work-around removed for 3.16w
  --====================================================================

  --#c package SI is new Unconstrained_Checked_SI (Float);
  --#u package SI is new Unconstrained_Unchecked_SI (Float);
  package U  is new SI.Generic_Units;
  package IO is new SI.Generic_Text_IO (U);
  use SI, U, IO;

  F: constant := 5;  -- degrees of freedom for bar-like (o-o) air molecules

  Adiabatic_Exp: constant Float := Float (F + 2) /  Float (F);

  Hour: constant Item := 60.0**2 * Second;

  T_Norm  : constant Item :=  288.15   * Kelvin;               -- 15 deg Celsius
  P_Norm  : constant Item := 1013.25   * hecto * Pascal;       -- 760 Torr = 1013 hPa
  Rho_Norm: constant Item :=    1.2255 * Kilogram / Meter**3;  -- kg/m**3

  Gas_Constant: constant Item :=
         287.05287 * Newton * Meter / (Kelvin * Kilogram);  -- N*m/(K*kg)

  Pressure: constant Item := 0.9 * P_Norm;
  OAT     : constant Item := 273.15 * Kelvin;  -- outside air temperature
  Rho     : constant Item := Pressure / (Gas_Constant * OAT);

  -- air density ratio (Rho/Rho_Norm)
  Sigma, Sqrt_Sigma: Item;

  IAS: constant Item := 200.0 * Kilo * Meter / Hour;  -- indicated and
  TAS: Item;                                          -- true air speed

  Sound: Item;

  -- Schottky-Langmuir equation

  Eps0: constant Item := 8.854 * Coulomb / (Volt * Meter);

  Electron_Mass  : constant Item := 9.10814E-28 * Gram;
  Electron_Charge: constant Item := 1.6020E-19 * Coulomb;

  Voltage : constant Item := 50.0 * Volt;
  Distance: constant Item := 20.0 * centi * Meter;

  Current_Density: Item;

begin

  Sigma      := Rho / Rho_Norm;
  Sqrt_Sigma := SQRT (Sigma);

  TAS := IAS / Sqrt_Sigma;

  Sound := Sqrt (Adiabatic_Exp * P_Norm / Rho_Norm);

  Current_Density := (4.0/9.0) * Eps0 * Sqrt (2.0 * Electron_Charge / Electron_Mass) *
                     Voltage**(3/2) / Distance**2;

  Put (Gas_Constant);  New_Line;

  Put (T_Norm);    Put (OAT);         New_Line;
  Put (P_Norm);    Put (Pressure);    New_Line;
  Put (Rho_Norm);  Put (Rho);         New_Line;
  Put (Sigma);     Put (Sqrt_Sigma);  New_Line;
  Put (IAS);       Put (TAS);         New_Line;

  Put (Sound);  New_Line;

  Put (Current_Density);  New_Line;

end Physical_Computations_Unconstrained;
