------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

generic

  Min, Max: Item;

package Constrained_Checked_SI.Generic_Item_Subtype is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      15 November 2002
  --====================================================================
  -- A "subtype" of Item with the range constraint Min .. Max is
  -- defined. Its dimension is fixed to the one given by the generic
  -- parameters.
  --
  -- A type conversion of an unconstrained item to a constrained item
  -- is done with the function Constrain. A dimension check and a range
  -- check are performed and Unit_Error respectively Constraint_Error
  -- is propagated if the check fails.
  -- A type conversion of a constrained item to an unconstrained item
  -- is done with the unary plus operator.
  --
  -- A set of operations is defined so that a constrained item can be
  -- used in almost all places where an unconstrained item can be used
  -- so that a type conversion should rarely be necessary.
  --
  -- Example:
  -- package Car_Speed is new Generic_Item_Subtype
  --                            (Min  =>   0.0 * Kilo*Meter/Hour,
  --                             Max  => 160.0 * Kilo*Meter/Hour);
  -- Discussion:
  -- A function identifier has been used rather than an operator symbol
  -- for a type conversion to Constrained in order to avoid overload
  -- resolution conflicts.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  09.11.2002
  --  C.G.    1.1  15.11.2002 Use Real'Base
  --====================================================================

  pragma Elaborate_Body;

  type Constrained is private;  -- "subtype" of Item

  function Constrain (X: Item       ) return Constrained;
  function Value     (X: Constrained) return Real;

  -- Operations

  function "+" (Right: Constrained) return Item;
  function "-" (Right: Constrained) return Item;

  function "+" (Left: Item       ; Right: Constrained) return Item;
  function "+" (Left: Constrained; Right: Item       ) return Item;
  function "-" (Left: Item       ; Right: Constrained) return Item;
  function "-" (Left: Constrained; Right: Item       ) return Item;

  function "*" (Left: Item       ; Right: Constrained) return Item;
  function "*" (Left: Constrained; Right: Item       ) return Item;
  function "/" (Left: Item       ; Right: Constrained) return Item;
  function "/" (Left: Constrained; Right: Item       ) return Item;

  function "*" (Left: Constrained     ; Right: Real'Base  ) return Item;
  function "*" (Left: Real'Base       ; Right: Constrained) return Item;
  function "/" (Left: Constrained     ; Right: Real'Base  ) return Item;
  function "/" (Left: Real'Base       ; Right: Constrained) return Item;

  function "**" (Base: Constrained; Exponent: Whole    ) return Item;
  function "**" (Base: Constrained; Exponent: Rational ) return Item;
  function "**" (Base: Constrained; Exponent: Real'Base) return Item;

private

  pragma Inline (Constrain, Value,
                 "+", "-", "*", "/", "**");

  type Constrained is record
    Value: Item (Min.m_N, Min.kg_N, Min.s_N, Min.A_N, Min.K_N, Min.cd_N, Min.mol_N,
                 Min.m_D, Min.kg_D, Min.s_D, Min.A_D, Min.K_D, Min.cd_D, Min.mol_D);
  end record;

end Constrained_Checked_SI.Generic_Item_Subtype;