------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

generic
package Constrained_Checked_SI.Generic_Temperatures is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      8 October 2002
  --====================================================================
  -- Kelvin and Celsius use the same unit interval.
  -- So do Rankine and Fahrenheit.
  -- Kelvin and Rankine are thermodynamic temperatures (they use the
  -- absolute zero as zero points).
  -- Celsius and Raumur use the same fundamental distance between the
  -- ice point and the vapour point, but divide it into 100 resp. 80
  -- unit intervals.
  --
  -- Only Celsius and Fahrenheit are in use nowadays.
  --
  -- Fahrenheit:  0F = -17.78C   32F = 0C     xC = (9/5*x + 32)F
  -- Rankine:     0R = 0 K                       x K = (9/5)*xR
  -- Celsius:     0C = 273.15 K                  x K = (x - 273.15)C
  -- Raumur:     0C = 0R        80R = 100C   xR = (5/4)*xC
  --
  -- Note: Rankine and Raumur use the same symbol R in literature
  --       (formerly, Rankine was also denoted by Rank resp. deg R).
  --       Since Raumur has long been out of use, this does not lead
  --       to ambiguities.
  --
  -- The conversion functions from Kelvin raise Unit_Error if Item does
  -- not have the unit Kelvin.
  -- All operators raise Unit_Error if the Left and Right arguments do
  -- not have the same scale.
  --
  -- Discussion:
  -- Temperatures measured in non-thermodynamic units have a very
  -- limited range of applications. They basically serve only as a
  -- scale, but cannot be used in physical equations.
  -- If defined naively, nonsensical equations can be written. Also
  -- their conversion to Kelvin has to take into account whether scaled
  -- values or differences are converted.
  -- Only the operations defined here seem to make sense to me.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  08.10.2002
  --====================================================================

  type Scale_Name is (Celsius, Fahrenheit, Rankine, Reaumur);

  type Temperature            (Scale: Scale_Name) is private;
  type Temperature_Difference (Scale: Scale_Name) is private;

  -- Constructors

  function to_Celsius    (T: Real) return Temperature;
  function to_Celsius    (T: Real) return Temperature_Difference;
  function to_Fahrenheit (T: Real) return Temperature;
  function to_Fahrenheit (T: Real) return Temperature_Difference;
  function to_Rankine    (T: Real) return Temperature;
  function to_Rankine    (T: Real) return Temperature_Difference;
  function to_Reaumur    (T: Real) return Temperature;
  function to_Reaumur    (T: Real) return Temperature_Difference;

  -- Conversions

  function to_Kelvin (T: Temperature           ) return Item;
  function to_Kelvin (T: Temperature_Difference) return Item;

  function to_Celsius    (T: Item) return Temperature;
  function to_Celsius    (T: Item) return Temperature_Difference;
  function to_Fahrenheit (T: Item) return Temperature;
  function to_Fahrenheit (T: Item) return Temperature_Difference;
  function to_Rankine    (T: Item) return Temperature;
  function to_Rankine    (T: Item) return Temperature_Difference;
  function to_Reaumur    (T: Item) return Temperature;
  function to_Reaumur    (T: Item) return Temperature_Difference;

  -- Rescaling

  function to_Celsius    (T: Temperature           ) return Temperature;
  function to_Celsius    (T: Temperature_Difference) return Temperature_Difference;
  function to_Fahrenheit (T: Temperature           ) return Temperature;
  function to_Fahrenheit (T: Temperature_Difference) return Temperature_Difference;
  function to_Rankine    (T: Temperature           ) return Temperature;
  function to_Rankine    (T: Temperature_Difference) return Temperature_Difference;
  function to_Reaumur    (T: Temperature           ) return Temperature;
  function to_Reaumur    (T: Temperature_Difference) return Temperature_Difference;

  -- Numeric value

  function Value (T: Temperature           ) return Real;
  function Value (T: Temperature_Difference) return Real;

  -- Operators

  function "+" (Left: Temperature           ; Right: Temperature_Difference) return Temperature;
  function "+" (Left: Temperature_Difference; Right: Temperature           ) return Temperature;
  function "-" (Left: Temperature           ; Right: Temperature_Difference) return Temperature;
  function "-" (Left: Temperature           ; Right: Temperature           ) return Temperature_Difference;

  function "+" (Left, Right: Temperature_Difference) return Temperature_Difference;
  function "-" (Left, Right: Temperature_Difference) return Temperature_Difference;

  function "<"  (Left, Right: Temperature           ) return Boolean;
  function "<"  (Left, Right: Temperature_Difference) return Boolean;
  function "<=" (Left, Right: Temperature           ) return Boolean;
  function "<=" (Left, Right: Temperature_Difference) return Boolean;
  function ">=" (Left, Right: Temperature           ) return Boolean;
  function ">=" (Left, Right: Temperature_Difference) return Boolean;
  function ">"  (Left, Right: Temperature           ) return Boolean;
  function ">"  (Left, Right: Temperature_Difference) return Boolean;

private

  pragma Inline (to_Celsius, to_Fahrenheit, to_Rankine, to_Reaumur, to_Kelvin,
                 Value,
                 "+", "-",
                 "<", "<=", ">=", ">");

  type Temperature (Scale: Scale_Name) is record
    Value: Real;
  end record;

  type Temperature_Difference (Scale: Scale_Name) is record
    Value: Real;
  end record;

end Constrained_Checked_SI.Generic_Temperatures;
