------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

package body Constrained_Checked_SI.Generic_Vector_Space is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      17 November 2002
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  17.11.2002 Coded
  --====================================================================

  use Vector_Space, SI;

  function Dimension_of (X: Vector) return Dimension is
  begin
    return (m   => X.  m_N/X.  m_D,
            kg  => X. kg_N/X. kg_D,
            s   => X.  s_N/X.  s_D,
            A   => X.  A_N/X.  A_D,
            K   => X.  K_N/X.  K_D,
            cd  => X. cd_N/X. cd_D,
            mol => X.mol_N/X.mol_D);
  end Dimension_of;

  function Dimension_of (X: Matrix) return Dimension is
  begin
    return (m   => X.  m_N/X.  m_D,
            kg  => X. kg_N/X. kg_D,
            s   => X.  s_N/X.  s_D,
            A   => X.  A_N/X.  A_D,
            K   => X.  K_N/X.  K_D,
            cd  => X. cd_N/X. cd_D,
            mol => X.mol_N/X.mol_D);
  end Dimension_of;

  function Transform (X: SI.Item) return Item is
    Dim: Dimension renames Dimension_of (X);
  begin
    return (m_N   => Numerator   (Dim.m),
            m_D   => Denominator (Dim.m),
            kg_N  => Numerator   (Dim.kg),
            kg_D  => Denominator (Dim.kg),
            s_N   => Numerator   (Dim.s),
            s_D   => Denominator (Dim.s),
            A_N   => Numerator   (Dim.A),
            A_D   => Denominator (Dim.A),
            K_N   => Numerator   (Dim.K),
            K_D   => Denominator (Dim.K),
            cd_N  => Numerator   (Dim.cd),
            cd_D  => Denominator (Dim.cd),
            mol_N => Numerator   (Dim.mol),
            mol_D => Denominator (Dim.mol),
            Inner => X);
  end Transform;

  function Transform (X: Vector_Space.Vector) return Vector is
    Dim: Dimension renames Dimension_of (X (1));
  begin
    return (m_N   => Numerator   (Dim.m),
            m_D   => Denominator (Dim.m),
            kg_N  => Numerator   (Dim.kg),
            kg_D  => Denominator (Dim.kg),
            s_N   => Numerator   (Dim.s),
            s_D   => Denominator (Dim.s),
            A_N   => Numerator   (Dim.A),
            A_D   => Denominator (Dim.A),
            K_N   => Numerator   (Dim.K),
            K_D   => Denominator (Dim.K),
            cd_N  => Numerator   (Dim.cd),
            cd_D  => Denominator (Dim.cd),
            mol_N => Numerator   (Dim.mol),
            mol_D => Denominator (Dim.mol),
            Inner => X);
  end Transform;

  function Transform (X: Vector_Space.Matrix) return Matrix is
    Dim: Dimension renames Dimension_of (X (1, 1));
  begin
    return (m_N   => Numerator   (Dim.m),
            m_D   => Denominator (Dim.m),
            kg_N  => Numerator   (Dim.kg),
            kg_D  => Denominator (Dim.kg),
            s_N   => Numerator   (Dim.s),
            s_D   => Denominator (Dim.s),
            A_N   => Numerator   (Dim.A),
            A_D   => Denominator (Dim.A),
            K_N   => Numerator   (Dim.K),
            K_D   => Denominator (Dim.K),
            cd_N  => Numerator   (Dim.cd),
            cd_D  => Denominator (Dim.cd),
            mol_N => Numerator   (Dim.mol),
            mol_D => Denominator (Dim.mol),
            Inner => X);
  end Transform;

  -----------------------------------------------------------------------------

  function "*" (Left: Proto_Vector; Right: Item) return Vector is
    V: constant Vector_Space.Vector := (Left (1) * SI.One, Left (2) * SI.One, Left (3) * SI.One);
  begin
    return Transform (V * Right.Inner);
  end "*";

  function "*" (Left: Item; Right: Proto_Vector) return Vector is
  begin
    return Right * Left;
  end "*";

  function Normalize (X: Proto_Vector) return Vector is
    V: constant Vector := X * One;
  begin
    return V / abs V;
  end Normalize;

  function Get (V: Vector; A: Axis) return Item is
  begin
    return Transform (V.Inner (Vector_Space.Axis (A)));
  end Get;

  procedure Set (V: in out Vector; A: in Axis; It: in Item) is
  begin
    if Dimension_of (V) /= Dimension_of (It.Inner) then
      raise Unit_Error;
    end if;
    V.Inner (Vector_Space.Axis (A)) := It.Inner;
  end Set;

  function "abs" (Right: Vector) return Item is
  begin
    return Transform (abs Right.Inner);
  end "abs";

  function "+" (Right: Vector) return Vector is
  begin
    return Right;
  end "+";

  function "-" (Right: Vector) return Vector is
    X: Vector := Right;
  begin
    X.Inner := -X.Inner;
    return X;
  end "-";

  function "+" (Left, Right: Vector) return Vector is
    X: Vector := Right;
  begin
    if Dimension_of (Left) /= Dimension_of (Right) then
      raise Unit_Error;
    end if;
    X.Inner := Left.Inner + Right.Inner;
    return X;
  end "+";

  function "-" (Left, Right: Vector) return Vector is
    X: Vector := Right;
  begin
    if Dimension_of (Left) /= Dimension_of (Right) then
      raise Unit_Error;
    end if;
    X.Inner := Left.Inner - Right.Inner;
    return X;
  end "-";

  function "*" (Left: Real'Base; Right: Vector) return Vector is
    X: Vector := Right;
  begin
    X.Inner := Left * Right.Inner;
    return X;
  end "*";

  function "*" (Left: Item; Right: Vector) return Vector is
  begin
    return Transform (Left.Inner * Right.Inner);
  end "*";

  function "*" (Left: Vector; Right: Real'Base) return Vector is
  begin
    return Right * Left;
  end "*";

  function "*" (Left: Vector; Right: Item) return Vector is
  begin
    return Right * Left;
  end "*";

  function "/" (Left: Vector; Right: Real'Base) return Vector is
    X: Vector := Left;
  begin
    X.Inner := X.Inner / Right;
    return X;
  end "/";

  function "/" (Left: Vector; Right: Item) return Vector is
  begin
    return Transform (Left.Inner / Right.Inner);
  end "/";

  function Inner (Left, Right: Vector) return Item is
  begin
    return Transform (Inner (Left.Inner, Right.Inner));
  end Inner;

  function Cross (Left, Right: Vector) return Vector is
  begin
    return Transform (Cross (Left.Inner, Right.Inner));
  end Cross;

  function "*" (Left: Proto_Matrix; Right: Item) return Matrix is
    M: constant Vector_Space.Matrix := ((Left (1, 1) * SI.One, Left (1, 2) * SI.One, Left (1, 3) * SI.One),
                                        (Left (2, 1) * SI.One, Left (2, 2) * SI.One, Left (2, 3) * SI.One),
                                        (Left (3, 1) * SI.One, Left (3, 2) * SI.One, Left (3, 3) * SI.One));
  begin
    return Transform (M * Right.Inner);
  end "*";

  function "*" (Left: Item; Right: Proto_Matrix) return Matrix is
  begin
    return Right * Left;
  end "*";

  function Get (M: Matrix; A, B: Axis) return Item is
  begin
    return Transform (M.Inner (Vector_Space.Axis (A), Vector_Space.Axis (B)));
  end Get;

  procedure Set (M: in out Matrix; A, B: in Axis; It: in Item) is
  begin
    if Dimension_of (M) /= Dimension_of (It.Inner) then
      raise Unit_Error;
    end if;
    M.Inner (Vector_Space.Axis (A), Vector_Space.Axis (B)) := It.Inner;
  end Set;

  function "+" (Right: Matrix) return Matrix is
  begin
    return Right;
  end "+";

  function "-" (Right: Matrix) return Matrix is
    X: Matrix := Right;
  begin
    X.Inner := -X.Inner;
    return X;
  end "-";

   function "+" (Left, Right: Matrix) return Matrix is
    X: Matrix := Right;
  begin
    if Dimension_of (Left) /= Dimension_of (Right) then
      raise Unit_Error;
    end if;
    X.Inner := Left.Inner + Right.Inner;
    return X;
  end "+";

  function "-" (Left, Right: Matrix) return Matrix is
    X: Matrix := Right;
  begin
    if Dimension_of (Left) /= Dimension_of (Right) then
      raise Unit_Error;
    end if;
    X.Inner := Left.Inner - Right.Inner;
    return X;
  end "-";

  function "*" (Left: Real'Base; Right: Matrix) return Matrix is
    X: Matrix := Right;
  begin
    X.Inner := Left * Right.Inner;
    return X;
  end "*";

  function "*" (Left: Matrix; Right: Real'Base) return Matrix is
  begin
    return Right * Left;
  end "*";

  function "/" (Left: Matrix; Right: Real'Base) return Matrix is
    X: Matrix := Left;
  begin
    X.Inner := X.Inner / Right;
    return X;
  end "/";

  function "*" (Left: Item; Right: Matrix) return Matrix is
  begin
    return Transform (Left.Inner * Right.Inner);
  end "*";

  function "*" (Left: Matrix; Right: Item) return Matrix is
  begin
    return Right * Left;
  end "*";

  function "/" (Left: Matrix; Right: Item) return Matrix is
  begin
    return Transform (Left.Inner / Right.Inner);
  end "/";

  function "*" (Left, Right: Matrix) return Matrix is
  begin
    return Transform (Left.Inner * Right.Inner);
  end "*";

  function Det (M: Matrix) return Item is
  begin
    return Transform (Det (M.Inner));
  end Det;

  function Inv (M: Matrix) return Matrix is
  begin
    return Transform (Inv (M.Inner));
  end Inv;

  function Tran (M: Matrix) return Matrix is
    X: Matrix := M;
  begin
    X.Inner := Tran (X.Inner);
    return X;
  end Tran;

  function "*" (Left: Matrix; Right: Vector) return Vector is
  begin
    return Transform (Left.Inner * Right.Inner);
  end "*";

  function "*" (Left: Vector; Right: Matrix) return Vector is
  begin
    return Transform (Left.Inner * Right.Inner);
  end "*";

end Constrained_Checked_SI.Generic_Vector_Space;