------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Numerics;

with Rational_Arithmetics;
use  Rational_Arithmetics;

with Unconstrained_Checked_SI;

generic

  type Real is digits <>;

package Constrained_Checked_SI is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.2
  -- Date      21 February 2003
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  24.07.2002
  --  C.G.    1.1  15.11.2002 Use Real'Base
  --  C.G.    1.2  21.02.2003 Add Dimension_as
  --====================================================================

  pragma Elaborate_Body;

  type Item (<>) is private;

  Zero: constant Item;
  One : constant Item;

  function Dimension_as (X: Item) return Item;  -- Value 1.0, dimension of X
  function Value        (X: Item) return Real;

  -- Base units

  Radian   : constant Item;  -- rad
  Steradian: constant Item;  -- sr
  Meter    : constant Item;  -- m
  Kilogram : constant Item;  -- kg
  Gram     : constant Item;  -- g
  Second   : constant Item;  -- s
  Ampere   : constant Item;  -- A
  Kelvin   : constant Item;  -- K
  Candela  : constant Item;  -- cd
  Mole     : constant Item;  -- mol

  -- Prefixes
  -- Any base unit may be prefixed with one of the constants below.
  -- Only the Kilogram is an exception: It is illegal to write
  -- Mkg (Megakilogram), the correct designation is Gg (Gigagram).

  -- http://physics.nist.gov/cuu/Units/prefixes.html
  yocto: constant := 1.0E-24;  -- y
  zepto: constant := 1.0E-21;  -- z
  atto : constant := 1.0E-18;  -- a
  femto: constant := 1.0E-15;  -- f
  pico : constant := 1.0E-12;  -- p
  nano : constant := 1.0E-09;  -- n
  micro: constant := 1.0E-06;  --  (u)
  milli: constant := 1.0E-03;  -- m
  centi: constant := 1.0E-02;  -- c
  deci : constant := 1.0E-01;  -- d
  deka : constant := 1.0E+01;  -- da
  hecto: constant := 1.0E+02;  -- h
  kilo : constant := 1.0E+03;  -- k
  mega : constant := 1.0E+06;  -- M
  giga : constant := 1.0E+09;  -- G
  tera : constant := 1.0E+12;  -- T
  peta : constant := 1.0E+15;  -- P
  exa  : constant := 1.0E+18;  -- E
  zetta: constant := 1.0E+21;  -- Z
  yotta: constant := 1.0E+24;  -- Y

  -- http://physics.nist.gov/cuu/Units/binarys.html
  kibi: constant := 2**10;  -- Ki = 1024 ~ 1000 = k
  mebi: constant := 2**20;  -- Mi
  gibi: constant := 2**30;  -- Gi
  tebi: constant := 2**40;  -- Ti
  pebi: constant := 2**50;  -- Pi
  exbi: constant := 2**60;  -- Ei

  -- Operations

  function "abs" (Right: Item) return Item;

  function "+" (Right: Item) return Item;
  function "-" (Right: Item) return Item;

  function "+" (Left, Right: Item) return Item;
  function "-" (Left, Right: Item) return Item;

  function "*" (Left, Right: Item) return Item;
  function "/" (Left, Right: Item) return Item;

  function "*" (Left: Item     ; Right: Real'Base) return Item;
  function "*" (Left: Real'Base; Right: Item     ) return Item;
  function "/" (Left: Item     ; Right: Real'Base) return Item;
  function "/" (Left: Real'Base; Right: Item     ) return Item;

  function "**" (Base: Item; Exponent: Whole    ) return Item;
  function "**" (Base: Item; Exponent: Rational ) return Item;
  function "**" (Base: Item; Exponent: Real'Base) return Item;

  function "<"  (Left, Right: Item) return Boolean;
  function "<=" (Left, Right: Item) return Boolean;
  function ">=" (Left, Right: Item) return Boolean;
  function ">"  (Left, Right: Item) return Boolean;

  Unit_Error: exception;

  -- Mathematics

  Argument_Error: exception renames Ada.Numerics.Argument_Error;

  Pi: constant := Ada.Numerics.Pi;
  e : constant := Ada.Numerics.e;

  function Sqrt (X: Item) return Item;  -- X**(1/2)
  function Cbrt (X: Item) return Item;  -- X**(1/3)

  function Log (X      : Item) return Item;
  function Log (X, Base: Item) return Item;
  function Exp (X      : Item) return Item;

  function Sin (X       : Item) return Item;
  function Sin (X, Cycle: Item) return Item;
  function Cos (X       : Item) return Item;
  function Cos (X, Cycle: Item) return Item;
  function Tan (X       : Item) return Item;
  function Tan (X, Cycle: Item) return Item;
  function Cot (X       : Item) return Item;
  function Cot (X, Cycle: Item) return Item;

  function Arcsin (X       : Item) return Item;
  function Arcsin (X, Cycle: Item) return Item;
  function Arccos (X       : Item) return Item;
  function Arccos (X, Cycle: Item) return Item;
  function Arctan (Y       : Item;
                   X       : Item := One)
                                   return Item;
  function Arctan (Y       : Item;
                   X       : Item := One;
                   Cycle   : Item) return Item;
  function Arccot (X       : Item;
                   Y       : Item := One)
                                   return Item;
  function Arccot (X       : Item;
                   Y       : Item := One;
                   Cycle   : Item) return Item;

  function Sinh    (X: Item) return Item;
  function Cosh    (X: Item) return Item;
  function Tanh    (X: Item) return Item;
  function Coth    (X: Item) return Item;
  function Arcsinh (X: Item) return Item;
  function Arccosh (X: Item) return Item;
  function Arctanh (X: Item) return Item;
  function Arccoth (X: Item) return Item;

private

  pragma Inline (Dimension_as, Value,
                 "abs", "+", "-", "*", "/", "**",
                 "<", "<=", ">=", ">",
                 Sqrt, Cbrt,
                 Log, Exp,
                 Sin, Cos, Tan, Cot,
                 Arcsin, Arccos, Arctan, Arccot,
                 Sinh, Cosh, Tanh, Coth,
                 Arcsinh, Arccosh, Arctanh, Arccoth);

  package SI is new Unconstrained_Checked_SI (Real);

  type Item (m_N, kg_N, s_N, A_N, K_N, cd_N, mol_N: Whole;
             m_D, kg_D, s_D, A_D, K_D, cd_D, mol_D: Positive_Whole) is record
    Inner: SI.Item;
  end record;

  Zero     : constant Item := (m_N   | kg_N | s_N | A_N | K_N | cd_N | mol_N => 0,
                               m_D   | kg_D | s_D | A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.Zero);
  One      : constant Item := (m_N   | kg_N | s_N | A_N | K_N | cd_N | mol_N => 0,
                               m_D   | kg_D | s_D | A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.One);
  Meter    : constant Item := (m_N   | m_D   => 1,
                                       kg_N | s_N | A_N | K_N | cd_N | mol_N => 0,
                                       kg_D | s_D | A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.Meter);
  Kilogram : constant Item := (kg_N  | kg_D  => 1,
                               m_N   |        s_N | A_N | K_N | cd_N | mol_N => 0,
                               m_D   |        s_D | A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.Kilogram);
  Gram     : constant Item := (kg_N  | kg_D  => 1,
                               m_N   |        s_N | A_N | K_N | cd_N | mol_N => 0,
                               m_D   |        s_D | A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.Gram);
  Second   : constant Item := (s_N   | s_D   => 1,
                               m_N   | kg_N |       A_N | K_N | cd_N | mol_N => 0,
                               m_D   | kg_D |       A_D | K_D | cd_D | mol_D => 1,
                               Inner => SI.Second);
  Ampere   : constant Item := (A_N   | A_D   => 1,
                               m_N   | kg_N | s_N |       K_N | cd_N | mol_N => 0,
                               m_D   | kg_D | s_D |       K_D | cd_D | mol_D => 1,
                               Inner => SI.Ampere);
  Kelvin   : constant Item := (K_N   | K_D   => 1,
                               m_N   | kg_N | s_N | A_N |       cd_N | mol_N => 0,
                               m_D   | kg_D | s_D | A_D |       cd_D | mol_D => 1,
                               Inner => SI.Kelvin);
  Candela  : constant Item := (cd_N  | cd_D  => 1,
                               m_N   | kg_N | s_N | A_N | K_N |        mol_N => 0,
                               m_D   | kg_D | s_D | A_D | K_D |        mol_D => 1,
                               Inner => SI.Candela);
  Mole     : constant Item := (mol_N | mol_D => 1,
                               m_N   | kg_N | s_N | A_N | K_N | cd_N         => 0,
                               m_D   | kg_D | s_D | A_D | K_D | cd_D         => 1,
                               Inner => SI.Mole);
  Radian   : constant Item := One;
  Steradian: constant Item := One;

  function Transform (X: SI.Item) return Item;
  pragma Inline (Transform);

end Constrained_Checked_SI;
