------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

package body Constrained_Unchecked_SI.Generic_Polynomial_Numerics is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   3.0
  -- Date      24 February 2003
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  07.12.1994
  --  C.G.    1.1  13.05.1997 Interface changed
  --  C.G.    2.0  31.07.1997 Interface rewrought
  --  C.G.    3.0  24.02.2003 Adapted to SI package
  --====================================================================

  function Polynomial (Coefficients: Table; X: Real) return Real is
    -- Algorithm: Use Horner's scheme.
    F: Real := 0.0;
  begin
    for I in reverse Coefficients'Range loop
      F := F * X + Coefficients (I);
    end loop;
    return F * X ** Integer (Coefficients'First);
  end Polynomial;

  function Define (X, Y: Table) return Interpolation_Table is
    -- Check conditions on input tables.
    -- Make the internal table increasing and store direction in sign:
    --   Sign := +1 (increasing);
    --   Sign :=  0 (table error);
    --   Sign := -1 (decreasing).
    -- Liniear interpolation will be called with signed argument "Sign * X".
    Table: Interpolation_Table (X'Length);
  begin
    if X'Length <= 1 or X'Length /= Y'Length then
      raise Table_Error;
    end if;
    if X (X'First) < X (X'First + 1) then
      Table.Sign := +1.0;  -- increasing
    else
      Table.Sign := -1.0;  -- decreasing
    end if;
    Table.X (1) := Table.Sign * X (X'First);
    for I in 2 .. Table.Length loop
      Table.X (I) := Table.Sign * X (X'First + I - 1);
      if Table.X (I - 1) >= Table.X (I) then
        raise Table_Error;
      end if;
    end loop;
    Table.Y := Y;
    return Table;
  end Define;

  function Linear_Interpolation
               (Table: Interpolation_Table; X: Item) return Item is
    -- Check whether table is defined.
    -- Use a binary search to find X's position in the table.
    Signed: constant Real := Table.Sign * Value (X);
    Lower, Upper: Positive_Whole;
    function Binary_Search return Positive_Whole is
      L: Positive_Whole := 1;
      U: Positive_Whole := Table.Length;
      M: Positive_Whole;
    begin
      while U - L > 1 loop
        M := Whole (Integer (L + U) / 2);
        if Signed <= Table.X (M) then
          U := M;
        else
          L := M;
        end if;
      end loop;
      return L;
    end Binary_Search;
  begin
    if Table.Sign = 0.0 then
      raise Table_Error;
    end if;
    if Signed < Table.X (1) then
      Lower := 1;
    elsif Signed > Table.X (Table.Length) then
      Lower := Table.Length - 1;
    else
      Lower := Binary_Search;
    end if;
    Upper := Lower + 1;
    return One *  -- type conversion
           (Table.Y (Lower) +
              (Table.Y (Upper) - Table.Y (Lower)) * ((Signed - Table.X (Lower)) /
              (Table.X (Upper) - Table.X (Lower))));
  end Linear_Interpolation;

  function Second_Order_Curve (X, Y: Table_2nd_Order) return Table_2nd_Order is
    -- Solve the linear equations for a parabola going thru three points.
    Dx01: constant Real := X (0) - X (1);
    Dx02: constant Real := X (0) - X (2);
    F1  : constant Real := (Y (0) - Y (1)) / Dx01;
    F2  : constant Real := (Y (0) - Y (2)) / Dx02;
    G1  : constant Real := (X (0) ** 2 - X (1) ** 2) / Dx01;
    G2  : constant Real := (X (0) ** 2 - X (2) ** 2) / Dx02;
    A, B, C: Real;
  begin
    C := (F1 - F2) / (G1 - G2);
    B := F1 - C * G1;
    A := Polynomial ((0 => Y (0), 1 => -B, 2 => -C), X (0));
    return (0 => A, 1 => B, 2 => C);
  end Second_Order_Curve;

end Constrained_Unchecked_SI.Generic_Polynomial_Numerics;
