------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

generic
package Constrained_Unchecked_SI.Generic_Polynomial_Numerics is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   3.0
  -- Date      21 February 2003
  --====================================================================
  -- A polynomial has the general form
  --    f(x) := sum (Coeff (i) * X**i)
  -- where i is in a subrange of Natural, say A .. B. The coefficients
  -- all have different dimensions. This cannot be handled with
  -- constrained items, but if you take out the lowest exponent term,
  -- the polynomial becomes
  --    (Coeff (A) * X**A) * sum (Coeff'(i) * X**i)
  -- where i runs in 0 .. B-A, Coeff'(i) = Coeff (i)/Coeff (A).
  -- The first factor Coeff (A) * X**A carries the dimension of the
  -- expression, the second is dimensionless. Thus, although neither the
  -- new coefficients Coeff'(i) nor X**i are dimensionless, their
  -- combination is and you can as well treat them as dimensionless as
  -- well.
  --
  -- Polynomial defines the function
  --    f(x) := sum (Coeff (i) * X**i)
  -- Coefficients' range need not include 0, e.g.
  --    Coefficients = (1=> 1.0, 2=> 1.0)
  -- defines the term x*(x+1).
  --
  -- Define creates a table for linear interpolation. The length of the
  -- table returned is X'Length.
  -- X/Y denote the abscissa and ordinate values of sample points.
  -- X must be strictly monotonic (increasing or decreasing).
  -- Both tables must have the same length > 1 or else Table_Error is
  -- propagated.
  --
  -- Linear_Interpolation interpolates linearly for x in the Table
  -- range and extrapolates outside.
  -- Table_Error is propagated if the table is undefined.
  --
  -- Second_Order_Curve calculates the coefficients of 2nd order curves
  --    Y = A + B * X + C * X**2
  -- with defining points: (Xi,Yi), i in 0..2.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  07.12.1994
  --  C.G.    1.1  13.05.1997 Interface changed
  --  C.G.    2.0  31.07.1997 Interpolation table made private
  --  C.G.    3.0  21.02.2003 Imported into SI package
  --====================================================================

  type Table is array (Natural_Whole range <>) of Real;
  subtype Table_2nd_Order is Table (0 .. 2);

  type Interpolation_Table (Length: Positive_Whole) is private;

  function Polynomial (Coefficients: Table; X: Real) return Real;

  function Define (X, Y: Table) return Interpolation_Table;

  function Linear_Interpolation
               (Table: Interpolation_Table; X: Item) return Item;

  Table_Error: exception;

  function Second_Order_Curve (X, Y: Table_2nd_Order) return Table_2nd_Order;

private

  type Interpolation_Table (Length: Positive_Whole) is record
    Sign: Real := 0.0;  -- means undefined
    X, Y: Table (1 .. Length);
  end record;

end Constrained_Unchecked_SI.Generic_Polynomial_Numerics;
