------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Text_IO;
use  Ada.Text_IO;

generic
package Constrained_Unchecked_SI.Generic_Text_IO is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   2.0
  -- Date      4 February 2003
  --====================================================================
  -- Put and Get work like the conventional operations in Text_IO for
  -- numbers, the formatting parameters apply to the numeric value only,
  -- not to the dimension.
  --
  -- For external representation, an Item is considered to be a value
  -- multiplied by its dimension. Dimensions are products of units,
  -- units are written with their symbolic names (e.g. m for Meter).
  --
  -- The default representation of a dimension is a product of all
  -- unit factors (m, kg, s, A, K, cd, mol) with an exponent different
  -- from 0 in that sequence. If the dimension is 1, the default
  -- representation is empty.
  -- Each factor is given either as the pure unit (if the exponent is
  -- 1), or as the unit with integer exponent (if the exponent is a
  -- whole number), or as the unit with a fractional exponent. The
  -- exponent is included in parentheses as necessary according to
  -- normal Ada rules.
  -- For the exponent, the caret '^' is used rather than the double-star
  -- "**" since it enhances readability (dimensions look a bit ugly
  -- anyway).
  -- In the non-default representation, units may be given in any
  -- sequence and with any exponents. Also the names rad, sr and any
  -- derived dimension defined in SI.Units with a genuine name (e.g. J
  -- but not Nm, which has to be written as N*m) may be used (note that
  -- dimension names are case-sensitive). Unit symbols may be modified
  -- by any of the non-binary prefixes (without multiplication sign)
  -- like km or ns. For micro, 'u' or '' may be used.
  -- Additionally, the division operator '/' may be used instead of the
  -- multiplication operator '*'.
  -- Dimension format modifiers must be valid representations as defined
  -- above, where the first operator, when omitted, is implied as a '*',
  -- else Unit_Error will be propagated.
  --
  -- Put proceeds as follows:
  -- If no dimension format (Dim) is given, Put writes the value in the
  -- numeric format given (Fore, Aft, Exp), no dimension is written.
  -- If a dimension format is given, Unit_Error is propagated if the
  -- string is no valid dimension.
  -- Then the Item is divided by the dimension given, the resulting
  -- value is written in the numeric format given, followed by the di-
  -- mension format string.
  --
  -- Get proceeds as follows:
  -- First it reads the value like Get of the standard Text_IO.Float_IO.
  -- Data_Error may be raised in this step.
  -- In case Data_Error was not raised, reading stops if no '*' or '/'
  -- follows.
  -- Else the dimension is read as a product of units. Reading stops if
  -- a unit is not followed by a '*' or '/', in any case at the end of
  -- the line.
  -- Each unit factor is read as follows:
  -- Characters are input only so long as the sequence input is an
  -- initial sequence of a literal having the syntax of a unit factor
  -- (in particular, input ceases when a line terminator is encounter-
  -- ed). The character or line terminator that causes input to cease
  -- remains available for subsequent input. [These two sentences mimic
  -- the behaviour for enumeration literals in RM A.10.6(5).]
  -- The exception Data_Error is propagated if the sequence input does
  -- not have the required syntax, or if the characters input for the
  -- unit symbol do not form an existing symbol.
  --
  -- This is the syntax of a unit factor:
  --   unit_factor ::= [prefix] unit_symbol [^ exponent]
  --   prefix      ::= any in the list of defined prefixes
  --                     (case sensitive)
  --   unit_symbol ::= letter {letter}
  --   letter      ::= any letter in the range a..z or A..Z
  --                     (case sensitive)
  --   exponent    ::= ( [sign] rational ) | numeral
  --   sign        ::= + | -
  --   rational    ::= numeral [ / numeral]
  --   numeral     ::= digit {digit}
  -- The leading letters of a unit factor are interpreted as a prefix
  -- only if they match one of the official prefixes (case sensitive)
  -- and are followed by at least one other letter, else they are con-
  -- sidered part of the unit symbol.
  --
  -- Examples:
  --   S: Item := 1.0 * Meter;         output as  1.0
  --   V: Item := 1.0 * Meter/Second;             1.0
  --   B: Item := 1.0 * Tesla;                    1.0
  --
  -- The speed V above, when output with a Dim format modifier "km/s" or
  -- "*km/s", results in "1.0E-3*km/s".
  -- It may also be read as
  --  10.0*dm*s^(-1)
  --   1.0*m/s
  --   1.0*s^(-1)/m^(-1)
  --   1.0*s^(-1)*m
  --   1.0*m^(2/2)*Hz
  --   1.0/dpt*Bq
  --
  -- The magnetic field B above may also be read as
  --   1.0E+3*mT
  --   1.0*m^(-2)*Wb
  --
  -- "1.0K" will be read a an item with value 1.0 and dimension 1,
  -- the following character 'K' will be left unconsumed (no '*' after
  -- the value).
  -- "1.0*mB " will raise Data_Error with the offending character 'B'
  -- (there is no unit symbol 'B'); the blank character will be left
  -- unconsumed. It will not be read as one meter with the following
  -- character 'B' left unconsumed.
  -- "1.0*mBqA " will raise Data_Error with the offending character 'A'
  -- (there is no unit symbol 'BqA'); the blank character will be left
  -- unconsumed. It will not be read as one Millibecquerel with the
  -- following character 'A' left unconsumed.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  25.07.2002
  --  C.G.    1.1  02.09.2002 Dim format modifier for Put added
  --  C.G.    2.0  04.02.2003 Syntax for reading unit symbols changed
  --====================================================================

  pragma Elaborate_Body;

  Default_Fore: Field := 2;
  Default_Aft : Field := Real'Digits - 1;
  Default_Exp : Field := 3;

  function Get                      return Item;
  function Get (File: in File_Type) return Item;

  procedure Put (X   : in Item;
                 Fore: in Field := Default_Fore;
                 Aft : in Field := Default_Aft;
                 Exp : in Field := Default_Exp;
                 Dim : in String := "");
  procedure Put (File: in File_Type;
                 X   : in Item;
                 Fore: in Field := Default_Fore;
                 Aft : in Field := Default_Aft;
                 Exp : in Field := Default_Exp;
                 Dim : in String := "");

end Constrained_Unchecked_SI.Generic_Text_IO;
