------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Unconstrained_Unchecked_SI.Generic_Vector_Space;

generic

package Constrained_Unchecked_SI.Generic_Vector_Space is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      19 November 2002
  --====================================================================
  -- This is not a linear algebraics package. All elements of a vector
  -- and a matrix must have the same dimension.
  --
  -- Vectors can be defined in several ways, e.g.:
  --
  -- V: Vector := (1.0, 3.9, 0.0) * Meter/Second;
  -- V: Vector := (1.0 * U1 + 2.0 * U2) * Meter/Second;
  -- V: Vector := P1 * (1.0 * Meter/Second) + P2 * (2.0 * Meter/Second);
  -- V: Vector := 1.0 * Meter/Second * P1 + 2.0 * Meter/Second * P2;
  --
  -- Similarly for matrices.
  --
  -- (The inner and the cross products are not defined as operators
  -- since they are not associative with themselves or with one another
  -- and having both as operators would lead to ambiguities or errors.)
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  18.11.2002
  --====================================================================

  type Axis is range 1 .. 3;

  -- Prototypes

  type Proto_Vector is array (Axis) of Real;

  Null_Proto_Vector: constant                 Proto_Vector :=  (0.0, 0.0, 0.0);
  Unit_Proto_Vector: constant array (Axis) of Proto_Vector := ((1.0, 0.0, 0.0),
                                                               (0.0, 1.0, 0.0),
                                                               (0.0, 0.0, 1.0));
  P1: Proto_Vector renames Unit_Proto_Vector (1);
  P2: Proto_Vector renames Unit_Proto_Vector (2);
  P3: Proto_Vector renames Unit_Proto_Vector (3);

  type Proto_Matrix is array (Axis, Axis) of Real;

  Null_Proto_Matrix: constant Proto_Matrix := ((0.0, 0.0, 0.0),
                                               (0.0, 0.0, 0.0),
                                               (0.0, 0.0, 0.0));
  Proto_Unity      : constant Proto_Matrix := ((1.0, 0.0, 0.0),
                                               (0.0, 1.0, 0.0),
                                               (0.0, 0.0, 1.0));
  -- Vector

  type Vector (<>) is private;

  type Vector_Ptr is access constant Vector;
  type Vectors    is array (Axis) of Vector_Ptr;

  Null_Vector:         constant Vector;
  U1, U2, U3 : aliased constant Vector;  -- the unit vectors
  Unit_Vector: constant Vectors;         -- Unit_Vector (i).all = Ui

  function "*" (Left: Proto_Vector; Right: Item        ) return Vector;  -- constructors
  function "*" (Left: Item        ; Right: Proto_Vector) return Vector;

  function Normalize (X: Proto_Vector) return Vector;  -- unit vector

  function  Get (V:        Vector; A:    Axis) return Item;
  procedure Set (V: in out Vector; A: in Axis; It: in Item);

  -- Matrix

  type Matrix (<>) is private;

  Null_Matrix, Unity: constant Matrix;

  function "*" (Left: Proto_Matrix; Right: Item        ) return Matrix;  -- constructors
  function "*" (Left: Item        ; Right: Proto_Matrix) return Matrix;

  function  Get (M:        Matrix; A, B:    Axis) return Item;
  procedure Set (M: in out Matrix; A, B: in Axis; It: in Item);

  -- Vector algebra

  function "abs" (Right: Vector) return Item;

  function "+" (Right: Vector) return Vector;
  function "-" (Right: Vector) return Vector;

  function "+" (Left, Right: Vector) return Vector;
  function "-" (Left, Right: Vector) return Vector;

  function "*" (Left: Real'Base; Right: Vector   ) return Vector;
  function "*" (Left: Item     ; Right: Vector   ) return Vector;
  function "*" (Left: Vector   ; Right: Real'Base) return Vector;
  function "*" (Left: Vector   ; Right: Item     ) return Vector;
  function "/" (Left: Vector   ; Right: Real'Base) return Vector;
  function "/" (Left: Vector   ; Right: Item     ) return Vector;

  function Inner (Left, Right: Vector) return Item;    -- inner product
  function Cross (Left, Right: Vector) return Vector;  -- cross product

  -- Matrix algebra

  function "+" (Right: Matrix) return Matrix;
  function "-" (Right: Matrix) return Matrix;

  function "+" (Left, Right: Matrix) return Matrix;
  function "-" (Left, Right: Matrix) return Matrix;

  function "*" (Left: Real'Base; Right: Matrix   ) return Matrix;
  function "*" (Left: Item     ; Right: Matrix   ) return Matrix;
  function "*" (Left: Matrix   ; Right: Real'Base) return Matrix;
  function "*" (Left: Matrix   ; Right: Item     ) return Matrix;
  function "/" (Left: Matrix   ; Right: Real'Base) return Matrix;
  function "/" (Left: Matrix   ; Right: Item     ) return Matrix;

  function "*" (Left, Right: Matrix) return Matrix;

  function Det  (M: Matrix) return Item;    -- determinant
  function Inv  (M: Matrix) return Matrix;  -- inverse
  function Tran (M: Matrix) return Matrix;  -- transposed

  -- Mixed operations

  function "*" (Left: Matrix; Right: Vector) return Vector;  -- Mij*Vj
  function "*" (Left: Vector; Right: Matrix) return Vector;  -- Vi*Mij

private

  package Vector_Space is new SI.Generic_Vector_Space;

  type Vector is record
    Inner: Vector_Space.Vector;
  end record;

  Null_Vector: constant Vector := (Inner => (Zero.Inner, Zero.Inner, Zero.Inner));

  U1: aliased  constant Vector := (Inner => ( One.Inner, Zero.Inner, Zero.Inner));
  U2: aliased  constant Vector := (Inner => (Zero.Inner,  One.Inner, Zero.Inner));
  U3: aliased  constant Vector := (Inner => (Zero.Inner, Zero.Inner,  One.Inner));

  Unit_Vector: constant Vectors := (U1'Access, U2'Access, U3'Access);

  type Matrix is record
    Inner: Vector_Space.Matrix;
  end record;

  Null_Matrix: constant Matrix := (Inner => ((SI.Zero, SI.Zero, SI.Zero),
                                             (SI.Zero, SI.Zero, SI.Zero),
                                             (SI.Zero, SI.Zero, SI.Zero)));
  Unity      : constant Matrix := (Inner => ((SI. One, SI.Zero, SI.Zero),
                                             (SI.Zero, SI. One, SI.Zero),
                                             (SI.Zero, SI.Zero, SI. One)));

end Constrained_Unchecked_SI.Generic_Vector_Space;
