------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Invisible;  -- is taboo everywhere else

package Rational_Arithmetics is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      5 July 2002
  --====================================================================
  -- Define rational (n/m) and mixed (i + n/m) numbers.
  --
  -- Image returns the argument as a string.
  -- If the denominator is 1, the string has the format "sn", else it
  -- has the format "sn/m" where there is no white space around '/',
  -- and the sign 's' is blank for non-negative numbers.
  -- Value returns the argument into a rational number ignoring any
  -- leading or trailing blanks.
  -- It raises Constraint_Error if the string does not have the
  -- aforementioned format.
  --
  -- Notes:
  -- Whole must be a subtype so that unwanted operations like
  --   function "/" (Left, Right: Whole) return Whole;
  -- can be hidden.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  30.06.2002
  --  C.G.    1.1  05.07.2002 added mixed relational operators,
  --                          Image and Value
  --====================================================================

  pragma Elaborate_Body;

  -- Whole numbers

  subtype Whole is Invisible.Whole;
  subtype Natural_Whole  is Whole range 0 .. Whole'Last;
  subtype Positive_Whole is Whole range 1 .. Whole'Last;

  function "=" (Left, Right: Whole) return Boolean renames Invisible."=";

  function "+" (Left, Right: Whole) return Whole renames Invisible."+";
  function "-" (Left, Right: Whole) return Whole renames Invisible."-";

  function "*" (Left, Right: Whole) return Whole renames Invisible."*";

  function "<"  (Left, Right: Whole) return Boolean renames Invisible."<";
  function "<=" (Left, Right: Whole) return Boolean renames Invisible."<=";
  function ">=" (Left, Right: Whole) return Boolean renames Invisible.">=";
  function ">"  (Left, Right: Whole) return Boolean renames Invisible.">";

  -- Rational numbers

  type Rational is private;

  function Numerator   (R: Rational) return Whole;
  function Denominator (R: Rational) return Whole;

  -- Constructor

  function "+" (Right: Whole) return Rational;
  function "-" (Right: Whole) return Rational;

  function "/" (Left, Right: Whole) return Rational;

  -- Rational operations

  function "+" (Right: Rational) return Rational;
  function "-" (Right: Rational) return Rational;

  function "+" (Left, Right: Rational) return Rational;
  function "-" (Left, Right: Rational) return Rational;

  function "*" (Left, Right: Rational) return Rational;
  function "/" (Left, Right: Rational) return Rational;

  -- Relational

  function "=" (Left: Whole   ; Right: Rational) return Boolean;
  function "=" (Left: Rational; Right: Whole   ) return Boolean;

  function "<"  (Left, Right: Rational) return Boolean;
  function "<=" (Left, Right: Rational) return Boolean;
  function ">=" (Left, Right: Rational) return Boolean;
  function ">"  (Left, Right: Rational) return Boolean;

  function "<"  (Left: Whole   ; Right: Rational) return Boolean;
  function "<"  (Left: Rational; Right: Whole   ) return Boolean;
  function "<=" (Left: Whole   ; Right: Rational) return Boolean;
  function "<=" (Left: Rational; Right: Whole   ) return Boolean;
  function ">=" (Left: Whole   ; Right: Rational) return Boolean;
  function ">=" (Left: Rational; Right: Whole   ) return Boolean;
  function ">"  (Left: Whole   ; Right: Rational) return Boolean;
  function ">"  (Left: Rational; Right: Whole   ) return Boolean;

  function Max (Left, Right: Rational) return Rational;
  function Min (Left, Right: Rational) return Rational;

  function Max (Left: Whole   ; Right: Rational) return Rational;
  function Max (Left: Rational; Right: Whole   ) return Rational;
  function Min (Left: Whole   ; Right: Rational) return Rational;
  function Min (Left: Rational; Right: Whole   ) return Rational;

  function "abs" (Right: Rational) return Rational;

  -- Mixed operations

  function "+" (Left: Whole   ; Right: Rational) return Rational;
  function "+" (Left: Rational; Right: Whole   ) return Rational;
  function "-" (Left: Whole   ; Right: Rational) return Rational;
  function "-" (Left: Rational; Right: Whole   ) return Rational;

  function "*" (Left: Whole   ; Right: Rational) return Rational;
  function "*" (Left: Rational; Right: Whole   ) return Rational;
  function "/" (Left: Whole   ; Right: Rational) return Rational;
  function "/" (Left: Rational; Right: Whole   ) return Rational;

  -- Conversion to String

  function Image (X: Rational) return String;
  function Value (X: String  ) return Rational;

private

  pragma Inline (Numerator, Denominator,
                 "+", "-", "*", "/");

  type Rational is record
    Numerator  : Whole;
    Denominator: Whole range 1 .. Whole'Last;
  end record;

end Rational_Arithmetics;
