------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

package body Unconstrained_Checked_SI.Generic_Temperatures is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      8 October 2002
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  08.10.2002
  --====================================================================

  Rankine_Kelvin: constant := 5.0/9.0;  Kelvin_Rankine: constant := 9.0/5.0;
  Reaumur_Kelvin: constant := 5.0/4.0;  Kelvin_Reaumur: constant := 4.0/5.0;

  Rankine_Reaumur: constant := 4.0/9.0;  Reaumur_Rankine: constant := 9.0/4.0;

  Zero_Celsius_Kelvin    : constant := 273.15;
  Zero_Celsius_Fahrenheit: constant :=  32.0;
  Zero_Celsius_Rankine   : constant := Zero_Celsius_Kelvin * Kelvin_Rankine;
  Zero_Fahrenheit_Rankine: constant := Zero_Celsius_Rankine - Zero_Celsius_Fahrenheit;

  function to_Celsius (T: Real) return Temperature is
  begin
    return (Scale => Celsius, Value => T);
  end to_Celsius;

  function to_Celsius (T: Real) return Temperature_Difference is
  begin
    return (Scale => Celsius, Value => T);
  end to_Celsius;

  function to_Fahrenheit (T: Real) return Temperature is
  begin
    return (Scale => Fahrenheit, Value => T);
  end to_Fahrenheit;

  function to_Fahrenheit (T: Real) return Temperature_Difference is
  begin
    return (Scale => Fahrenheit, Value => T);
  end to_Fahrenheit;

  function to_Rankine (T: Real) return Temperature is
  begin
    return (Scale => Rankine, Value => T);
  end to_Rankine;

  function to_Rankine (T: Real) return Temperature_Difference is
  begin
    return (Scale => Rankine, Value => T);
  end to_Rankine;

  function to_Reaumur (T: Real) return Temperature is
  begin
    return (Scale => Reaumur, Value => T);
  end to_Reaumur;

  function to_Reaumur (T: Real) return Temperature_Difference is
  begin
    return (Scale => Reaumur, Value => T);
  end to_Reaumur;

  function to_Kelvin (T: Temperature) return Item is
  begin
    case T.Scale is
      when Celsius    => return Kelvin * (T.Value + Zero_Celsius_Kelvin);
      when Fahrenheit => return Kelvin * Rankine_Kelvin * (T.Value + Zero_Fahrenheit_Rankine);
      when Rankine    => return Kelvin * Rankine_Kelvin *  T.Value;
      when Reaumur    => return Kelvin * (T.Value * Reaumur_Kelvin + Zero_Celsius_Kelvin);
    end case;
  end to_Kelvin;

  function to_Kelvin (T: Temperature_Difference) return Item is
  begin
    case T.Scale is
      when Celsius    => return T.Value * Kelvin;
      when Fahrenheit => return T.Value * Kelvin * Rankine_Kelvin;
      when Rankine    => return T.Value * Kelvin * Rankine_Kelvin;
      when Reaumur    => return T.Value * Kelvin * Reaumur_Kelvin;
    end case;
  end to_Kelvin;

  function to_Celsius (T: Item) return Temperature is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Celsius,
            Value => T.Value - Zero_Celsius_Kelvin);
  end to_Celsius;

  function to_Celsius (T: Item) return Temperature_Difference is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Celsius,
            Value => T.Value);
  end to_Celsius;

  function to_Fahrenheit (T: Item) return Temperature is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Fahrenheit,
            Value => T.Value * Kelvin_Rankine - Zero_Fahrenheit_Rankine);
  end to_Fahrenheit;

  function to_Fahrenheit (T: Item) return Temperature_Difference is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Fahrenheit,
            Value => T.Value * Kelvin_Rankine);
  end to_Fahrenheit;

  function to_Rankine (T: Item) return Temperature is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Rankine,
            Value => T.Value * Kelvin_Rankine);
  end to_Rankine;

  function to_Rankine (T: Item) return Temperature_Difference is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Rankine,
            Value => T.Value * Kelvin_Rankine);
  end to_Rankine;

  function to_Reaumur (T: Item) return Temperature is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Reaumur,
            Value => (T.Value - Zero_Celsius_Kelvin) * Kelvin_Reaumur);
  end to_Reaumur;

  function to_Reaumur (T: Item) return Temperature_Difference is
  begin
    if T.Unit /= Kelvin.Unit then
      raise Unit_Error;
    end if;
    return (Scale => Reaumur,
            Value => T.Value * Kelvin_Reaumur);
  end to_Reaumur;

  function to_Celsius (T: Temperature) return Temperature is
  begin
    case T.Scale is
      when Celsius    => return T;
      when Fahrenheit => return (Scale => Celsius, Value => Rankine_Kelvin * (T.Value - Zero_Celsius_Fahrenheit));
      when Rankine    => return (Scale => Celsius, Value => Rankine_Kelvin * (T.Value - Zero_Celsius_Rankine));
      when Reaumur    => return (Scale => Celsius, Value => Reaumur_Kelvin * T.Value);
    end case;
  end to_Celsius;

  function to_Celsius (T: Temperature_Difference) return Temperature_Difference is
  begin
    case T.Scale is
      when Celsius    => return T;
      when Fahrenheit => return (Scale => Celsius, Value => Rankine_Kelvin * T.Value);
      when Rankine    => return (Scale => Celsius, Value => Rankine_Kelvin * T.Value);
      when Reaumur    => return (Scale => Celsius, Value => Reaumur_Kelvin * T.Value);
    end case;
  end to_Celsius;

  function to_Fahrenheit (T: Temperature) return Temperature is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Fahrenheit, Value => T.Value * Kelvin_Rankine + Zero_Celsius_Fahrenheit);
      when Fahrenheit => return T;
      when Rankine    => return (Scale => Fahrenheit, Value => T.Value - Zero_Fahrenheit_Rankine);
      when Reaumur    => return (Scale => Fahrenheit, Value => T.Value * Reaumur_Rankine + Zero_Celsius_Fahrenheit);
    end case;
  end to_Fahrenheit;

  function to_Fahrenheit (T: Temperature_Difference) return Temperature_Difference is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Fahrenheit, Value => T.Value * Kelvin_Rankine);
      when Fahrenheit => return T;
      when Rankine    => return (Scale => Fahrenheit, Value => T.Value);
      when Reaumur    => return (Scale => Fahrenheit, Value => T.Value * Reaumur_Rankine);
    end case;
  end to_Fahrenheit;

  function to_Rankine (T: Temperature) return Temperature is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Rankine, Value => (T.Value + Zero_Celsius_Kelvin) * Kelvin_Rankine);
      when Fahrenheit => return (Scale => Rankine, Value => T.Value + Zero_Fahrenheit_Rankine);
      when Rankine    => return T;
      when Reaumur    => return (Scale => Rankine, Value => T.Value * Reaumur_Rankine + Zero_Celsius_Rankine);
    end case;
  end to_Rankine;

  function to_Rankine (T: Temperature_Difference) return Temperature_Difference is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Rankine, Value => T.Value * Kelvin_Rankine);
      when Fahrenheit => return (Scale => Rankine, Value => T.Value);
      when Rankine    => return T;
      when Reaumur    => return (Scale => Rankine, Value => T.Value * Reaumur_Rankine);
    end case;
  end to_Rankine;

  function to_Reaumur (T: Temperature) return Temperature is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Reaumur, Value => Kelvin_Reaumur  * T.Value);
      when Fahrenheit => return (Scale => Reaumur, Value => Rankine_Reaumur * (T.Value - Zero_Celsius_Fahrenheit));
      when Rankine    => return (Scale => Reaumur, Value => Rankine_Reaumur * (T.Value - Zero_Celsius_Rankine));
      when Reaumur    => return T;
    end case;
  end to_Reaumur;

  function to_Reaumur (T: Temperature_Difference) return Temperature_Difference is
  begin
    case T.Scale is
      when Celsius    => return (Scale => Reaumur, Value => T.Value * Kelvin_Reaumur);
      when Fahrenheit => return (Scale => Reaumur, Value => T.Value * Rankine_Reaumur);
      when Rankine    => return (Scale => Reaumur, Value => T.Value * Rankine_Reaumur);
      when Reaumur    => return T;
    end case;
  end to_Reaumur;

  function Value (T: Temperature) return Real is
  begin
    return T.Value;
  end Value;

  function Value (T: Temperature_Difference) return Real is
  begin
    return T.Value;
  end Value;

  function "+" (Left: Temperature; Right: Temperature_Difference) return Temperature is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value + Right.Value);
  end "+";

  function "+" (Left: Temperature_Difference; Right: Temperature) return Temperature is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value + Right.Value);
  end "+";

  function "-" (Left: Temperature; Right: Temperature_Difference) return Temperature is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value - Right.Value);
  end "-";

  function "-" (Left: Temperature; Right: Temperature) return Temperature_Difference is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value - Right.Value);
  end "-";

  function "+" (Left, Right: Temperature_Difference) return Temperature_Difference is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value + Right.Value);
  end "+";

  function "-" (Left, Right: Temperature_Difference) return Temperature_Difference is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return (Scale => Left.Scale,
            Value => Left.Value - Right.Value);
  end "-";

  function "<" (Left, Right: Temperature) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value < Right.Value;
  end "<";

  function "<" (Left, Right: Temperature_Difference) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value < Right.Value;
  end "<";

  function "<=" (Left, Right: Temperature) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value <= Right.Value;
  end "<=";

  function "<=" (Left, Right: Temperature_Difference) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value <= Right.Value;
  end "<=";

  function ">=" (Left, Right: Temperature) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value >= Right.Value;
  end ">=";

  function ">=" (Left, Right: Temperature_Difference) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value >= Right.Value;
  end ">=";

  function ">" (Left, Right: Temperature) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value > Right.Value;
  end ">";

  function ">" (Left, Right: Temperature_Difference) return Boolean is
  begin
    if Left.Scale /= Right.Scale then
      raise Unit_Error;
    end if;
    return Left.Value > Right.Value;
  end ">";

end Unconstrained_Checked_SI.Generic_Temperatures;