------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Strings.Maps.Constants;
use  Ada.Strings.Maps, Ada.Strings.Maps.Constants;

separate (Unconstrained_Checked_SI.Generic_Text_IO)
function Generic_Read_Dimension return Item is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   2.1
  -- Date      11 March 2003
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  06.08.2002 Made generic and separate
  --  C.G.    1.1  14.09.2002 Added Katal
  --  C.G.    1.2  15.11.2002 Use Real'Base
  --  C.G.    2.0  04.02.2003 Syntax for reading unit symbols changed
  --  C.G.    2.1  11.03.2003 IO customization
  --====================================================================

  use Units;

  type Symbol_Map is record
    Symbol: String (1 .. 3);
    Length: Positive;
    Value : Item;
  end record;

  Valid_Symbol_Map: constant array (1 .. 33) of Symbol_Map :=
    (("A  ", 1, Ampere),
     ("Bq ", 2, Becquerel),
     ("C  ", 1, Coulomb),
     ("F  ", 1, Farad),
     ("Gy ", 2, Gray),
     ("H  ", 1, Henry),
     ("Hz ", 2, Hertz),
     ("J  ", 1, Joule),
     ("K  ", 1, Kelvin),
     ("L  ", 1, Liter),
     ("N  ", 1, Newton),
     ("Ohm", 3, Ohm),
     ("Pa ", 2, Pascal),
     ("S  ", 1, Siemens),
     ("Sv ", 2, Sievert),
     ("T  ", 1, Tesla),
     ("V  ", 1, Volt),
     ("W  ", 1, Watt),
     ("Wb ", 2, Weber),
     --                       "alphabetically" ordered
     ("cd ", 2, Candela),
     ("dpt", 3, Dioptry),
     ("g  ", 1, Gram),
     ("h  ", 1, Hour),
     ("kat", 3, Katal),
     ("l  ", 1, Liter),
     ("lm ", 2, Lumen),
     ("lx ", 2, Lux),
     ("m  ", 1, Meter),
     ("min", 3, Minute),
     ("mol", 3, Mole),
     ("rad", 3, Radian),
     ("s  ", 1, Second),
     ("sr ", 2, Steradian));

  function Search (Symbol: String) return Item is
    -- Binary search for Symbol in Valid_Symbol. If not found, Data_Error
    -- is raised.
    First: Positive := Valid_Symbol_Map'First;
    Last : Natural  := Valid_Symbol_Map'Last;
  begin
    loop
      declare
        Mean : constant Positive := (First + Last) / 2;
        Valid: Symbol_Map renames Valid_Symbol_Map (Mean);
      begin
        if Symbol = Valid.Symbol (1 .. Valid.Length) then
          return Valid.Value;
        elsif Symbol > Valid.Symbol (1 .. Valid.Length) then
          First := Mean + 1;
        else
          Last := Mean - 1;
        end if;
      end;
      if Last < First then
        raise Data_Error;
      end if;
    end loop;
  end Search;

  ----------------------------------------------------------------------

  function Read_Unit return Item is

    function Read_Symbol return Item is

      -- There must be a symbol at this place.
      -- We need more than one look-ahead to properly separate
      -- prefix and symbol proper.

      -- There are at most three character is a symbol. If more are
      -- read, Data_Error has to be raised.

      subtype Symbol_Range is Integer range 1 .. 3;
      Symbol: String (Symbol_Range);

      subtype Storable is Integer range 1 .. 2;
      Store : String renames Symbol (Storable);
      Stored: Natural := 0;

      -- Up to two characters might be unconsumed in Store.
      -- As many characters are consumed as follow the syntax of a unit
      -- name, e.g. "hPa_" is evaluated as Hecto-Pascal, the trailing
      -- (and incorrect) '_' stays in the input stream.

      function Read_Prefix return Real'Base is
        -- A string of letters is only interpreted as a prefix if it is
        -- followed by another letter (it is irrelevant whether that letter
        -- can match a unit name or not).
        -- Upon return, the prefix (if any) has been consumed; there may be
        -- up to two letters in Store, but only if these letters are the
        -- possible begin of a symbol (actually this can only occur for
        -- "da" deka when there is no letter following).
        Factor: Real'Base := 1.0;
        procedure All_Letters (Value    : in Real'Base;
                               Exclude  : in Character_Set := Null_Set) is
          -- All letters read so far form a prefix with Value or no prefix
          -- at all.
          C: Character;
        begin
          Stored := Stored + 1;             -- store previous
          Store (Stored) := Get_Character;  -- letter
          C := Look_Ahead;
          if Is_In (C, Letter_Set - Exclude) then
            Stored := 0;
            Factor := Value;
          end if;
        end All_Letters;
        procedure Keep_Last (Value    : in Real'Base;
                             Exclude  : in Character_Set) is
          -- All letters but the last read so far form a prefix with Value
          -- or no prefix at all.
          C: Character;
        begin
          C := Look_Ahead;
          if Is_In (C, Letter_Set - Exclude) then
            Stored := 0;
            Factor := Value;
          end if;
        end Keep_Last;
        C: Character;
      begin
        C := Look_Ahead;
        if C = LF then
          return Factor;
        end if;
        case C is
          when 'y' => All_Letters (yocto);
          when 'z' => All_Letters (zepto);
          when 'a' => All_Letters (atto);
          when 'f' => All_Letters (femto);
          when 'p' => All_Letters (pico);
          when 'n' => All_Letters (nano);
          when '' |
               'u' => All_Letters (micro);
          when 'm' => -- might also denote Minute or Mole
                      All_Letters (milli, Exclude => To_Set ("io"));
          when 'c' => -- might also denote Candela
                      All_Letters (centi, Exclude => To_Set ('d'));
          when 'd' => -- might also denote Dioptry
                      Stored := 1;                      -- store first
                      Store (Stored) := Get_Character;  -- letter
                      C := Look_Ahead;
                      if Is_In (C, Letter_Set) then
                        if C = 'a' then
                          All_Letters (deka);
                        else
                          Keep_Last (deci, Exclude => To_Set ('p'));
                        end if;
                      end if;
          when 'h' => All_Letters (hecto);
          when 'k' => -- might also denote Katal
                      All_Letters (kilo, Exclude => To_Set ('a'));
          when 'M' => All_Letters (mega);
          when 'G' => -- might also denote Gray
                      All_Letters (giga, Exclude => To_Set ('y'));
          when 'T' => All_Letters (tera);
          when 'P' => -- might also denote Pascal
                      All_Letters (peta, Exclude => To_Set ('a'));
          when 'E' => All_Letters (exa);
          when 'Z' => All_Letters (zetta);
          when 'Y' => All_Letters (yotta);
          when others => null;  -- C has not been consumed
        end case;
        return Factor;
      end Read_Prefix;

      procedure Read_Symbol_Characters is
        -- Remember, there may be up to two letters in Symbol, but only
        -- if these letters are the possible begin of a symbol.
        -- At return, when Data_Error is not raised, Symbol (1..Stored)
        -- hold the symbol letters read.
        C: Character;
        Too_long: Boolean := False;
      begin
        for S in Stored + 1 .. Symbol_Range'Last loop
          if Is_In (Look_Ahead, Letter_Set) then
            Symbol (S) := Get_Character;  -- consume
            Stored     := S;
          else
            exit;
          end if;
        end loop;
        while Is_In (Look_Ahead, Letter_Set) loop
          C := Get_Character;  -- consume and throw away
          Too_long := True;    -- (it's illegal)
        end loop;
        if Too_long then
          raise Data_Error;
        end if;
      end Read_Symbol_Characters;

      Prefix: Real'Base renames Read_Prefix;

    begin

      Read_Symbol_Characters;

      declare
        Found: String renames Symbol (1 .. Stored);
      begin
        if (Found = "h" or Found = "min") and Prefix /= 1.0 then
          raise Data_Error;
        end if;
        return Prefix * Search (Found);
      end;

    end Read_Symbol;

    --------------------------------------------------------------------

    function Read_Exponent return Rational is
      Exponent   : Rational;
      Parentheses: Boolean;
      C: Character;
    begin
      -- Is there an exponent?
      C := Look_Ahead;
      if C /= '^' then
        return +1;
      end if;
      C := Get_Character;  -- consume '^'
      -- Is there a parenthesis?
      C := Look_Ahead;
      if not Is_In (C, To_Set ("(+-") or Decimal_Digit_Set) then
        raise Data_Error;
      end if;
      Parentheses := C = '(';
      if Parentheses then
        C := Get_Character;  -- consume '('
      elsif C = '+' or C = '-' then
        raise Data_Error;
      end if;
      if Parentheses then
        Exponent := Get_Rational;
        C := Look_Ahead;
        if C /= ')' then
          raise Data_Error;
        end if;
        C := Get_Character;  -- consume ')'
      else  -- only whole number may follow (m^2/s^2)
        Exponent := +Get_Whole;
      end if;
      return Exponent;
    end Read_Exponent;

    Symbol  : Item     renames Read_Symbol  ;  -- force sequence
    Exponent: Rational renames Read_Exponent;  -- of evaluation

  begin

    return Symbol ** Exponent;

  end Read_Unit;

  C: Character;
  X: Item := One;

  use IO_Customization, Ada.Strings.Maps;

begin

  loop
    C := Look_Ahead;
    if not Is_In (C, Multiply) and C /= '/' then
      return X;
    end if;
    C := Get_Character;  -- consume '*' resp. '/'
    if C = '/' then
      X := X / Read_Unit;
    else  -- C = '*'
      X := X * Read_Unit;
    end if;
  end loop;

end Generic_Read_Dimension;
