------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

generic

package Unconstrained_Checked_SI.Generic_Vector_Space is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.2
  -- Date      17 November 2002
  --====================================================================
  -- This is not a linear algebraics package. All elements of a vector
  -- and a matrix should have the same dimension (not checked).
  --
  -- (The inner and the cross products are not defined as operators
  -- since they are not associative with themselves or with one another
  -- and having both as operators would lead to ambiguities or errors.)
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  17.09.2002
  --  C.G.    1.1  08.11.2002 Added Det, Inv, Tran
  --  C.G.    1.2  17.11.2002 Use Real'Base
  --====================================================================

  type Axis is range 1 .. 3;

  type Vector is array (Axis) of Item;

  Null_Vector: constant                 Vector :=  (Zero, Zero, Zero);
  Unit_Vector: constant array (Axis) of Vector := (( One, Zero, Zero),
                                                   (Zero,  One, Zero),
                                                   (Zero, Zero,  One));
  U1: Vector renames Unit_Vector (1);
  U2: Vector renames Unit_Vector (2);
  U3: Vector renames Unit_Vector (3);

  type Matrix is array (Axis, Axis) of Item;

  Null_Matrix: constant Matrix := ((Zero, Zero, Zero),
                                   (Zero, Zero, Zero),
                                   (Zero, Zero, Zero));
  Unity      : constant Matrix := (( One, Zero, Zero),
                                   (Zero,  One, Zero),
                                   (Zero, Zero,  One));

  -- Vector algebra

  function "abs" (Right: Vector) return Item;

  function "+" (Right: Vector) return Vector;
  function "-" (Right: Vector) return Vector;

  function "+" (Left, Right: Vector) return Vector;
  function "-" (Left, Right: Vector) return Vector;

  function "*" (Left: Real'Base; Right: Vector   ) return Vector;
  function "*" (Left: Item     ; Right: Vector   ) return Vector;
  function "*" (Left: Vector   ; Right: Real'Base) return Vector;
  function "*" (Left: Vector   ; Right: Item     ) return Vector;
  function "/" (Left: Vector   ; Right: Real'Base) return Vector;
  function "/" (Left: Vector   ; Right: Item     ) return Vector;

  function Inner (Left, Right: Vector) return Item;    -- inner product
  function Cross (Left, Right: Vector) return Vector;  -- cross product

  -- Matrix algebra

  function "+" (Right: Matrix) return Matrix;
  function "-" (Right: Matrix) return Matrix;

  function "+" (Left, Right: Matrix) return Matrix;
  function "-" (Left, Right: Matrix) return Matrix;

  function "*" (Left: Real'Base; Right: Matrix   ) return Matrix;
  function "*" (Left: Item     ; Right: Matrix   ) return Matrix;
  function "*" (Left: Matrix   ; Right: Real'Base) return Matrix;
  function "*" (Left: Matrix   ; Right: Item     ) return Matrix;
  function "/" (Left: Matrix   ; Right: Real'Base) return Matrix;
  function "/" (Left: Matrix   ; Right: Item     ) return Matrix;

  function "*" (Left, Right: Matrix) return Matrix;

  function Det  (M: Matrix) return Item;    -- determinant
  function Inv  (M: Matrix) return Matrix;  -- inverse
  function Tran (M: Matrix) return Matrix;  -- transposed

  -- Mixed operations

  function "*" (Left: Matrix; Right: Vector) return Vector;  -- Lij*Rj
  function "*" (Left: Vector; Right: Matrix) return Vector;  -- Li*Rij

end Unconstrained_Checked_SI.Generic_Vector_Space;