------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Numerics.Generic_Elementary_Functions;

package body Unconstrained_Checked_SI is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   2.3
  -- Date      21 February 2003
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  05.07.2002 with Gnat 3.14p work-around
  --  C.G.    2.0  15.09.2002 Of course the argument of trigonometric
  --                          functions with cycle can be dimensioned
  --  C.G.    2.1  27.09.2002 Gnat work-around removed for 3.16w
  --  C.G.    2.2  15.11.2002 Use Real'Base
  --  C.G.    2.3  21.02.2003 Add Dimension_as
  --====================================================================

  function Dimension_of (X: Item) return Dimension is
  begin
    return X.Unit;
  end Dimension_of;

  function Dimension_as (X: Item) return Item is
  begin
    return (X.Unit, 1.0);
  end Dimension_as;

  function Value (X: Item) return Real is
  begin
    return X.Value;
  end Value;

  package Elementary_Functions is new Ada.Numerics.Generic_Elementary_Functions (Real'Base);
  use     Elementary_Functions;

  -- Dimension arithmetic

  procedure Check (X, Y: in Dimension) is
    pragma Inline (Check);
  begin
    if X /= Y then
      raise Unit_Error;
    end if;
  end Check;

  function "-" (Right: Dimension) return Dimension is
    pragma Inline ("-");
  begin
    return (- Right.m,
            - Right.kg,
            - Right.s,
            - Right.A,
            - Right.K,
            - Right.cd,
            - Right.mol);
  end "-";

  function "*" (Left, Right: Dimension) return Dimension is
  begin
    return (Left.m   + Right.m,
            Left.kg  + Right.kg,
            Left.s   + Right.s,
            Left.A   + Right.A,
            Left.K   + Right.K,
            Left.cd  + Right.cd,
            Left.mol + Right.mol);
  end "*";

  function "/" (Left, Right: Dimension) return Dimension is
  begin
    return (Left.m   - Right.m,
            Left.kg  - Right.kg,
            Left.s   - Right.s,
            Left.A   - Right.A,
            Left.K   - Right.K,
            Left.cd  - Right.cd,
            Left.mol - Right.mol);
  end "/";

  function "**" (Base: Dimension; Exponent: Whole) return Dimension is
  begin
    return (Base.m   * Exponent,
            Base.kg  * Exponent,
            Base.s   * Exponent,
            Base.A   * Exponent,
            Base.K   * Exponent,
            Base.cd  * Exponent,
            Base.mol * Exponent);
  end "**";

  function "**" (Base: Dimension; Exponent: Rational) return Dimension is
  begin
    return (Base.m   * Exponent,
            Base.kg  * Exponent,
            Base.s   * Exponent,
            Base.A   * Exponent,
            Base.K   * Exponent,
            Base.cd  * Exponent,
            Base.mol * Exponent);
  end "**";

  -- Real arithmetic

  function to_Real (X: Rational) return Real'Base is
    pragma Inline (to_Real);
  begin
    return Real'Base (Numerator (X)) / Real'Base (Denominator (X));
  end to_Real;

  function "**" (Base: Real'Base; Exponent: Whole) return Real'Base is
    pragma Inline ("**");
  begin
    return Base ** Integer (Exponent);
  end "**";

  function "**" (Base: Real'Base; Exponent: Rational) return Real'Base is
    pragma Inline ("**");
  begin
    return Base ** to_Real (Exponent);
  end "**";

  -- Operators

  function "abs" (Right: Item) return Item is
  begin
    return (Right.Unit, abs Right.Value);
  end "abs";

  function "+" (Right: Item) return Item is
  begin
    return Right;
  end "+";

  function "-" (Right: Item) return Item is
  begin
    return (Right.Unit, -Right.Value);
  end "-";

  function "+" (Left, Right: Item) return Item is
    R: Item := Right;
  begin
    Check (Left.Unit, Right.Unit);
    R.Value := Left.Value + Right.Value;
    return R;
  end "+";

  function "-" (Left, Right: Item) return Item is
    R: Item := Right;
  begin
    Check (Left.Unit, Right.Unit);
    R.Value := Left.Value - Right.Value;
    return R;
  end "-";

  function "*" (Left, Right: Item) return Item is
  begin
    return (Left.Unit  * Right.Unit ,
            Left.Value * Right.Value);
  end "*";

  function "/" (Left, Right: Item) return Item is
  begin
    return (Left.Unit  / Right.Unit ,
            Left.Value / Right.Value);
  end "/";

  function "*" (Left: Item; Right: Real'Base) return Item is
    L: Item := Left;
  begin
    L.Value := Left.Value * Right;
    return L;
  end "*";

  function "*" (Left: Real'Base; Right: Item) return Item is
  begin
    return Right * Left;
  end "*";

  function "/" (Left: Item ; Right: Real'Base) return Item is
    L: Item := Left;
  begin
    L.Value := Left.Value / Right;
    return L;
  end "/";

  function "/" (Left: Real'Base; Right: Item) return Item is
  begin
    return (- Right.Unit,
            Left / Right.Value);
  end "/";

  function "**" (Base: Item; Exponent: Whole) return Item is
  begin
    return (Base.Unit  ** Exponent,
            Base.Value ** Exponent);
  end "**";

  function "**" (Base: Item; Exponent: Rational) return Item is
  begin
    return (Base.Unit  ** Exponent,
            Base.Value ** Exponent);
  end "**";

  function "**" (Base: Item; Exponent: Real'Base) return Item is
  begin
    Check (Base.Unit, None);
    return (None,
            Base.Value ** Exponent);
  end "**";

  function "<" (Left, Right: Item) return Boolean is
  begin
    Check (Left.Unit, Right.Unit);
    return Left.Value < Right.Value;
  end "<";

  function "<=" (Left, Right: Item) return Boolean is
  begin
    Check (Left.Unit, Right.Unit);
    return Left.Value <= Right.Value;
  end "<=";

  function ">=" (Left, Right: Item) return Boolean is
  begin
    Check (Left.Unit, Right.Unit);
    return Left.Value >= Right.Value;
  end ">=";

  function ">" (Left, Right: Item) return Boolean is
  begin
    Check (Left.Unit, Right.Unit);
    return Left.Value > Right.Value;
  end ">";

  -- Mathematics

  function Sqrt (X: Item) return Item is
  begin
    return (X.Unit ** (1/2),
            SQRT (X.Value));
  end Sqrt;

  function Cbrt (X: Item) return Item is
  begin
    return (X.Unit  ** (1/3),
            X.Value ** to_Real (1/3));
  end Cbrt;

  function Log (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Log (X.Value));
  end Log;

  function Log (X, Base: Item) return Item is
  begin
    Check (X   .Unit, None);
    Check (Base.Unit, None);
    return (None,
            Log (X.Value, Base.Value));
  end Log;

  function Exp (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Exp (X.Value));
  end Exp;

  function Sin (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Sin (X.Value));
  end Sin;

  function Sin (X, Cycle: Item) return Item is
  begin
    Check (X    .Unit, Cycle.Unit);
    return (None,
            Sin (X.Value, Cycle.Value));
  end Sin;

  function Cos (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Cos (X.Value));
  end Cos;

  function Cos (X, Cycle: Item) return Item is
  begin
    Check (X    .Unit, Cycle.Unit);
    return (None,
            Cos (X.Value, Cycle.Value));
  end Cos;

  function Tan (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Tan (X.Value));
  end Tan;

  function Tan (X, Cycle: Item) return Item is
  begin
    Check (X    .Unit, Cycle.Unit);
    return (None,
            Tan (X.Value, Cycle.Value));
  end Tan;

  function Cot (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Cot (X.Value));
  end Cot;

  function Cot (X, Cycle: Item) return Item is
  begin
    Check (X    .Unit, Cycle.Unit);
    return (None,
            Cot (X.Value, Cycle.Value));
  end Cot;

  function Arcsin (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arcsin (X.Value));
  end Arcsin;

  function Arcsin (X, Cycle: Item) return Item is
  begin
    Check (X.Unit, None);
    return (Cycle.Unit,
            Arcsin (X.Value, Cycle.Value));
  end Arcsin;

  function Arccos (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arccos (X.Value));
  end Arccos;

  function Arccos (X, Cycle: Item) return Item is
  begin
    Check (X.Unit, None);
    return (Cycle.Unit,
            Arccos (X.Value, Cycle.Value));
  end Arccos;

  function Arctan (Y: Item;
                   X: Item := One) return Item is
  begin
    Check (X.Unit, Y.Unit);
    return (None,
            Arctan (Y.Value, X.Value));
  end Arctan;

  function Arctan (Y    : Item;
                   X    : Item := One;
                   Cycle: Item) return Item is
  begin
    Check (X.Unit, Y.Unit);
    return (Cycle.Unit,
            Arctan (Y.Value, X.Value, Cycle.Value));
  end Arctan;

  function Arccot (X: Item;
                   Y: Item := One) return Item is
  begin
    Check (X.Unit, Y.Unit);
    return (None,
            Arccot (X.Value, Y.Value));
  end Arccot;

  function Arccot (X    : Item;
                   Y    : Item := One;
                   Cycle: Item) return Item is
  begin
    Check (X.Unit, Y.Unit);
    return (Cycle.Unit,
            Arccot (X.Value, Y.Value, Cycle.Value));
  end Arccot;

  function Sinh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Sinh (X.Value));
  end Sinh;

  function Cosh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Cosh (X.Value));
  end Cosh;

  function Tanh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Tanh (X.Value));
  end Tanh;

  function Coth (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Coth (X.Value));
  end Coth;

  function Arcsinh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arcsinh (X.Value));
  end Arcsinh;

  function Arccosh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arccosh (X.Value));
  end Arccosh;

  function Arctanh (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arctanh (X.Value));
  end Arctanh;

  function Arccoth (X: Item) return Item is
  begin
    Check (X.Unit, None);
    return (None,
            Arccoth (X.Value));
  end Arccoth;

end Unconstrained_Checked_SI;
