------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Characters.Latin_1, Ada.Strings.Maps, Ada.Strings.Fixed;
with IO_Customization;

with Rational_Arithmetics.Text_IO;

package body Unconstrained_Unchecked_SI.Generic_Text_IO is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.5
  -- Date      12 March 2003
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  25.07.2002 Done
  --  C.G.    1.1  29.07.2002 Also offending: unit^+1 unit^1/3
  --  C.G.    1.2  03.08.2002 Allow '/' when reading dimension
  --  C.G.    1.3  03.09.2002 Dim format modifier for Put added
  --  C.G.    1.4  15.11.2002 Use Real'Base
  --  C.G.    1.5  12.03.2003 IO customization
  --====================================================================

  LF: Character renames Ada.Characters.Latin_1.LF;

  generic
    with function Get_Character return Character is <>;
    with function Look_Ahead    return Character is <>;
    with function Get_Whole     return Whole     is <>;
    with function Get_Rational  return Rational  is <>;
  function Generic_Read_Dimension return Item;
  function Generic_Read_Dimension return Item is separate;

  ------------------------------------------------------------------------

  package Whole_Text_IO is new Ada.Text_IO.Integer_IO (Whole);
  package Real_Text_IO  is new Ada.Text_IO.Float_IO   (Real'Base);
  use Whole_Text_IO, Real_Text_IO;

  procedure Get (X: out Item) is
  begin
    Get (Current_Input, X);
  end Get;

  procedure Get (File: in File_Type; X: out Item) is
    function Get_Character return Character is
      C: Character;
    begin
      Get (File, C);
      return C;
    end Get_Character;
    function Look_Ahead return Character is
      C  : Character;
      EoL: Boolean;
    begin
      Look_Ahead (File, C, EoL);
      if EoL then
        return LF;
      else
        return C;
      end if;
    end Look_Ahead;
    function Get_Whole return Whole is
      W: Whole;
    begin
      Get (File, W);
      return W;
    end Get_Whole;
    function Get_Rational return Rational is
      R: Rational;
    begin
      Rational_Arithmetics.Text_IO.Get (File, R);
      return R;
    end Get_Rational;
    function Read_Dimension is new Generic_Read_Dimension;
    Value: Real'Base;
  begin
    Get (File, Value);
    X := Value * Read_Dimension;
  end Get;

  procedure Put (X   : in Item;
                 Fore: in Field  := Default_Fore;
                 Aft : in Field  := Default_Aft;
                 Exp : in Field  := Default_Exp;
                 Dim : in String := "") is
  begin
    Put (Current_Output, X, Fore, Aft, Exp, Dim);
  end Put;

  procedure Put (File: in File_Type;
                 X   : in Item;
                 Fore: in Field  := Default_Fore;
                 Aft : in Field  := Default_Aft;
                 Exp : in Field  := Default_Exp;
                 Dim : in String := "") is
    use IO_Customization, Ada.Strings.Maps;
  begin
    if Dim = "" then
      Put (File, X.Value, Fore, Aft, Exp);
    elsif not Is_In (Dim (Dim'First), Multiply) and Dim (Dim'First) /= '/' then
      Put (File, X, Fore, Aft, Exp, Default_Multiply & Dim);
    else
      Dimension_Format:
      declare
        P: Natural := Dim'First - 1;  -- last consumed character
        function Get_Character return Character is
        begin
          P := P + 1;
          return Dim (P);
        end Get_Character;
        function Look_Ahead return Character is
        begin
          if P = Dim'Last then
            return LF;
          else
            return Dim (P + 1);
          end if;
        end Look_Ahead;
        function Get_Whole return Whole is
          B: Positive := P + 1;
          T: Positive := B + 1;
        begin
          P := B;
          while T <= Dim'Last and then Dim (T) in '0' .. '9' loop
            P := T;
            T := T + 1;
          end loop;
          return Whole'Value (Dim (B .. P));
        end Get_Whole;
        function Get_Rational return Rational is
          N: Whole renames Get_Whole;
          D: Positive := P + 1;
        begin
          if D <= Dim'Last and then Dim (D) = '/' then
            P := D;
            return N/Get_Whole;
          else
            return +N;
          end if;
        end Get_Rational;
        function Read_Dimension is new Generic_Read_Dimension;
        Req_Dim: Item;
        use Ada.Strings.Fixed;
      begin  -- Dimension_Format
        Req_Dim := Read_Dimension;  -- must be called here because of exception handler below
        P := P + 1;
        while P <= Dim'Last loop  -- only blanks may trail
          if Dim (P) /= ' ' then
            raise Unit_Error;
          end if;
          P := P + 1;
        end loop;
        Put (File, X.Value / Req_Dim.Value, Fore, Aft, Exp);
        Put (Translate (Dim, To_Mapping (From => To_Sequence (Multiply),
                                         To   => 2 * Default_Multiply)));
      exception
        when Data_Error => raise Unit_Error;
      end Dimension_Format;
    end if;
  end Put;

end Unconstrained_Unchecked_SI.Generic_Text_IO;
