------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002, 2003 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

generic
package Unconstrained_Unchecked_SI.Generic_Units is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.2
  -- Date      4 February 2003
  --====================================================================
  -- Derived units (liter, minute and hour have been added as very
  -- important non-SI units).
  --
  -- For more information see:
  --   National Institute of Standards and Technology
  --   <http://physics.nist.gov/cuu/Units/>
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  24.07.2002
  --  C.G.    1.1  14.09.2002 Added Katal and reference to NIST
  --  C.G.    1.2  04.02.2003 Minute, Hour, Liter
  --====================================================================

  Meter_2: constant Item := Meter**2;                       -- m**2
  Meter_3: constant Item := Meter**3;                       -- m**3

  p_Meter: constant Item := 1.0/Meter;                      -- 1/m
  Dioptry:          Item renames p_Meter;                   -- dpt = 1/m

  Meter_p_Second    : constant Item :=  Meter/Second;       -- m/s
  Meter_2_p_Second_2: constant Item := (Meter/Second)**2;   -- (m/s)**2
  Meter_p_Second_2  : constant Item :=  Meter/Second**2;    -- m/s**2

  p_Second : constant Item := 1.0/Second;
  Hertz    :          Item renames p_Second;                -- Hz = 1/s
  Becquerel:          Item renames p_Second;                -- Bq = 1/s

  Newton      : constant Item := Kilogram*Meter/Second**2;  -- N = m*kg/s**2
  Joule       : constant Item := Newton*Meter;              -- J = Nm = Ws = VC
  Newton_Meter:          Item renames Joule;                -- Nm = J
  Watt_Second :          Item renames Joule;                -- Ws = J

  Pascal: constant Item := Newton/Meter**2;                 -- Pa = N/m**2

  Watt       : constant Item := Joule/Second;               -- W = J/s = VA
  Volt       : constant Item := Watt/Ampere;                -- V = W/A
  Volt_Ampere:          Item renames Watt;                  -- VA = W

  Coulomb      : constant Item := Ampere*Second;            -- C = As
  Ampere_Second:          Item renames Coulomb;             -- As = C

  Volt_p_Meter: constant Item := Volt / Meter;              -- V/m

  Ohm    : constant Item := Volt/Ampere;                    -- Ohm = V/A = 1/S
  Siemens: constant Item := 1.0/Ohm;                        -- S = 1/Ohm

  Farad: constant Item := Coulomb/Volt;                     -- F = C/V
  Henry: constant Item := Volt*Second/Ampere;               -- H = Vs/A
  Weber: constant Item := Volt*Second;                      -- Wb = Vs
  Tesla: constant Item := Weber/Meter**2;                   -- T = Wb/m**2

  Gray   : constant Item := Joule/Kilogram;                 -- Gy = J/kg
  Sievert:          Item renames Gray;                      -- Sv = Gy

  Lumen: constant Item := Candela/Steradian;                -- lm = cd/sr
  Lux  : constant Item := Lumen/Meter**2;                   -- lx = lm/m**2

  Katal: constant Item := Mole/Second;                      -- kat = mol/s

  -- Important no-SI units

  Minute: constant Item := 60.0*Second;  -- min (may not be prefixed)
  Hour  : constant Item := 60.0*Minute;  -- h   (may not be prefixed)

  Liter: constant Item := (Deci*Meter)**3;  -- l, L (may be prefixed like ml, hl)

end Unconstrained_Unchecked_SI.Generic_Units;
