------------------------------------------------------------------------------
-- Checked and Unchecked Computation with SI Units
-- Copyright (C) 2002 Christoph Karl Walter Grein
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--
-- As a special exception,  if other files  instantiate  generics from this
-- unit, or you link this unit with other files to produce an executable,
-- this unit does not by itself cause the resulting executable to be
-- covered by the GNU General Public License.  This exception does not
-- however invalidate any other reasons why the executable file might be
-- covered by the GNU Public License.
--
-- Author's homepage and email address:
--   http://home.T-Online.de/home/Christ-Usch.Grein/
--   Christ-Usch.Grein@T-Online.de
------------------------------------------------------------------------------

with Ada.Numerics.Generic_Elementary_Functions;

package body Unconstrained_Unchecked_SI is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      15 November 2002
  --====================================================================
  --
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  24.07.2002
  --  C.G.    1.1  15.11.2002 Use Real'Base
  --====================================================================

  function Value (X: Item) return Real is
  begin
    return X.Value;
  end Value;

  package Elementary_Functions is new Ada.Numerics.Generic_Elementary_Functions (Real'Base);
  use     Elementary_Functions;

  -- Real arithmetic

  function to_Real (X: Rational) return Real'Base is
    pragma Inline (to_Real);
  begin
    return Real'Base (Numerator (X)) / Real'Base (Denominator (X));
  end to_Real;

  function "**" (Base: Real'Base; Exponent: Whole) return Real'Base is
    pragma Inline ("**");
  begin
    return Base ** Integer (Exponent);
  end "**";

  function "**" (Base: Real'Base; Exponent: Rational) return Real'Base is
    pragma Inline ("**");
  begin
    return Base ** to_Real (Exponent);
  end "**";

  -- Operators

  function "abs" (Right: Item) return Item is
  begin
    return (Value => abs Right.Value);
  end "abs";

  function "+" (Right: Item) return Item is
  begin
    return Right;
  end "+";

  function "-" (Right: Item) return Item is
  begin
    return (Value => -Right.Value);
  end "-";

  function "+" (Left, Right: Item) return Item is
  begin
    return (Value => Left.Value + Right.Value);
  end "+";

  function "-" (Left, Right: Item) return Item is
  begin
    return (Value => Left.Value - Right.Value);
  end "-";

  function "*" (Left, Right: Item) return Item is
  begin
    return (Value => Left.Value * Right.Value);
  end "*";

  function "/" (Left, Right: Item) return Item is
  begin
    return (Value => Left.Value / Right.Value);
  end "/";

  function "*" (Left: Item; Right: Real'Base) return Item is
  begin
    return (Value => Left.Value * Right);
  end "*";

  function "*" (Left: Real'Base; Right: Item) return Item is
  begin
    return Right * Left;
  end "*";

  function "/" (Left: Item ; Right: Real'Base) return Item is
  begin
    return (Value => Left.Value / Right);
  end "/";

  function "/" (Left: Real'Base; Right: Item) return Item is
  begin
    return (Value => Left / Right.Value);
  end "/";

  function "**" (Base: Item; Exponent: Whole) return Item is
  begin
    return (Value => Base.Value ** Exponent);
  end "**";

  function "**" (Base: Item; Exponent: Rational) return Item is
  begin
    return (Value => Base.Value ** Exponent);
  end "**";

  function "**" (Base: Item; Exponent: Real'Base) return Item is
  begin
    return (Value => Base.Value ** Exponent);
  end "**";

  function "<" (Left, Right: Item) return Boolean is
  begin
    return Left.Value < Right.Value;
  end "<";

  function "<=" (Left, Right: Item) return Boolean is
  begin
    return Left.Value <= Right.Value;
  end "<=";

  function ">=" (Left, Right: Item) return Boolean is
  begin
    return Left.Value >= Right.Value;
  end ">=";

  function ">" (Left, Right: Item) return Boolean is
  begin
    return Left.Value > Right.Value;
  end ">";

  -- Mathematics

  function Sqrt (X: Item) return Item is
  begin
    return (Value => SQRT (X.Value));
  end Sqrt;

  function Cbrt (X: Item) return Item is
  begin
    return (Value => X.Value ** to_Real (1/3));
  end Cbrt;

  function Log (X: Item) return Item is
  begin
    return (Value => Log (X.Value));
  end Log;

  function Log (X, Base: Item) return Item is
  begin
    return (Value => Log (X.Value, Base.Value));
  end Log;

  function Exp (X: Item) return Item is
  begin
    return (Value => Exp (X.Value));
  end Exp;

  function Sin (X: Item) return Item is
  begin
    return (Value => Sin (X.Value));
  end Sin;

  function Sin (X, Cycle: Item) return Item is
  begin
    return (Value => Sin (X.Value, Cycle.Value));
  end Sin;

  function Cos (X: Item) return Item is
  begin
    return (Value => Cos (X.Value));
  end Cos;

  function Cos (X, Cycle: Item) return Item is
  begin
    return (Value => Cos (X.Value, Cycle.Value));
  end Cos;

  function Tan (X: Item) return Item is
  begin
    return (Value => Tan (X.Value));
  end Tan;

  function Tan (X, Cycle: Item) return Item is
  begin
    return (Value => Tan (X.Value, Cycle.Value));
  end Tan;

  function Cot (X: Item) return Item is
  begin
    return (Value => Cot (X.Value));
  end Cot;

  function Cot (X, Cycle: Item) return Item is
  begin
    return (Value => Cot (X.Value, Cycle.Value));
  end Cot;

  function Arcsin (X: Item) return Item is
  begin
    return (Value => Arcsin (X.Value));
  end Arcsin;

  function Arcsin (X, Cycle: Item) return Item is
  begin
    return (Value => Arcsin (X.Value, Cycle.Value));
  end Arcsin;

  function Arccos (X: Item) return Item is
  begin
    return (Value => Arccos (X.Value));
  end Arccos;

  function Arccos (X, Cycle: Item) return Item is
  begin
    return (Value => Arccos (X.Value, Cycle.Value));
  end Arccos;

  function Arctan (Y: Item;
                   X: Item := One) return Item is
  begin
    return (Value => Arctan (Y.Value, X.Value));
  end Arctan;

  function Arctan (Y    : Item;
                   X    : Item := One;
                   Cycle: Item) return Item is
  begin
    return (Value => Arctan (Y.Value, X.Value, Cycle.Value));
  end Arctan;

  function Arccot (X: Item;
                   Y: Item := One) return Item is
  begin
    return (Value => Arccot (X.Value, Y.Value));
  end Arccot;

  function Arccot (X    : Item;
                   Y    : Item := One;
                   Cycle: Item) return Item is
  begin
    return (Value => Arccot (X.Value, Y.Value, Cycle.Value));
  end Arccot;

  function Sinh (X: Item) return Item is
  begin
    return (Value => Sinh (X.Value));
  end Sinh;

  function Cosh (X: Item) return Item is
  begin
    return (Value => Cosh (X.Value));
  end Cosh;

  function Tanh (X: Item) return Item is
  begin
    return (Value => Tanh (X.Value));
  end Tanh;

  function Coth (X: Item) return Item is
  begin
    return (Value => Coth (X.Value));
  end Coth;

  function Arcsinh (X: Item) return Item is
  begin
    return (Value => Arcsinh (X.Value));
  end Arcsinh;

  function Arccosh (X: Item) return Item is
  begin
    return (Value => Arccosh (X.Value));
  end Arccosh;

  function Arctanh (X: Item) return Item is
  begin
    return (Value => Arctanh (X.Value));
  end Arctanh;

  function Arccoth (X: Item) return Item is
  begin
    return (Value => Arccoth (X.Value));
  end Arccoth;

end Unconstrained_Unchecked_SI;
