--                              -*- Mode: Ada -*-
-- Filename        : math-generics-arrays-matrix_operations.ads
-- Description     : generic matrix operations
-- Author          :  J. Schrer
-- Created On      : Wed May 21 19:04:37 1997
-- Last Modified By:
-- Last Modified On: Thu May 22 10:35:40 1997
-- Update Count    : 2
-- Status          :
--
-- Remarks: Only simple operations that are similar for all
--          2-dimensional array types are gathered here. All more
--          specific algorithms as matrix inversion are located in
--          different packages.
--          There are other generic packages for other array types as:
--
--          arrays.vector_operations
--          arrays.matrix_vector_operations
--          arrays.diagonal_matrix_operations
--          arrays.diagonal_matrix_vector_operations
--          arrays.diagonal_matrix_matrix_operations
--
--          One may instantiate these packages for integer, float,
--          complex, boolean, symbolic expression (e.g. implemented
--          as a binary tree) or other types.
-------------------------------------------------------------------------

generic
  type Element is private;
  type Index   is range <>;
  type Matrix  is array (Index range <>, Index range <>) of Element;

  Zero, One : in Element;

  with function "+" (Left, Right : in Element) return Element is <>;
  with function "-" (Right       : in Element) return Element is <>;
  with function "-" (Left, Right : in Element) return Element is <>;
  with function "*" (Left, Right : in Element) return Element is <>;
  with function "/" (Left, Right : in Element) return Element is <>;

package Math.Generics.Arrays.Matrix_Operations is

  Matrix_Dimension_Error : exception;
  Singular_Matrix        : exception;
  
  -- Normal operators; raise matrix_dimension_error if left and right are
  -- incompatible for the operation.

  function "+" (Left, Right : in Matrix) return Matrix;
  function "-" (Right       : in Matrix) return Matrix;
  function "-" (Left, Right : in Matrix) return Matrix;
  function "*" (Left, Right : in Matrix) return Matrix;

  function "*" (Left : in Element; Right : in Matrix)  return Matrix;
  function "*" (Left : in Matrix;  Right : in Element) return Matrix;
  function "/" (Left : in Matrix;  Right : in Element) return Matrix;

  function Submatrix(The_Matrix : in Matrix;
                     Row, Col   : in Index) return Matrix;

  -- R = submatrix(The_Matrix, row, col) = The_Matrix(i, j), i /= row, j /= col

  function Adjoint  (The_Matrix : in Matrix;
                     Row, Col   : in Index) return Element;

  -- e = adjoint(the_Matrix, row, col)
  --   = (-1) ** (row + col) * abs submatrix(the_matrix, row, col)

  function "abs"     (Right      : in Matrix) return Element; 
  
  -- Determinant

  function Inverse   (The_Matrix : in Matrix) return Matrix;
  
  -- Raises singular_matrix if the_matrix is singular.

  function Transpose (The_Matrix : in Matrix) return Matrix;
  
  -- R(I,J) = The_Matrix(J,I)

  function Elementwise_Multiplication(Left, Right : in Matrix) return Matrix;
  
  -- R(I,J) = Left(I,J) * Right(I,J)

  function  Copy (Outoff          : in Matrix;
                  From_First_Row,
                  To_Last_Row,
                  From_First_Col,
                  To_Last_Col     : in Index;
                  Multiply_With   : in Element := One) return Matrix;
                  
  --  B : Matrix := Copy(Outoff         => A,
  --                     From_First_Row => I1,
  --                     To_Last_Row    => I2, 
  --                     From_First_Col => J1,
  --                     To_Last_Col    => J2,
  --                     Multiply_With  => X); 
  -- 
  -- copies following slice of A into B                                  
  --  
  --  B = A(I1 .. I2, J1 .. J2) * X                  

  procedure Copy (Into           : in out Matrix;
                  Source         : in     Matrix;
                  First_Row,
                  First_Col      : in     Index;
                  Multiply_With  : in     Element := One);
                  
  --  Copy(Into          => A,
  --       Source        => B,
  --       First_Row     => I,
  --       First_Col     => J,
  --       Multiply_With => X); 
  --  
  -- copies B * X into A
  --  
  --  A(I .. I + B'Length(1) - 1, J .. J + B'Length(2) - 1) := B * X;

  procedure Copy (Into           : in out Matrix;
                  Outoff         : in     Matrix;
                  From_First_Row,
                  To_Last_Row,
                  From_First_Col,
                  To_Last_Col,
                  First_Row,
                  First_Col      : in     Index;
                  Multiply_With  : in     Element := One);

  --  Copy(Into           => A,
  --       Outoff         => B,
  --       From_First_Row => I1,
  --       To_Last_Row    => I2, 
  --       From_First_Col => J1,
  --       To_Last_Col    => J2, 
  --       First_Row      => I,
  --       First_Col      => J,
  --       Multiply_With  => X);
  --  
  --  copies B(I1 .. I2, J1 .. J2) * X into A
  --  
  --  A(I .. I + I2 - I1, J .. J + J2 - J1) := B(I1 .. I2, J1 .. J2) * X;
                  
  procedure Accumulate (In_Matrix      : in out Matrix;
                        Source         : in     Matrix;
                        First_Row,
                        First_Col      : in     Index;
                        Positive_Accu  : in     Boolean := True;
                        Multiply_With  : in     Element := One);

  --  Accumulate(In_Matrix      => A,
  --             Source         => B,
  --             First_Row      => I,
  --             First_Col      => J,
  --             Positive_Accu  => True,
  --             Multiply_With  => X);
  --  
  --  Adds B * X to slice of A
  --  
  --  A(I .. I + B'Length(1) - 1, J .. J + B'Length(2) - 1) := 
  --    A(I .. I + B'Length(1) - 1, J .. J + B'Length(2) - 1) + B * X;
                      
  procedure Accumulate (In_Matrix      : in out Matrix;
                        Outoff         : in     Matrix;
                        From_First_Row,
                        To_Last_Row,
                        From_First_Col,
                        To_Last_Col,
                        First_Row,
                        First_Col      : in     Index;
                        Positive_Accu  : in     Boolean := True;
                        Multiply_With  : in     Element := One);

  --  Accumulate(In_Matrix      => A,
  --             Outoff         => B,
  --             From_First_Row => I1,
  --             To_Last_Row    => I2, 
  --             From_First_Col => J1,
  --             To_Last_Col    => J2, 
  --             First_Row      => I,
  --             First_Col      => J,
  --             Positive_Accu  => False,
  --             Multiply_With  => X);
  --  
  --  Subtracts slice B(I1 .. I2, J1 .. J2) * X from part of A
  --  
  --  A(I .. I + I2 - I1, J .. J + J2 - J1) := 
  --     A( I.. I + I2 - I1, J .. J + J2 - J1) - B(I1 .. I2, J1 .. J2) * X;

  procedure Zero_Matrix     (The_Matrix : out Matrix);
  
  -- The_Matrix := (others => (others => Zero));

  function  Zero_Matrix     (Of_Order   : in Index) return Matrix;

  -- The_Matrix(1 .. Of_Order, 1 .. Of_Order) := (others => (others => Zero));
  
  procedure Identity_Matrix (The_Matrix : out Matrix);
  
  -- The_Matrix(I,I) := One; 
  -- The_Matrix(I,J) := Zero; I /= J

  function  Identity_Matrix (Of_Order   : in Index) return Matrix;

  -- The_Matrix(I,I) := One; 
  -- The_Matrix(I,J) := Zero; I /= J
  
end Math.Generics.Arrays.Matrix_Operations;
