--                              -*- Mode: Ada -*-
-- Filename        : math-generics-arrays-matrix_vector_operations.ads
-- Description     : Mixed matrix vector operations
-- Author          : J. Schrer
-- Created On      : Wed May 21 19:06:47 1997
-- Last Modified By:
-- Last Modified On: Thu May 22 10:35:39 1997
-- Update Count    : 2
-- Status          :

generic
  type Element    is private;
  type Index      is range <>;
  type Vector     is array (Index range <>) of Element;
  type Row_Vector is new Vector;
  type Matrix     is array (Index range <>, Index range <>) of Element;

  Zero, One : in Element;

  with function "+" (Left, Right : in Element) return Element is <>;
  with function "*" (Left, Right : in Element) return Element is <>;

package Math.Generics.Arrays.Matrix_Vector_Operations is

  Matrix_Vector_Dimension_Error : exception;

  -- Normal operators; raise Matrix_Vector_Dimension_Error if left and right
  -- are incompatible for the operation.
  
  function "*" (Left : in Matrix;     Right : in Vector)     return Vector;
  function "*" (Left : in Row_Vector; Right : in Matrix)     return Row_Vector;
  
  function "*" (Left : in Vector;     Right : in Row_Vector) return Matrix;

  -- Dyadic product: 
  --
  -- |X1| * |Y1 Y2| = |X1*Y1 X1*Y2|
  -- |X2|             |X2*Y1 X2*Y2|
  
  function Copy_Row (Outoff        : in Matrix;
                     From_First,
                     To_Last,
                     Of_Row        : in Index;
                     Multiply_With : in Element := One) return Row_Vector;

  --  V : Vector := Copy_Row(Outoff        => A,
  --                         From_First    => J1,
  --                         To_Last       => J2, 
  --                         Of_Row        => I,
  --                         Multiply_With => X);
  --  
  --  assigns row I of matrix A multiplied by X to vector V
  --  
  --  V := A(I, J1 .. J2) * X;

  function Copy_Col (Outoff        : in Matrix;
                     From_First,
                     To_Last,
                     Of_Col        : in Index;
                     Multiply_With : in Element := One) return Vector;

  --  V : Vector := Copy_Col(Outoff        => A,
  --                         From_First    => I1,
  --                         To_Last       => I2, 
  --                         Of_Col        => J,
  --                         Multiply_With => X);
  --  
  --  assigns col J of matrix A multiplied by X to vector V
  --  
  --  V := A(I1 .. I2, J) * X;

  function Copy_Row (Outoff        : in Matrix;
                     Of_Row        : in Index;
                     Multiply_With : in Element := One) return Row_Vector;

  --  V : Vector := Copy_Row(Outoff        => A,
  --                         Of_Row        => I,
  --                         Multiply_With => X);
  --  
  --  assigns row I of matrix A multiplied by X to vector V
  --  
  --  V := A(I, A'range(2)) * X;

  function Copy_Col (Outoff        : in Matrix;
                     Of_Col        : in Index;
                     Multiply_With : in Element := One) return Vector;

  --  V : Vector := Copy_Col(Outoff        => A,
  --                         Of_Col        => J,
  --                         Multiply_With => X);
  --  
  --  assigns col J of matrix A multiplied by X to vector V
  --  
  --  V := A(A'range(1), J) * X;

end Math.Generics.Arrays.Matrix_Vector_Operations;
