--                              -*- Mode: Ada -*-
-- Filename        : utilities-calendar_utilities.ads
-- Description     : Taken from "Software components with Ada" by Grady Booch
-- Author          : J. Schroeer
-- Created On      : Wed Mar 26 16:55:03 1997
-- Last Modified By:
-- Last Modified On: Wed Jul  2 13:37:09 1997
-- Update Count    : 3
-- Status          :

with Portable.Standard; use Portable.Standard;

with Ada.Calendar;

package Utilities.Calendar_Utilities is

  subtype Year        is Ada.Calendar.Year_Number;
  subtype Month       is Positive range 1 .. 12;
  subtype Day         is Positive range 1 .. 31;
  subtype Hour        is Natural  range 0 .. 23;
  subtype Minute      is Natural  range 0 .. 59;
  subtype Second      is Natural  range 0 .. 59;
  subtype Millisecond is Natural  range 0 .. 999;

  type Month_Name is (January, February, March,
                      April,   May,      June,
                      July,    August,   September,
                      October, November, December);

  type Day_Name is (Monday,   Tuesday, Wednesday,
                    Thursday, Friday,  Saturday, Sunday);

  type Time is
    record
      The_Weekday     : Day_Name    := Day_Name'First;
      The_Year        : Year        := Year'First;
      The_Month       : Month       := Month'First;
      The_Day         : Day         := Day'First;
      The_Hour        : Hour        := Hour'First;
      The_Minute      : Minute      := Minute'First;
      The_Second      : Second      := Second'First;
      The_Millisecond : Millisecond := Millisecond'First;
    end record;

  type Interval is
    record
      Elapsed_Days         : Natural     := 0;
      Elapsed_Hours        : Hour        := Hour'First;
      Elapsed_Minutes      : Minute      := Minute'First;
      Elapsed_Seconds      : Second      := Second'First;
      Elapsed_Milliseconds : Millisecond := Millisecond'First;
    end record;

  subtype Year_Day is Positive range 1 .. 366;

  type Date_Format is (Full,            -- june 30, 1988
                       Month_Day_Year); -- 06/30/88

  function Is_Leap_Year  (The_Year : in Year) return Boolean;

  function Days_In  (The_Year  : in Year) return Year_Day;

  function Days_In  (The_Month : in Month;
                     The_Year  : in Year)       return Day;

  function Year_Day_Of (The_Day   : in Day;
                        The_Month : in Month;
                        The_Year  : in Year) return Year_Day;
                          
  function Month_Of (The_Month : in Month)      return Month_Name;

  function Month_Of (The_Month : in Month_Name) return Month;

  function Day_Of   (The_Year  : in Year;
                     The_Day   : in Year_Day) return Day_Name;

  function Day_Of   (The_Time  : in Time)     return Year_Day;

  function Time_Of  (The_Year  : in Year;
                     The_Day   : in Year_Day) return Time;

  function Time_Of  (The_Time  : in Time)     return Ada.Calendar.Time;

  function Time_Of  (The_Time  : in Ada.Calendar.Time) return Time;

  function Time_Image_Of (The_Time : in Time;
                          Hundreds : in Boolean := True)   return String;

  function Time_Image_Of (The_Time : in Ada.Calendar.Time;
                          Hundreds : in Boolean := True)   return String;
                          
  -- The time image of a variable of type time delivers a string with the
  -- constant number of 11 characters of the following format.
  -- 12345678901
  -- Hundreds = true                        
  -- hh:mm:ss:hh
  -- 16:13:08:00
  -- Hundreds = false                        
  -- hh:mm:ss
  -- 16:13:08

  function Date_Image_Of (The_Time  : in Time;
                          Date_Form : in Date_Format := Full) return String;

  -- Date_Format = full:
  -- The time image of a variable of type time delivers a string with the
  -- constant number of 38 characters of the following format.
  -- 1234567890123456789012345678
  -- wwwwwwwww mmmmmmmmm dd, yyyy
  -- FRIDAY    JUNE      23, 1995

  -- Date_Format = Month_Day_Year:
  -- The time image of a variable of type time delivers a string with the
  -- constant number of 38 characters of the following format.
  -- 12345678
  -- mm/dd/yy
  -- 06/23/95

  function Value_Of    (The_Date     : in String;
                        The_Time     : in String;
                        Date_Form    : in Date_Format := Full) return Time;

  -- The function value_of needs strings of the above formats.
  -- Otherwise exception lexical_error is raised.

  function Image_Of    (The_Interval : in Interval;
                        Milliseconds : in Boolean := True;
                        Days         : in Boolean := True) return String;
                        
  function Image_Of    (The_Interval : in Duration;
                        Milliseconds : in Boolean := True;
                        Days         : in Boolean := True) return String;

  -- The intervall image of a variable of type time delivers a string with the
  -- constant number of 15 characters of the following format.
  -- 123456789012345
  -- dd:hh:mm:ss:mmm
  -- 12:08:35:10:123
  -- If days = false, days is only displayed, when it is > 0.                        

  function Value_Of    (The_Interval : in String)   return Interval;

  -- The function value_of needs a string of the above format.
  -- Otherwise exception lexical_error is raised.

  Lexical_Error : exception;
  
  ----------------------------------------------------------------------------
  -- Some functions for interval.
  
  function Duration_Of (The_Interval : in Interval) return Duration;
  function Interval_Of (The_Duration : in Duration) return Interval;  
  function Interval_Of (The_Seconds  : in Float_64) return Interval;
  -- The interval_of functions may raise ada.calendar.time_error.
  
  function "-" (Left, Right : in Ada.Calendar.Time) return Interval;
  function "-" (Left, Right : in Time)              return Interval;  
  
  function "-" (Left, Right : in Ada.Calendar.Time) return Float_64;
  function "-" (Left, Right : in Time)              return Float_64;      
  
  -- The result is in seconds.

  function Seconds_Of (The_Interval : in Interval)  return Float_64;
  
  -- Gives the number of seconds between 2 intervals.

  -----------------------------------------------------------------------------
  subtype Gregorian_Calendar_Year is Natural range 1582 .. Natural'Last;

  function Day_Of (The_Year  : in Gregorian_Calendar_Year;
                   The_Month : in Month;
                   The_Day   : in Day) return Day_Name;

  -- Schroer, 14.8.92: the additional routine IMAGE_OF delivers a image by
  --                   separating the date- and the time-image by " ,".

  function  Image_Of (The_Time  : in Time;
                      Date_Form : in Date_Format := Full;
                      Hundreds  : in Boolean := True) return String;
  
  function  Image_Of (The_Time  : in Ada.Calendar.Time;
                      Date_Form : in Date_Format := Full;
                      Hundreds  : in Boolean := True) return String;

  -- The time image of a variable of type time delivers a string with the
  -- constant number of 41 characters of the following format (for full, true).
  -- 12345678901234567890123456789012345678901
  -- wwwwwwwww mmmmmmmmm dd, yyyy, hh:mm:ss:hh
  -- FRIDAY    JUNE      23, 1995, 16:13:08:00

  function  Value_Of (The_Time : in String) return Time;

  -- The function value_of needs a string of the above format.
  -- Otherwise exception lexical_error is raised.

end Utilities.Calendar_Utilities;
