--                    -*- Mode: Ada -*-
-- Filename        : opengl-bitmaps.adb
-- Description     : Package for the definition of bitmaps. 
-- Author          : J. Schrer
-- Created On      : 21. April 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Ada.Streams.Stream_Io;
with Ada.Exceptions;
with Ada.Unchecked_Deallocation;
--with Ada.Unchecked_Conversion;
with Interfaces.C;
with Opengl.Glu;
with Opengl.Colors;
with Opengl.Logging;
with Opengl.Messages;
--with Opengl.Timer;
--with Opengl.Glaux;
with Utilities.Code_Generator;

package body Opengl.Bitmaps is
  
  use type Byte;
  use type Interfaces.C.Int;
  use type Interfaces.C.Unsigned_Long;
  use type Interfaces.C.Unsigned_Short;
  ---------------------------------------------------------------------------- 
  procedure Free(Item : in out Byte_Array_Access) is
	  procedure Deallocate is new Ada.Unchecked_Deallocation
	    (Object => Byte_Array, Name => Byte_Array_Access);
  begin
    Deallocate(Item);
  end Free;          
  ---------------------------------------------------------------------------- 
  procedure Free(Item : in out Pixel_Array_Access) is
    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Pixel_Array, Name => Pixel_Array_Access);
  begin
    Deallocate(Item);
  end Free;          
  ----------------------------------------------------------------------------     
  procedure Free(Item : in out Bitmap_Type) is
  begin
    Free(Item.Bytes);
    Free(Item.Pixels);
    Ada.Strings.Unbounded.Free(Item.Filename);
  end Free;
  ----------------------------------------------------------------------------    
  procedure Read(The_Bitmap :    out Bitmap_Type;
                 Filename   : in     String) is                       
    ----------------------------------------------------------------------------
    function Bitmap_Offset_Of
	    (Bitmap_Info_Header : in Win32.Wingdi.Bitmapinfoheader) return Positive is
	
      Num_Colors       : Integer := 0;
      Color_Table_Size : Integer;
    begin
      -- Use number-of-colors-in-image field from header, if present
      if Bitmap_Info_Header.Bisize >= 36 then
        Num_Colors := Integer(Bitmap_Info_Header.Biclrused);
      elsif Bitmap_Info_Header.Bibitcount /= 24 then
        -- Compute number-of-colors from color-bits-per-pixel field
        Num_Colors := 2 ** Integer(Bitmap_Info_Header.Bibitcount);
      end if;
      -- Compute size of color table in DIB
      Color_Table_Size := Num_Colors * (Win32.Wingdi.Rgbquad'Size / 8);
	    return Integer(Bitmap_Info_Header.Bisize) + Color_Table_Size + 1;
    end Bitmap_Offset_Of;
    --------------------------------------------------------------------------
    Bm     : constant Win32.Word := 16#4D42#;
    Stream : Ada.Streams.Stream_Io.Stream_Access;
    File   : Ada.Streams.Stream_Io.File_Type;    
  begin
    Ada.Streams.Stream_Io.Open(File, Ada.Streams.Stream_Io.In_File, Filename);
    Stream := Ada.Streams.Stream_Io.Stream(File);

    Win32.Wingdi.Bitmapfileheader'Read(Stream, The_Bitmap.File_Header);
    
    -- Check for first two bytes in bitmap file header = "BM"
    if The_Bitmap.File_Header.Bftype = Bm then
      Win32.Wingdi.Bitmapinfoheader'Read(Stream, The_Bitmap.Info_Header);

      The_Bitmap.Bytes := new Byte_Array
        (1 .. Natural(The_Bitmap.File_Header.Bfsize) -
             (The_Bitmap.File_Header'Size + The_Bitmap.Info_Header'Size) / 8);
             
      Byte_Array'Read(Stream, The_Bitmap.Bytes.all);                                 

      The_Bitmap.Pixels       := null;
      The_Bitmap.Index_Offset := Bitmap_Offset_Of(The_Bitmap.Info_Header);
      The_Bitmap.Scale        := (others => 1.0);          
      The_Bitmap.Filename     := new String'(Ada.Streams.Stream_Io.Name(File));  
      
      The_Bitmap.Offset(X) := - 0.5 * 
                 Gl.Glfloat(The_Bitmap.Info_Header.Biwidth) / 
                 Gl.Glfloat(Screen.Pixel_Width)  * Screen.Width;
      The_Bitmap.Offset(Y) := - 0.5 * 
                 Gl.Glfloat(The_Bitmap.Info_Header.Biheight) / 
                 Gl.Glfloat(Screen.Pixel_Height) * Screen.Height;                                      
    else    
      The_Bitmap := (Bytes        => null, 
                     Pixels       => null, 
                     File_Header  => The_Bitmap.File_Header,
                     Info_Header  => The_Bitmap.Info_Header,
                     Index_Offset => 1, 
                     Offset       => (0.0, 0.0),
                     Scale        => (1.0, 1.0),
                     Filename     => null);
    end if;
    
    if Ada.Text_Io.Is_Open(Log_File) then
      Utilities.Code_Generator.Put_Lines(Log_File,
        "Opengl.Bitmaps.Read(File => " & '"' & Ada.Streams.Stream_Io.Name(File) & '"' & ")",
        "Index_Offset               :" & Integer'Image(The_Bitmap.Index_Offset),              
        "Bytes'size                 :" & Integer'Image(The_Bitmap.Bytes.all'Size),           
        "File_Header.BfType         :" & Win32.Word'Image(The_Bitmap.File_Header.Bftype),
        "File_Header.BfSize         :" & Win32.Dword'Image(The_Bitmap.File_Header.Bfsize),
        "File_Header.BfOffBits      :" & Win32.Dword'Image(The_Bitmap.File_Header.Bfoffbits),
        "Info_Header.BiSize         :" & Win32.Dword'Image(The_Bitmap.Info_Header.Bisize), 
        "Info_Header.BiWidth        :" & Win32.Long'Image(The_Bitmap.Info_Header.Biwidth),               
        "Info_Header.BiHeight       :" & Win32.Long'Image(The_Bitmap.Info_Header.Biheight),                   
        "Info_Header.BiPlanes       :" & Win32.Word'Image(The_Bitmap.Info_Header.Biplanes),                     
        "Info_Header.BiBitCount     :" & Win32.Word'Image(The_Bitmap.Info_Header.Bibitcount),                   
        "Info_Header.BiCompression  :" & Win32.Dword'Image(The_Bitmap.Info_Header.Bicompression),                  
        "Info_Header.BiSizeImage    :" & Win32.Dword'Image(The_Bitmap.Info_Header.Bisizeimage),                
        "Info_Header.BiXpelsPerMeter:" & Win32.Long'Image(The_Bitmap.Info_Header.Bixpelspermeter),                   
        "Info_Header.BiYpelsPerMeter:" & Win32.Long'Image(The_Bitmap.Info_Header.Biypelspermeter),                  
        "Info_Header.BiCLRUsed      :" & Win32.Dword'Image(The_Bitmap.Info_Header.Biclrused),                
        "Info_Header.BiCLRImportant :" & Win32.Dword'Image(The_Bitmap.Info_Header.Biclrimportant),
        "---------------------------------------------------------------------");    
      Ada.Text_Io.Flush(Log_File);             
    end if;  
    
    Ada.Streams.Stream_Io.Close(File);    
  exception
    when E: others =>      
      Free(The_Bitmap.Bytes);
      The_Bitmap := (Bytes        => null, 
                     Pixels       => null, 
                     File_Header  => The_Bitmap.File_Header,
                     Info_Header  => The_Bitmap.Info_Header,
                     Index_Offset => 1, 
                     Offset       => (0.0, 0.0),
                     Scale        => (1.0, 1.0),
                     Filename     => null);
      if Ada.Streams.Stream_Io.Is_Open(File) then
        Ada.Streams.Stream_Io.Close(File);
      end if;                    
      Ada.Exceptions.Reraise_Occurrence(E);
  end Read;                   
  ----------------------------------------------------------------------------  
  procedure Generate_Pixels(The_Bitmap : in out Bitmap_Type) is
    -------------------------------------------------------------------------
    function Next_Power_Of_2_greater(Number : in Natural) return Natural is
      Pow : Natural := 2;
    begin
      while Pow < Number loop
        Pow := 2 * Pow;
      end loop;
      return Pow; 
    end Next_Power_Of_2_greater;  
    ----------------------------      
    Error_Code : Interfaces.C.Int;
  begin
    if The_Bitmap.Pixels /= null then
      return;
    end if;
    
    The_Bitmap.Pixels := new Pixel_Array
      (1 .. Next_Power_Of_2_greater(Natural(The_Bitmap.Info_Header.Biwidth)), 
       1 .. Next_Power_Of_2_greater(Natural(The_Bitmap.Info_Header.Biheight)));    
      
    Error_Code := Glu.Gluscaleimage
                  (Format    => Gl.Gl_Rgb,
                   Widthin   => Gl.Glint(The_Bitmap.Info_Header.Biwidth),
                   Heightin  => Gl.Glint(The_Bitmap.Info_Header.Biheight),
                   Typein    => Gl.Gl_Unsigned_Byte,
                   Datain    => The_Bitmap.Bytes.all'Address,
                   Widthout  => Gl.Glint(The_Bitmap.Pixels'Length(1)),
                   Heightout => Gl.Glint(The_Bitmap.Pixels'Length(2)),
                   Typeout   => Gl.Gl_Unsigned_Byte,
                   Dataout   => The_Bitmap.Pixels.all'Address);
                   
    if Error_Code = 0 then                                                                  
      Free(The_Bitmap.Bytes);                                   
    
      Exchange_Colors(The_Bitmap,
                      First_Color  => Red,
                      Second_Color => Blue);                                 
    else   
      Ada.Exceptions.Raise_Exception(Opengl_Error'Identity,  
        Message => Messages.Error_Message
        (Subprogram => "Bitmaps.Generate_Pixels('" & The_Bitmap.Filename.all & "')",
         Message    => "GluScaleImage failed," & 
                       Messages.Error_String(Gl.Glenum(Error_Code))));                   
    end if;
  end Generate_Pixels;
  ----------------------------------------------------------------------------                                         
  procedure Exchange_Colors(The_Bitmap   : in out Bitmap_Type;
                            First_Color,
                            Second_Color : in     Color_Index_Type) is
    Temp : Byte;                          
  begin
    if The_Bitmap.Pixels = null then
      Generate_Pixels(The_Bitmap);
    end if;
    for I in The_Bitmap.Pixels'range(1) loop
      for J in The_Bitmap.Pixels'range(2) loop
        Temp := The_Bitmap.Pixels(I,J)(Second_Color);
        The_Bitmap.Pixels(I,J)(Second_Color) := 
          The_Bitmap.Pixels(I,J)(First_Color);
        The_Bitmap.Pixels(I,J)(First_Color)  := Temp;
      end loop;
    end loop;    
  end Exchange_Colors;                                
  ---------------------------------------------------------------------------- 
  procedure Add_Color_Bias (The_Bitmap   : in out Bitmap_Type;                         
                            Red_Bias     : in     Byte := 0;
                            Green_Bias   : in     Byte := 0;
                            Blue_Bias    : in     Byte := 0) is
    ----------------------------------------------------------                          
    procedure Add_Component(Color_Index : in Color_Index_Type;
                            Bias        : in Byte) is
    begin
	    for I in The_Bitmap.Pixels'range(1) loop
	      for J in The_Bitmap.Pixels'range(2) loop
	        The_Bitmap.Pixels(I,J)(Color_Index) := 
            The_Bitmap.Pixels(I,J)(Color_Index) + Bias;
	      end loop;
	    end loop;    
    end Add_Component;
    ------------------                        
  begin
    if The_Bitmap.Pixels = null then
      Generate_Pixels(The_Bitmap);
    end if;
    if Red_Bias /= 0 then
      Add_Component(Red, Red_Bias);
    end if;  
    if Green_Bias /= 0 then
      Add_Component(Green, Green_Bias);
    end if;  
    if Blue_Bias /= 0 then
      Add_Component(Blue, Blue_Bias);
    end if;  
  end Add_Color_Bias;      
  ----------------------------------------------------------------------------    
  procedure Filter(The_Bitmap : in out Bitmap_Type;                            
                   The_Filter : in     Filter_Type) is
  begin                           
    if The_Bitmap.Pixels = null then
      Generate_Pixels(The_Bitmap);
    end if;

    for I in The_Bitmap.Pixels'range(1) loop
  	  for J in The_Bitmap.Pixels'range(2) loop 
        if The_Bitmap.Pixels(I,J)(Red)   in 
           The_Filter(Red).  First .. The_Filter(Red).  Last and then
           The_Bitmap.Pixels(I,J)(Green) in 
           The_Filter(Green).First .. The_Filter(Green).Last and then
           The_Bitmap.Pixels(I,J)(Blue)  in 
           The_Filter(Blue). First .. The_Filter(Blue). Last then
  
          The_Bitmap.Pixels(I,J)(Red)   := The_Bitmap.Pixels(I,J)(Red)   +
                                           The_Filter(Red).Bias;
          The_Bitmap.Pixels(I,J)(Green) := The_Bitmap.Pixels(I,J)(Green) +
                                           The_Filter(Green).Bias;
          The_Bitmap.Pixels(I,J)(Blue)  := The_Bitmap.Pixels(I,J)(Blue)  +
                                           The_Filter(Blue).Bias;          
        end if;                                                          
      end loop;  
    end loop;  
  end Filter;                          
  ----------------------------------------------------------------------------    
  procedure Filter(The_Bitmap : in out Bitmap_Type;                            
                   The_Filter : in     Filter_Color_Procedure) is
  begin                           
    if The_Bitmap.Pixels = null then
      Generate_Pixels(The_Bitmap);
    end if;

    for I in The_Bitmap.Pixels'range(1) loop
  	  for J in The_Bitmap.Pixels'range(2) loop 
        The_Filter(The_Bitmap.Pixels(I,J));
      end loop;  
    end loop;  
  end Filter;                          
  ----------------------------------------------------------------------------    
  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d := (X => Screen.Width  / 2.0, 
     	                                 Y => Screen.Height / 2.0,
                                       Z => 0.0);
     Attitude : in Gl.Glfloat      := 0.0;
     Scale    : in Gl.Glfloat      := Screen.Height / 2.0) return Model is
       
    The_Bitmap : Bitmap_Type;
  begin
    Read(The_Bitmap, Filename);         
--    Generate_Pixels(The_Bitmap);
    
    return (Models.Root_Model with
            Size_Scaling => Scale,
            Color        => Colors.White,
            Position     => Position,
            Attitude     => Attitude,
            Bitmap       => The_Bitmap);
  end Model_With;
  ----------------------------------------------------------------------------         
  procedure Render(The_Model : in out Model) is    
--    subtype Vector_4 is Types.Float_Vector(1 .. 4);    
--    function Raster_Position return Vector_4 is
--      Pos : Vector_4;
--    begin
--      Gl.Glgetfloatv(Pname  => Gl.Gl_Current_Raster_Position, 
--                     Params => Pos(Pos'First)'Unchecked_Access);                       
--      return Pos;
--    end Raster_Position;
  begin
    Models.Initialize_Render(The_Model);

    Gl.Glpixelstorei(Pname   => Gl.Gl_Unpack_Alignment, Param => 4);
    Gl.Glrasterpos2f(X       => The_Model.Size_Scaling * 
                                The_Model.Bitmap.Offset(X), 
                     Y       => The_Model.Size_Scaling * 
                                The_Model.Bitmap.Offset(Y));
    Gl.Glpixelzoom  (Xfactor => The_Model.Size_Scaling * 
                                The_Model.Bitmap.Scale(X),  
                     Yfactor => The_Model.Size_Scaling *
                                The_Model.Bitmap.Scale(Y));
    begin
      if    The_Model.Bitmap.Pixels /= null then                  
        Gl.Gldrawpixels
          (Width  => Gl.Glsizei(The_Model.Bitmap.Pixels'Length(1)),
           Height => Gl.Glsizei(The_Model.Bitmap.Pixels'Length(2)),
           Format => Gl.Gl_Rgb,
           C_Type => Gl.Gl_Unsigned_Byte,
           Pixels => The_Model.Bitmap.Pixels.all'Address);
      elsif The_Model.Bitmap.Bytes /= null then
        Gl.Gldrawpixels
          (Width  => Gl.Glsizei(The_Model.Bitmap.Info_Header.Biwidth),
           Height => Gl.Glsizei(The_Model.Bitmap.Info_Header.Biheight),
           Format => Gl.Gl_Rgb,
           C_Type => Gl.Gl_Unsigned_Byte,
           Pixels => The_Model.Bitmap.Bytes.all'Address);
      end if;     
    exception
      when E : others => Logging.Put(Ada.Exceptions.Exception_Information(E));
    end;     

    Models.Finalize_Render(The_Model);
  end Render;
  ---------------------------------------------------------------------------- 
  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d := (X => Screen.Width  / 2.0, 
     	                                 Y => Screen.Height / 2.0,
                                       Z => 0.0);
     Attitude : in Gl.Glfloat      := 0.0;
     Scale    : in Gl.Glfloat      := Screen.Height / 2.0)
     return Model_With_Filter is
       
    The_Model : constant Model := Model_With(Filename, Position, Attitude, Scale);   
  begin
    return Model_With_Filter'(The_Model with Filter            => null, 
                                             Unfiltered_Pixels => null);
  end Model_With;    
  ---------------------------------------------------------------------------- 
  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d        := (X => Screen.Width  / 2.0, 
            	                                Y => Screen.Height / 2.0,
                                              Z => 0.0);
     Attitude : in Gl.Glfloat             := 0.0;
     Scale    : in Gl.Glfloat             := Screen.Height / 2.0;
     Filter   : in Filter_Color_Procedure := null)
     return Model_With_Filter is
     
    The_Model : constant Model := Model_With(Filename, Position, Attitude, Scale);   
  begin
    return Model_With_Filter'(The_Model with Filter, Unfiltered_Pixels => null);
  end Model_With;    
  ---------------------------------------------------------------------------- 
  procedure Render(The_Model : in out Model_With_Filter) is
  begin
    if The_Model.Bitmap.Pixels = null then
      Generate_Pixels(The_Model.Bitmap);
    end if;
    
    if The_Model.Unfiltered_Pixels = null then
      The_Model.Unfiltered_Pixels := 
        new Pixel_Array'(The_Model.Bitmap.Pixels.all);    
    end if;
    
    The_Model.Bitmap.Pixels.all := The_Model.Unfiltered_Pixels.all;
    Filter(The_Bitmap => The_Model.Bitmap, The_Filter => The_Model.Filter);                  
    Render(Model(The_Model));
  end Render;
  ---------------------------------------------------------------------------- 
end Opengl.Bitmaps;

