--                    -*- Mode: Ada -*-
-- Filename        : opengl-bitmaps.ads
-- Description     : Package for the definition of bitmaps. 
-- Author          : J. Schrer
-- Created On      : 21. April 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Ada.Strings.Unbounded;
with Opengl.Gl;
with Opengl.Screen;
with Opengl.Types;
with Opengl.Models;
with Interfaces;
with Win32.Wingdi;

package Opengl.Bitmaps is
  
  -- Simple bitmap type 
  
  Red   : constant := 1;
  Green : constant := 2;
  Blue  : constant := 3;
  Alpha : constant := 4;
    
  subtype Color_Index_Type is Positive range Red .. Blue;


  subtype Byte is Interfaces.Unsigned_8;
  
  type Byte_Array is array(Positive range <>) of Byte;      
  type Byte_Array_Access is access Byte_Array;
  
  type Color is array(Red .. Alpha) of Byte;
  type Rgb_Color is array(Red .. Blue) of Byte;
  subtype Pixel is Rgb_Color;
  
  type Pixel_Array is array(Positive range <>, Positive range <>) of Pixel;
  type Pixel_Array_Access is access Pixel_Array;
  
  type Bitmap_Type is record
    Bytes        : Byte_Array_Access;
    Pixels       : Pixel_Array_Access;
    File_Header  : Win32.Wingdi.Bitmapfileheader;
    Info_Header  : Win32.Wingdi.Bitmapinfoheader;
    Index_Offset : Positive        := 1;
    Offset       : Types.Vector_2d := (X => 0.0, Y => 0.0);
                   -- Half the size of the bitmap.
    Scale        : Types.Vector_2d := (X => 1.0, Y => 1.0);
                   -- Scale is set after resizing the window. 
    Filename     : Ada.Strings.Unbounded.String_Access;                   
  end record;
                     
  
  procedure Free(Item : in out Byte_Array_Access);
  procedure Free(Item : in out Pixel_Array_Access);
    
  procedure Free(Item : in out Bitmap_Type);
  
  -- Frees heap storage of item.
    
    
  procedure Read(The_Bitmap :    out Bitmap_Type;
                 Filename   : in     String);
                 
  -- Reads an the_bytes in the_bitmap from filename.                                                   

  procedure Generate_Pixels(The_Bitmap : in out Bitmap_Type);                            
  
  -- Allocates and fills an the_pixels in the_bitmap, frees the_bytes.
  -- Calls Gluscaleimage and enlarges the bitmap to powers of two for the 
  -- number of pixels for width and height.
  -- Exchanges the green and blue component of all pixels.
        
  procedure Exchange_Colors(The_Bitmap   : in out Bitmap_Type;
                            First_Color,
                            Second_Color : in     Color_Index_Type);
                            
  -- Exchanges the two colors in the_bitmap.                            
                            
  procedure Add_Color_Bias (The_Bitmap   : in out Bitmap_Type;                            
                            Red_Bias     : in     Byte := 0;
                            Green_Bias   : in     Byte := 0;
                            Blue_Bias    : in     Byte := 0);
                            
  -- Manipulates colors in the_bitmap.                           
  
  type Filter_Color_Type is record
    First : Byte := Byte'First; -- 0
    Last  : Byte := Byte'Last;  -- 255
    Bias  : Byte := 0;
  end record;  
  
  type Filter_Type is array(Color_Index_Type) of Filter_Color_Type;
  
  procedure Filter(The_Bitmap : in out Bitmap_Type;                            
                   The_Filter : in     Filter_Type);
                            
  -- If The_Bitmat.pixels /= null generates them.
  -- For all pixels filters them 
  --
  -- if The_Bitmap.Pixels(I,J)(Red)   in The_Filter(Red).  First .. The_Filter(Red).  Last and then
  --    The_Bitmap.Pixels(I,J)(Green) in The_Filter(Green).First .. The_Filter(Green).Last and then
  --    The_Bitmap.Pixels(I,J)(Blue)  in The_Filter(Blue). First .. The_Filter(Blue). Last then
  --
  --   The_Bitmap.Pixels(I,J)(Red)   := The_Bitmap.Pixels(I,J)(Red)   + The_Filter(Red).bias;
  --   The_Bitmap.Pixels(I,J)(Green) := The_Bitmap.Pixels(I,J)(Green) + The_Filter(Green).bias;
  --   The_Bitmap.Pixels(I,J)(Blue)  := The_Bitmap.Pixels(I,J)(Blue)  + The_Filter(Blue).bias;
  -- end if;            
                   
  type Filter_Color_Procedure is access procedure(The_Pixel : in out Pixel);                                                                 
  
  procedure Filter(The_Bitmap : in out Bitmap_Type;                            
                   The_Filter : in     Filter_Color_Procedure);
                   
  -- Filters all Pixels in The_Bitmap using The_Filter procedure.                   
  ----------------------------------------------------------------------------
                            
  type Model is new Models.Model_2d_Z with record
    Bitmap : Bitmap_Type;    
  end record;

  use type Gl.Glfloat;
  
  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d := (X => Screen.Width  / 2.0, 
     	                                 Y => Screen.Height / 2.0,
                                       Z => 0.0);
     Attitude : in Gl.Glfloat      := 0.0;
     Scale    : in Gl.Glfloat      := Screen.Height / 2.0) 
     return Model;
     
  -- Reads the bitmap from file filename and returns it in the model.   
    
  procedure Render(The_Model : in out Model);                   
  
  ----------------------------------------------------------------------------

  type Model_With_Filter is new Model with record    
    Filter            : Filter_Color_Procedure;
    Unfiltered_Pixels : Pixel_Array_Access;
  end record;

  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d := (X => Screen.Width  / 2.0, 
     	                                 Y => Screen.Height / 2.0,
                                       Z => 0.0);
     Attitude : in Gl.Glfloat      := 0.0;
     Scale    : in Gl.Glfloat      := Screen.Height / 2.0)
     return Model_With_Filter;
     
  function Model_With
    (Filename : in String;
     Position : in Types.Vector_3d        := (X => Screen.Width  / 2.0, 
            	                                Y => Screen.Height / 2.0,
                                              Z => 0.0);
     Attitude : in Gl.Glfloat             := 0.0;
     Scale    : in Gl.Glfloat             := Screen.Height / 2.0;
     Filter   : in Filter_Color_Procedure := null)
     return Model_With_Filter;
     
    
  procedure Render(The_Model : in out Model_With_Filter);                   
  
end Opengl.Bitmaps;  

