--                              -*- Mode: Ada -*-
-- Filename        : opengl-clip_planes.adb
-- Description     : Package for planes. The number of available clip-planes is 
--                   dependent on the OS. MS-windows has 6 clip-planes.
-- Author          : J. Schrer
-- Created On      : 22. May 1999
-- Last Modified By: 
-- Last Modified On: 
-- Update Count    : 
-- Status          :

with Ada.Exceptions;
with Opengl.Float_32.Elementary_Functions;
with Opengl.Messages;
with Opengl.Win_Stuff;

package body Opengl.Clip_Planes is

  use type Gl.Glfloat, Gl.Gldouble;
  use Opengl.Float_32.Elementary_Functions;
  
  Clip_Plane_Indices : constant array(1 .. 6) of Gl.Glenum :=
      (1 => Gl.Gl_Clip_Plane0,
       2 => Gl.Gl_Clip_Plane1,
       3 => Gl.Gl_Clip_Plane2,
       4 => Gl.Gl_Clip_Plane3,
       5 => Gl.Gl_Clip_Plane4,
       6 => Gl.Gl_Clip_Plane5);
 
  Counter            : Natural := 0;                 
  Max_Counter        : Natural := 0;
  ----------------------------------------------------------------------------
  function Clip_Plane(Axis    : in Types.Axis_Type;
                      Visible : in Direction_Type;
                      Cutoff  : in Gl.Glfloat := 0.0) return Plane is
                        
    The_Plane : Plane := (others => 0.0);                        
  begin
    case Visible is
      when Positive_Direction => The_Plane(Axis) :=   1.0;
                                 The_Plane(D)    := Gl.Gldouble(- Cutoff);
      when Negative_Direction => The_Plane(Axis) := - 1.0;
                                 The_Plane(D)    := Gl.Gldouble(Cutoff);
    end case;
    return The_Plane;
  end Clip_Plane;                            
  ----------------------------------------------------------------------------
  function Rotated_Clip_Plane(Axis    : in Types.Axis_Type;
                              Visible : in Direction_Type;
                              Angle   : in Angle_Type := 0.0) return Plane is
                                
    Axis_1    : constant Natural    := (Axis mod 3) + 1;
    Axis_2    : constant Natural    := Axis_1 + 1;                                                   
    Sin_Angle : constant Gl.Glfloat := Sin(Angle, Cycle => 360.0);
    Cos_Angle : constant Gl.Glfloat := Cos(Angle, Cycle => 360.0);    
    The_Plane : Plane := (others => 0.0);       
  begin
    case Visible is
      when Positive_Direction => The_Plane(Axis_1) := Gl.Gldouble(-Sin_Angle);
		                             The_Plane(Axis_2) := Gl.Gldouble( Cos_Angle);
      when Negative_Direction => The_Plane(Axis_1) := Gl.Gldouble( Sin_Angle);
		                             The_Plane(Axis_2) := Gl.Gldouble(-Cos_Angle);
    end case;
    return The_Plane;
  end Rotated_Clip_Plane;
  ----------------------------------------------------------------------------                              
  procedure Enable(The_Plane : in Plane) is
  begin
    Counter := Counter + 1;
    if Counter <= Clip_Plane_Indices'Last then
      Gl.Glclipplane(Plane    => Clip_Plane_Indices(Counter),
                     Equation => The_Plane(A)'Unchecked_Access);
      Gl.Glenable(Clip_Plane_Indices(Counter));
      Max_Counter := Natural'Max(Max_Counter, Counter);
    else        
      Ada.Exceptions.Raise_Exception(Opengl_Error'Identity, Message =>
        Messages.Error_Message
          (Subprogram => "Opengl.Clip_Planes.Enable",
           Message    => "Tried to enable" & Natural'Image(Counter) & 
                         "-th clip-plane. Max number is" &
                         Natural'Image(Clip_Plane_Indices'Length)));                     
    end if;                     
  end Enable;    
  ----------------------------------------------------------------------------
  procedure Disable(Number : in Positive := 1) is
  begin
    if Number <= Counter then           
      for I in 1 .. Number loop
        Gl.Gldisable(Clip_Plane_Indices(Counter));    
        Counter := Counter - 1;        
      end loop;  
    else  
      for I in 1 .. Counter loop
        Gl.Gldisable(Clip_Plane_Indices(I));    
      end loop;  
      Counter := 0;
      Ada.Exceptions.Raise_Exception(Opengl_Error'Identity, Message =>
        Messages.Error_Message
          (Subprogram => "Opengl.Clip_Planes.Disable",
           Message    => "No more planes enabled")); 
    end if;       
  end Disable;    
  ----------------------------------------------------------------------------
  function Planes_In_Use return Natural is
  begin
    return Counter;
  end Planes_In_Use;
  ----------------------------------------------------------------------------
  function Max_Planes_Used return Natural is
  begin
    return Max_Counter;
  end Max_Planes_Used;
  ----------------------------------------------------------------------------
begin
  if Win_Stuff.Max_Clip_Planes > Clip_Plane_Indices'Length then
    Ada.Exceptions.Raise_Exception(Opengl_Error'Identity, Message =>
        Messages.Error_Message
          (Subprogram => "Opengl.Clip_Planes (body elaboration)",
           Message    => Natural'Image(Win_Stuff.Max_Clip_Planes) & 
                         " clip-planes may be declared in OpenGL.Clip_Planes."));
  end if;      
end Opengl.Clip_Planes;
