--                              -*- Mode: Ada -*-
-- Filename        : opengl-clip_planes.ads
-- Description     : Package for planes. The number of available clip-planes is 
--                   dependent on the OS. MS-windows has 6 clip-planes
--                   (see package body).                   
-- Author          : J. Schrer
-- Created On      : 22. May 1999
-- Last Modified By: 
-- Last Modified On: 
-- Update Count    : 
-- Status          :

with Opengl.Gl;
with Opengl.Types;

package Opengl.Clip_Planes is

  -- Plane coefficients -----------------------------------------------
  
  A : constant := 1;
  B : constant := 2;
  C : constant := 3;
  D : constant := 4;

  -- Plane-type and simple clip-plane constants -----------------------

  type Plane is array(A .. D) of aliased Gl.Gldouble;
  
  type Plane_Array is array(Positive range <>) of Plane;

  X_Clip_Plane : constant Plane := (A => 1.0, B => 0.0, C => 0.0, D => 0.0);
  Y_Clip_Plane : constant Plane := (A => 0.0, B => 1.0, C => 0.0, D => 0.0);
  Z_Clip_Plane : constant Plane := (A => 0.0, B => 0.0, C => 1.0, D => 0.0);

  -- Constructor functions for simple clip-planes.

  type Direction_Type is (Positive_Direction, Negative_Direction);
  
  function Clip_Plane(Axis    : in Types.Axis_Type;
                      Visible : in Direction_Type;
                      Cutoff  : in Gl.Glfloat := 0.0) return Plane;
                         
  -- Returns a plane that (when enabled as clipping plane) lets visible all
  -- in the positive or negative direction of axis (x,y,z) measured from the
  -- cutoff point.                            
  --                         
  -- Examples: 
  --                         
  -- Clip_Plane(Axis    => X, 
  --            Visible => Positive_Direction, 
  --            Cutoff  => Screen.Width / 2.0)
  --
  --   shows the right half of the screen.                                                
  --                         
  -- Clip_Plane(Axis    => X, 
  --            Visible => Negative_Direction, 
  --            Cutoff  => Screen.Width / 2.0)                                                
  --
  --   shows the left half of the screen.                                                
  
  subtype Angle_Type is Gl.Glfloat range 0.0 .. 180.0;
  
  function Rotated_Clip_Plane(Axis    : in Types.Axis_Type;
                              Visible : in Direction_Type;
                              Angle   : in Angle_Type := 0.0) return Plane;
                         
  -- Returns a plane that (when enabled as clipping plane) is rotated about
  -- Axis by Angle and lets Visible all in the positive or negative 
  -- direction of the remaining two axis.
  -- The visible direction refers to mathematical angle.
  -- Positive_Direction is counter clockwise, Negative_Direction is clockwise.
  --                         
  -- Examples: 
  --                         
  -- Rotated_Clip_Plane(Axis    => Z,
  --                    Visible => Positive_Direction, 
  --                    Angle   => 90.0)
  --
  --   shows the left half of the screen.                                                
  --                         
  -- Rotated_Clip_Plane(Axis    => Z, 
  --                    Visible => Negative_Direction,
  --                    Angle   => 90.0)                                                
  --
  --   shows the right half of the screen.                                                
  --  
  -- Rotated_Clip_Plane(Axis    => Z, 
  --                    Visible => Negative_Direction, 
  --                    Angle   => 45.0)                                                
  --
  --   shows the lower right diagonal part of the screen.                                                

  ----------------------------------------------------------------------------
  -- Plane-handling subprograms. 
                          
  procedure Enable(The_Plane : in Plane);  
  
  -- Enables The_Plane with the next free clip-plane index.
  -- Raises opengl_error if no more planes are available. 
  
  procedure Disable(Number : in Positive := 1);  
  
  -- Disables the Number last enabled planes and frees that clip-plane indices. 
  -- Raises opengl_error if Number > the number of actually enabled clip-planes
  -- but frees all enabled clip-planes before.
  
  -- Enable and disable act as a stack.
  
  function Planes_In_Use return Natural;
  
  -- Gives the actual used clip-plane number.
  
  function Max_Planes_Used return Natural;

  -- Gives the maximum number of used clip-planes.
  
private

  pragma Inline(Enable);
  pragma Inline(Disable);
    
end Opengl.Clip_Planes;
