--                              -*- Mode: Ada -*-
-- Filename        : opengl-command_line.ads
-- Description     : Package with all constants read from the command-line.
-- Author          : J. Schrer
-- Created On      : 15. dez. 1998
-- Last Modified By: 
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with Ada.Real_Time;
with Opengl.Screen;
with Utilities.Command_Line;
with System;

package Opengl.Command_Line is
  
  -- The parameters are used from within subprograms in the closure of opengl.win_main.
  
  Test              : constant Boolean := Utilities.Command_Line.Has_Argument("-test");
    -- Render models defined in Opengl.Test_Screen.
  
  Config_File       : constant String  := Utilities.Command_Line.Argument    ("-file");
    -- The filename of the file to read the instrument configuration from.
  
  Navdb_File        : constant String  := Utilities.Command_Line.Argument 
                                          ("-navdb", "efis-worldwide");
    -- The filename of the navigation database file for the moving map.
  
  Flight_Plan_File  : constant String := Utilities.Command_Line.Argument     ("-flightplan");
    -- Filename of the flightplan that is read at program-start. 
    -- The plane than immediatly flies along the route.
                                
  Map_Search_Timeout: constant Ada.Real_Time.Time_Span := 
                      Ada.Real_Time.Milliseconds
                        (Utilities.Command_Line.Argument("-mapsearchtimeout", 1));                                                              
    -- Increment for limiting the search time in the moving-map.
    -- After Map_Search_Timeout ms in one render step the searching process 
    -- is interrupted.
                        
  Fpl_Search_Inc    : constant Positive := Utilities.Command_Line.Argument
                                           ("-fplsearchinc", 100);     
    -- Increment for limiting the search time in the flight_plan editing.
    -- Only Fpl_Search_Inc elements from the database are processed during one
    -- render step.                                                                                                                                  
    
  Test_Inputs       : constant Boolean := Utilities.Command_Line.Has_Argument("-testinputs");
    -- Generate test-inputs or not.
  
  Generator_Delta   : constant Ada.Real_Time.Time_Span := 
                      Ada.Real_Time.Milliseconds
                        (Utilities.Command_Line.Argument("-generatordelta", 20));
    -- Time increment for task that generates inputs.
                        
  Timer_Delta       : constant Ada.Real_Time.Time_Span := 
                      Ada.Real_Time.Milliseconds
                        (Utilities.Command_Line.Argument("-timerdelta", 1));                        
    -- Time increment for task that generates render-events.
                        
  Corrupted_Inputs  : constant Boolean := Utilities.Command_Line.Has_Argument("-corruptedinputs");
    -- (Read by task simulator.instruments.input_generator.manager)
    -- Generate inputs that are far outside the valid range for the instruments.
  
  Timer             : constant Boolean := Utilities.Command_Line.Has_Argument("-timer");
    -- Show the timer test output.
  
  Timer_Fontsize    : constant Float   := Utilities.Command_Line.Argument("-timerfontsize",4.0);
    -- The fontsize of the timer output.
  
  Model_Timing      : constant Boolean := Utilities.Command_Line.Has_Argument("-timing");
    -- For each model the render-time is calculated and written to log-file.
  
  Clock             : constant Natural := Utilities.Command_Line.Argument("-clock",0);
    -- If -clock is given with no argument shows one Opengl-Ada clock.
    -- If an earth is shown, the sun is always in the cenit (like xearth on Unix).
    -- If -clock is given with an argument between 1 .. 24 shows up to 24 
    -- Opengl-Ada clocks with timezone-offset.
    -- If an earth is shown, the viewpoint is always perpendicular to one
    -- location in the timezone and the sun generates a shadow.    
  
  Clock_24_H        : constant Natural := Utilities.Command_Line.Argument("-clock24h",0);
    -- Show an Opengl-Ada clock with a 24h dialface.
    -- The meaning of the parameter is as with clock.

  Earth_Scale       : constant Float := Utilities.Command_Line.Argument("-earthscale",0.66);
    -- Render the earth in the middle of the clock dialface. 0.0 means no earth is rendered.
  
  Earth             : constant Boolean := Utilities.Command_Line.Has_Argument("-earth");
    -- Render an earth/globe similar to xearth.
  
  Earth_File        : constant String  := Utilities.Command_Line.Argument("-earthfile","world.bmp");  
    -- Bitmap to use for the earth.
  
  Time_Offset       : constant Duration := Duration
                      (Utilities.Command_Line.Argument("-timeoffseth",0.0));
    -- Timezone offset of the clock and earth in hours.
  Time_Factor       : constant float := Utilities.Command_Line.Argument("-timefactor",1.0);
    -- Time factor for earth-rotation. 
    -- Normal angular velocity is multiplied by this number.

  Earth_Location    : constant String := Utilities.Command_Line.Argument("-earthloc","");
    -- A string that is written on the earth or clock to show the time zone.
  
  Latitude          : constant Float := Utilities.Command_Line.Argument("-lat",0.0);
  Longitude         : constant Float := Utilities.Command_Line.Argument("-lon",0.0);
    -- Position of the earth-location.
  
  Anti_Aliasing     : constant Boolean := Utilities.Command_Line.Has_Argument("-anti");
    -- Anti-aliasing on or off.
  
  Show_Mouse        : constant Boolean := Utilities.Command_Line.Has_Argument("-mouse");
    -- Show the mouse pointer on the opengl-screen or hide it.
  
  Write_No_Log_File : constant Boolean := Utilities.Command_Line.Has_Argument("-nologfile");
    -- Create a log file and write to it.
  
  Font_Info         : constant Boolean := Utilities.Command_Line.Has_Argument("-fontinfo");
    --  Writes full font info comprising length and height of all characters in the font to 
    --  the log file. If font info is False only short info is written.
  
  Show_Errors       : constant Boolean := Utilities.Command_Line.Has_Argument("-showerrors");
    -- Show a popup window with the Ada.Exceptions.Exception_Information or
    -- ignore the exception (handler in Opengl.Win.General_C_called_Windows_Callback).
  
  Full_Screen_Window: constant Boolean := Utilities.Command_Line.Has_Argument("-fullwindow");
    -- If true, a full window is drawn, else the window has no border and buttons.  

  -- The window coordinates at program start, X and Y refer to the upper left corner.  
  Window_X          : constant Natural := Utilities.Command_Line.Argument("-windowx",0);
  Window_Y          : constant Natural := Utilities.Command_Line.Argument("-windowy",0);
  Window_Width      : constant Natural := Utilities.Command_Line.Argument
                      ("-windoww", Screen.Pixel_Width);
  Window_Height     : constant Natural := Utilities.Command_Line.Argument
                      ("-windowh", Screen.Pixel_Height);
  
  Background_Color  : constant String := Utilities.Command_Line.Argument("-background","black");
    -- Defines the Background_Color of the opengl-screen, has to be one of colors.palette. 
  
    
  -- The following parameters are read by Simulator.Network_Interface body
  Network_Time_Limit: constant Ada.Real_Time.Time_Span := 
                      Ada.Real_Time.Milliseconds
                        (Utilities.Command_Line.Argument("-networklimit", 1000));
    -- If after Network_Time_Limit ms no new network data has received, 
    -- a messgage is written to timer.line(4) and to log_file.                                                
                        
  Network_Down_Beep : constant Boolean := Utilities.Command_Line.Has_Argument("-networkbeep");
    -- Beep, if after Network_Time_Limit ms no new data arrives.                      
  
  -- The following parameters are read by OpenGL.Win.Set_Priotity
  Priority          : constant string := Utilities.Command_Line.Argument("-priority","-idle");
  
  -- On WinNT there are 7 priorities available for Ada 
  -- (see Cross Development Mini-Reference, page 7).
  -- The priotity of the main and all tasks should not be fixed by using theese
  -- command-line parameters. The following priority is set by the 
  -- priority pragma in opengl.win_main and the timer task(s) in opengl.win. 
  -- The Realtime_Priority should not be used. It blocks the system and it's
  -- difficult to shut the main down. 
  
  Winnt_Priority    : constant System.Priority := 16;
      
  -- The following parameters are read by OpenGL.Win.Callbacks.Shutdown_System
  Logoff            : constant Boolean := Utilities.Command_Line.Has_Argument("-logoff");
  Poweroff          : constant Boolean := Utilities.Command_Line.Has_Argument("-poweroff");
  Reboot            : constant Boolean := Utilities.Command_Line.Has_Argument("-reboot");
  Shutdown          : constant Boolean := Utilities.Command_Line.Has_Argument("-shutdown");

end Opengl.Command_Line;  
