--                              -*- Mode: Ada -*-
-- Filename        : opengl-logging.adb
-- Description     : Package with utilities for logging of opengl programs
-- Author          : J. Schrer
-- Created On      : 05. Jan 2000
-- Last Modified By: J. Schrer
-- Last Modified On: 
-- Update Count    : 1
-- Status          :

with System;

with Ada.Calendar;
with Ada.Text_Io;
with Ada.Task_Identification;
with Ada.Tags;

with Opengl.Gl;
with Opengl.Command_Line;
with Opengl.Clip_Planes;
with Opengl.Display_Lists;
with Opengl.Messages;
with Opengl.Screen;
with Opengl.Types;
with Opengl.Timer;
with Opengl.Float_32.Text_Io;

with Utilities.Command_Line;
with Utilities.Calendar_Utilities;
with Utilities.String_Utilities;
with Utilities.Files;

with Portable.Operating_System;
with Portable.Utils;

package body Opengl.Logging is
  
  use type Gl.Glfloat;
  use type Models.Model_Group_Access;
  
  Newline : constant String  := Portable.Operating_System.New_Line;
  ----------------------------------------------------------------------------
  procedure Initialize_Log_File(Os_Info : in String := "") is
    ----------------------------------------
    function Listing_Name return String is
    begin
      if Command_Line.Test or else Command_Line.Config_File = "" then
        return Utilities.Command_Line.Command_Name(Full_Path => False) & ".log";
      else
        return Utilities.Files.Prefix_Of(Command_Line.Config_File) & ".lst";
      end if;
    end Listing_Name;
    --------------------------------------
  begin
    if Command_Line.Write_No_Log_File or else
       Ada.Text_Io.Is_Open(Log_File) then
      return;
    end if;

    begin  
      Ada.Text_Io.Create(File => Log_File,
                         Mode => Ada.Text_Io.Out_File,
                         Name => Listing_Name);
    exception
      when Ada.Text_Io.Name_Error =>
        Ada.Text_Io.Create(File => Log_File,
                           Mode => Ada.Text_Io.Out_File,
                           Name => "opengl-main.log");
    end;

    Ada.Text_Io.Put_Line(Log_File, "Start: " &
       Utilities.Calendar_Utilities.Image_Of
      (Utilities.Calendar_Utilities.Time_Of(Ada.Calendar.Clock)));
    Ada.Text_Io.Put_Line(Log_File,
      "---------------------------------------------------------------------");
    Ada.Text_Io.Put_Line(Log_File, "Logfile: " & Ada.Text_Io.Name(Log_File));
    Ada.Text_Io.Put_Line(Log_File,
      "Command: " & Utilities.Command_Line.Total_Command_Line);
    Ada.Text_Io.Put_Line(Log_File, "Compiler: " & 
                         system.name'image(system.system_name));
    Ada.Text_Io.New_Line(Log_File);
    Ada.Text_Io.Put(Log_File, "Screen-width :");
    Opengl.Float_32.Text_Io.Put(Log_File, Screen.Width, 
                                Fore => 4, Aft => 3, Exp => 0);
    Ada.Text_Io.New_Line(Log_File);                      
    Ada.Text_Io.Put(Log_File, "Screen-height:");
    Opengl.Float_32.Text_Io.Put(Log_File, Screen.Height, 
                                Fore => 4, Aft => 3, Exp => 0);
    Ada.Text_Io.New_Line(Log_File);                      
    Ada.Text_Io.Put_Line(Log_File, Os_Info);      
    Ada.Text_Io.Put_Line(Log_File, "Vendor    : " & 
      Messages.Glgetstring(Gl.Gl_Vendor));
    Ada.Text_Io.Put_Line(Log_File, "Renderer  : " & 
      Messages.Glgetstring(Gl.Gl_Renderer));
    Ada.Text_Io.Put_Line(Log_File, "Version   : " & 
      Messages.Glgetstring(Gl.Gl_Version));
    Ada.Text_Io.Put_Line(Log_File, "Extensions: " &
      Messages.Glgetstring(Gl.Gl_Extensions));
  end Initialize_Log_File;
  ----------------------------------------------------------------------------  
  procedure Finalize_Log_File is
  begin
    if not Ada.Text_Io.Is_Open(Log_File) then
      return;
    end if;
--    Windows.Messages.Display_Message
--       (The_Title           => Exe_Name & " finished",
--        The_Message         => 
--          "Log_file '" & Ada.Text_Io.Name(Log_File) & "'" & Newline & 
--          "with " & Ada.Text_Io.Positive_Count'Image
--                   (Ada.Text_Io.Line(Log_File)) & " lines created!",
--        Add_Windows_Message => True);
    Ada.Text_Io.Put_Line(Log_File, "Stop:  " &
       Utilities.Calendar_Utilities.Image_Of
      (Utilities.Calendar_Utilities.Time_Of(Ada.Calendar.Clock)));
    Ada.Text_Io.Close(Log_File);
  end Finalize_Log_File;
  ----------------------------------------------------------------------------
  procedure put(text : in string) is
  begin
    if Ada.Text_Io.Is_Open(Log_File) then
      Ada.Text_Io.Put_Line(Log_File, text);
      ada.text_io.flush(log_file);
    end if;
  end put;  
  ----------------------------------------------------------------------------
  procedure Write_Current_Task_Id is
  begin    
    put("Main-task: " & Ada.Task_Identification.Image
                       (Ada.Task_Identification.Current_Task));
  end Write_Current_Task_Id; 
  ----------------------------------------------------------------------------
  procedure Write_Timing_Info(The_Models   : in Models.Window_Models_Access;
                              Window_Index : in Positive := 1) is
    --------------------------------------------------------------------
    function Sum(The_Array : in Types.Float_Vector) return Gl.Glfloat is
      The : Gl.Glfloat := 0.0;
    begin
      for I in The_Array'range loop
        The := The + The_Array(I);
      end loop;
      return The;
    end Sum;
    ---------------------------------------------------                               
    procedure Output(The_Group : in Models.Model_Group;
                     Name      : in String;
                     Total     : in Gl.Glfloat) is
    begin
      Ada.Text_Io.Put(Log_File, Name & ", Render-Counter: ");
      Ada.Text_Io.Set_Col (Log_File, 35);
      Ada.Text_Io.Put_Line(Log_File, Natural'Image(The_Group.Render_Counter));
      for I in The_Group.Models'range loop
        Ada.Text_Io.Put(Log_File, "  " &
          Ada.Tags.Expanded_Name(The_Group.Models(I)'Tag));
        if The_Group.Render_Counter > 0 then
          Ada.Text_Io.Set_Col(Log_File, 50);
          Opengl.Float_32.Text_Io.Put(Log_File, The_Group.Render_Time(I),
            Fore => 6, Aft => 3, Exp => 0);
          Ada.Text_Io.Set_Col(Log_File, 60);
          Opengl.Float_32.Text_Io.Put(Log_File, 
            1000.0 * The_Group.Render_Time(I) / 
              Gl.Glfloat(The_Group.Render_Counter),
            Fore => 5, Aft => 3, Exp => 0);
          Ada.Text_Io.Set_Col(Log_File, 70);
          Opengl.Float_32.Text_Io.Put(Log_File, 
            The_Group.Render_Time(I) / Total * 100.0,
            Fore => 5, Aft => 3, Exp => 0);
        end if;
        Ada.Text_Io.New_Line(Log_File);
      end loop;
    end Output;
    --------------------------------------------------------------    
    Total : Gl.Glfloat := 0.0;        
  begin  
    if Ada.Text_Io.Is_Open(Log_File) then
      if Command_Line.Model_Timing then
        if The_Models.Permanent_Group /= null then
          Ada.Text_Io.Put_Line(Log_File, "Counter =" &
            Natural'Image(The_Models.Permanent_Group.Render_Counter) &
            "  (Render_Counter) for Window_Index =>" & 
            Natural'Image(Window_Index));
          Total := Sum(The_Models.Permanent_Group.Render_Time);
        end if;
        for I in The_Models.Model_Groups'range loop
          Total := Total + Sum(The_Models.Model_Groups(I).Render_Time);
        end loop;
        Ada.Text_Io.Put_Line(Log_File,
          "---------------------------------------------------------------------");
        Ada.Text_Io.Put_Line(Log_File,
         --12345678901234567890123456789012345678901234567890123456789012345678901234567890  
          "Model Statistics:                                     T/s     E[T]/ms      %");
        if The_Models.Permanent_Group /= null then
          Output(The_Group => The_Models.Permanent_Group.all,
                 Name      => "Permanent Group",
                 Total     => Total);
        end if;
        for I in The_Models.Model_Groups'range loop
          Output(The_Group => The_Models.Model_Groups(I).all,
                 Name      => "Group" & Positive'Image(I),
                 Total     => Total);
        end loop;
        Ada.Text_Io.Put     (Log_File, "Total");
        Ada.Text_Io.Set_Col (Log_File, 50);
        Opengl.Float_32.Text_Io.Put(Log_File, Total, Fore => 6, Aft => 3, Exp => 0);
        Ada.Text_Io.New_Line(Log_File);
      end if;
      Ada.Text_Io.Put_Line(Log_File,
        "---------------------------------------------------------------------");
      Ada.Text_Io.Put_Line(Log_File, 
        "Used number of display-lists (the first 1000 are reserved for OpenGL):" &
        Gl.Gluint'Image(Display_Lists.Lists_Used));
      Ada.Text_Io.Put_Line(Log_File, 
        "Maximum number of used clip-planes at one time                       :" &
        Natural'Image(Clip_Planes.Max_Planes_Used));
    end if;

    if Command_Line.Timer then
      Timer.Print_Statistics
        (The_Model => Timer.Model(The_Models.Permanent_Group.Models
                     (The_Models.Permanent_Group.Models'Last).all));
    end if;
  end Write_Timing_Info;  
  ----------------------------------------------------------------------------
  procedure Write_Error_File(E : in Ada.Exceptions.Exception_Occurrence) is
    Error_File : Ada.Text_Io.File_Type;
    Time       : constant Utilities.Calendar_Utilities.Time :=
                          Utilities.Calendar_Utilities.Time_Of(Ada.Calendar.Clock);
    Date_Image : constant String := Utilities.String_Utilities.Stripped
      (From_The_String => Utilities.Calendar_Utilities.Date_Image_Of
                          (The_Time  => Time,
                           Date_Form => Utilities.Calendar_Utilities.Month_Day_Year),
       The_Character   => '/');
    Time_Image : constant String := Utilities.String_Utilities.Stripped
      (From_The_String => Utilities.Calendar_Utilities.Time_Image_Of
                          (The_Time  => Time),
       The_Character   => ':');
    Filename   : constant String := Utilities.String_Utilities.Stripped
      (Date_Image & Time_Image(Time_Image'First .. Time_Image'First + 5) & ".err");
  begin
    Ada.Text_Io.Create(Error_File, Ada.Text_Io.Out_File, Filename);
    Ada.Text_Io.Put_Line(Error_File, Utilities.Calendar_Utilities.Image_Of(Time));
    Ada.Text_Io.Put_Line(Error_File, Portable.Utils.Exception_Information(E) & Newline &
                                     Ada.Exceptions.Exception_Message(E));
    Ada.Text_Io.Close(Error_File);
  end Write_Error_File;
  ----------------------------------------------------------------------------
end Opengl.Logging;

