--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-forms.ads
-- Description     : Package for the definition of form-models. 
-- Author          : H. Altmann, A. Gaisbauer, J. Schrer
-- Created On      : 09. Feb. 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Opengl.Glu;
with Opengl.Clip_Planes;
with Opengl.Display_Lists;

package Opengl.Models.Forms is
  
  -- Simple lines

  procedure Render_2d_Line(From, To : in Types.Vector_2d);
  pragma Inline(Render_2d_Line);
  
  procedure Render_3d_Line(From, To : in Types.Vector_3d);
  pragma Inline(Render_3d_Line);

  procedure Render    (The_Points : in out Types.Points_2d);
  pragma Inline(Render);
  -- Auxiliar routine for all forms containing points_2d.

  procedure Render(The_Square_Width : in Gl.Glfloat;
    							 Angle            : in Gl.Glfloat := 0.0);
                   
  -- Renders a square with width = The_Square_Width.
                   
  type Rectangle is record
    Lower_Left  : Types.Vector_2d := (X => 0.0, Y => 0.0);
    Upper_Right : Types.Vector_2d := (X => 1.0, Y => 1.0);
  end record;

 	Default_Rectangle : Rectangle;

  procedure Render(The_Rectangle : in Rectangle;
    							 Angle         : in Gl.Glfloat := 0.0);

  -- Render a given rectangle with a given orientation.
  -- Angle is in degree.                                    

  function Display_List_Of(Rectangle : in Forms.Rectangle;
                           Color     : in Types.Color_Type) return Gl.Gluint;
   
  -- Returns a rectangle as a display_list.                         
                                                   
  function Display_List_Model_Of(Rectangle : in Forms.Rectangle;
                                 Color     : in Types.Color_Type) 
   return Display_List_Model_2d_Z;
   
  -- Returns a rectangle as a display_list model.                         


  type rectangle_and_frame_2d is new model_2d_z with record
    Width            : Gl.Glfloat       := screen.width;
    Height           : Gl.Glfloat       := screen.height;
    Frame_Line_Width : Gl.Glfloat       := 1.0;
    Frame_Color      : Types.Color_Type := Colors.White; 
  end record;
       
  function rectangle_and_frame
          (Position         : in Types.Vector_3d  := (X => 0.0, Y => 0.0, z => 0.0);
           attitude         : in gl.glfloat       := 0.0;
           Width            : in gl.glfloat       := screen.width;
           Height           : in Gl.Glfloat       := screen.height;
           Frame_Line_Width : in Gl.Glfloat       := 1.0;
           Background_Color : in Types.Color_Type := Colors.Black;
           Frame_Color      : in Types.Color_Type := Colors.White) 
    return rectangle_and_frame_2d;    

  procedure render(the_model : in out rectangle_and_frame_2d);
  
  
  function Rectangle_And_Frame
          (Position         : in Types.Vector_2d  := (X => 0.0, Y => 0.0);
           Width            : in gl.glfloat       := screen.width;
           Height           : in Gl.Glfloat       := screen.height;
           Frame_Line_Width : in Gl.Glfloat       := 1.0;
           Background_Color : in Types.Color_Type := Colors.Black;
           Frame_Color      : in Types.Color_Type := Colors.White) 
    return Gl.Gluint;    
    
  -- Generates a display_list for a rectangle at position with width and height.
  -- If Frame_Line_Width > 0 then a frame is drawn around the rectangle.    
    
  -- Circle as a static model  

  type Circle_2d_Z is new Model_Points_2d_Z with record
    Radius       : Gl.Glfloat := 1.0;
    Point_Number : Positive range 36 .. 60 * 60 * 360 := 360;
                   -- Granularity from 1 point each 10 degs .. 1 point each sec.
  end record;

  procedure Initialize(The_Model : in out Circle_2d_Z);

  -- Allocates the points-component and calculates the values.

  procedure Render(The_Model : in out Circle_2d_Z);

  ----------------------------------------------------------------------------
  -- A square and a rectangle
  
  type Square_2d_Z is new Model_2d_Z with record
    Width : Gl.Glfloat := 1.0;
  end record;                                     
  
  -- The 4 edge points of the square are at +- width/2 in x- and y-direction
  -- from the position.
  
  procedure Render(The_Model : in out Square_2d_Z);
  

  type Rectangle_2d_Z is new Model_2d_Z with record
    Dimension : Types.Rectangle := Types.Default_Rectangle;
    Points    : Rectangle       := Default_Rectangle;
  end record;

  -- The rectangle dimension (width and height) are specified by
  -- dimension(x) and dimension(y). The position component refers to the 
  -- center of the rectangle.
  -- When declaring an object of type rectangle only the dimension-component 
  -- has to be specified, the points-component is calculated by initialize
  -- and adjust.

  procedure Initialize(The_Model : in out Rectangle_2d_Z);
  procedure Adjust    (The_Model : in out Rectangle_2d_Z);

  procedure Render    (The_Model : in out Rectangle_2d_Z);
  
  type Striped_Rectangle_Parts is private;
  
  Default_Striped_Rectangle_Parts : constant Striped_Rectangle_Parts;  
    
  type Striped_Rectangle_2d_Z is new Model_2d_Z with record
    Height         : Gl.Glfloat       := 0.5;
    Stripe_Width   : Gl.Glfloat       := 0.2;
    Stripe_Spacing : Gl.Glfloat       := 0.2;
    First_Stripe   : Gl.Glfloat       := 0.0;
    Stripe_Angle   : Gl.Glfloat       := 45.0;
    Stripe_Color   : Types.Color_Type := Colors.Red;
    Hidden_Data    : Striped_Rectangle_Parts;
  end record;       
    
  -- Height, Stripe_Width, Stripe_Spacing, First_Stripe are related to the 
  -- length of the rectangle which is size_scaling.
  -- the position component refers to the lower left corner of the rectangle.
    
  procedure Initialize(The_Model : in out Striped_Rectangle_2d_Z);
  procedure Adjust    (The_Model : in out Striped_Rectangle_2d_Z);

  procedure Render    (The_Model : in out Striped_Rectangle_2d_Z);
  
  ----------------------------------------------------------------------------
  -- A disk

  type Disk_2d_Z is new Model_2d_Z with record
    Inner_Radius       : Gl.Glfloat := 0.5;
    Outer_Radius       : Gl.Glfloat := 1.0;
    Slices             : Positive range 6 .. 60 * 60 * 360 := 36;
    Loops              : Positive   := 1;
    Glu_Quadric_Object : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  -- see gluDisk
  -- For a circle inner_radius = outer_radius and loops = 1

  procedure Initialize(The_Model : in out Disk_2d_Z);
  procedure Adjust    (The_Model : in out Disk_2d_Z);
  procedure Finalize  (The_Model : in out Disk_2d_Z);

  -- Initializing and finalizing using the Glu Quadric routines.

  procedure Render    (The_Model : in out Disk_2d_Z);

  type Partial_Disk_2d_Z is new Model_2d_Z with record
    Inner_Radius       : Gl.Glfloat := 0.5;
    Outer_Radius       : Gl.Glfloat := 1.0;
    Slices             : Positive range 1 .. 60 * 60 * 360 := 36;
    Loops              : Positive   := 1;
    Start_Angle        : Gl.Glfloat := 0.0;
    Sweep_Angle        : Gl.Glfloat := 90.0;
    Glu_Quadric_Object : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  -- see gluPartialDisk
  -- For a partial circle inner_radius = outer_radius and loops = 1

  procedure Initialize(The_Model : in out Partial_Disk_2d_Z);
  procedure Adjust    (The_Model : in out Partial_Disk_2d_Z);
  procedure Finalize  (The_Model : in out Partial_Disk_2d_Z);

  -- Initializing and finalizing using the Glu Quadric routines.

  procedure Render    (The_Model : in out Partial_Disk_2d_Z);


  type Arc_2d_Z is new Model_2d_Z with record
    Inner_Radius       : Gl.Glfloat := 0.5;
    Outer_Radius       : Gl.Glfloat := 1.0;
    Slices             : Positive range 1 .. 60 * 60 * 360 := 36;
    Loops              : Positive   := 1;
    Start_Angle        : Gl.Glfloat := 0.0;
    End_Angle          : Gl.Glfloat := 90.0;
    Glu_Quadric_Object : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  -- is build on top of gluPartialDisk.
  -- Start_angle and end_angle refer to normal OpenGL notation.
  -- An Angle = 0 around Z is the positive x-axis.
  -- For a circle inner_radius = outer_radius and loops = 1

  procedure Initialize(The_Model : in out Arc_2d_Z);
  procedure Adjust    (The_Model : in out Arc_2d_Z);
  procedure Finalize  (The_Model : in out Arc_2d_Z);

  -- Initializing and finalizing using the Glu Quadric routines.

  procedure Render    (The_Model : in out Arc_2d_Z);
    
  -- Dashed arc / partial disk
  
  type Dashed_Arc_2d_Z is new Arc_2d_Z with record
    Dash_Angle : Gl.Glfloat := 1.0;
  end record;
  
  -- Slices are slices per dash.
  
  procedure Render    (The_Model : in out Dashed_Arc_2d_Z);
          
  ----------------------------------------------------------------------------
  -- Triangles

  -- Auxiliar type.

  type Triangle is record
    Points : Types.Points_2d(1 .. 3) := (others => (0.0, 0.0));
  end record;

  procedure Render    (The_Triangle : in out Triangle);


  type Triangle_2d_Z is new Model_2d_Z with record
    Points : Types.Points_2d(1 .. 3) := (others => (0.0, 0.0));
  end record;

  procedure Initialize(The_Model : in out Triangle_2d_Z);
  procedure Adjust    (The_Model : in out Triangle_2d_Z);

  procedure Render    (The_Model : in out Triangle_2d_Z);

  -- An equal sided and a symmetric triangle attached to a circle.
  -- Base_Radius and Peak_Radius refer to the x-axis.
  -- A triangle with attitude = 0 is placed on the x-axis.

  type Equal_Sided_Triangle_2d_Z is new Triangle_2d_Z with record
    Base_Radius : Gl.Glfloat := 0.8;
    Peak_Radius : Gl.Glfloat := 1.0;
  end record;
  
  procedure Initialize(The_Model : in out Equal_Sided_Triangle_2d_Z);
  procedure Adjust    (The_Model : in out Equal_Sided_Triangle_2d_Z);
  

  type Symmetric_Triangle_2d_Z is new Equal_Sided_Triangle_2d_Z with record
    Base_Width  : Gl.Glfloat := 0.4;
  end record;

  -- For an Equal_Sided_Triangle holds: 
  -- Base_Width = 2 / Sqrt(3) * abs(Peak_Radius - Base_Radius). 
  
  procedure Initialize(The_Model : in out Symmetric_Triangle_2d_Z);
  procedure Adjust    (The_Model : in out Symmetric_Triangle_2d_Z);

  type Polygon_Type(Point_Number : Positive) is record
    Points : Types.Points_2d(1 .. Point_Number) := (others => (0.0, 0.0));
  end record;

  procedure Render    (The_Polygon : in out Polygon_Type);

  type Polygon_2d_Z(Point_Number : Positive) is new Model_2d_Z with record
    Points : Types.Points_2d(1 .. Point_Number) := (others => (0.0, 0.0));
	end record;

  procedure Initialize(The_Model : in out Polygon_2d_Z);
  procedure Adjust    (The_Model : in out Polygon_2d_Z);
  procedure Render    (The_Model : in out Polygon_2d_Z);

  type Gen_Param_Type is array ( 0 .. 3) of aliased Gl.Glfloat;

  type Striped_Polygon_2d_Z(Point_Number : Positive) is
    	new Polygon_2d_Z(Point_Number) with record
    Stripe_Color  : Types.Color_Type := Colors.Red;
    Stripe_Number : Gl.Glfloat       := 4.0;
    Stripe_Angle  : Gl.Glfloat       := 45.0; --in deg
    -- internal use, set by prepare
    Texture_Id : aliased Gl.Gluint := 0;
    Gen_Param  : Gen_Param_Type;
	end record;

--  procedure Initialize(The_Model : in out Striped_Polygon_2d_Z);
--  procedure Adjust    (The_Model : in out Striped_Polygon_2d_Z);
  procedure Prepare   (The_Model : in out Striped_Polygon_2d_Z);
  procedure Render    (The_Model : in out Striped_Polygon_2d_Z);

  ----------------------------------------------------------------------------
  -- A cylinder

  type Cylinder_3d is new Model_3d with record
    Base_Radius        : Gl.Glfloat := 1.0;
    Top_Radius         : Gl.Glfloat := 1.0;
    Height             : Gl.Glfloat := 1.0;
    Slices             : Positive range 6 .. 60 * 60 * 360 := 36;
    Stacks             : Positive   := 1;
    Glu_Quadric_Object : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  -- see gluCylinder
  -- For a cylinder base_radius  = top_radius.
  -- For a cone     base_radius /= top_radius.

  procedure Initialize(The_Model : in out Cylinder_3d);
  procedure Adjust    (The_Model : in out Cylinder_3d);
  procedure Finalize  (The_Model : in out Cylinder_3d);

  procedure Render    (The_Model : in out Cylinder_3d);

  ----------------------------------------------------------------------------
  -- A sphere

  type Sphere_3d is new Model_3d with record
    Radius             : Gl.Glfloat := 1.0;
    Slices             : Positive range 6 .. 60 * 60 * 360 := 36;
    Stacks             : Positive   := 1;
    Glu_Quadric_Object : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  -- see gluSphere

  procedure Initialize(The_Model : in out Sphere_3d);
  procedure Adjust    (The_Model : in out Sphere_3d);
  procedure Finalize  (The_Model : in out Sphere_3d);

  procedure Render    (The_Model : in out Sphere_3d);

  ----------------------------------------------------------------------------
  -- Lines and line lists

  -- enumeration type of line rendering

  type Line_Type is (Gl_Points, Gl_Lines, Gl_Line_Strip, Gl_Line_Loop);

  -- generic rendering of lines with different styles

  type Line_List_2d_Z(Point_Number : Positive) is new
       Polygon_2d_Z(Point_Number) with record

    Kind        : Line_Type  := Gl_Line_Strip;
    Line_Width  : Gl.Glfloat := 5.0;
    Initialized : Boolean := False;
  end record;

  procedure Render(The_Model : in out Line_List_2d_Z);
  
  procedure Initialize(The_Model: in out Line_List_2d_Z);
  procedure Adjust(The_Model: in out Line_List_2d_Z);

  ----------------------------------------------------------------------------

  type Line_List_Type(Point_Number : Positive) is record
    Points     : Types.Points_2d(1 .. Point_Number) := (others => (0.0, 0.0));
    Kind       : Line_Type  := Gl_Line_Strip;
    Line_Width : Gl.Glfloat := 5.0;
  end record;

  ----------------------------------------------------------------------------
  -- Arrows

  -- An arrow specified by radius.

  type Radial_Arrow_Parts is private;

  Default_Radial_Arrow_Parts : constant Radial_Arrow_Parts;


  type Radial_Arrow_2d_Z is new Model_2d_Z with record
    Inner_Radius         : Gl.Glfloat := 0.0;
    Rectangle_Width      : Gl.Glfloat := 0.1;
    Triangle_Base_Radius : Gl.Glfloat := 0.8;
    Triangle_Peak_Radius : Gl.Glfloat := 1.0;
    Triangle_Base_Width  : Gl.Glfloat := 0.4;
    Parts                : Radial_Arrow_Parts;-- := Default_Radial_Arrow_Parts;
  end record;

  procedure Initialize(The_Model : in out Radial_Arrow_2d_Z);
  procedure Adjust    (The_Model : in out Radial_Arrow_2d_Z);

  procedure Render    (The_Model : in out Radial_Arrow_2d_Z);


  -- A round arrow.

  type Round_Arrow_Parts is private;

  Default_Round_Arrow_Parts : constant Round_Arrow_Parts;


  type Round_Arrow_2d_Z is new Model_2d_Z with record
    Arc_Radius          : Gl.Glfloat := 10.0;
    Arc_Width           : Gl.Glfloat := 1.0;
    Arc_Start_Angle     : Gl.Glfloat := 0.0;
    Arc_Delta_Angle     : Gl.Glfloat := 45.0;
    Triangle_Base_Width : Gl.Glfloat := 0.4;
    Triangle_Height     : Gl.Glfloat := 0.8;
    Parts               : Round_Arrow_Parts;-- := Default_Round_Arrow_Parts;
  end record;

  procedure Initialize(The_Model : in out Round_Arrow_2d_Z);
  procedure Adjust    (The_Model : in out Round_Arrow_2d_Z);

  procedure Render    (The_Model : in out Round_Arrow_2d_Z);
    
private ---------------------------------------------------------------------
  
  type Striped_Rectangle_Parts is record  
    Points           : Rectangle         := Default_Rectangle;
    Stripe_Width     : Gl.Glfloat        := 0.2;
    Stripe_Spacing   : Gl.Glfloat        := 0.2;
    First_Stripe     : Gl.Glfloat        := 0.0;
    X_Offset         : Gl.Glfloat        := 0.2;
    Number           : Natural           := 1;
    Left_Clip_Plane  : Clip_Planes.Plane := Clip_Planes.X_Clip_Plane;
    Right_Clip_Plane : Clip_Planes.Plane := Clip_Planes.X_Clip_Plane;
  end record;  
  
  Default_Striped_Rectangle_Parts : constant Striped_Rectangle_Parts :=
    (Points           => Default_Rectangle,
     Stripe_Width     => 0.2,
     Stripe_Spacing   => 0.2,
     First_Stripe     => 0.0,
     X_Offset         => 0.2,
     Number           => 1,
     Left_Clip_Plane  => Clip_Planes.X_Clip_Plane,
     Right_Clip_Plane => Clip_Planes.X_Clip_Plane);       

  type Radial_Arrow_Parts is record
--    Triangle_Part  : Triangle;
--    Rectangle_Part : Rectangle;
    Display_List : Gl.Gluint := Display_Lists.Next_List_Id;
  end record;

  The_Radial_Arrow_Parts : Radial_Arrow_Parts;

  Default_Radial_Arrow_Parts : constant Radial_Arrow_Parts := 
                                        The_Radial_Arrow_Parts;

  ----------------------------------------------------------------------------

  type Symmetric_Triangle_2d_Z_access is access Symmetric_Triangle_2d_Z;

  type Round_Arrow_Parts is record
    Triangle_Part    : Symmetric_Triangle_2d_Z_access;
    Arc_Outer_Radius : Gl.Glfloat := 10.0;
    Arc_Inner_Radius : Gl.Glfloat := 9.0;
    Arc_Start_Angle  : Gl.Glfloat := 0.0;
    Arc_Sweep_Angle  : Gl.Glfloat := 90.0;
    Arc_Slices       : Positive   := 36;
    Arc_Object       : Glu.Pgluquadricobj := Glu.Glunewquadric;
  end record;

  The_Round_Arrow_Parts : Round_Arrow_Parts;

  Default_Round_Arrow_Parts : constant Round_Arrow_Parts := The_Round_Arrow_Parts;

end Opengl.Models.Forms;
