--                    -*- Mode: Ada -*-
-- Filename        : opengl-models-lighting.adb
-- Description     : Package for the definition of lighting 
--                   and models with material properties. 
-- Author          : J. Schrer
-- Created On      : 17.3.2001
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Ada.Exceptions;
with Opengl.Messages;

package body Opengl.Models.Lighting is

  use type Gl.Glenum;
  use type Types.Float_Vector;

  Light_Ids : constant array(1 .. 8) of Gl.Glenum :=
              (1 => Gl.Gl_Light0,
               2 => Gl.Gl_Light1,
               3 => Gl.Gl_Light2,
               4 => Gl.Gl_Light3,
               5 => Gl.Gl_Light4,
               6 => Gl.Gl_Light5,
               7 => Gl.Gl_Light6,
               8 => Gl.Gl_Light7);
               
  Light_Index : Natural := 0;             

  subtype Vector_4 is Types.Float_Vector(1 .. 4);
  ----------------------------------------------------------------------------
  function To_Array(Item : in Types.Color_Type) return Vector_4 is
  begin
    return (Item.Red, Item.Green, Item.Blue, Item.Alpha);
  end To_Array; 
  ----------------------------------------------------------------------------  
  procedure Init (The_Model : in Lighting_Model) is
  begin
    if not The_Model.Initialized then
      Gl.Gllightmodelfv(Pname  => Gl.Gl_Light_Model_Ambient,
                        Params => To_Array(The_Model.Ambient_Color)(1)'
                                           Unchecked_Access);
      Gl.Gllightmodeli (Pname  => Gl.Gl_Light_Model_Local_Viewer,
                        Param  => Gl.Glint(The_Model.Local_Viewer));
      Gl.Gllightmodeli (Pname  => Gl.Gl_Light_Model_Two_Side,
                        Param  => Gl.Glint(The_Model.Two_Side));    
--      Gl.Gllightmodelf (Pname  => Gl.Gl_Light_Model_Local_Viewer,
--                        Param  => The_Model.Local_Viewer);
--      Gl.Gllightmodelf (Pname  => Gl.Gl_Light_Model_Two_Side,
--                        Param  => The_Model.Two_Side);
    end if;                      
  end Init;
  ----------------------------------------------------------------------------
  procedure Enable_Lighting is
  begin
    Gl.Glshademodel(Gl.Gl_Smooth);                      
    Gl.Glenable(Gl.Gl_Depth_Test);
    Gl.Glenable(Gl.Gl_Lighting);    
  end Enable_Lighting;
  ----------------------------------------------------------------------------
  procedure Disable_Lighting is
  begin
    Gl.Gldisable(Gl.Gl_Depth_Test);
    Gl.Gldisable(Gl.Gl_Lighting);
  end Disable_Lighting;
  ----------------------------------------------------------------------------
  procedure Init (The_Model : in out Light_Source) is
  begin
    if The_Model.Id = Gl.Glenum'Last then
      Light_Index := Light_Index + 1;
      if Light_Index in Light_Ids'range then
        The_Model.Id := Light_Ids(Light_Index);
      else
        Ada.Exceptions.Raise_Exception(Opengl_Error'Identity, Message =>
          Messages.Error_Message
            (Subprogram => "Opengl.Models.Lighting.Init(Light_Source)",
             Message    => "Tried to enable" & Natural'Image(Light_Index) & 
                           "-th light-source. Max number is" &
                           Natural'Image(Light_Ids'Length)));                     
      end if;  
    elsif not The_Model.Initialized then  
      Gl.Gllightfv(Light  => The_Model.Id,
                   Pname  => Gl.Gl_Ambient,
                   Params => To_Array(The_Model.Ambient_Color)(1)'
                                      Unchecked_Access);
      Gl.Gllightfv(Light  => The_Model.Id,
                   Pname  => Gl.Gl_Diffuse,
                   Params => To_Array(The_Model.Diffuse_Color)(1)'
                                      Unchecked_Access);
      Gl.Gllightfv(Light  => The_Model.Id,
                   Pname  => Gl.Gl_Specular,
                   Params => To_Array(The_Model.Specular_Color)(1)'
                                      Unchecked_Access);
      Gl.Gllightf (Light  => The_Model.Id,
                   Pname  => Gl.Gl_Constant_Attenuation,
                   Param  => The_Model.Constant_Attenuation);
      Gl.Gllightf (Light  => The_Model.Id,
                   Pname  => Gl.Gl_Linear_Attenuation,
                   Param  => The_Model.Linear_Attenuation);
      Gl.Gllightf (Light  => The_Model.Id,
                   Pname  => Gl.Gl_Quadratic_Attenuation,
                   Param  => The_Model.Quadratic_Attenuation);
      Gl.Gllightf (Light  => The_Model.Id,
                   Pname  => Gl.Gl_Spot_Cutoff,
                   Param  => The_Model.Spot_Cutoff);
      Gl.Gllightf (Light  => The_Model.Id,
                   Pname  => Gl.Gl_Spot_Exponent,
                   Param  => The_Model.Spot_Exponent);
      Gl.Gllightfv(Light  => The_Model.Id,
                   Pname  => Gl.Gl_Spot_Direction,
                   Params => The_Model.Spot_Direction(X)'Unchecked_Access);                     
      The_Model.Initialized := True;             
    end if;  
  end Init;
  ----------------------------------------------------------------------------
  procedure Enable (The_Model : in out Light_Source) is
  begin
    Gl.Glenable(The_Model.Id);
  end Enable;
  ----------------------------------------------------------------------------
  procedure Disable (The_Model : in out Light_Source) is
  begin
    Gl.Gldisable(The_Model.Id);
  end Disable;
  ----------------------------------------------------------------------------
  procedure Render (The_Model : in out Light_Source) is  
    Position : constant Vector_4 := The_Model.Position & The_Model.Position_W;
  begin
    The_Model.Position := (0.0, 0.0, 0.0);
    Initialize_Render(The_Model);

--    Init(The_Model);
    Gl.Gllightfv(Light  => The_Model.Id,
                 Pname  => Gl.Gl_Position,
                 Params => Position(X)'Unchecked_Access);                         
--    Enable(The_Model);

    Finalize_Render  (The_Model);
    The_Model.Position := Position(X .. Z);
  end Render;
  ----------------------------------------------------------------------------
  procedure Init (The_Model : in out Real_Model'Class) is
  begin
    if not The_Model.Initialized then
      Gl.Glmaterialfv(Face   => The_Model.Face,
                      Pname  => Gl.Gl_Ambient,
                      Params => To_Array(The_Model.Ambient_Color)(1)'
                                         Unchecked_Access);
      Gl.Glmaterialfv(Face   => The_Model.Face,
                      Pname  => Gl.Gl_Diffuse,
                      Params => To_Array(The_Model.Diffuse_Color)(1)'
                                         Unchecked_Access);
      Gl.Glmaterialfv(Face   => The_Model.Face,
                      Pname  => Gl.Gl_Specular,
                      Params => To_Array(The_Model.Specular_Color)(1)'
                                         Unchecked_Access);
      Gl.Glmaterialfv(Face   => The_Model.Face,
                      Pname  => Gl.Gl_Emission,
                      Params => To_Array(The_Model.Emission_Color)(1)'
                                         Unchecked_Access);                    
      Gl.Glmaterialf (Face   => The_Model.Face,
                      Pname  => Gl.Gl_Shininess,
                      Param  => The_Model.Shininess);
      The_Model.Initialized := True;
    end if;                
  end Init;
  ----------------------------------------------------------------------------
  procedure Initialize_Render(The_Model : in out Real_Model'Class) is
  begin
    Initialize_Render(Model_3d(The_Model));
    Init(The_Model);
  end Initialize_Render;
  ----------------------------------------------------------------------------
end Opengl.Models.Lighting;
