--                    -*- Mode: Ada -*-
-- Filename        : opengl-models.ads
-- Description     : Package for the definition of basic opengl-models. 
--                   higher level packages based on Opengl
-- Author          : J. Schrer
-- Created On      : 27. Jan. 1998
-- Last Modified By:
-- Last Modified On: ... Feb. 1998
-- Update Count    : 
-- Status          :

with Ada.Exceptions;
with Ada.Text_Io.Text_Streams;
with Ada.Unchecked_Deallocation;
with Opengl.Display_Lists;
with Opengl.Command_Line;
with Opengl.Messages;
with Opengl.Logging;

package body Opengl.Models is

  use type Ada.Real_Time.Time,
           Gl.Glfloat,
           Gl.Glbitfield,
           Gl.Glboolean,
           Types.Color_Type,
           Types.Points_2d,        Types.Points_3d,
           Types.Points_2d_Access, Types.Points_3d_Access,
           Types.Vector_3d;
  ----------------------------------------------------------------------------
  procedure Antialiasing_On is
  begin           
    Gl.Glenable(Gl.Gl_Line_Smooth);
    Gl.Glenable(Gl.Gl_Point_Smooth);
    Gl.Glenable(Gl.Gl_Polygon_Smooth);
  end Antialiasing_On;
  ----------------------------------------------------------------------------    
  Rotation_Axes : constant array(X .. Z) of Types.Vector_3d :=
      		    (X => (X => 1.0, Y => 0.0, Z => 0.0),
			     Y => (X => 0.0, Y => 1.0, Z => 0.0),
      		     Z => (X => 0.0, Y => 0.0, Z => 1.0));

  Actual_Color  : Types.Color_Type := Opengl.Colors.Black;
  ----------------------------------------------------------------------------
  procedure Set_Color(The_Color : in Types.Color_Type;
                      Force     : in Boolean := False) is
  begin
    if Force or else Actual_Color /= The_Color then
      Gl.Glcolor4f(Red   => Gl.Glfloat(The_Color.Red),
                   Green => Gl.Glfloat(The_Color.Green),
                   Blue  => Gl.Glfloat(The_Color.Blue),
                   Alpha => Gl.Glfloat(The_Color.Alpha));
      Actual_Color := The_Color;
    end if;
  end Set_Color;
  ----------------------------------------------------------------------------
  procedure Set_Color(The_Model : in Model'Class;
                      Force     : in Boolean := False) is
  begin
    Set_Color(The_Model.Color, Force);
  end Set_Color;
  ----------------------------------------------------------------------------
  procedure Translate(The_Vector : in Types.Vector_3d) is
  begin
    Gl.Gltranslatef(X => The_Vector(X),
  	                Y => The_Vector(Y),
  	                Z => The_Vector(Z));
  end Translate;
  ----------------------------------------------------------------------------
  procedure Rotate   (The_Angle  : in Gl.Glfloat;
                      The_Axis   : in Types.Vector_3d) is
  begin
    Gl.Glrotatef(Angle => The_Angle,
                 X     => The_Axis(X),
                 Y     => The_Axis(Y),
                 Z     => The_Axis(Z));
  end Rotate;
  ----------------------------------------------------------------------------
  procedure Translate_X(The_Value : in Gl.Glfloat) is
  begin
    Gl.Gltranslatef(X => The_Value, Y => 0.0, Z => 0.0);
  end Translate_X;
  ----------------------------------------------------------------------------    
  procedure Translate_Y(The_Value : in Gl.Glfloat) is
  begin
    Gl.Gltranslatef(X => 0.0, Y => The_Value, Z => 0.0);
  end Translate_Y;
  ---------------------------------------------------------------------------- 
  procedure Translate_Z(The_Value : in Gl.Glfloat) is
  begin
    Gl.Gltranslatef(X => 0.0, Y => 0.0, Z => The_Value);
  end Translate_Z;
  ---------------------------------------------------------------------------- 
  procedure Rotate_X (The_Angle  : in Gl.Glfloat) is
  begin
    Rotate(The_Angle, Rotation_Axes(X));
  end Rotate_X;
  ----------------------------------------------------------------------------
  procedure Rotate_Y (The_Angle  : in Gl.Glfloat) is
  begin
    Rotate(The_Angle, Rotation_Axes(Y));
  end Rotate_Y;
  ----------------------------------------------------------------------------
  procedure Rotate_Z (The_Angle  : in Gl.Glfloat) is
  begin
    Rotate(The_Angle, Rotation_Axes(Z));
  end Rotate_Z;
  ----------------------------------------------------------------------------
  procedure Translate(The_Model  : in Model_2d_Z'Class) is
  begin
    Translate(The_Model.Position);
  end Translate;
  ----------------------------------------------------------------------------
  procedure Rotate_Z (The_Model  : in Model_2d_Z'Class) is
  begin
    Rotate_Z(The_Model.Attitude);
  end Rotate_Z;
  ----------------------------------------------------------------------------
  procedure Translate(The_Model  : in Model_3d'Class) is
  begin
    Translate(The_Model.Position);
  end Translate;
  ----------------------------------------------------------------------------
  procedure Rotate   (The_Model  : in Model_3d'Class;
			    Rotation_1 : in Rotation_Axis := X;
			    Rotation_2 : in Rotation_Axis := Y;
			    Rotation_3 : in Rotation_Axis := Z) is
  begin
    Rotate(The_Angle => The_Model.Attitude(Rotation_1),
           The_Axis  => Rotation_Axes     (Rotation_1));
    Rotate(The_Angle => The_Model.Attitude(Rotation_2),
           The_Axis  => Rotation_Axes     (Rotation_2));
    Rotate(The_Angle => The_Model.Attitude(Rotation_3),
           The_Axis  => Rotation_Axes     (Rotation_3));
  end Rotate;
  ----------------------------------------------------------------------------
  procedure Rotate_X (The_Model  : in Model_3d'Class) is
  begin
    Rotate_X(The_Model.Attitude(X));
  end Rotate_X;
  ----------------------------------------------------------------------------
  procedure Rotate_Y (The_Model  : in Model_3d'Class) is
  begin
    Rotate_Y(The_Model.Attitude(Y));
  end Rotate_Y;
  ----------------------------------------------------------------------------
  procedure Rotate_Z (The_Model  : in Model_3d'Class) is
  begin
    Rotate_Z(The_Model.Attitude(Z));
  end Rotate_Z;
  ----------------------------------------------------------------------------
  procedure Initialize_Render(The_Model : in out Model_2d_Z'Class) is
  begin
    if Model_Name > 0 then
      Gl.Glloadname(Gl.Gluint(Model_Name));
    end if;
    if The_Model not in Display_List_Model_2d_Z then
      Set_Color(The_Model);
    end if;  

    Gl.Glpushmatrix;

    if Scale_Model then
      Gl.Glscalef(X => The_Model.Size_Scaling,
                  Y => The_Model.Size_Scaling,
                  Z => 1.0);
    end if;

    case Motion_Sequence is
      when Translate_First =>
        Translate(The_Model);
	      Rotate_Z (The_Model);
      when Rotate_First    =>
        Rotate_Z (The_Model);
	      Translate(The_Model);
    end case;
  end Initialize_Render;
  ----------------------------------------------------------------------------
  procedure Finalize_Render(The_Model : in out Model_2d_Z'Class) is
  begin
    Model_Name := 0;
    Gl.Glpopmatrix;
  end Finalize_Render;
  ----------------------------------------------------------------------------
  procedure Initialize_Render(The_Model : in out Model_3d'Class) is
  begin
    if Model_Name > 0 then
      Gl.Glloadname(Gl.Gluint(Model_Name));
    end if;
    if The_Model not in Display_List_Model_3d then
      Set_Color(The_Model);
    end if;  

    Gl.Glpushmatrix;

    if Scale_Model then
      Gl.Glscalef(X => The_Model.Size_Scaling,
                  Y => The_Model.Size_Scaling,
                  Z => The_Model.Size_Scaling);
    end if;

    case Motion_Sequence is
      when Translate_First =>
        Translate(The_Model);
        Rotate(The_Model, Rotation_1, Rotation_2, Rotation_3);
      when Rotate_First    =>
        Rotate(The_Model, Rotation_1, Rotation_2, Rotation_3);
        Translate(The_Model);
    end case;
  end Initialize_Render;
  ----------------------------------------------------------------------------
  procedure Finalize_Render(The_Model : in out Model_3d'Class) is
  begin
    Model_Name := 0;
    Gl.Glpopmatrix;
  end Finalize_Render;
  ----------------------------------------------------------------------------
  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   : in     Model_Input_Access) is
  begin
    if Item = null then
      Boolean'Write(Stream, False);
    else
      Boolean'Write(Stream, True);
      Model_Input'Class'Output(Stream, Item.all);
    end if;
  end Write;
  ----------------------------------------------------------------------------
  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out Model_Input_Access) is
    Is_Not_Null : Boolean;
  begin
    Boolean'Read(Stream, Is_Not_Null);
    if Is_Not_Null then
      Item := new Model_Input'Class'(Model_Input'Class'Input(Stream));
    else
      Item := null;
    end if;
  end Read;
  ----------------------------------------------------------------------------          
  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   : in     Root_Model_Access) is
  begin
    if Item = null then
      Boolean'Write(Stream, False);
    else
      Boolean'Write(Stream, True);
      Root_Model'Class'Output(Stream, Item.all);
    end if;
  end Write;
  ----------------------------------------------------------------------------
  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out Root_Model_Access) is
    Is_Not_Null : Boolean;
  begin
    Boolean'Read(Stream, Is_Not_Null);
    if Is_Not_Null then
      Item := new Root_Model'Class'(Root_Model'Class'Input(Stream));
    else
      Item := null;
    end if;
  end Read;
  ----------------------------------------------------------------------------          
  procedure Initialize(The_Model : in out Model_Points_2d_Z) is
  begin
    null;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust    (The_Model : in out Model_Points_2d_Z) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Finalize  (The_Model : in out Model_Points_2d_Z) is
  begin
    Types.Free(The_Model.Points);
  end Finalize;
  ----------------------------------------------------------------------------
  procedure Initialize(The_Model : in out Model_Points_3d) is
  begin
    null;
  end Initialize;
  ----------------------------------------------------------------------------
  procedure Adjust    (The_Model : in out Model_Points_3d) is
  begin
    Initialize(The_Model);
  end Adjust;
  ----------------------------------------------------------------------------
  procedure Finalize  (The_Model : in out Model_Points_3d) is
  begin
    Types.Free(The_Model.Points);
  end Finalize;
  ----------------------------------------------------------------------------
  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   : in     Model_Group_Access) is
  begin
    if Item = null then
      Boolean'Write(Stream, False);
    else
      Boolean'Write(Stream, True);
      Model_Group'Output(Stream, Item.all);
    end if;
  end Write;
  ----------------------------------------------------------------------------
  procedure Read  (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   :    out Model_Group_Access) is
    Is_Not_Null : Boolean;
  begin
    Boolean'Read(Stream, Is_Not_Null);
    if Is_Not_Null then
      Item := new Model_Group'(Model_Group'Input(Stream));
    else
      Item := null;
    end if;
  end Read;
  ----------------------------------------------------------------------------                         
  procedure Write(The      : in Window_Models;
                  Filename : in String) is
    Stream : Ada.Text_Io.Text_Streams.Stream_Access;
    File   : Ada.Text_Io.File_Type;
  begin
    Ada.Text_Io.Create(File, Ada.Text_Io.Out_File, Filename);
    Stream := Ada.Text_Io.Text_Streams.Stream(File);
    Window_Models'Output(Stream, The);
    Ada.Text_Io.Close(File);
  end Write;
  ----------------------------------------------------------------------------
  function Read(Filename : in String) return Window_Models is
    Stream : Ada.Text_Io.Text_Streams.Stream_Access;
    File   : Ada.Text_Io.File_Type;
  begin
    Ada.Text_Io.Open(File, Ada.Text_Io.In_File, Filename);
    Stream := Ada.Text_Io.Text_Streams.Stream(File);
    declare
      The_Models : Window_Models := Window_Models'Input(Stream);
    begin
      Ada.Text_Io.Close(File);
      return The_Models;
    end;
  end Read;
  ----------------------------------------------------------------------------
  procedure Render(The_Model : in out Display_List_Root_Model) is
  begin
    Gl.Glcalllist(The_Model.Display_List);
  end Render;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Model      : in Root_Model'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Root_Model is

    The : Display_List_Root_Model :=
         (Ada.Finalization.Controlled with
          Display_List => The_List);

    Local_Model : Root_Model'Class := The_Model;
  begin
    Display_Lists.Generate(The.Display_List);
    Gl.Glnewlist(List => The.Display_List, Mode => Gl.Gl_Compile);
    for I in 1 .. Render_Counter loop
      Render(Local_Model);
    end loop;  
    Gl.Glendlist;
    
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Root_Model'Class is
             
    The : Display_List_Root_Model :=
         (Ada.Finalization.Controlled with
          Display_List => The_Display_List);
  begin
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------                 
  procedure Render(The_Models : in out Display_List_Root_Model_Array) is
  begin
    for I in The_Models'range loop
      Render(The_Models(I));
    end loop;
  end Render;
  ----------------------------------------------------------------------------    
  procedure Render(The_Model : in out Display_List_Model_2d_Z) is
  begin
    Initialize_Render(The_Model);
    Gl.Glcalllist(The_Model.Display_List);
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Model      : in Model_2d_Z'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Model_2d_Z is

    The : Display_List_Model_2d_Z :=
         (Ada.Finalization.Controlled with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Position     => The_Model.Position,
          Attitude     => The_Model.Attitude,
          Display_List => The_List);

    Local_Model : Model_2d_Z'Class := The_Model;
  begin
    Local_Model.Position := (0.0, 0.0, 0.0);
    Local_Model.Attitude := 0.0;
    Display_Lists.Generate(The.Display_List);
    Gl.Glnewlist(List => The.Display_List, Mode => Gl.Gl_Compile);
    Set_Color(The_Model.Color, Force => True);
    for I in 1 .. Render_Counter loop
      Render(Local_Model);
    end loop;  
    Gl.Glendlist;
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Model_2d_Z'Class is
             
    The : Display_List_Model_2d_Z;
  begin
    The.Display_List := The_Display_List;
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------                     
  procedure Render(The_Models : in out Display_List_Model_2d_Z_array) is
  begin
    for I in The_Models'range loop
      Render(The_Models(I));
    end loop;
  end Render;
  ----------------------------------------------------------------------------    
  procedure Render(The_Model : in out Display_List_Model_3d) is
  begin
    Initialize_Render(The_Model);
    Gl.Glcalllist(The_Model.Display_List);
    Finalize_Render(The_Model);
  end Render;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Model      : in Model_3d'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Model_3d is

    The : Display_List_Model_3d :=
         (Ada.Finalization.Controlled with
          Size_Scaling => The_Model.Size_Scaling,
          Color        => The_Model.Color,
          Position     => The_Model.Position,
          Attitude     => The_Model.Attitude,
          Display_List => The_List);

    Local_Model : Model_3d'Class := The_Model;
  begin
    Local_Model.Position := (0.0, 0.0, 0.0);
    Local_Model.Attitude := (0.0, 0.0, 0.0);
    Display_Lists.Generate(The.Display_List);
    Gl.Glnewlist(List => The.Display_List, Mode => Gl.Gl_Compile);
    Set_Color(The_Model.Color, Force => True);
    for I in 1 .. Render_Counter loop
      Render(Local_Model);
    end loop; 
    Gl.Glendlist;
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------
  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Model_3d'Class is
             
    The : Display_List_Model_3d;
  begin
    The.Display_List := The_Display_List;
    return The;
  end Display_List_Model_Of;
  ----------------------------------------------------------------------------                     
  procedure Render(The_Models : in out Display_List_Model_3d_Array) is
  begin
    for I in The_Models'range loop
      Render(The_Models(I));
    end loop;
  end Render;
  ----------------------------------------------------------------------------    
  procedure Prepare_Render(Background_Color : in Types.Color_Type) is
    pragma Inline(Prepare_Render);
  begin
    Gl.Glclearcolor(Red   => Gl.Glclampf(Background_Color.Red),
 		        Green => Gl.Glclampf(Background_Color.Green),
			  Blue  => Gl.Glclampf(Background_Color.Blue),
			  Alpha => Gl.Glclampf(Background_Color.Alpha));
    Gl.Glclear(Mask => Gl.Glbitfield(Gl.Gl_Color_Buffer_Bit) or 
                       Gl.Glbitfield(Gl.Gl_Depth_Buffer_Bit));
                      
    -- Rectangle around screen             
--    Gl.Glpolygonmode(Gl.Gl_Front_And_Back, Gl.Gl_Line);
--    Gl.Glbegin(Gl.Gl_Quads);
--    Gl.Glvertex2f(0.0, 0.0);
--    Gl.Glvertex2f(0.0, Screen.Height);
--    Gl.Glvertex2f(Screen.Width, Screen.Height);
--    Gl.Glvertex2f(Screen.Width, 0.0);
--    Gl.Glend;    

--    Gl.Glenable(Gl.Gl_Cull_Face);

    Gl.Glpolygonmode(Gl.Gl_Front_And_Back, Gl.Gl_Fill);
    Gl.Glmatrixmode(Gl.Gl_Modelview);
    Gl.Glloadidentity;

    Gl.Glenable(Gl.Gl_Blend);
    Gl.Glblendfunc(Sfactor => Gl.Gl_Src_Alpha,
                   Dfactor => Gl.Gl_One_Minus_Src_Alpha);
  end Prepare_Render;
  ----------------------------------------------------------------------------
  procedure Render(The_Group : in out Model_Group) is
    Start_Time : Ada.Real_Time.Time;
  begin
    if The_Group.Render = null or else The_Group.Render.all then
      for I in The_Group.Models'range loop
        if The_Group.Models(I) /= null then
          if The_Group.Input_Calculation(I) /= null then
--        begin -- The input-procedures fail e.g. if 
                -- NaNs come from the network buffer.
	      The_Group.Input_Calculation(I)
                    (The_Model => The_Group.Models(I).all);
--        exception
--          when others => null;
--        end;
          end if;
          if Command_Line.Model_Timing then
            Start_Time := Ada.Real_Time.Clock;
          end if;
          
          Render(The_Group.Models(I).all);
          
          if Command_Line.Model_Timing then
            The_Group.Render_Time(I) := The_Group.Render_Time(I) +
                                        Gl.Glfloat(Ada.Real_Time.To_Duration
                                        (Ada.Real_Time.Clock - Start_Time));
          end if;
        end if;
      end loop;

      begin
        The_Group.Render_Counter := The_Group.Render_Counter + 1;
      exception
        when Constraint_Error =>
          The_Group.Render_Counter := 1;
          Logging.Put("Opengl.Models.Render, Render_Counter reset to 0!");
      end;
    end if;
  end Render;
  ----------------------------------------------------------------------------    
  procedure Render(The : in out Window_Models) is
  begin
    Prepare_Render(The.Background_Color);

    for I in The.Model_Groups'range loop
      if The.Model_Groups(I) /= null then
        Render(The.Model_Groups(I).all);
      end if;
    end loop;

    if The.Permanent_Group /= null then
      Render(The.Permanent_Group.all);
    end if;
  exception            -- In the windows-implementation this handler is in 
    when E : others => -- function OpenGL.Win.General_C_Called_Windows_Callback. 
      Messages.Log(The_Exception => E);
      Ada.Exceptions.Reraise_Occurrence(E); 
  end Render;
  ----------------------------------------------------------------------------    
  procedure Free(Item  : in out Model_Input_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Model_Input'Class, Name => Model_Input_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      Finalize(Item.all);
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------          
  procedure Free(Item : in out Model_Input_Array_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Model_Input_Array, Name => Model_Input_Array_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      for I in Item'range loop
        Free(Item(I));
      end loop;
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------
  procedure Free(Item : in out Root_Model_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Root_Model'Class, Name => Root_Model_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      Finalize(Item.all);
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------          
  procedure Free(Item : in out Root_Model_Array_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Root_Model_Array, Name => Root_Model_Array_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      for I in Item'range loop
        Free(Item(I));
      end loop;
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------
  procedure Free(Item : in out Model_Parameter_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Model_Parameter'Class, Name => Model_Parameter_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      Finalize(Item.all);
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------      
  procedure Free(Item : in out Model_Group_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Model_Group, Name => Model_Group_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      for I in Item.Models'range loop
        Free(Item.Models(I));
      end loop;
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------          
  procedure Free(Item : in out Window_Models_Access) is

    procedure Deallocate is new Ada.Unchecked_Deallocation
      (Object => Window_Models, Name => Window_Models_Access);
    pragma Inline(Deallocate);
  begin
    if Item /= null then
      Free(Item.Permanent_Group);
      for I in Item.Model_Groups'range loop
        Free(Item.Model_Groups(I));
      end loop;
      Deallocate(Item);
    end if;
  end Free;
  ----------------------------------------------------------------------------          
end Opengl.Models;
