--                    -*- Mode: Ada -*-
-- Filename        : opengl-models.ads
-- Description     : Package for the definition of basic opengl-models. 
--                   higher level packages based on Opengl
-- Author          : J. Schrer
-- Created On      : 27. Jan. 1998
-- Last Modified By:
-- Last Modified On: 09. June 1998
-- Update Count    : 
-- Status          : 

with Ada.Streams;
with Ada.Finalization;
with Ada.Real_Time;
with Opengl.Gl;
with Opengl.Types;
with Opengl.Colors;
with Opengl.Screen;

package Opengl.Models is

  use type Gl.Glfloat;

  type Model_Input is abstract new Ada.Finalization.Controlled with null record;

  type Model_Input_Access       is access all Model_Input'Class;
  type Model_Input_Array        is array(Natural range <>) of Model_Input_Access;
  type Model_Input_Array_Access is access all Model_Input_Array;

  procedure Free(Item : in out Model_Input_Access);
  -- To make a deep free, the finalize procedure for derived types, that 
  -- contain access components have to be overwritten.

  procedure Free(Item : in out Model_Input_Array_Access);

  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   : in     Model_Input_Access);

  procedure Read  (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   :    out Model_Input_Access);

  for Model_Input_Access'Write  use Write;
  for Model_Input_Access'Read   use Read;

  -- Inputs alter the state of an animated_model

  -- Classes for basic models (leafs) ----------------------------------------

  type Root_Model is abstract new Ada.Finalization.Controlled with null record;

  -- Primitive subprograms of class root_model.

  procedure Render(The_Model : in out Root_Model) is abstract;

  -- Draw the model with current state.
  -- Render has to be owerwritten for each concrete subclass.
  -- Therefore it is abstract.
  -- (see e.g.: Cohen, Ada as a second language,
  --  chapter 12 Classwide programming, page 499 .. 590,
  --  subchapter 12.4.5 Abstract types and subprograms, top of page 553.)  

  type Root_Model_Access       is access all Root_Model'Class;
  type Root_Model_Array        is array(Natural range <>) of Root_Model_Access;
  type Root_Model_Array_Access is access all Root_Model_Array;

  procedure Free(Item : in out Root_Model_Access);
  -- To make a deep free, the finalize procedure for derived types, that 
  -- contain access components have to be overwritten.

  procedure Free(Item : in out Root_Model_Array_Access);

  procedure Write(Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   : in     Root_Model_Access);

  procedure Read (Stream : access Ada.Streams.Root_Stream_Type'Class;
                  Item   :    out Root_Model_Access);

  for Root_Model_Access'Write  use Write;
  for Root_Model_Access'Read   use Read;


  type Model_Input_Procedure is access
    procedure(The_Model : in out Root_Model'Class);

  type Model_Input_Procedures is array (Natural range <>) of
       Model_Input_Procedure;

  -- To calculate / assign the input of a model.
  ----------------------------------------------------------------------

  type Model is abstract new Root_Model with record
    Size_Scaling    : Gl.Glfloat       := 1.0;
    Color           : Types.Color_Type := Opengl.Colors.White;
--    outer_Rectangle : Types.Screen_Rectangle;
  end record;

  ----------------------------------------------------------------------
  -- Types in class-tree model are normally used for not moving objects.

  type Model_2d_Z is abstract new Model with record
    Position : Types.Vector_3d  := (X => 0.0, Y => 0.0, Z => 0.0);
               -- Origin of reference point
    Attitude : Gl.Glfloat := 0.0;
               -- Angle about z-axis
  end record;

  -- 2D model with Z-position component.


  type Model_3d is abstract new Model with record
    Position : Types.Vector_3d  := (X => 0.0, Y => 0.0, Z => 0.0);
               -- Origin of reference point
    Attitude : Types.Vector_3d  := (X => 0.0, Y => 0.0, Z => 0.0);
              -- Angles about all 3 axes
  end record;

  -- Full 3D model.

  ----------------------------------------------------------------------------
  -- Display-list models

  type Display_List_Root_Model is new Root_Model with record
    Display_List : Gl.Gluint := 0;
  end record;

  procedure Render(The_Model : in out Display_List_Root_Model);

  function Display_List_Model_Of(The_Model      : in Root_Model'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Root_Model;

  -- Generates a display-list for a static model by calling render for The_Model.           
  -- If The_List = 0 a new list-id is created. If The_List /= 0 there alredy
  -- exists a list-id for The_Model which is reused.         
           
  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Root_Model'Class;

  -- Assigns the_display_list into the returned Display_List_Root_Model.           
           
  Default_Display_List_Root_Model : Display_List_Root_Model;
  
  
  type Display_List_Root_Model_Array is array(Integer range <>) of
       Display_List_Root_Model;

  procedure Render(The_Models : in out Display_List_Root_Model_Array);

             
  type Display_List_Model_2d_Z is new Model_2d_Z with record
    Display_List : Gl.Gluint := 0;
  end record;  

  procedure Render(The_Model : in out Display_List_Model_2d_Z);

  function Display_List_Model_Of(The_Model      : in Model_2d_Z'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Model_2d_Z;

  -- Generates a display-list by calling render for The_Model.           
  -- If The_List = 0 a new list-id is created. If The_List /= 0 there alredy
  -- exists a list-id for The_Model which is reused.         
  
  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Model_2d_Z'Class;

  -- Assigns the_display_list into the returned Display_List_Root_Model.           
           
  Default_Display_List_Model_2d_Z : Display_List_Model_2d_Z;


  type Display_List_Model_2d_Z_array is array(Integer range <>) of
       Display_List_Model_2d_Z;

  procedure Render(The_Models : in out Display_List_Model_2d_Z_array);


  type Display_List_Model_3d is new Model_3d with record
    Display_List : Gl.Gluint := 0;
  end record;
  
  procedure Render(The_Model : in out Display_List_Model_3d);

  function Display_List_Model_Of(The_Model      : in Model_3d'Class;
                                 Render_Counter : in Positive  := 1;
                                 The_List       : in Gl.Gluint := 0)
           return Display_List_Model_3d;

  -- Generates a display-list by calling render for The_Model.
  -- If The_List = 0 a new list-id is created. If The_List /= 0 there alredy
  -- exists a list-id for The_Model which is reused.         

  function Display_List_Model_Of(The_Display_List : in Gl.Gluint)
           return Display_List_Model_3d'Class;

  -- Assigns the_display_list into the returned Display_List_Model_3d.           
           
  Default_Display_List_Model_3d : Display_List_Model_3d;
  

  type Display_List_Model_3d_Array is array(Integer range <>) of
       Display_List_Model_3d;

  procedure Render(The_Models : in out Display_List_Model_3d_Array);

  ----------------------------------------------------------------------------
  -- Access types to model constructor functions.
  -- All other parameters than position, attitude, scale and color that are
  -- necessary to define a model shall be defined in a type derived from
  -- Model_Parameter.
  -- For complex models consisting of multiple parts, the color parameter
  -- will be ignored and is defined in the actual constructor function.


  type Model_Parameter is new Ada.Finalization.Controlled with null record;
  type Model_Parameter_Access is access all Model_Parameter'Class;

  procedure Free(Item : in out Model_Parameter_Access);
  -- To make a deep free, the finalize procedure for derived types, that 
  -- contain access components have to be overwritten.

  No_Model_Parameter : constant Model_Parameter :=
    (Ada.Finalization.Controlled with null record);


  type Model_2d_Z_constructor_function is access
    function (Position  : in Types.Vector_3d := (X => Screen.Width  / 2.0,
                                                 Y => Screen.Height / 2.0,
                                                 Z => 0.0);
					    Attitude  : in Gl.Glfloat      := 0.0;
              Scale     : in Gl.Glfloat      := Screen.Height / 2.0;
              Parameter : in Model_Parameter'Class := No_Model_Parameter)
    return Root_Model_Access;

  -- Actual functions assigned to Model_2d_Z_constructor_function return
  -- access values pointing to models in the subclass Model_2d_Z'class.     

  type Model_3d_Constructor_Function is access
    function (Position  : in Types.Vector_3d := (X => Screen.Width  / 2.0,
                                                 Y => Screen.Height / 2.0,
                                                 Z => 0.0);
					    Attitude  : in Types.Vector_3d := (X => 0.0,
                                                 Y => 0.0,
                                                 Z => 0.0);
              Scale     : in Gl.Glfloat      := Screen.Height / 2.0;
              Parameter : in Model_Parameter'Class := No_Model_Parameter)
    return Root_Model_Access;

  -- Actual functions assigned to Model_3d_Constructor_function return
  -- access values pointing to models in the subclass Model_3d'class.     

  ----------------------------------------------------------------------
  -- Types in class animated_model are normally used for moving objects, 
  -- i.e. models with changing state.
  -- The state of the model depends on the input.

  type Animated_Root_Model(Input : Model_Input_Access)
       is abstract new Root_Model with null record;

  type Animated_Model_2d_Z(Input : Model_Input_Access)
       is abstract new Model_2d_Z with null record;

  type Animated_Model_3d(Input : Model_Input_Access)
       is abstract new Model_3d with null record;

  --------------------------------------------------------------------------------
  -- Models with points to be precalculated by an initialize routine

  type Model_Points_2d_Z is abstract new Model_2d_Z with record
    Points   : Types.Points_2d_Access;
  end record;

  procedure Initialize(The_Model : in out Model_Points_2d_Z);-- is abstract;
  procedure Adjust    (The_Model : in out Model_Points_2d_Z);
  procedure Finalize  (The_Model : in out Model_Points_2d_Z);

  -- Initialize allocates the access object for the points vector. 
  -- Adjust calls initialize to perform a copy of the allocated points and is 
  -- inherited by each subclass of model_2d_points.
  -- Finalize deallocates the points component and is inherited by
  -- each subclass of model_2d_points.

  type Model_Points_3d is abstract new Model_3d with record
    Points   : Types.Points_3d_Access;
  end record;

  procedure Initialize(The_Model : in out Model_Points_3d);-- is abstract;
  procedure Adjust    (The_Model : in out Model_Points_3d);
  procedure Finalize  (The_Model : in out Model_Points_3d);

  ----------------------------------------------------------------------------
  -- Types for all models on screen
  ----------------------------------------------------------------------------

  type Render_Group_Function is access function return Boolean;

  type Model_Group(Model_Number : Natural) is record
    Render            : Render_Group_Function := null;
    Models            : Root_Model_Array      (1 .. Model_Number);
    Input_Calculation : Model_Input_Procedures(1 .. Model_Number);
    Render_Time       : Types.Float_Vector    (1 .. Model_Number) :=
                                              (others => 0.0);
    Render_Counter    : Natural := 0;
  end record;

  type Model_Group_Access is access all Model_Group;
  type Model_Group_Array  is array(Positive range <>) of Model_Group_Access;

  procedure Free(Item : in out Model_Group_Access);

  procedure Write (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   : in     Model_Group_Access);

  procedure Read  (Stream : access Ada.Streams.Root_Stream_Type'Class;
                   Item   :    out Model_Group_Access);

  for Model_Group_Access'Write  use Write;
  for Model_Group_Access'Read   use Read;


  procedure Render(The_Group : in out Model_Group);

  -- Renders whole group if the component Render = null or if it returns true. 

  -- The model_group type is introduced for the case, that groups of models are
  -- rendered only under certain circumstances, e.g. if they represent all 
  -- instruments of a certain (aircraft) model, and more than one model can be 
  -- simulated at a time. Then the render_group_function indicates if a group 
  -- is to render or not.

  Default_Render_Interval : constant
    Ada.Real_Time.Time_Span := Ada.Real_Time.Milliseconds(10);

  type Window_Models(Group_Number : Natural) is record
    Background_Color : Types.Color_Type := Colors.Black; -- of window
    Render_Intervall : aliased Ada.Real_Time.Time_Span :=
                               Default_Render_Interval;
    Permanent_Group  : Model_Group_Access;
    Model_Groups     : Model_Group_Array(1 .. Group_Number);
  end record;

  type Window_Models_Access is access all Window_Models;

  procedure Free(Item : in out Window_Models_Access);

  procedure Render(The : in out Window_Models);

  -- Renders all groups. 


  procedure Write(The      : in Window_Models;
                  Filename : in String);

  -- Writes the_model to file "filename".

  function Read  (Filename : in String) return Window_Models;

  -- Returns the models from file "filename".
  -- IO_Exceptions.Name_Error is raised if file with name filename does not exist.                               

  ----------------------------------------------------------------------------
  
  procedure Antialiasing_On;
  
  -- Switches antialiasing on.
  
  ----------------------------------------------------------------------------
  -- Color procedures

  procedure Set_Color(The_Color : in Types.Color_Type;
                      Force     : in Boolean := False);

  -- If     force, then glcolor3f is called in each case.                      
  -- If not force, then glcolor3f is called only, when the 
  -- current color (a package body permanent variable) is /= The_Color.                      
  -- Force has to be set to true, if called to generate a display-list.

  pragma Inline(Set_Color);

  procedure Set_Color(The_Model : in Model'Class;
                      Force     : in Boolean := False);

  -- Sets color to the one of The_Model.

  pragma Inline(Set_Color);

  ----------------------------------------------------------------------------
  -- General motion procedures

  procedure Translate(The_Vector : in Types.Vector_3d);

  -- Translates in the three dimensions of The_Vector.

  procedure Rotate   (The_Angle  : in Gl.Glfloat;
                      The_Axis   : in Types.Vector_3d);

  -- Rotates The_Angle degrees about The_Axis.                    

  pragma Inline(Translate, Rotate);

  procedure Translate_X(The_Value : in Gl.Glfloat);
  procedure Translate_Y(The_Value : in Gl.Glfloat);
  procedure Translate_Z(The_Value : in Gl.Glfloat);

  pragma Inline(Translate_X, Translate_Y, Translate_Z);
  
  
  procedure Rotate_X (The_Angle  : in Gl.Glfloat);
  procedure Rotate_Y (The_Angle  : in Gl.Glfloat);
  procedure Rotate_Z (The_Angle  : in Gl.Glfloat);

  -- Rotate The_Angle about x,y,z axis.

  pragma Inline(Rotate_X, Rotate_Y, Rotate_Z);


  -- Classwide subprograms that set the origin corresponding to 
  -- the position and attitude components of The_Model.

  procedure Translate(The_Model  : in Model_2d_Z'Class);
  procedure Rotate_Z (The_Model  : in Model_2d_Z'Class);

  pragma Inline(Translate, Rotate_Z);

  procedure Translate(The_Model  : in Model_3d'Class);


  subtype Rotation_Axis is Positive range X .. Z;

  procedure Rotate   (The_Model  : in Model_3d'Class;
									    Rotation_1 : in Rotation_Axis := X;
									    Rotation_2 : in Rotation_Axis := Y;
									    Rotation_3 : in Rotation_Axis := Z);

  -- The rotation sequence is significant when rotating about more than one
  -- axis. Rotates in the sequence Rotation_1, Rotation_2, Rotation_3.      

  procedure Rotate_X (The_Model  : in Model_3d'Class);
  procedure Rotate_Y (The_Model  : in Model_3d'Class);
  procedure Rotate_Z (The_Model  : in Model_3d'Class);

  pragma Inline(Translate, Rotate);
  pragma Inline(Rotate_X, Rotate_Y, Rotate_Z);


  -- Other classwide subprograms that set the origin corresponding to 
  -- position and attitude

  Render_Count : Positive := 1;

  -- This global variable is used to inform some render-procedures about 
  -- how often they should repeat their rendering code.
  -- Render_Count = 1 renders once, Render_Count = 2 renders twice etc..

  ----------------------------------------------------------------------------
  -- Two pairs of class-wide procedures for model_2d_z'class and 
  -- model_3d'class that are necessary to manipulate nested coordinate-systems
  -- in hierarchically nested models.
  -- The Render procedures of hierarchical models all have the same skeleton.
  -- 
  -- procedure Render(The_Model : in out Some_Type_In_Animated_Model_2D_Z_Class_Tree) is
  --   Input : Some_Input_In_Model_Input_Class_Tree renames 
  --           Some_Input_In_Model_Input_Class_Tree(The_Model.Input.all);
  -- begin
  --   The_Model.Attitude := Some_Function   (Input.Some_Component);
  --   The_Model.Position := Another_Function(Input.Another_Component);
  --
  --   Initialize_Render(The_Model);
  --   Render(The_Model.First_Part);
  --   ...
  --   Render(The_Model.Last_Part);
  --   Finalize_Render(The_Model);
  -- end Render;
  --
  -- To calculate The_Model.Attitude and The_Model.Position from model inputs
  -- is only possible for models in the Animated_Model_2d_Z'Class and in the
  -- Animated_Model_3d'Class.
  -- 
  -- Initialize_Render has to be called at the beginning of a Render procedure,
  -- Finalize_Render has to be called at the end. Initialize_Render pushes the 
  -- actual model-view-matrix (the one of the enclosing model) on an 
  -- OpenGL-internal stack, than translates and rotates the local 
  -- cordinate-system defined by the position and attitude components of 
  -- the_model using the global variables Motion_Sequence and Rotation_1, 
  -- Rotation_2, Rotation_3 in the 3D case.
  -- Finalize_Render simply pops the model-view-matrix of the enclosing model.
  --
  -- The position and attitude components of submodel components shall
  -- always be given locally, relative to the enclosing model.
  -- All Rotations and Translations are accumulative. 
  

  Model_Name  : Natural := 0;

  -- If model_name /= 0 then the following call to initialize_model calls
  -- glLoadName(model_name) for later selection.

  Scale_Model : Boolean := False;

  -- If Scale_Model = true, then calls gl.glscalef for x, y 
  -- (and z for model_3d) with The_Model.Size_Scaling in Initilize_Render.

  type Motion_Sequence_Type is (Translate_First, Rotate_First);

  Motion_Sequence : Motion_Sequence_Type := Translate_First;

  -- If the motion_sequence of the Initialize_Render procedures 
  -- inside the Render procedures shall be altered from outside,
  -- this variable has to be used.

  procedure Initialize_Render(The_Model : in out Model_2d_Z'Class);

  procedure Finalize_Render  (The_Model : in out Model_2d_Z'Class);


  Rotation_1 : Rotation_Axis := X;
  Rotation_2 : Rotation_Axis := Y;
  Rotation_3 : Rotation_Axis := Z;

  -- If the rotation-sequence of the Initialize_Render procedure for 3D models
  -- inside the Render procedure shall be altered from outside,
  -- this variables have to be used.

  procedure Initialize_Render(The_Model : in out Model_3d'Class);

  procedure Finalize_Render  (The_Model : in out Model_3d'Class);

end Opengl.Models;
