--                    -*- Mode: Ada -*-
-- Filename        : opengl-timer.ads
-- Description     : Package for measuring opengl graphics performance. 
-- Author          : J. Schrer
-- Created On      : 23. Feb. 1998
-- Last Modified By:
-- Last Modified On:
-- Update Count    : 1
-- Status          :

with Ada.Real_Time;
with Ada.Strings.Unbounded;
with Opengl.Models;
with Opengl.Models.Strings;
with Opengl.Gl;

package Opengl.Timer is
  
  Output    : aliased Boolean := True;
  -- If output = true, renders a timer model, else supresses rendering.

  subtype Line_Number_Range is Natural range 0 .. 4;

  Line      : array(Line_Number_Range'First + 1 .. Line_Number_Range'Last) 
              of String(1 .. 132) := (others => (others => ' '));
  Info_Text : String renames Line(1);

  -- Is displayed at the bottom of the screen.

  procedure Set(The_Item : in String;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True);

  procedure Set(The_Item : in Integer;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True);

  procedure Set(The_Item : in Gl.Glfloat;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True);

  procedure Set(The_Item : in Boolean;
                To_Line  : in Line_Number_Range := 0;
                Clear    : in Boolean           := True;
                Append   : in Boolean           := True);

  -- Copies The_Text to Line(To_Line). 
  -- If To_Line = 0, then increments an internal counter and writes to the
  -- next line. If the last line that was written to was 4 then writes to 
  -- line 1. If To_Line in 1 .. 4 writes to line(To_Line).
  -- If Clear then Line(To_Line)(The_Text'Length + 1 .. Line(To_Line)'Last)
  -- is set blank.
  -- If Append, then appends to line, else starts at first character.                
                        
  function Sec_Since_Start return Float;
  
  -- Returns the number of seconds till program startup in sec.                                    


  type Model(Screen_Update_Dt_Sec : Positive;
             Increment            : Positive;
             Max_Iterations       : Positive;
             Include_Render_Time  : Boolean;
             Text                 : Ada.Strings.Unbounded.String_Access)
      is new Opengl.Models.Model_2d_Z with private;

  -- Screen_Update_Dt_Sec : The timer text-fields are updated at this rate.
  -- Increment            : The calculations are performed at each increment'th
  --                        call of Timer.Render. If a render function with very 
  --                        few calculations shall be measured, 
  --                        then increment shall be set to a greater value than 1, 
  --                        because the overhead of a call to Timer.Render 
  --                        may falsify the result.
  -- Max_Iterations       : After max_iterations calls to Timer.Render, the 
  --                        timer text-fields are no longer updated;
  -- Include_Render_Time  : If false, the time necessary for executing 
  --                        Timer.Render is subtracted. 
  --                        This shall be the normal use of the timer.
  -- Text                 : The text to display (may be null).    


  procedure Print_Statistics(The_Model : in Model);

  -- If Ada.Text_Io.Is_Open(Opengl.Log_File), 
  -- writes some statistics to that file.

  procedure Render(The_Model : in out Model);

  pragma Inline(Render);

private ----------------------------------------------------------------------
  
  Number_Image_Length : constant := 10;

  type Strings is array(Line'range) of
    Models.Strings.String_2d(Length => Line(1)'Length);

  type Model(Screen_Update_Dt_Sec : Positive;
             Increment            : Positive;
             Max_Iterations       : Positive;
             Include_Render_Time  : Boolean;
             Text                 : Ada.Strings.Unbounded.String_Access)
       is new Opengl.Models.Model_2d_Z with record
         
    Screen_Update_Dt        : Ada.Real_Time.Time_Span := 
                              Ada.Real_Time.Milliseconds(1000);
    Counter                 : Natural                 := 0;
    Last_Time,
    Last_Screen_Update_Time : Ada.Real_Time.Time      := Ada.Real_Time.Clock;
    Mean_Dt,
    Max_Dt                  : Float                   := 0.0;
    Counter_Image,
    Dt_Image,
    Mean_Dt_Image,
    Max_Dt_Image            : Models.Strings.String_2d
                              (Length => Number_Image_Length);
    Text_Image              : Models.Strings.String_2d_Access;
    Line_Images             : Strings;
    Display_List_Index      : Gl.Gluint := 0;
  end record;

  procedure Initialize(The_Model : in out Model);

end Opengl.Timer;
