--                              -*- Mode: Ada -*-
-- Filename        : opengl-utils.adb
-- Description     : Package for utilities to simplify the use of the
--                   thin Opengl-bindings
-- Author          : J. Schrer
-- Created On      : 17 Nov. 1997
-- Last Modified By: J. Schrer
-- Last Modified On:
-- Update Count    : 1
-- Status          :

--with Ada.Unchecked_Conversion;
with Portable.Standard; use Portable.Standard;
with Ada.Strings.Fixed;
with Math.Float_64.Text_Io;

with Opengl.Screen;
with Opengl.Float_32.Arrays.Vector_Operations;
with Opengl.Models.dialfaces;
with Opengl.Colors.Palette;
with Opengl.Command_Line;

package body Opengl.Utils is

  use type Gl.Glfloat, Gl.Gldouble, Gl.Glsizei;
  use Opengl.Float_32.Arrays.Vector_Operations;
  --------------------------------------------------------------
  function Number_Array_Of(First           : in Integer; 
                           Increment       : in Integer; 
                           Number          : in Natural;
                           Index_First     : in Positive := 1;
                           Index_Increment : in Positive := 1)
    return Types.Integer_Array is
      
    The   : Types.Integer_Array(1 .. Number) := 
           (others => Opengl.Models.dialfaces.Invalid_Number);
    Index : Positive := Index_First;     
  begin
    The(Index_First) := First;
    loop
      Index := Index + Index_Increment;
      exit when Index > Number;
      The(Index) := The(Index - Index_Increment) + Increment;
    end loop;
    return The;
  end Number_Array_Of;
  -------------------------------------------------------------------------- 
  Pixels_Per_Mm_Width  : constant Gl.Glfloat :=
    Gl.Glfloat(Screen.Pixel_Width) / Screen.Width;

  Pixels_Per_Mm_Height : constant Gl.Glfloat :=
    Gl.Glfloat(Screen.Pixel_Height) / Screen.Height;
  ----------------------------------------------------------------------------
  procedure Resize_Window(New_Width, New_Height : in Natural) is

    Height          : constant natural := natural'max(New_Height, 1);
    Width           : constant natural := natural'max(New_Width,  1);
    Screen_Width,
    Screen_Height   : Gl.Gldouble;
    Height_To_Width : Gl.Gldouble;
  begin
    Gl.Glviewport(X      => 0,                 Y      => 0, 
                  Width  => gl.glsizei(Width), Height => gl.glsizei(Height));
    Gl.Glmatrixmode(Gl.Gl_Projection);
    Gl.Glloadidentity;

    Height_To_Width := Gl.Gldouble(Height) / Gl.Gldouble(Width);

    if Width <= Height then
      Screen_Height := Gl.Gldouble(Screen.Width)  * Height_To_Width;
      Screen_Width  := Gl.Gldouble(Screen.Width);
    else
      Screen_Height := Gl.Gldouble(Screen.Height);
      Screen_Width  := Gl.Gldouble(Screen.Height) / Height_To_Width;
    end if;

--    gl.Glortho(Left   => - Screen_Width,  Right => Screen_Width, 
--               Bottom => - Screen_Height, Top   => Screen_Height,
--               Znear  => 1000.0,          Zfar  => - 1000.0);

    Gl.Glortho(Left   => 0.0,    Right => Screen_Width,
               Bottom => 0.0,    Top   => Screen_Height,
               Znear  => 1000.0, Zfar  => - 1000.0);

    Gl.Glmatrixmode(Gl.Gl_Modelview);
    Gl.Glloadidentity;
  end Resize_Window;
  ----------------------------------------------------------------------------
  function Center_Of(Rectangle : in Types.Screen_Rectangle)
           return Types.Vector_2d is
  begin
    return (Rectangle.Lower_Left + Rectangle.Upper_Right) / 2.0;
  end Center_Of;
  ----------------------------------------------------------------------------
  function Center_Of(Rectangle : in Types.Pixel_Rectangle)
           return Types.Pixel_Point is
  begin
    return Types.Pixel_Point'
      (X => (Rectangle.Lower_Left(X) + Rectangle.Upper_Right(X)) / 2,
       Y => (Rectangle.Lower_Left(Y) + Rectangle.Upper_Right(Y)) / 2);
  end Center_Of;
  ----------------------------------------------------------------------------
  function "+" (Left  : in Types.Screen_Rectangle;
                Right : in Types.Vector_2d)        return Types.Screen_Rectangle is
  begin
    return Types.Screen_Rectangle'
      (Lower_Left  => Left.Lower_Left  + Right,
       Upper_Right => Left.Upper_Right + Right);
  end "+";
  ----------------------------------------------------------------------------                
  function "+" (Left  : in Types.Vector_2d;
                Right : in Types.Screen_Rectangle) return Types.Screen_Rectangle is
  begin
    return Right + Left;
  end "+";
  ----------------------------------------------------------------------------                
  function "+" (Left  : in Types.Rectangle;
                Right : in Types.Vector_2d)        return Types.Screen_Rectangle is
  begin
    return Types.Screen_Rectangle'
      (Lower_Left  => Right,
       Upper_Right => Types.Vector_2d(Left) + Right);
  end "+";
  ----------------------------------------------------------------------------                
  function "+" (Left  : in Types.Vector_2d;
                Right : in Types.Rectangle)        return Types.Screen_Rectangle is
  begin
    return Types.Screen_Rectangle'
      (Lower_Left  => Left,
       Upper_Right => Left + Types.Vector_2d(Right));
  end "+";
  ----------------------------------------------------------------------------                
  function Screen_Point_Of(The_Point         : in Types.Pixel_Point)
           return Types.Vector_2d is
  begin
    return Types.Vector_2d'
      (X => Gl.Glfloat(The_Point(X)) / Pixels_Per_Mm_Width,
       Y => Gl.Glfloat(The_Point(Y)) / Pixels_Per_Mm_Height);
  end Screen_Point_Of;
  ----------------------------------------------------------------------------           
  function Pixel_Point_Of(The_Point          : in Types.Vector_2d;
                          Round              : in Rounding := Up)
           return Types.Pixel_Point is
  begin
    case Round is
      when Down =>
        return Types.Pixel_Point'
		      (X => Natural(Gl.Glfloat'Floor(The_Point(X) * Pixels_Per_Mm_Width)),
		       Y => Natural(Gl.Glfloat'Floor(The_Point(Y) * Pixels_Per_Mm_Height)));
      when Up   =>
        return Types.Pixel_Point'
		      (X => Natural(Gl.Glfloat'Ceiling(The_Point(X) * Pixels_Per_Mm_Width)),
		       Y => Natural(Gl.Glfloat'Ceiling(The_Point(Y) * Pixels_Per_Mm_Height)));
    end case;
  end Pixel_Point_Of;
  ----------------------------------------------------------------------------
  function Screen_Rectangle_Of(The_Rectangle : in Types.Pixel_Rectangle)
           return Types.Screen_Rectangle is
  begin
    return Types.Screen_Rectangle'
      (Lower_Left  => Screen_Point_Of(The_Rectangle.Lower_Left),
       Upper_Right => Screen_Point_Of(The_Rectangle.Upper_Right));
  end Screen_Rectangle_Of;
  ----------------------------------------------------------------------------           
  function Pixel_Rectangle_Of (The_Rectangle : in Types.Screen_Rectangle;
                               Next_Power_2  : in Boolean := False)
           return Types.Pixel_Rectangle is

    Pixels : Types.Pixel_Rectangle :=
	    (Lower_Left  => Pixel_Point_Of(The_Rectangle.Lower_Left,  Round => Down),
	     Upper_Right => Pixel_Point_Of(The_Rectangle.Upper_Right, Round => Up));
    Pixel_Length,
    Pixel_Length_Pow_2,
    Half_Pixel_Length_Diff : Types.Pixel_Point;
    -----------------------------------------------------
    function Next_Pow_2(N : in Natural) return Natural is
      Pow : Positive := 1;
    begin
      while Pow < N loop
        Pow := Pow * 2;
      end loop;
      return Pow;
    end Next_Pow_2;
    ---------------
  begin
    if Next_Power_2 then
      Pixel_Length :=
        (X => Pixels.Upper_Right(X) - Pixels.Lower_Left(X),
         Y => Pixels.Upper_Right(Y) - Pixels.Lower_Left(Y));
      Pixel_Length_Pow_2 :=
        (X => Next_Pow_2(Pixel_Length(X)),
         Y => Next_Pow_2(Pixel_Length(Y)));
      Half_Pixel_Length_Diff :=
        (X => (Pixel_Length_Pow_2(X) - Pixel_Length(X)) / 2,
         Y => (Pixel_Length_Pow_2(Y) - Pixel_Length(Y)) / 2);
	    return Types.Pixel_Rectangle'
        (Lower_Left  => (X => Pixels.Lower_Left(X) - Half_Pixel_Length_Diff(X),
                         Y => Pixels.Lower_Left(Y) - Half_Pixel_Length_Diff(Y)),
         Upper_Right => (X => Pixels.Lower_Left(X) + Pixel_Length_Pow_2(X),
                         Y => Pixels.Lower_Left(Y) + Pixel_Length_Pow_2(Y)));
    else
	    return Types.Pixel_Rectangle'
	      (Lower_Left  => Pixel_Point_Of(The_Rectangle.Lower_Left),
	       Upper_Right => Pixel_Point_Of(The_Rectangle.Upper_Right));
    end if;
  end Pixel_Rectangle_Of;
  ----------------------------------------------------------------------------
  function To_Math_Angle (Compass_Angle : in Gl.Glfloat) return Gl.Glfloat is
  begin
    return 90.0 - Compass_Angle;
  end To_Math_Angle;
  ----------------------------------------------------------------------------
  function To_Compass_Angle (Math_Angle : in Gl.Glfloat) return Gl.Glfloat is
  begin
    return 90.0 - Math_Angle;
  end To_Compass_Angle;
  ----------------------------------------------------------------------------
  function To_Delta_Math_Angle (Delta_Compass_Angle : in Gl.Glfloat)
           return Gl.Glfloat is
  begin
    return - Delta_Compass_Angle;
  end To_Delta_Math_Angle;
  ----------------------------------------------------------------------------
  function To_Delta_Compass_Angle (Delta_Math_Angle : in Gl.Glfloat)
           return Gl.Glfloat is
  begin
    return - Delta_Math_Angle;
  end To_Delta_Compass_Angle;
  ----------------------------------------------------------------------------
  function Limitation (Input_Value : in Gl.Glfloat;
                       Lower_Limit : in Gl.Glfloat;
                       Upper_Limit : in Gl.Glfloat ) return Gl.Glfloat is

    Output_Value : Gl.Glfloat;
  begin
    if    Input_Value > Upper_Limit then
      Output_Value := Upper_Limit;
    elsif Input_Value < Lower_Limit then
      Output_Value := Lower_Limit;
    else
      Output_Value := Input_Value;
    end if;
    return Output_Value;
  end Limitation;
  ----------------------------------------------------------------------------
  procedure Set_Integer_Image(Image      : in out String;
                              Value      : in     Gl.Glfloat;
                              Right_Text : in     String := "") is
                  
    Number : constant String := Ada.Strings.Fixed.Trim
            (Integer'Image(Integer(Value)), Ada.Strings.Both);              
  begin
    if Number'Length <= Image'Length - Right_Text'Length then
      Image(Image'First .. Image'First + Image'Length - 
            Number'Length - Right_Text'Length - 1) := (others => ' ');
      Image(Image'First + Image'Length - Number'Length - Right_Text'Length .. 
            Image'Last - Right_Text'Length) := Number;
      Image(Image'Last - Right_Text'Length + 1 .. Image'Last) := Right_Text;     
    else
      Image(Image'First .. Image'Last - Right_Text'Length) := (others => '-');
      Image(Image'Last - Right_Text'Length + 1 .. Image'Last) := Right_Text;     
    end if;
  end Set_Integer_Image;                      
  ----------------------------------------------------------------------------  
  procedure Set_Float_Image(Image      : in out String;
                            Value      : in     Gl.Glfloat;
                            Aft        : in     Ada.Text_Io.Field := 
                                                Opengl.Float_32.Text_Io.Default_Aft;
                            Exp        : in     Ada.Text_Io.Field := 0;
                            Right_Text : in     String := "") is
                              
    Fore     : Integer := 1;                          
    Text     : String(1 .. 32); -- Workaround for GNAT 3.13p
    Gnat_Aft : constant := 16;
    Last     : constant Natural := Text'Length - Gnat_Aft + Aft;
  begin   
    Image(Image'Last - Right_Text'Length + 1 .. Image'Last) := Right_Text;                     
    Math.Float_64.Text_Io.Put
      (To   => Text, --Image(Image'First .. Image'Last - Right_Text'Length), 
       Item => float_64(Value), 
       Aft  => Gnat_Aft, 
       Exp  => Exp);
    Image(Image'First .. Image'Last - Right_Text'Length) := 
      Text(Last - Image'Length + Right_Text'Length + 1 .. Last);   
  exception
    when Ada.Text_Io.Layout_Error =>
      if Exp = 0 then
        Fore := Image'Length - Right_Text'Length - Aft - 1;
        Image(Image'First..Image'First+Fore-1)          := (others => '-');
        Image(Image'First+Fore)                         := '.';
        Image(Image'First+Fore+1..Image'First+Fore+Aft) := (others => '-');
      else
        Fore := Image'Length - Right_Text'Length - Aft - Exp - 3;
        Image(Image'First..Image'First+Fore-1)               := (others => '-');
        Image(Image'First+Fore)                              := '.';
        Image(Image'First+Fore+1..Image'First+Fore+Aft)      := (others => '-');
        Image(Image'First+Fore+Aft+1..Image'First+Fore+Aft+2):= "E+";
        Image(Image'First+Fore+Aft+3..
              Image'First+Fore+Aft+2+Exp)                    := (others => '-');
      end if;
  end Set_Float_Image;                      
  ----------------------------------------------------------------------------            
  function Position_Of(Value : in Gl.Glfloat;
                       Slope : in Gl.Glfloat;
                       Min   : in Gl.Glfloat;
                       Max   : in Gl.Glfloat) return Gl.Glfloat is
  begin
    if    Value < Min then
      return Min * Slope;
    elsif Value > Max then
      return Max * Slope;
    else
      return Value * Slope;
    end if;
  end Position_Of;
  ----------------------------------------------------------------------------                         
  function Background_Color return Types.Color_Type is     
    the_color : types.color_type;
  begin
    the_color := Colors.Palette.Color_Array(Colors.Palette.Color_Type'Value
                                           (Command_Line.Background_Color));
    the_color.alpha := 0.0;
    return the_color;                                       
  exception
    when others => return Colors.Black;                                     
  end Background_Color;
  ----------------------------------------------------------------------------                         
--  function To_Auxmainproc(The_Procedure : in System.Address) 
--           return Glaux.Auxmainproc is
--             
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxmainproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxmainproc;
--  ----------------------------------------------------------------------------
--  function To_Auxexposeproc(The_Procedure : in System.Address) 
--           return Glaux.Auxexposeproc is
--             
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxexposeproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxexposeproc;
--  ----------------------------------------------------------------------------
--  function To_Auxreshapeproc(The_Procedure : in System.Address)
--           return Glaux.Auxreshapeproc is
--             
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxreshapeproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxreshapeproc;
--  ----------------------------------------------------------------------------
--  function To_Auxidleproc(The_Procedure : in System.Address)
--            return Glaux.Auxidleproc is
--              
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxidleproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxidleproc;
--  ----------------------------------------------------------------------------
--  function To_Auxkeyproc(The_Procedure : in System.Address)
--           return Glaux.Auxkeyproc is
--             
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxkeyproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxkeyproc;
--  ----------------------------------------------------------------------------
--  function To_Auxmouseproc(The_Procedure : in System.Address)
--           return Glaux.Auxmouseproc is
--             
--    function Convert is new Ada.Unchecked_Conversion
--            (Source => System.Address, Target => Glaux.Auxmouseproc);
--  begin
--    return Convert(The_Procedure);
--  end To_Auxmouseproc;
--  ----------------------------------------------------------------------------
end Opengl.Utils;
